#!/usr/local/bin/perl
#
#Base class for HPCC:Cluster:XXX

package Glue::HPCC::Cluster;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);
use DBI;
use Module::Load;

use Glue::HostinfoVars qw($CLUSTER $HOSTNAME);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $VERBOSE;
my $DRYRUN;
my $DEBUG;
$DEBUG=1;

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @rw_accessors = qw();
my @ro_accessors = qw( 
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);


my @required_parms = qw(
	clustername
	db_clustername
	unixgrp_prefix
	netgroup
	homeroot_user
	homeroot_admin
);

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my ($arg, $meth, $val);
	RWARG: foreach $arg (@rw_accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	ROARG: foreach $arg (@ro_accessors)
	{	next ROARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next ROARG unless defined $val;
		$meth = $arg;
		$obj->set($meth,$val);
	}


	#Warn about unknown arguments
	if ( scalar(%args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;

	return;
}

sub _init($)
{	my $obj = shift;

	my $name = $obj->clustername;
	croak "Missing required argument clustername at " unless $name;

	my ($fld, $meth, $val);
	foreach $fld (@required_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld for cluster $name at ";
		}
	}

}

sub DESTROY($)
#Cleanup
#For now, just disconnect from DBs if connected
{	my $obj = shift;

	$obj->disconnect_from_projects_db;
}

#-------------------------------------------------------------------
#	Common Methods: verbosity and dryrun mode
#-------------------------------------------------------------------

sub verbose($;$)
#Sets verbosity on all of our subordinate modules
{	my $self = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		$VERBOSE=$val;

		my $util = $self->_hpcUtil;
		$util->verbose($VERBOSE) if ( $util && ref($util) );
	}
	return $VERBOSE;
}

sub dryrun($$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		$DRYRUN=$val;

		my $util = $class->_hpcUtil;
		$util->dryrun($DRYRUN) if ( $util && ref($util) );
	}
	return $DRYRUN;
}

#-------------------------------------------------------------------
#	Common Methods: hpcUtil
#-------------------------------------------------------------------

sub _hpcUtil($)
{	my $self = shift;
	return $self->get('_hpcUtil');
}

sub _hpcUtil_class($)
{	my $self = shift;
	my $schedtype = $self->scheduler_type;

	return 'Glue::HPCC::ClusterUtil::Slurm' if $schedtype eq 'slurm';
	return 'Glue::HPCC::ClusterUtil::Dummy' if $schedtype eq 'dummy';
	die "Unrecognized scheduler type '$schedtype' at ";
}

sub get_hpcUtil($)
{	my $self = shift;
	my $me = 'get_hpcUtil';

	#Use cached version if possible
	my $hpcUtil = $self->_hpcUtil;
	return $hpcUtil if $hpcUtil && ref($hpcUtil);

	#Nothing cached, make new instance
	my $hUclass = $self->_hpcUtil_class;
	load $hUclass;
	$hpcUtil = $hUclass->new;
	croak "$me: Error instantiating hpcUtil instance: $hpcUtil at "
		unless $hpcUtil && ref($hpcUtil);

	#Inherit our verbose/dryrun settings
	$hpcUtil->verbose($self->verbose);
	$hpcUtil->dryrun($self->dryrun);

	#Inherit our scheduler_version/rootdir settings
	my $tmp = $self->scheduler_version;
	$hpcUtil->scheduler_version($tmp) if defined $tmp;
	$tmp = $self->scheduler_rootdir;
	$hpcUtil->scheduler_rootdir($tmp) if defined $tmp;
	
	$self->set('_hpcUtil', $hpcUtil);
	return $hpcUtil;
}
	
	
#-------------------------------------------------------------------
#	Common Methods: database related
#-------------------------------------------------------------------

sub db_connection_attribs($)
#Attributes for DB connection.  Can be overridden
{	my $obj = shift;
	my $attribs = {	AutoCommit => 0, RaiseError => 1 };
	return $attribs;
}

sub get_projects_dbh($)
#Returns db handle to projects DB
{	my $obj = shift;

	my $fld = "_projects_dbh";
	my $dbh = $obj->get($fld);
	return $dbh if $dbh && ref($dbh) && $dbh->ping;

	#No dbh (or need a new one)
	#my ($user, $pw) = $obj->get_db_creds;

	#We use kerberos now
	my ($user, $pw);
	my $connstr = $obj->db_connection_string;

	my $attribs = $obj->db_connection_attribs;
	$dbh = DBI->connect($connstr, $user, $pw, $attribs);

	unless ( $dbh && ref($dbh) && $dbh->ping )
	{	die "Unable to connect to DB: $connstr with error: " .
			$DBI::errstr . "\n";
	}
	$obj->set($fld, $dbh);
	return $dbh;
}

sub disconnect_from_projects_db($)
#Disconnects from projects DB, if needed
{	my $obj = shift;
	my $fld = "_projects_dbh";
	my $dbh = $obj->get($fld);
	return unless $dbh && ref($dbh);

	$dbh->disconnect;
}

sub db_commit($)
{	my $obj = shift;
	my $dbh = $obj->get_projects_dbh;
	$dbh->commit;
}

#-------------------------------------------------------------------
#	Methods which might want to be overridden
#-------------------------------------------------------------------

sub scheduler_version($)
#Returns the version of the scheduler to use.
#Interpretation will depend on scheduler_type()
#Default version returns default based on scheduler_type()
#figure out how to default.
{	my $self = shift;
	my $schedtype = $self->scheduler_type;
	if ( $schedtype eq 'slurm' )
	{	return '14';
	}
	return;
}

sub scheduler_rootdir($)
#Returns the root directory for the scheduler being used.
#Interpretation will depend on scheduler_type()
#Default version returns based on scheduler_type;
#figure out how to default.
{	my $self = shift;
	my $schedtype = $self->scheduler_type;
	if ( $schedtype eq 'slurm' )
	{	return '/usr/local/slurm';
	}
	return;
}


#-------------------------------------------------------------------
#	Methods needing to be overridden
#-------------------------------------------------------------------

sub _missing_overload($$)
{	my $obj = shift;
	my $method = shift;
	my $class = $obj;
	$class = ref($class) if ref($class);

	die "Failed to overload '$method' for class $class ";
}

sub db_connection_string($)
{	my $obj = shift;
	$obj->_missing_overload('db_connection_string');
}

sub db_password_file($)
{	my $obj = shift;
	$obj->_missing_overload('db_password_file');
}

sub clusters_sharing_pdb_tables($)
{	my $class = shift;
	return [];
}

sub clusters_sharing_unixgrp_prefix($)
{	my $class = shift;
	return [];
}

sub headnode($)
{	my $class = shift;
	$class->_missing_overload('headnode');
}

sub domain($)
{	my $class = shift;
	$class->_missing_overload('domain');
}

sub clustername($)
{	my $class = shift;
	$class->_missing_overload('clustername');
}
	
sub clustershort($)
{	my $class = shift;
	$class->_missing_overload('clustershort');
}
	
sub db_clustername($)
{	my $class = shift;
	$class->_missing_overload('db_clustername');
}
	
sub unixgrp_prefix($)
{	my $class = shift;
	$class->_missing_overload('unixgrp_prefix');
}
	
sub netgroup($)
{	my $class = shift;
	$class->_missing_overload('netgroup');
}
	
sub homeroot_user($)
{	my $class = shift;
	$class->_missing_overload('homeroot_user');
}
	
sub homeroot_admin($)
{	my $class = shift;
	$class->_missing_overload('homeroot_admin');
}

sub home_softquota_logfile($)
{	my $class = shift;
	$class->_missing_overload('home_softquota_logfile');
}

sub scheduler_type($)
{	my $class = shift;
	$class->_missing_overload('scheduler_type');
}

sub lustre_filesystems($)
#Returns a list ref of lustre filesystems 
{       my $class = shift;
	$class->_missing_overload('lustre_filesystems');
}

sub nfs_data_filesystems($)
#Returns a list ref of lustre filesystems 
{       my $class = shift;
	$class->_missing_overload('lustre_filesystems');
}

sub mailinglist_name($)
#Returns the name of the mailing list for this cluster
#This should return undef or empty string if no mailing list for cluster
{       my $class = shift;
	$class->_missing_overload('mailinglist_name');
}

sub mailinglist_pwdfile($)
#Returns the name of the file containing the mailinglist's password
{       my $class = shift;
	$class->_missing_overload('mailinglist_pwdfile');
}

sub mailinglist_password($)
#Returns the mailing list password (from the mailinglist_pwdfile)
{	my $class = shift;
	my $me = __PACKAGE__ . '::mailinglist_password';

	my $pwdfile = $class->mailinglist_pwdfile;
	unless ( $pwdfile )
	{	carp "$me: called but no pwdfile defined at ";
		return;
	}
	unless ( -f $pwdfile && -r $pwdfile )
	{	carp "$me: No/Unable to read pwdfile $pwdfile at ";
		return;
	}

	my $FILE;
	open($FILE, "<$pwdfile") 
		or die "$me: Error opening pwdfile $pwdfile: $! at ";
	my @lines = <$FILE>;
	close($FILE);
	chomp @lines;
	my $pwd = $lines[0];
	carp "$me: No password found on first line of $pwdfile at "
		unless $pwd;
	return $pwd;
}

sub mailinglist_listserv($)
{	my $class = shift;
	return 'listserv@listserv.umd.edu';
}

sub clusters_sharing_mailinglist($)
{	my $class = shift;
	return [];
}

sub should_users_get_db_access($)
{	my $class = shift;
	$class->_missing_overload('should_users_get_db_access');
}


#-------------------------------------------------------------------
#	Common cluster/hostname utilities
#-------------------------------------------------------------------

sub cluster_class_from_clustername($$)
#Returns name of Cluster class given the name of the cluster
{	my $class = shift;
	my $cluster = lc shift;
	my $me = 'cluster_class_from_clustername';
	
	my $prefix = "Glue::HPCC::Cluster::";
	return $prefix . 'Deepthought2'
		if $cluster eq 'dt2' || $cluster eq 'deepthought2';
	return $prefix . 'Deepthought'
		if $cluster eq 'dt' || $cluster eq 'deepthought';
	return $prefix . 'Bswift'
		if $cluster eq 'bswift';
	return $prefix . 'Evergreen'
		if $cluster eq 'eg' || $cluster eq 'evergreen';
	return $prefix . 'BlueCrab'
		if $cluster eq 'bc' || $cluster eq 'bluecrab' 
			|| $cluster eq 'marcc' ;
	return $prefix . 'Pippin'
		if $cluster eq 'pippin';

	croak "$me: do not recognize cluster '$cluster' at ";
}
	
sub new_from_clustername($;$@)
#Special constructor.  Returns the appropriate child class instance of 
#Glue::HPCC::Cluster for the named cluster, or the cluster which this
#host belongs to if not named.
{	my $class = shift;
	my $clustername = shift;
	my @extra = @_;
	my $me = 'new_from_clustername';

	unless ( $clustername )
	{	#Try to get from CLUSTER in hostinfo
		$clustername = $CLUSTER;
		my $hostname = $HOSTNAME;

		#Hack for Tom's desktops
		$clustername='hpcc-dt2' if $HOSTNAME eq 'proxima.umd.edu';
		$clustername='hpcc-dt2' if $HOSTNAME eq 'procyon.umd.edu';
		$clustername='hpcc-dt2' if $HOSTNAME eq 'barnards.umd.edu';
		$clustername='hpcc-dt2' if $HOSTNAME eq 'wolf359.umd.edu';

		unless ( $clustername =~ /^hpcc-/ )
		{	warn "$me: This host ($HOSTNAME) does not appear to " .
				"be part of " .
				"an HPC cluster ($clustername)\n";
			croak "$me: Unable to set cluster from host, and " .
				"none specified, aborting at ";
		}
		$clustername =~ s/^hpcc-//;
	}

	my $newclass = $class->cluster_class_from_clustername($clustername);
	croak "$me: Unable to get classname for cluster $clustername at "
		unless $newclass;

	load $newclass;

	my $hpc = $newclass->new(@extra);
}

sub running_on_headnode($)
#Returns true if we are running on headnode of cluster,
#e.g. $HOSTNAME from hostinfo == headnode
{	my $class = shift;
	my $me='running_on_headnode';
	my $headnode = $class->headnode;
	croak "$me: Error obtaining headnode for $class at "
		unless $headnode;
	my $hostname = $HOSTNAME;
	croak "$me: Cannot obtain hostname from HOSTNAME hostinfo var at "
		unless $hostname;

	return $headnode eq $hostname;
}

sub running_in_cluster($)
#Returns true if we are running on a host in the HPC cluster
{	my $class = shift;
	my $me = 'running_in_cluster';
	my $cluster = $class->clustername;
	my $wanted_cluster = "hpcc-${cluster}";

	return $CLUSTER eq $wanted_cluster;
}
	
#-------------------------------------------------------------------
#	QoS utilities
#-------------------------------------------------------------------

sub get_qos_list($$$)
#Returns the qos list (as list ref) associated with the given 
#tag.  Basically just alls the method qos_list_$tag, except when tag
#is std+ext, in which case returns the combination of standard + extended.
{	my $class = shift;
	my $tag = shift;
	my $projname = shift;
	my $me = 'get_qos_list';
	croak "$me: Missing required parameter 'tag' at " unless $tag;
	croak "$me: Missing required parameter 'project' at " unless $projname;
	$tag = lc $tag;

	my $qoses;
	if ( $tag eq 'std+ext' || $tag eq 'ext+std' )
	{	$qoses = $class->qos_list_standard($projname) || [];
		my $tmp = $class->qos_list_extended($projname) || [];
		push @$qoses, @$tmp;
		return $qoses;
	}
	

	my $meth = "qos_list_${tag}";
	if ( $class->can($meth) )
	{	$qoses = $class->$meth($projname) || [];
		return $qoses;
	} else
	{	croak "$me: Illegal tag $tag; unable to find $meth at ";
	}
}

sub qos_list_standard($$)
{	my $class = shift;
	my $projname = shift;
	return [ 	'narrow-long',
			'narrow-med',
			'wide-med',
			'wide-short',
		];
}

sub qos_list_extended($$)
{	my $class = shift;
	my $projname = shift;
	return [	'med-extended',
			'narrow-extended',
		];
}

sub qos_list_debug($$)
{	my $class = shift;
	my $projname = shift;
	return	[	'debug' ];
}

sub qos_list_scavenger($$)
{	my $class = shift;
	my $projname = shift;
	return	[	'scavenger' ];
}

sub qos_list_condo($$)
{	my $class = shift;
	my $projname = shift;
	my $me = 'qos_list_condo';
	croak "$me: Missing required parameter 'projname' at " unless $projname;
	
	my $qosstd = "${projname}-std";
	my $qoshi = "${projname}-hi";
	return	[	$qosstd, $qoshi ];
}


1;
__END__

=head1 NAME

Glue::HPCC::Cluster - Base class for Glue HPC cluster objects

=head1 SYNOPSIS

  package Glue::HPCC::Cluster::FooCluster;
  use base qw(Glue::HPCC::Cluster::Base);

  ...


=head1 DESCRIPTION

This is the base class for Glue::HPCC::Cluster::XXX cluster objects.
It provides some common routines and default values.  

Most of the heavy lifting for managing the cluster has been refactored
into Glue::HPCC::ClusterUtil, which subclasses into Slurm, etc. specific
methods. 

The primary methods of interest are:

=over 4

=item B<new_from_clustername($clustername, @args)>

This class method takes an optional clustername (will attempt to default
based on the hostinfo cluster of the host you are running on), and figures
out what is the corresponding Glue::HPCC::Cluster::FOO class corresponding
to it.  And then run and return B<new> with the specified @args on it.
You MUST include 'undef' for $clustername if you want to force it to be
defaulted and still want to give args to it, but since currently no args
are recognized, this is not very onerous.

=item B<get_project_dbh>

Returns the DBI DB handle to the projects/hpcc database, using a cached one
if possible, otherwise getting a new one.

=item B<get_hpcUtil>

Returns the Glue::HPCC::ClusterUtil instance for this cluster, caching when
possible.  This is now the interface to all the heavy lifting of creating
projects, etc.

=item B<verbose>

Gets/sets the verbosity level.  This gets propagated to hpcUtil if/when
created.

=item B<dryrun>

Gets/sets the dryrun mode.  This gets propagated to hpcUtil if/when
created.

=item B<get_qos_list($tag, $projname)>

Returns the QoS list corresponding to B<$tag>, as a list reference.  The
project name is required, as the QoS names might depend on that.
Recognized tags include:
B<standard>, B<extended>, B<std+ext>, B<condo>, B<debug>, B<scavenger>.    
This wrapper calls
B<qos_list_$tag> (except for B<std+ext>, which is concatenation of the
B<standard> and B<extended> lists) to obtain the list.  The base class
provides defaults for these, but allows for them to be overridden in child
classes.

=item B<qos_list_standard>, B<qos_list_extended>, B<qos_list_condo>, B<qos_list_debug>, B<qos_list_scavenger>:

Returns the standard, extended, or condo QoS list for the cluster, as a 
list reference.  The base class provides defaults, but child classes can
override.

=back

In addition, the cluster specific child classes must define the following 
methods as well:

=over 4

=item B<db_connection_string>

The DBI connection string for connecting the 
Project/HPCC database for this cluster.

=item B<clusters_sharing_pdb_tables>

Returns a listref of names (suitable for B<new_from_clustername>) for HPC
clusters that share some tables in the projects database.  The base class
version just returns [].  The list returned may or may not include the current
clustername.  This is needed so e.g. if user johndoe is removed from DT cluster,
we need to determine whether he exists on DT2 before disabling in common tables.

=item B<clusters_sharing_unixgrp_prefix>

Returns a listref of names (suitable for B<new_from_clustername) for HPC
clusters that share the same Unix group prefix as this cluster.  The base
class version just returns [].  The list returned may or may not include the
current clustername.  Used to determine if it is safe to delete the unix
group corresponding to this cluster.

=item B<headnode>

The hostname of the headnode for the cluster (e.g. deepthought.umd.edu);

=item B<domain>

The domain name for the cluster (e.g. deepthought.umd.edu)

=item B<clustername>

The clustername, long (e.g. deepthought2)

=item B<clustershort>

The clustername, short (e.g. dt2)

=item B<db_clustername>

The clustername as used in the DB (e.g. dt2)

=item B<netgroup>

The netgroup controlling login access to the login nodes

=item B<homeroot_user>

The root directory for normal user home directories.

=item B<homeroot_admin>

The root directory for admin user home directories.

=item B<scheduler_type>

The scheduler type for this cluster (e.g. slurm)

=item B<unixgrp_prefix>

The prefix for the unix groups for projects in this cluster (e.g. 'dt-')

=item B<lustre_filesystems>

Returns a list reference of mountpoints for the lustre filesystems on this
cluster.  This is so that we can autocreate $username subdirectories for
the users of the cluster (and restrict access on toplevel of lustre).
And for searching for user data when deleting accounts.

=item B<nfs_data_filesystems>

Returns a list reference of mountpoints for the NFS data filesystems on this
cluster.  This is so that we can autocreate $username subdirectories for
the users of the cluster (and restrict access on toplevel of lustre).
And for searching for user data when deleting accounts.

=item B<mailinglist_name>

Returns the name of the mailing list for the cluster.  If empty/not-defined,
no mailing list and users are NOT automatically subscribed/unsubscribed as
added/removed.

=item B<mailinglist_pwdfile>

Returns the path to a file containing the password for the mailing list.
Required if B<mailinglist_name> is set.

=item B<mailinglist_listserv>

Returns the email address to the listserv for B<mailinglist_name>.  Defaults
to C<listserv@listserv.umd.edu>

=item B<clusters_sharing_mailinglist>

Returns a listref of names (suitable for B<new_from_clustername>) for HPC
clusters that share the same mailing list as this cluster.  The base class
version just returns [].  The list returned may or may not include the current
clustername.  This is needed so e.g. if user johndoe is in the mailing list
because has accounts on both DTs, removing him from only one cluster should
NOT remove from mailing list.

=item B<should_users_get_db_access>

Booleans.  Returns true if users should get access to a read-only view
in the Projects DB.

=back

The following methods are defined:

=over 4

=item B<mailinglist_password>

This reads the password in the file specified by B<mailinglist_pwdfile>
and returns to caller.  On errors, a warning will be displayed and undef
will be returned.

=back
