#!/usr/local/bin/perl
#
#Base class for Glue::HPCC::ClusterUtil::XXX
#where XXX is a scheduler type
#
#Pretty much a dummy class.
#
#HPC scripts will get a child subclass Glue::HPCC::Cluster
#for the specific cluster.
#The get_hpcUtil method will return a child class of 
#Glue::HPCC::ClusterUtil (this class).
#That child class should be for the specific scheduler
#used by that cluster.
#
#Actually, this class (and children) only provide 2 main class
#methods:
#
#       user => returns Glue::HPCC::ClusterUtil::Slurm::User
#       project => returns Glue::HPCC::ClusterUtil::Slurm::Project



package Glue::HPCC::ClusterUtil;
use strict;
use warnings;
use Carp qw(croak);

use Glue::HPCC::ClusterUtil::Project;
use Glue::HPCC::ClusterUtil::User;

#Globals
my $_sched_version;
my $_sched_rootdir;

sub new($@)
{	my $class = shift;
	my @args = @_;
	my $me = "${class}::new";

	my $obj = {};
	bless $obj, $class;

	#Parse arguments
	my %args = (@args);
	my @known_args = qw( 
		dryrun 
		verbose 
		scheduler_version 
		scheduler_rootdir
	);
	foreach my $arg (@known_args)
	{	my $val = delete $args{$arg};
		next unless defined $val;
		my $meth = $arg;
		$obj->$meth($val);
	}
	if ( scalar(@args) )
	{	my $tmp = join ", ", @args;
		croak "$me: Unexpected arguments [ $tmp ], aborting at ";
	}

	#Return instance
	return $obj;
}

sub user($)
{	my $class = shift;
	$class = ref($class) if ref($class);
	my $uclass =  $class . '::User';
	my $sch_rootdir = $class->scheduler_rootdir;
	my $sch_version = $class->scheduler_version;
	$uclass->scheduler_rootdir($sch_rootdir) if $sch_rootdir;
	$uclass->scheduler_version($sch_version) if $sch_version;
	return $uclass;
}

sub project($)
{	my $class = shift;
	$class = ref($class) if ref($class);
	my $pclass = $class . '::Project';
	my $sch_rootdir = $class->scheduler_rootdir;
	my $sch_version = $class->scheduler_version;
	$pclass->scheduler_rootdir($sch_rootdir) if $sch_rootdir;
	$pclass->scheduler_version($sch_version) if $sch_version;
	return $pclass;
}

sub verbose($;$)
{	my $class = shift;
	my $new = shift;

	my $user = $class->user;
	my $proj = $class->project;
	if ( defined $new )
	{	$user->verbose($new);
		$proj->verbose($new);
	}
	return $user->verbose || $proj->verbose;
}

sub dryrun($;$)
{	my $class = shift;
	my $new = shift;

	my $user = $class->user;
	my $proj = $class->project;
	if ( defined $new )
	{	$user->dryrun($new);
		$proj->dryrun($new);
	}
	return $user->dryrun || $proj->dryrun;
}

sub scheduler_version($;$)
{	my $class = shift;
	my $new = shift;
	if ( defined $new )
	{	$_sched_version = $new;
	}
	return $_sched_version;
}

sub scheduler_rootdir($;$)
{	my $class = shift;
	my $new = shift;
	if ( defined $new )
	{	$_sched_rootdir = $new;
	}
	return $_sched_rootdir;
}

	
1;
__END__

=head1 NAME

Glue::HPCC::ClusterUtil - Base class for Glue HPC cluster utility objects

=head1 SYNOPSIS

	...

=head1 DESCRIPTION

This is a pretty basic package, just to enable us to get
some scheduler independence in scripts.

HPC scripts will get a child subclass of B<Glue::HPCC::Cluster>
for the specific cluster being run against.  Usually from the
host they are running on.

B<Glue::HPCC::Cluster> has a B<get_hpcUtil> method that returns
the name of a class for HPC cluster utilities for a specific
scheduler.  That class, in turn, has two main class methods:

=over 4

=item B<project>: 

Returns the name of a child of the B<Glue::HPCC::ClusterUtil::Project> class.
This class handles creation, deletion, replenishment, etc. of projects and
accounts.

=item B<user>: 

Returns the name of a child of the B<Glue::HPCC::ClusterUtil::User> class
This class handles adding and removing users from projects, login node access,
etc.

=back

The above two classes contain methods to do all the HPC management.  The
B<Glue:HPCC::ClusterUtil::Project> and B<User> classes define the scheduler
independent stuff, and the scheduler specific subclasses handle the 
scheduler dependent stuff.

There are also a handful of accessors/mutators for classwide data members,
namely:

=over 4

=item B<verbose>: Boolean, gets/sets the verbosity level

=item B<dryrun>: Boolean, gets/sets the dryrun flag

=item B<scheduler_version>: gets/sets to version of selected scheduler.

=item B<scheduler_rootdir>: gets/sets root directory of scheduler install.

=back

The above are mainly for passing information to child classes, so e.g.
they inherit verbose/dryrun settings, or know which version of sacctmgr to
use.

This particular package just defines the B<user> and B<project> methods
and the above accessors.

