#!/usr/local/bin/perl
#
#Base class for Glue::HPCC::ClusterUtil::* classes
#
#All methods are class methods, 
#OO only to keep namespace clean and for inheritance
#
#This just defines barebones constructors, etc.

package Glue::HPCC::ClusterUtil::Base;
use strict;
use warnings;
use Carp qw(carp croak);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $VERBOSE;
my $DRYRUN;
my $DEBUG;
$DEBUG=1;

my $_sched_version;
my $_sched_rootdir;

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	if ( scalar(%args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;
	return;
}

sub _init($)
{	my $obj = shift;
}

#-------------------------------------------------------------------
#	Common Methods: user and project
#	These pass the buck to User and Project classes, which
#	handle user and project management tasks, respectively
#-------------------------------------------------------------------

sub user($)
#Should return Glue::HPCC::ClusterUtil::XXX::User where XXX is our
#scheduler type
{	my $class = shift;
	die "Class $class did not overload 'user' method at ";
}

sub project($)
#Should return Glue::HPCC::ClusterUtil::XXX::Project where XXX is our
#scheduler type
{	my $class = shift;
	die "Class $class did not overload 'project' method at ";
}


#-------------------------------------------------------------------
#	Common Methods: verbosity and dryrun mode
#-------------------------------------------------------------------

sub verbose($;$)
#Sets verbosity on our subordinate (user, project) modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		$VERBOSE=$val;
		$class->user->verbose($val);
		$class->project->verbose($val);
	}
	return $VERBOSE;
}

sub dryrun($$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		$DRYRUN=$val;
		$class->user->dryrun($val);
		$class->project->dryrun($val);
	}
	return $DRYRUN;
}

#-------------------------------------------------------------------
#	Common Methods: scheduler version and root dir
#-------------------------------------------------------------------

sub scheduler_version($;$)
{	my $class = shift;
	my $new = shift;
	if ( defined $new )
	{	$_sched_version = $new;
	}
	return $_sched_version;
}

sub scheduler_rootdir($;$)
{	my $class = shift;
	my $new = shift;
	if ( defined $new )
	{	$_sched_rootdir = $new;
	}
	return $_sched_rootdir;
}

#-------------------------------------------------------------------
#       Common boring utilities
#-------------------------------------------------------------------

sub _validate_hpc($$$)
#Validates hpc argument for given method
{       my $class = shift;
        my $me = shift;
        my $hpc = shift;

        croak "$me: Missing required parameter hpc at "
                unless $hpc;
        croak "$me: Invalid value '$hpc' for hpc at "
                unless $hpc && ref($hpc);
}

sub _no_extraneous_args($$@)
{	my $class = shift;
	my $me = shift;
	my %args = @_;

	if ( scalar(%args))
	{	my @tmp = keys %args;
		my $tmp = join ', ', @tmp;
		croak "$me: Unrecognized arguments [ $tmp ] at ";
	}
}

	
1;
__END__

=head1 NAME

Glue::HPCC::ClusterUtil::Base - Base for Glue::HPCC::ClusterUtil::* classes

=head1 SYNOPSIS

  ...


=head1 DESCRIPTION

This is the base class for Glue::HPCC::ClusterUtil::XXX classes
It provides some common routines and default values.








