#!/usr/local/bin/perl
#
#Does the heavy lifting for Cluster management
#project creation, user management, etc.
#For Dummy cluster type: i.e., just PDB tracking, no actual lifting

package Glue::HPCC::ClusterUtil::Dummy::Project;
use strict;
use warnings;
use base qw(Glue::HPCC::ClusterUtil::Project);
use Carp qw(carp croak);
use DBI;
use POSIX qw(floor);

use Glue::HPCC::Cluster;

use Glue::HPCC::Project::Log;
use Glue::HPCC::Project::Project;
use Glue::HPCC::Project::ProjectType;


#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $DEBUG;
$DEBUG=1;
my $VERBOSE;
my $DRYRUN;

#-------------------------------------------------------------------
#	Common Methods: verbosity and dryrun mode
#-------------------------------------------------------------------

sub verbose($;$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

        if ( defined $val )
        {       $val = $val?1:0;
                $VERBOSE=$val;
        }
        return $VERBOSE;
}

sub dryrun($$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

        if ( defined $val )
        {       $val = $val?1:0;
                $DRYRUN=$val;
        }
        return $DRYRUN;
}

sub sched_list_accounts($@)
#Wrapper for sacctmgr_list_accounts
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_list_accounts';

	return;
}
	
		

sub sched_create_project($@)
#Creates named project in given cluster, creating the normal and/or
#high priority accounts needed for this type.
#Takes key=>value pairs as arguments, recognizing
#	hpc: Glue::HPCC:Cluster instance, REQUIRED
#	project: name of project, REQUIRED
#	type: type of project in this cluster, REQUIRED
#	su: Value of allocation in this cluster, in SU, Required for 
#		metered types
#	reason: for logging, REQUIRED
#	FIXME: operate in FIXME mode
#
#Returns false on success.
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_create_project';

	return;
}

sub sched_delete_project($@)
#Deletes named project in given cluster (on slurm side)
#Takes key=>value pairs as arguments, recognizing
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	project: name of project to delete, REQUIRED
#	type : name of project type, REQUIRED
#	reason: for logging, REQUIRED
#
#This only does the Dummy sacctmgr part.  Normally expect it to be
#called from Glue::HPCC::Cluster::delete_project
#
#Returns false on success.
#On error, either croaks or returns true (error message).
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_delete_project';

	return;
}

sub sched_update_project($@)
#Updates project (scheduler dependent pieces)
#Takes key => value pairs, recognizing:
#	hpc : Glue::HPCC::Cluster object, REQUIRED
#	project: name of project, REQUIRED,
#	ksu => new ksu
#	active => new active
#	type => new type
#	reason => for logging, REQUIRED
#
#Caller is expected to log.
#Returns false on success, croaks or returns error message on error
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_update_project';

	return;
}


#-------------------------------------------------------------------
#	Autoreplenishment methods
#
# All autoreplenishment methods take as arguments key=>value pairs, as follows:
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	project => name of project, REQUIRED
#	reason => for logging.  REQUIRED
# The 'mode' is specific to each routine.  The 'prorated' mode may also
# accept an argument 'tstamp'; if omitted defaults to current time.
#
# All autoreplenishment methods return as follows:
# 	( $errflag, $msgs, $usage, $usage2 ) 
# On success, 
#	$errflag is false
#	$msgs is a listref of text strings describing what happened.
#		Suitable for emailing to project owner.
#	$usage and $usage2 are account usages for normal and hi-pri
#	    accounts, resp. assoc with project just before replenishment
#	    For use in reports to PoC.
#	    If only one account, $usage2 will be undef.  Either may be undef
#	    These are hash refs with keys
#		account: name of account
#		cluster: name of cluster
#		limit_ksu: the limit on the account, kSU
#		used_ksu: the total usage on the account, kSU
#		avail_ksu: the amount of time available, kSU
#		byuser : Hash ref with per user stats.  
#			Keyed on username
#			Value is usage by this user, in kSU.
# On error, $errflag is true (error message).  $msgs should contain what
# was done up to error, as well as error message.  Rest is likely undef.
#
#-------------------------------------------------------------------

#---------------	coop projects -------------------------

sub sched_quarterly_replenish_coop_proj($@)
#Does quarterly autoreplenishment for 'coop' type projects
#See above for inputs/outputs.
#
#This code _tries_ not to abort on run-time like errors.
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_quarterly_replenish_coop_proj';

	return (0, []);
}

sub sched_monthly_replenish_coop_proj($@)
#Does monthly autoreplenishment for projects of type 'coop', 
#does the sacctmgr bits
#Takes key => value pairs
#	hpc => Glue::HPCC:Cluster instance, REQUIRED
#	project => project name, REQUIRED
#	reason => for logging, REQUIRED
#
#Does monthly autoreplenishment of named project in cluster
#	1) Get usage and limits for normal and high priority accounts, and
#		computes balances, and monthly allotment (1/3 of after tax 
#		quarterly amount)
#	2) If normal balance <= 0, zero normal usage and limits.  Zero hipri
#		usage and set limit to current balance.
#	3) If hipri balance => monthly allotment, Zero hipri usage and set
#		limit to monthly allotment if not already there
#	4) Compute the amount of funds $xfer need to xfer to bring hipri 
#		account balance back to monthly allotment.
#	5) Reduce $xfer to available funds in normal balance if available
#		funds less than what was computed in 4
#	6) Zero usage on normal account, set limit to previous balance less
#		the amount transferred $xfer.
#	7) Zero usage on hipri account, set limit to prev balance + $xfer
#
#Returns ( $errflag, $msgs, $usage, $usage2 ) where $errflag is false on 
#success, true (error #message) on errors, and $msgs is a listref of text 
#strings describing what #was done (suitable for emailing to project owner).
# $usage and $usage2 are the account usages (from sshare_get_account_usage)
#for the normal and high-priority accounts associated with this project,
#taken just before the account replenished, for use in reports to the PoC
#
#This code _tries_ not to abort on run-time like errors.
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_monthly_replenish_coop_proj';

	return (0, []);
}

sub sched_prorated_replenish_coop_proj($@)
#Does prorated autoreplenishment for projects of type 'coop', 
#does the sacctmgr bits
#Takes key => value pairs
#	hpc => Glue::HPCC:Cluster instance, REQUIRED
#	project => project name, REQUIRED
#	tstamp => time to prorate to, defaults to now
#	reason => for logging, REQUIRED
#
#Does prorated autoreplenishment of named project in cluster
#Returns ( $errflag, $msgs, $usage, $usage2 ) where $errflag is false on 
#success, true (error #message) on errors, and $msgs is a listref of text 
#strings describing what #was done (suitable for emailing to project owner).
# $usage and $usage2 are the account usages (from sshare_get_account_usage)
#for the normal and high-priority accounts associated with this project,
#taken just before the account replenished, for use in reports to the PoC
#
#This code _tries_ not to abort on run-time like errors.
{	my $class = shift;
	my $me = 'sched_prorated_replenish_coop_proj';
	return (0, []);
}

sub sched_replenish_condo_proj($@)
#Autoreplenishes projects of type 'condo', does the sacctmgr bits
#Takes key => value pairs
#	hpc => Glue::HPCC:Cluster instance, REQUIRED
#	cluster => cluster name, defaults from $hpc->db_clustername
#	project => project name, REQUIRED
#	mode => mode, one of quarterly, monthly, prorate, REQUIRED
#	tstamp => seconds from epoch for use with prorate mode.  Defaults to
#		now.
#	reason => for logging, REQUIRED
#mode and tstamp are for consistency with other project types.  
#Currently, all that
#the autoreplenish of condo projects does is zero the usage, and return usage
#stats.  This is the same for monthly, quarterly, and prorated modes.
#
#Returns ( $errflag, $msgs, $usage, $usage2 ) where $errflag is false on 
#success, true (error #message) on errors, and $msgs is a listref of text 
#strings describing what #was done (suitable for emailing to project owner).
# $usage is the account usages (from sshare_get_account_usage) for
#the account associated with this project.
#taken just before the account replenished, for use in reports to the PoC
# $usage2 is undef; only there for consistency with other project types.
#
#This code _tries_ not to abort on run-time like errors.
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_replenish_condo_proj';

	return (0, []);
}

	
1;
__END__

=head1 NAME

Glue::HPCC::ClusterUtil::Dummy::Project - Project management under Dummy

=head1 SYNOPSIS

  ...


=head1 DESCRIPTION

Null interfaces for everything.

We only want to track in PDB.
