#!/usr/local/bin/perl
#
#Base class for Glue::HPCC::ClusterUtil::UserXXX
#where XXX is a scheduler type
#
#Does the heavy lifting for Cluster management,
#Project management related stuff
#
#This handles the scheduler-independent stuff.
#
#All methods are class methods, 
#OO only to keep namespace clean and for inheritance

package Glue::HPCC::ClusterUtil::Project;
use strict;
use warnings;
use base qw(Glue::HPCC::ClusterUtil::Base);
use Carp qw(carp croak);
use POSIX qw(strftime);

use Glue::HPCC::Project::Department;
use Glue::HPCC::Project::Log;
use Glue::HPCC::Project::Person;
use Glue::HPCC::Project::Project;
use Glue::HPCC::Project::ProjectType;

use Glue::HPCC::Utils::UnixGroup qw(
	unix_group_exists 
	delete_unix_group
);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $DEBUG;
$DEBUG=1;
my $VERBOSE;
my $DRYRUN;

my $GHPDept = 'Glue::HPCC::Project::Department';
my $GHPLog = 'Glue::HPCC::Project::Log';
my $GHPPers = 'Glue::HPCC::Project::Person';
my $GHPProj = 'Glue::HPCC::Project::Project';
my $GHPPType = 'Glue::HPCC::Project::ProjectType';

#-------------------------------------------------------------------
#	Common Methods: verbosity and dryrun mode
#-------------------------------------------------------------------

my @_subord_classes_for_dryrun = qw(
	Glue::HPCC::Project::Department
	Glue::HPCC::Project::Log
	Glue::HPCC::Project::Person
	Glue::HPCC::Project::Project
	Glue::HPCC::Project::ProjectType
);

sub verbose($;$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		my $sclass;
		foreach $sclass (@_subord_classes_for_dryrun)
		{	$sclass->verbose($val);
		}
		$VERBOSE=$val;
	}
	return $VERBOSE;
}

sub dryrun($$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		my $sclass;
		foreach $sclass (@_subord_classes_for_dryrun)
		{	$sclass->dryrun($val);
		}
		$DRYRUN=$val;
	}
	return $DRYRUN;
}

#-------------------------------------------------------------------
#       Method to get our "user" counterpart
#-------------------------------------------------------------------

sub user_class($)
{       my $class = shift;
        $class = ref($class) if ref($class);
        $class =~ s/Project$/User/;
        return $class;
}
 
#-------------------------------------------------------------------
#	Handling shared resources between clusters 
#  Eg: shared prefixes for unix groups
#-------------------------------------------------------------------

sub is_project_in_any_sister_clusters($@)
#Determines if specified project is is any of the listed HPC clusters.
#Used to determine if resources which might be shared between clusters
#should be deleted (e.g. unix groups, etc).
#
#Returns true (the name of a sister cluster belongs to) if belongs to
#a sister cluster, and false (undef) otherwise.
#
#Takes key => value pairs for explicit arguments, recognizing:
#	hpc => Glue::HPCC::Cluster instance for cluster we are add/deleting
#		user from, used to skip matching cluster in $clusters.  Optional.
#	clusters => A list ref of sister clusternames to check, REQUIRED
#	project => name of project, REQUIRED
#	noactive => Boolean.  By default, only active projects are considered.
#		If this is set, will consider inactive/disabled projects.
#	me => impersonate another routine
{	my $class = shift;
	my %args = @_;
	my $realme = __PACKAGE__ . '::is_project_in_any_sister_clusters';

	my $hpc = delete $args{hpc};
	my $clusters = delete $args{clusters};
	my $project = delete $args{project};
	my $noactive = delete $args{noactive};
	my $me = delete $args{me} || $realme;

	$class->_no_extraneous_args($me, %args);
	$class->_validate_hpc($me, $hpc) if $hpc;
	croak "$me: Missing required argument project at " unless $project;
	croak "$me: Missing required argument clusters at " unless $clusters;
	croak "$me: Invalid value '$clusters' for clusters, expecting list ref,  at " 
		unless ref($clusters) eq 'ARRAY';

	my $this_cluster;
	$this_cluster = $hpc->clustername if $hpc;
	#If no sister clusters, then we return false always
	return unless $clusters && ref($clusters) eq 'ARRAY'
		&& scalar(@$clusters);

	SISTERS: foreach my $sister (@$clusters)
	{	next SISTERS if $this_cluster && $sister eq $this_cluster;
		my $sishpc = Glue::HPCC::Cluster->new_from_clustername($sister);
		unless ( $sishpc && ref($sishpc) )
		{	croak "$me: Unable to instantiate cluster for $sister: $sishpc at ";
		}
		#Skip if this is our cluster
		next SISTERS if $this_cluster && $sishpc->clustername eq $this_cluster;

		#First check if project is even in PDB for this cluster
		my $proj = Glue::HPCC::Project::Project->project_by_name(
			$sishpc, $project);
		next SISTERS unless $proj; #Not in PDB, so not in this cluster
		croak "$me: Error looking up project $project in PDB for cluster " .
			"$sister: $proj at " unless ref($proj); 

		#Ignore inactive projects unless $noactive set
		next SISTERS unless $proj->active || $noactive;

		return $sister;
	}
	
	#We failed to find any sister cluster with named project
	return;
}

sub is_unixgrp_in_sister_cluster($@)
#Determine if any projects  exist in a sister cluster with which
#we share the same unixgrp prefix which have the specified unix group.
#Returns true (the matching sister clustername) if so, and false otherwise.
#Takes key => value pairs for explicit arguments, recognizing:
#	hpc => Glue::HPCC::Cluster instance for cluster we are add/deleting
#		user from, REQUIRED
#	unixgroup => name of unixgroup, REQUIRED
#	noactive => Boolean.  By default, only active projects are considered.
#		If this is set, will consider inactive projects as well.
{	my $class = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::is_unixgrp_sister_cluster';

	my $hpc = delete $args{hpc};
	my $unixgroup = delete $args{unixgroup};
	my $noactive = delete $args{noactive};

	$class->_no_extraneous_args($me, %args);
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required argument unixgroup at " unless $unixgroup;

	my $this_cluster;
	$this_cluster = $hpc->clustername if $hpc;
	my $sister_clusters = $hpc->clusters_sharing_unixgrp_prefix;
	#If no sister clusters, then we return false always
	return unless $sister_clusters && ref($sister_clusters) eq 'ARRAY'
		&& scalar(@$sister_clusters);

	SISTERS: foreach my $sister (@$sister_clusters)
	{	next SISTERS if $this_cluster && $sister eq $this_cluster;
		my $sishpc = Glue::HPCC::Cluster->new_from_clustername($sister);
		unless ( $sishpc && ref($sishpc) )
		{	croak "$me: Unable to instantiate cluster for $sister: $sishpc at ";
		}
		#Skip if this is our cluster
		next SISTERS if $this_cluster && $sishpc->clustername eq $this_cluster;

		my $projects = Glue::HPCC::Project::Project->pdb_query(
			$sishpc, unixgroup => $unixgroup);
		croak "$me: Error looking up projects with unixgroup=$unixgroup in " .
			"cluster $sister; $projects at "
			unless $projects && ref($projects) eq 'ARRAY';

		unless ( $noactive )
		{	#Remove inactive projects
			$projects = [ grep { $_->active } @$projects ];
		}
		return $sister if scalar(@$projects);
	}
	return;
}


#-------------------------------------------------------------------
#	Project creation/deletion
#-------------------------------------------------------------------

sub _add_project_to_pdb($@)
#Does the actual insert of specified project to project DB table
#Takes args:
#	hpc: Glue::Project::Cluster instance, REQUIRED
#	project: name of the project, REQUIRED
#	description: description of project.  Default is undef.
#	unixgroup: name of unix group for this project.  
#		Required for certain proejct types
#	parent: Name of parent project. Default is no parent.
#	department: Department code, REQUIRED
#	type: name of projtype, REQUIRED
#	su: value of allocation, Required for metered project types
#	active: boolean flag.  Defaults to TRUE if not defined.
#		Set to 0 to create inactive project.
#	expiration: expiration date for project.  Default is
#		now + default_expiry_days from projtype
#	thrhld_pct_over: Threshhold for warning if usage is over.  No default
#	thrhld_pct_under: Threshhold for warning if usage is under.  No default
#	me: method to report in error messages
#
#Caller is expected to log the transaction.
#Project must NOT already exist in DB
#Returns false on success, on error returns true (Error msg) or croaks
{	my $class = shift;
	my %args = @_;
	my $realme = '_add_project_to_pdb';

	#Validate parameters
	my $me = delete $args{me} || $realme;
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $desc = delete $args{description};
	my $unixgrp = delete $args{unixgroup};
	my $parentname = delete $args{parent};
	my $deptcode = delete $args{department};
	my $ptname = delete $args{type};
	my $su = delete $args{su};
	my $active = delete $args{active};
	my $exp = delete $args{expiration};
	my $thrhld_pct_over = delete $args{thrhld_pct_over};
	my $thrhld_pct_under = delete $args{thrhld_pct_under};

	$class->_validate_hpc($realme, $hpc);
	croak "$realme: Missing required parameter 'project' at " unless $projname;
	croak "$realme: Missing required parameter 'projtype' at " unless $ptname;
	croak "$realme: Missing required parameter 'dept' at " unless $deptcode;
	$class->_no_extraneous_args($realme, %args);

	my $error;
	my $dept = $GHPDept->dept_by_code($hpc, $deptcode);
	croak "$me: Unable to find dept $deptcode in PDB at "
		unless $dept;
	croak "$me: Error looking up dept $deptcode in PDB: $dept at "
		unless ref($dept);
	my $deptid = $dept->dept_pkey;

	my $parentid;
	if ( $parentname && $parentname ne 'NULL' )
	{	my $parent = $GHPProj->project_by_name($hpc, $parentname);
		croak "$me: Unable to find parent proj $parentname in PDB at "
			unless $parent;
		croak "$me: Error looking up parent proj $parentname " .
			"in PDB: $parent at " unless ref($parent);
		$parentid = $parent->prj_pkey;
	}

	my $ptype = $GHPPType->projecttype_by_name($hpc, $ptname);
	croak "$me: Invalid project type $ptname, not found in PDB at "
		unless $ptype;
	croak "$me: Error looking up project type $ptname: $ptype at "
		unless ref($ptype);
	my $ptypeid = $ptype->pkey;
		
	#Default missing values
	if ( $ptype->is_metered )
	{	croak "$me: Missing required parameter 'su' for project " .
			"type $ptname at " unless defined $su;
	}

	#$unixgrp = $hpc->unixgrp_prefix . $projname unless ( $unixgrp );
	if ( $ptype->requires_unix_group )
	{	croak "$me: Missing required parm 'unix_group', needed for " .
			" project type $ptname at " unless $unixgrp;
	}
	$active = 1 unless defined $active;
	$active = ($active)?1:0; #Canonicalize
	unless ( $exp )
	{	my $tstamp = time;
		$tstamp += 24*60*60*$ptype->default_expiry_days;
		$exp = strftime("%Y-%m-%d %H:%M:%S", localtime($tstamp) );
	}

	my $proj = $GHPProj->project_by_name($hpc, $projname);
	croak "$me: Error looking up project $projname in PDB: $proj at "
		if $proj && ! ref($proj);
	croak "$me: Project $projname already exists; prj_pkey = " . 
		$proj->prj_pkey . " at " if $proj;

	my @proj_fields = (	
		name => $projname,
		description => $desc,
		unixgroup => $unixgrp,
		parent => $parentid,
		department => $deptid,
		projtype => $ptypeid,
		su => $su,
		active => $active,
		expiration => $exp,
		thrhld_pct_over => $thrhld_pct_over,
		thrhld_pct_under => $thrhld_pct_under,
	);

	$error = $GHPProj->pdb_insert($hpc, @proj_fields);
	croak "$me: Error inserting project $projname: $error at "
		if $error;

	return;
}

sub _update_project_in_pdb($@)
#Does the actual update of specified project to project DB table
#Takes args:
#	hpc: Glue::Project::Cluster instance, REQUIRED
#	project: name of the project, REQUIRED
#	desc: description of project.  
#	unixgroup: name of unix group for this project.
#	parent: Name of parent project. Give the string
#		NULL to unset existing project
#	dept: Department code
#	projtype: name of projtype
#	su: value of allocation
#	active: boolean flag.  Set to 0 to change to inactive
#	expiration: expiration date for project.
#	thrhld_pct_over: Threshhold for warning if usage is over.
#	thrhld_pct_under: Threshhold for warning if usage is under
#	me: method to report in error messages
#
#Project and hpc are REQUIRED.  Any other arguments will cause the
#corresponding field in the specified project to be changed if needed.
#
#Caller is expected to log the transaction.
#Project must already exist in DB
#Returns false on success, on error returns true (Error msg) or croaks
{	my $class = shift;
	my %args = @_;
	my $realme = '_update_project_in_pdb';

	#Validate parameters
	my $me = delete $args{me} || $realme;
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $desc = delete $args{desc};
	my $unixgrp = delete $args{unixgroup};
	my $parentname = delete $args{parent};
	my $deptcode = delete $args{dept};
	my $ptname = delete $args{projtype};
	my $su = delete $args{su};
	my $active = delete $args{active};
	my $exp = delete $args{expiration};
	my $thrhld_pct_over = delete $args{thrhld_pct_over};
	my $thrhld_pct_under = delete $args{thrhld_pct_under};

	$class->_validate_hpc($realme, $hpc);
	croak "$realme: Missing required parameter 'project' at " unless $projname;

	my $error;
	my $deptid;
	if ( $deptcode )
	{	my $dept = $GHPDept->department_by_code($hpc, $deptcode);
		croak "$me: Unable to find dept $deptcode in PDB at "
			unless $dept;
		croak "$me: Error looking up dept $deptcode in PDB: $dept at "
			unless ref($dept);
		$deptid = $dept->dept_pkey;
	}

	my $parentid;
	if ( $parentname && $parentname ne 'NULL' )
	{	my $parent = $GHPProj->project_by_name($hpc, $parentname);
		croak "$me: Unable to find parent proj $parentname in PDB at "
			unless $parent;
		croak "$me: Error looking up parent proj $parentname " .
			"in PDB: $parent at " unless ref($parent);
		$parentid = $parent->prj_pkey;
	}

	my $ptypeid;
	if ( $ptname )
	{	my $ptype = $GHPPType->projecttype_by_name($hpc, $ptname);
		croak "$me: Invalid project type $ptname, not found in PDB at "
			unless $ptype;
		croak "$me: Error looking up project type $ptname: $ptype at "
			unless ref($ptype);
		$ptypeid = $ptype->pkey;
	}
		
	if ( defined $active )
	{	$active = $active?1:0; #Canonicalize
	}

	my $proj = $GHPProj->project_by_name($hpc, $projname);
	croak "$me: Project $projname not found in PDB at " unless $proj;
	croak "$me: Error looking up project $projname in PDB: $proj at "
		unless ref($proj);

	my %set_fields = ( );
	$set_fields{description} = $desc 
		if $desc && $desc ne $proj->description;
	$set_fields{unixgroup} = $unixgrp 
		if $unixgrp && $unixgrp ne $proj->unixgroup;
	$set_fields{parent} = $parentid
		if defined $parentid && $parentid ne $proj->parent;
	$set_fields{department} = $deptid 
		if defined $deptid && $deptid ne $proj->department;
	$set_fields{projtype} = $ptypeid
		if defined $ptypeid && $ptypeid ne $proj->projtype;
	$set_fields{su} = $su if defined $su && $su ne $proj->su;
	$set_fields{expiration} = $exp 
		if $exp && $exp ne $proj->expiration;
	$set_fields{thrhld_pct_over} = $thrhld_pct_over 
		if $thrhld_pct_over && $thrhld_pct_over ne $proj->thrhld_pct_over;
	$set_fields{thrhld_pct_under} = $thrhld_pct_under 
		if $thrhld_pct_under && $thrhld_pct_under ne $proj->thrhld_pct_under;
	my $tmpactive = defined $active;
	$tmpactive = 0 if $active && $proj->active; #Both true
	$tmpactive = 0 unless $active || $proj->active; #Both false
	$set_fields{active} =  $active if $tmpactive;

	$error = $proj->pdb_update_me($hpc, %set_fields);
	croak "$me: Error updating project $projname: $error at "
		if $error;

	return;
}

sub add_project_to_pdb($@)
#Creates the named project in PDB project table
#Takes a list of key => value pairs
#Recognized keys are:
#	hpc: Glue::Project::Cluster instance, REQUIRED
#	project: name of the project, REQUIRED
#	desc: description of project.  
#	unixgroup: name of unix group for this project.
#	parent: Name of parent project. Give the string
#		NULL to unset existing project
#	dept: Department code.  REQUIRED if new
#	projtype: name of projtype.  REQUIRED if new
#	su: value of allocation.  Required for new, metered project types
#	active: boolean flag.  Set to 0 to change to inactive
#	expiration: expiration date for project.
#	thrhld_pct_over: Threshhold for warning if usage is over.
#	thrhld_pct_under: Threshhold for warning if usage is under
#	reason: for logging.  REQUIRED
#	FIXME: boolean, if set allow to operate in "update" mode
#
#Returns false on success, 
#croaks or returns non-ref true (error message) on error
{	my $class = shift;
	my %args = @_;
	my $me = 'add_project_to_pdb';

	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $reason = delete $args{reason};
	my $FIXME = delete $args{FIXME};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'reason' at " unless $reason;

	my $proj = $GHPProj->project_by_name($hpc, $projname);
	my $error;
	
	if ( $proj )
	{	if ( ref($proj) )
		{	#Project already defined.  
			croak "$me: Project $projname already exists at "
				unless $FIXME;
			$error = $class->update_project_in_pdb(
				hpc => $hpc,
				project => $projname,
				reason => $reason,
				%args);
			croak "$me: Error updating project $projname: $error at "
				if $error;
		} else
		{	croak "$me: Error trying to lookup project $projname: $proj";
		}
	} else
	{	#Project needs to be created
		$error = $class->_add_project_to_pdb(
			hpc => $hpc,
			project => $projname,
			%args,
		);
		croak "$me: Error inserting project $projname: $error at "
			if $error;

		my $deptcode = $args{dept};
		$GHPLog->add_log_entry($hpc,
			cluster => $hpc->db_clustername,
			action_code => 'insert-project',
			project => $projname,
			tablename => 'project',
			reason => $reason,
			description => "insert project into PDB",
		) ;

		return;
	}
}

sub update_project_in_pdb($@)
#Updates the named project in PDB project table
#Takes a list of key => value pairs
#Recognized keys are:
#	hpc: Glue::Project::Cluster instance, REQUIRED
#	project: name of the project, REQUIRED
#	desc: description of project.  
#	unixgroup: name of unix group for this project.
#	parent: Name of parent project. Give the string
#		NULL to unset existing project
#	dept: Department code. 
#	projtype: name of projtype. 
#	su: value of allocation.
#	active: boolean flag.  Set to 0 to change to inactive
#	expiration: expiration date for project.
#	thrhld_pct_over: Threshhold for warning if usage is over.
#	thrhld_pct_under: Threshhold for warning if usage is under
#	reason: for logging.  REQUIRED
#
#Any fields given will be updated if differ from current.
#Any fields omitted will be left the same.
#NOTE: 
#	active: must set to 0 (not undef) to set to false.
#	parent: must set to the string "NULL" to unset previous.
#If either set to undef, will leave previous value unchanged.
#
#Returns false on success, 
#croaks or returns non-ref true (error message) on error
{	my $class = shift;
	my %args = @_;
	my $me = 'update_project_in_pdb';

	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $reason = delete $args{reason};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'reason' at " unless $reason;

	my $proj = $GHPProj->project_by_name($hpc, $projname);
	croak "$me: No project $projname found in PDB at " unless $proj;
	croak "$me: Error looking up proj $projname in PDB: $proj at" 
		unless ref($proj);
	
	my $error = $class->_update_project_in_pdb(
		hpc => $hpc,
		project => $projname,
		me => $me,
		%args,
	);
	croak "$me: Error updating project $projname: $error at "
		if $error;

	my @keys = keys %args;
	@keys = grep { defined $args{$_} } @keys;
	my @updates =  map { $_ => $args{$_} } (@keys);
	my $logstr = join ", ", @updates;

	$GHPLog->add_log_entry($hpc,
		action_code => 'update-project',
		cluster => $hpc->db_clustername,
		project => $projname,
		tablename => 'project',
		reason => $reason,
		description => "set [$logstr] on $projname",
	);

	return;
}

sub pdb_add_or_update_project($@)
#Creates or updates named project in PDB project table
#Takes a list of key => value pairs
#Recognized keys are:
#	hpc: Glue::Project::Cluster instance, REQUIRED
#	project: name of the project, REQUIRED
#	desc: description of project.  
#	unixgroup: name of unix group for this project.
#	parent: Name of parent project. Give the string
#		NULL to unset existing project
#	dept: Department code.  REQUIRED if new
#	projtype: name of projtype.  REQUIRED if new
#	su: value of allocation.  REQUIRED if new
#	active: boolean flag.  Set to 0 to change to inactive
#	expiration: expiration date for project.
#	thrhld_pct_over: Threshhold for warning if usage is over.
#	thrhld_pct_under: Threshhold for warning if usage is under
#	reason: for logging.  REQUIRED
#
#Returns false on success, 
#croaks or returns non-ref true (error message) on error
{	my $class = shift;
	my %args = @_;
	my $me = 'pdb_add_or_update_project';

	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $reason = delete $args{reason};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'reason' at " unless $reason;

	my $proj = $GHPProj->project_by_name($hpc, $projname);
	my $deptcode = $args{dept};
	my $error;
	
	if ( $proj )
	{	if ( ref($proj) )
		{	#Project already defined.  
			#Update fields as needed
			$error = $class->_update_project_in_pdb(
				hpc => $hpc,
				project => $projname,
				me => $me,
				%args,
			);
			croak "$me: Error updating project $projname: $error at "
				if $error;

			my @updates =  map { $_ => $args{$_} } (keys %args);
			my $logstr = join ", ", @updates;

			$GHPLog->add_log_entry($hpc,
				action_code => 'update-project',
				cluster => $hpc->db_clustername,
				project => $projname,
				tablename => 'project',
				department => $deptcode,
				reason => $reason,
				description => "set [$logstr] on $projname",
			);
			return;
		} else
		{	croak "$me: Error trying to lookup project $projname: $proj";
		}
	} else
	{	#Project needs to be created
		$error = $class->_add_project_to_pdb(
			hpc => $hpc,
			project => $projname,
			me => $me,
			%args,
		);
		croak "$me: Error inserting project $projname: $error at "
			if $error;

		$GHPLog->add_log_entry($hpc,
			cluster => $hpc->db_clustername,
			action_code => 'insert-project',
			project => $projname,
			tablename => 'project',
			department => $deptcode,
			reason => $reason,
			description => "insert project into PDB",
		);

		return;
	}
}

sub pdb_activate_project($@)
#Activates (set active flag to true) an existing project
#Takes key => value pairs, recognizing:
#	hpc:  Glue::HPCC::Cluster instance, REQUIRED
#	project: name of the project, REQUIRED
#	reason: reason to activate project, for logs, REQUIRED
#
#Returns false on success, 
#croaks or returns non-ref true (error message) on error
{	my $class = shift;
	my %args = @_;
	my $me = 'pdb_activate_project';

	#Validate arguments
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $reason = delete $args{reason};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	return $class->update_project_in_pdb(
		hpc => $hpc,
		project => $projname,
		active => 1,
		reason => $reason,
	);
}

sub pdb_deactivate_project($@)
#Deactivates (set active flag to false) an existing project
#Takes key => value pairs, recognizing:
#	hpc:  Glue::HPCC::Cluster instance, REQUIRED
#	project: name of the project, REQUIRED
#	reason: reason to activate project, for logs, REQUIRED
#
#Returns false on success, 
#croaks or returns non-ref true (error message) on error
{	my $class = shift;
	my %args = @_;
	my $me = 'pdb_deactivate_project';

	#Validate arguments
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $reason = delete $args{reason};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	return $class->update_project_in_pdb(
		hpc => $hpc,
		project => $projname,
		active => 0,
		reason => $reason,
	);
}

sub create_project($@)
#Creates named project for our cluster.
#Takes a list of key => value pairs
#Recognized keys are:
#	hpc:  Glue::HPCC::Cluster instance, REQUIRED
#	project: name of the project, REQUIRED
#	description: description of project, REQUIRED
#	unixgroup: name of unix group for this project, 
#		Required for certain project types
#	parent: name of parent project to this project.  Default
#		is no parent.
#	department: dept code for this project, REQUIRED.
#	type: type of project, REQUIRED.  See project_type table
#	ksu: value of this allocation in kSU.  Required for metered types
#	poc: username of poc for this project, REQUIRED.
#	expiration: expiration date for project, REQUIRED 
#	active: Boolean.  Defaults to true.  Use 0 to create inactive project.
#	thrhld_pct_over, thrhld_pct_under: Threshholds, in pct points,
#		for over/under usage of project warnings
#	reason: reason text for logs.  REQUIRED
#
#We also recognize:
#	FIXME: if set, do not abort if project partially exists, just try
#		to fix it.
#
#Unix group must already exist.  POC user must already exist in PDB persons 
#table.  kSU can be fractional.
#
#This code does the projects DB stuff, then calls sched_create_project with
#the same arguments to do the scheduler specific bits.
#
#Returns false on success.
#On error, either croaks or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'create_project';

	#Validate arguments
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $desc = delete $args{description};
	my $unixgrp = delete $args{unixgroup};
	my $parentname = delete $args{parent};
	my $deptcode = delete $args{department};
	my $ptname = delete $args{type};
	my $ksu = delete $args{ksu};
	my $poc = delete $args{poc};
	my $exp = delete $args{expiration};
	my $active = delete $args{active};
	my $thrhld_pct_over = delete $args{thrhld_pct_over};
	my $thrhld_pct_under = delete $args{thrhld_pct_under};
	my $reason = delete $args{reason};
	my $FIXME = delete $args{FIXME};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'description' at " unless $desc;
	croak "$me: Missing required parameter 'department' at " unless $deptcode;
	croak "$me: Missing required parameter 'type' at " unless $ptname;
	croak "$me: Missing required parameter 'poc' at " unless $poc;
	croak "$me: Missing required parameter 'expiration' at " unless $exp;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $dept = $GHPDept->dept_by_code($hpc, $deptcode);
	croak "$me: Unable to find dept $deptcode in PDB at " unless $dept;
	croak "$me: Error looking up dept $deptcode in PDB: $dept at "
			unless ref($dept);
	my $deptid = $dept->dept_pkey;

	my $parentid;
	if ( $parentname && $parentname ne 'NULL' )
	{	my $parent = $GHPProj->project_by_name($hpc, $parentname);
		croak "$me: Unable to find parent proj $parentname in PDB at "
			unless $parent;
		croak "$me: Error looking up parent proj $parentname " .
			"in PDB: $parent at " unless ref($parent);
		$parentid = $parent->prj_pkey;
	}

	my $ptype = $GHPPType->projecttype_by_name($hpc, $ptname);
	croak "$me: Invalid project type $ptname, not found in PDB at "
		unless $ptype;
	croak "$me: Error looking up project type $ptname: $ptype at "
		unless ref($ptype);
	my $ptypeid = $ptype->pkey;
		
	if ( $ptype->is_metered )
	{ 	croak "$me: Missing required parameter 'ksu' for project" .
			"type $ptname at " unless defined $ksu;
	}
	if ( $ptype->requires_unix_group )
	{ 	croak "$me: Missing required parameter 'unixgroup' for " .
			"project type $ptname at " unless $unixgrp;
	}

	#Default missing values
	my $cluster =  $hpc->db_clustername;
	#$unixgrp = $hpc->unixgrp_prefix . $projname unless ( $unixgrp );
	$active = 1 unless defined $active;
	$active = ($active)?1:0; #Canonicalize
	unless ( $exp )
	{	my $tstamp = time;
		$tstamp += 24*60*60*$ptype->default_expiry_days;
		$exp = strftime("%Y-%m-%d %H:%M:%S", localtime($tstamp) );
	}

	my $su = 1000 * $ksu if defined $ksu;
	if ( defined $su && $su < 0 )
	{	croak "$me: Invalid kSU=$ksu, must be numeric and positive, at ";
	}

	if ( $unixgrp && $hpc->scheduler_type ne 'dummy' )
	{	croak "$me: Invalid unix group $unixgrp (does not exist) at "
			unless unix_group_exists($unixgrp);
	}
	my $pocrec =  $GHPPers->person_by_username($hpc, $poc);
	croak "$me:  Invalid PoC $poc (does not exist in PDB) at "
		unless $pocrec || $class->dryrun;
	croak "$me: Error looking up PoC $poc: $pocrec at "
		if $pocrec && ! ref($pocrec);

        croak "$me: Must be runnning on cluster to create_project at "
                unless $hpc->running_in_cluster || $ptname eq 'ou';

	my $error;
	my $proj = $GHPProj->project_by_name($hpc, $projname);
	croak "$me: Error looking up project $projname: $proj at "
		if $proj && ! ref($proj);

	if ( $proj )
	{	#Project exists in PDB
		#Only OK if inactive OR if we are in FIXME mode
		if ( $proj->active )
		{	#Project is NOT inactive
			croak "$me: Error creating project $projname: already " .
				"exists/active and NOT FIXME mode, at "
				unless $FIXME;
		} else
		{	#Project is inactive, make sure really inactive
			#Check if any scheduler bits
			my $tmpAccts = [ $projname. "${projname}-hi" ];
			my $tmp = $class->sched_list_accounts(cluster=>$cluster,
				account => $tmpAccts);
			die "$me: Error checking for scheduler bits for account " .
				"$projname: $tmp at "
				unless ( $tmp && ref($tmp) eq 'ARRAY' );
			$tmpAccts = join ", ", @$tmp;
			die "$me: Error creating project $projname: exists, inactive " .
				"but has scheduler bits for [$tmpAccts] --- " .
				"something is wrong! at " if scalar(@$tmp);
		}
	}
			  
	if ( $proj )
	{	#croak "$me: Error creating project $projname; already exists at "
		#	unless $FIXME;
		warn "Updating project $projname in project table.\n" if $DEBUG;
		$error = $class->update_project_in_pdb(
			hpc => $hpc,
			project => $projname,
			description => $desc,
			unixgroup => $unixgrp,
			parent => $parentname,
			department => $deptcode,
			type => $ptname,
			su => $su,
			expiration => $exp,
			active => $active,
			thrhld_pct_over => $thrhld_pct_over,
			thrhld_pct_under => $thrhld_pct_under,
			reason => $reason,
		);
		croak "$me: Error updating project $projname in PDB: $error at "
			if $error;
	} else
	{	#Need to create project
		warn "Adding project $projname to project table.\n" if $DEBUG;
		$error = $class->add_project_to_pdb(
			hpc => $hpc,
			project => $projname,
			description => $desc,
			unixgroup => $unixgrp,
			parent => $parentname,
			department => $deptcode,
			type => $ptname,
			su => $su,
			expiration => $exp,
			active => $active,
			thrhld_pct_over => $thrhld_pct_over,
			thrhld_pct_under => $thrhld_pct_under,
			reason => $reason,
		);
		croak "$me: Error inserting project $projname into PDB: $error at "
			if $error;
	}

	#Add PoC to project
	warn "Adding poc $poc to project $projname.\n" if $DEBUG;
	my $usrclass = $class->user_class;
	$error = $usrclass->pdb_add_poc_to_project(
		hpc => $hpc,
		user => $poc, 
		project => $projname,
		reason => $reason);
	croak "$me: Error adding $poc as PoC of $projname: $error at " 
		if $error;


	if ( $ptname ne 'ou' )
	{	warn "Starting scheduler specific parts.\n" if $DEBUG;
		$class->sched_create_project(
			hpc => $hpc,
			project => $projname,
			type => $ptname,
			su => $su,
			reason => $reason,
			FIXME => $FIXME,
		);
		croak "$me: Error doing scheduler specific bits to " .
			"create $projname in $cluster: $error at " if $error;
	}

	$GHPLog->add_log_entry($hpc,
		action_code => 'create-project',
		project => $projname,
		cluster => $cluster, 
		reason => $reason,
		description => "created project $projname in $cluster with poc $poc",
	);

	return;
}
	
sub delete_project($@)
#Deletes named project for our cluster.
#Takes a list of key => value pairs
#Recognized keys are:
#	hpc:  Glue::HPCC::Cluster instance, REQUIRED
#	project: name of the project, REQUIRED
#	reason: reason text for logs.  REQUIRED
#	projtype: type of project.  Required if project
#		no longer exists in PDB.
#	DELETE_PDB: boolean, if set delete from PDB
#		Default is to only set active to false
#
#Deletes a project, both in PDB (default is only to deactivate
#in PDB) and in scheduler (via sched_delete_project).
#
#Will also delete the associated unix group, but under some
#restrictive conditions (only if the unix group is of the
#form ${PREFIX}-${PROJNAME} and no similar named project exists in sister clusters).
#
#Project must NOT have any children (active OR inactive) if to be deleted (DELETE_PDB)
#Project must NOT have any active children if to be disabled
#
#Project must have no members to be deactivated or deleted.
#Project must also have no PoCs or coordinators to be deleted (DELETE_PDB)
#
#Returns false on success.
#On error, either croaks or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'delete_project';

	#Validate arguments
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $reason = delete $args{reason};
	my $DELETE = delete $args{DELETE_PDB};
	my $ptname = delete $args{projtype};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	my $cluster = $hpc->db_clustername;

	my $proj = $GHPProj->project_by_name($hpc, $projname);
	croak "$me: Error looking up project $projname in PDB: $proj at "
		if $proj && ! ref($proj);
	my $ptype;
	my $users;
	my $usrclass = $class->user_class;
	if ( $proj )
	{	my $ptypeid = $proj->projtype;
		$ptype = $GHPPType->projecttype_by_pkey($hpc, $ptypeid);
		croak "$me: Error - unable to find ptype rec for proj $projname, " .
			"ptypeid=$ptypeid at " unless $ptype;
		croak "$me: Error looking up ptype $ptypeid: $ptype at "
			unless ref($ptype);
		croak "$me: Projtype mismatch for proj $projname: PDB shows " .
			"as being " . $ptype->name . " but called with $ptname at "
			if $ptname && $ptname ne $ptype->name;
		$ptname = $ptype->name;

		#Check for children
		my $children = Glue::HPCC::Project::Project->children_of_project_by_pkey(
			$hpc, $proj->prj_pkey);
		croak "$me: Error - unable to get children for proj $projname, " .
			"$children at " unless $children && ref($children) eq 'ARRAY';
		if ( $DELETE )
		{	#Must be no children, active or inactive
			if ( scalar(@$children) )
			{	my @cnames = map { $_->name . 
					"(" . ($_->active?'active':'inactive') . ")"
					 } @$children;
				my $cnames = join ", ", @cnames;
				croak "$me: Cannot delete project $projname, has children=" .
					"[ $cnames ], at ";
			}
		} else
		{	#Must be no active children
			my @children = grep { $_->active } @$children;
			if ( scalar(@children) )
			{	my @cnames = map { $_->name } @children;
				my $cnames = join ", ", @cnames;
				croak "$me: Cannot disable project $projname, has " .
					"active children [ $cnames ] at ";
			}
		}
				
		#Make sure no users for project
		$users = $usrclass->pdb_get_members_for_project(
			hpc => $hpc, project => $projname);
		if ( scalar(@$users) )
		{	$users = join ", ", @$users;
			croak "$me: Cannot delete project $projname, has members=" .
				"[ $users ], at ";
		}

		if ( $DELETE )
		{   $users = $usrclass->pdb_get_users_for_project(
			hpc => $hpc, project => $projname);
		    if ( scalar(@$users) )
		    {	$users = join ", ", @$users;
			croak "$me: Cannot DELETE project $projname, has users=" .
				"[ $users ], at ";
		    }
		}
		
	} else
	{	croak "$me: Error, proj $projname not in PDB " .
			"and projtype not given at " unless $ptname;
		$ptype = $GHPPType->projecttype_by_name($hpc, $ptname);
		croak "$me: Invalid projtype $ptname, not in PDB at " 
			unless $ptype;
		croak "$me: Error looking up ptype $ptname: $ptype at "
			unless ref($ptype);
	}

	#Check if exists in scheduler, and if any users
	my $acctstd = $projname;
	my $accthi = "${projname}-hi";
	my @acctnames = ( );
	my $error;
	push @acctnames, $acctstd if $ptype->has_normal_priority ||
			$ptype->has_special_partition;
	push @acctnames, $accthi if $ptype->has_high_priority;

	my $sched_accts = $class->sched_list_accounts(
		cluster => $cluster,
		account => [ @acctnames ],
	);

	if ( $sched_accts && ref($sched_accts) eq 'ARRAY' && @$sched_accts)
	{ 	my $sched_users = $usrclass->sched_list_users(
			cluster => $cluster,
			account => $sched_accts,
		);
		if ( $sched_users && ref($sched_users) eq 'ARRAY' && @$sched_users)
		{	$users = join ", ", @$sched_users;
			croak "$me: Cannot delete project $projname: has " . 
				"users = [ $users ] at ";
		}
	}
	
	#Save some values we might need later
	my $unixgrp = $proj->unixgroup;

	#Delete in PDB
	if ( $proj )
	{	$error = $class->pdb_deactivate_project(
			hpc => $hpc,
			project => $projname, 
			reason => $reason
		);
		#Try to continue
		carp "$me: Error deactivating project $projname: $error at "
			if $error;
			
		if ( $DELETE )
		{ 	my $pkey = $proj->prj_pkey;
			$error = $GHPProj->pdb_delete($hpc, prj_pkey => $pkey);
			carp "$me: Error DELETEing proj $projname: $error at "
				if $error;
			$GHPLog->add_log_entry($hpc,
				action_code => 'delete-project',
				project => $projname,
				cluster => $cluster, 
				reason => $reason,
				description => "deleted project $projname in $cluster from PDB",
			) unless $error;
		}
	}

	#Delete in scheduler
	$error = $class->sched_delete_project(
		hpc => $hpc,
		project => $projname,
		type => $ptname,
		reason => $reason,
	);

	#Delete the unix group
	my $ugrp_prefix = $hpc->unixgrp_prefix;
	my $nodelete_unixgrp;
	if ( $unixgrp eq  "${ugrp_prefix}${projname}" ) 
	{	#UnixGroup is of the standard form
		
		#Is it used elsewhere in our cluster?
		my $tmpprojs = Glue::HPCC::Project::Project->pdb_query(
			$hpc, unixgroup => $unixgrp);
		croak "$me: Error looking up projects with unixgroup=$unixgrp; $tmpprojs" 
			unless $tmpprojs && ref($tmpprojs) eq 'ARRAY';
		#Remove inactive projects
		$tmpprojs = [ grep { $_->active } @$tmpprojs ];
		
		if ( scalar(@$tmpprojs) )
		{	my @tmp = map { $_->name } @$tmpprojs;
			my $tmp = join ", ", @tmp;
			$nodelete_unixgrp = "in use for projects [$tmp]\n";
		} else
		{ 	my $sister = $class->is_unixgrp_in_sister_cluster(
				unixgroup => $unixgrp,
				noactive => 0,
				hpc => $hpc);
			if ( $sister )
			{	#The unix group was found in a sister cluster
				$nodelete_unixgrp = "in use in cluster $sister\n";
			}
		}
	} else
	{	$nodelete_unixgrp = "not of standard PREFIX-NAME form";
	}

	if ( $nodelete_unixgrp )
	{	warn "$me: Not deleting unix group $unixgrp: $nodelete_unixgrp\n";
	} else
	{	#Try to delete the unix group
		my $res = delete_unix_group(group => $unixgrp);
		if ( $res )
		{	die "$me: Error deleting unix group $unixgrp: $res at \n";
		} else
		{	$GHPLog->add_log_entry($hpc,
				action_code => 'delete-unix-group',
				cluster => $cluster,
				project => $projname,
				reason => $reason,
				description => "Deleting unix group $unixgrp",
				fieldname => $unixgrp,
			);
		}
	}
		
	return;
}
			

sub update_project($@)
#Creates named project for our cluster.
#Takes a list of key => value pairs
#Recognized keys are:
#	hpc:  Glue::HPCC::Cluster instance, REQUIRED
#	project: name of the project, REQUIRED
#	description: description of project
#	unixgroup: name of unix group for this project
#	parent: name of parent project to this project. 
#		Use string NULL to orphan project
#	department: dept code for this project
#	type: type of project
#	ksu: value of this allocation in kSU. 
#	expiration: expiration date for project
#	active: Boolean.  Defaults to true.  Use 0 to de-activate project.
#	thrhld_pct_over, thrhld_pct_under: Threshholds, in pct points,
#		for over/under usage of project warnings
#	reason: reason text for logs.  REQUIRED
#
#Updates a project, both in PDB and in scheduler.
#NOTE: only ksu, type, and active have any meaning outside of PDB.
#	Pass buck on what to do to sched_update_project
#	ksu: only affects PDB, because not sure what updating ksu should
#		mean in terms of existing balances
#	type: generates an error.  Changing type in sacctmgr is TRICKY.
#		Should fix this at some point.
#	active: deactivating should DELETE to sched accounts
#
#
#Returns false on success.
#On error, either croaks or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'update_project';


	#Validate arguments
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $desc = delete $args{description};
	my $unixgrp = delete $args{unixgroup};
	my $parentname = delete $args{parent};
	my $deptcode = delete $args{department};
	my $ptname = delete $args{type};
	my $ksu = delete $args{ksu};
	my $exp = delete $args{expiration};
	my $active = delete $args{active};
	my $thrhld_pct_over = delete $args{thrhld_pct_over};
	my $thrhld_pct_under = delete $args{thrhld_pct_under};
	my $reason = delete $args{reason};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	my $cluster = $hpc->db_clustername;

	my $proj = $GHPProj->project_by_name($hpc, $projname);
	croak "$me: Invalid project $projname, not in PDB at " unless $proj;
	croak "$me: Error looking up project $projname: $proj at "
		unless ref($proj);

	my ($parent, $parentid);
	if ( $parentname )
	{	$parent = $GHPProj->project_by_name($hpc, $parentname);
		croak "$me: Invalid project $parentname, not in PDB at " 
			unless $parent;
		croak "$me: Error looking up project $parentname: $parent at "
			unless ref($parent);
		$parentid = $parent->prj_pkey;
	}

	my ($dept, $deptid);
	if ( $deptcode )
	{	$dept = $GHPDept->dept_by_code($hpc, $deptcode);
		croak "$me: Invalid department $deptcode, not in PDB at "
			unless $dept;
		croak "$me: Error looking up dept $deptcode: $dept at "
			unless ref($dept);
		$deptid = $dept->dept_pkey;
	}

	if ( $unixgrp )
	{	croak "$me: Illegal unix group $unixgrp, not in hesiod at "
			unless unix_group_exists($unixgrp);
	}

	$active = $active?1:0 if defined $active;

	#Update scheduler dependent pieces first
	my @sched_changes = ();
	push @sched_changes, ksu => $ksu if defined $ksu;
	push @sched_changes, type => $ptname if $ptname;
	push @sched_changes, active => $active if defined $active;
	my $error;
	my (%temp, @temp, $temp);
	if ( @sched_changes)
	{	$error = $class->sched_update_project(
			hpc => $hpc,
			project => $projname,
			reason => $reason,
			@sched_changes,
		);
		croak "$me: Error updating project $projname in sched: $error at "
			if $error;
	}

	my @pdb_changes = ();
	push @pdb_changes, description => $desc 
		if $desc && $desc ne $proj->description;
	push @pdb_changes, unixgroup => $unixgrp 
		if $unixgrp && $unixgrp ne $proj->unixgroup;
	push @pdb_changes, parent => $parentid
		if $parentid && $parentid ne $proj->parent;
	push @pdb_changes, department => $deptid
		if $deptid && $deptid ne $proj->department;
	push @pdb_changes, ksu => $ksu
		if $ksu && $ksu != $proj->ksu;
	push @pdb_changes, projtype => $ptname
		if $ptname && $ptname ne $proj->projtype;
	push @pdb_changes, expiration => $exp
		if $exp && $exp ne $proj->expiration;
	push @pdb_changes, thrhld_pct_over => $thrhld_pct_over
		if $thrhld_pct_over 
		&& $thrhld_pct_over != $proj->thrhld_pct_over;
	push @pdb_changes, thrhld_pct_under => $thrhld_pct_under
		if $thrhld_pct_under && 
		$thrhld_pct_under != $proj->thrhld_pct_under;
	my $tmpactive = 0;
	$tmpactive = 1 if defined $active && $active && ! $proj->active;
	$tmpactive = 1 if defined $active && ! $active && $proj->active;
	push @pdb_changes, active => $active if $tmpactive;

	$error = $proj->update_me($hpc, @pdb_changes);
	croak "$me: Error updating project $projname in PDB: $error at "
		if $error;
	%temp = @pdb_changes;
	@temp = map  { "$_=$temp{$_}" } (keys %temp);
	$temp = join ", ", @temp;
	$GHPLog->add_log_entry($hpc,
		action_code => 'pdb-update-proj',
		cluster => $cluster,
		project => $projname,
		reason => $reason,
		description => "updating sched properties for proj ($temp)"
	);

	return;
}
	
#-------------------------------------------------------------------
#	Autoreplenishment methods
#-------------------------------------------------------------------

sub get_autoreplenish_projects($@)
#On success, returns a list ref of names of 
#projects for this cluster which
#are of types with autoreplenish flag set.
#
#On error, returns non-ref but true (error message), or
#croaks.
#
#Takes key=>value pairs, recognizing
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
{	my $class = shift;
	my %args = @_;
	my $me = 'get_autoreplenish_projects';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	$class->_no_extraneous_args($me, %args);

	#Get list of prject type names that autoreplenish
	my $ptypes = $GHPPType->autoreplenishing_project_types($hpc);
	croak "$me: Error getting autoreplenish ptypes: $ptypes at "
		unless $ptypes && ref($ptypes) eq 'ARRAY';
	my @ptypes = map { $_->pkey } @$ptypes;

	#Get list of Glue::HPCC::Project::Project's 
	my $projs = $GHPProj->pdb_query($hpc, active=>1);
	croak "$me: Error getting projects ($projs) at "
		unless $projs && ref($projs) eq 'ARRAY';

	#Find projects which match one of the autoreplenish ptypes
	my @projs = ();
	foreach my $ptype (@ptypes)
	{	my @tmp = grep { $_->projtype == $ptype } @$projs;
		push @projs, @tmp;
	}

	#Convert to a list of project names
	@projs = map { $_->name } @projs;

	return \@projs;
}

my @valid_mode_strings = qw(monthly quarterly prorated);
my %valid_mode_strings = map { $_ => undef } @valid_mode_strings;
my $valid_mode_strings = join ", ", @valid_mode_strings;

sub autoreplenish_project($@)
#Autoreplenishes the accounts for specified project
#Takes key => value pairs for arguments, recognizing:
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	project => name of project.  REQUIRED
#	mode => mode.  REQUIRED.  One of
#		monthly
#		quarterly
#		prorated
#	tstamp => Timestamp to use for mode=prorated.  Default is current
#		time.  Only used for prorated mode.
#	reason => reason for replenishment.  REQUIRED.
#
#Returns ($errflag, $msgs, $ptname, $usage, $usage2) where $errflag is false on 
#success, and true (error message) on error.  $msgs is list ref of 
#text strings describing what did, suitable for sending in email to PoC.  
#$ptname is the name of the project type for this project in this cluster.
#$usage and $usage2 are hash refs #describing usage stats before the 
#replenishment occured.  For projects with hi and std proirity accounts, 
#$usage will be the std priority, and $usage2 the high priority (but
#check the account name field!)
#The usage hashes should have keys:
#	account => name of the account	
#	cluster => name of cluster
#	limit_ksu => the limit on the account, in kSU
#	used_ksu  => the total usage on the account, in kSU
#	avail_ksu => the amount of time avail, in kSU
#	byuser => a hash ref containing usage stats by user, keyed on username
#		Values are the usage of this account by this user, in kSU
#Either or both of $usage, $usage2 may be undef.  NOTE: kSU values are NOT
#rounded
#
#Try not to abort if possible
{	my $class = shift;
	my %args = @_;
	my $me = 'autoreplenish_project';

	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $mode = delete $args{mode};
	my $tstamp = delete $args{tstamp};
	my $reason = delete $args{reason};

	#These are more 'compile' time errors, should die on
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parm 'project' at " unless $projname;
	croak "$me: Missing required parm 'mode' at " unless $mode;
	croak "$me: Missing required parm 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	$mode = lc $mode;
	croak "$me: Illegal value '$mode' for mode, expecting one of "
		. "[$valid_mode_strings] at " 
		unless exists $valid_mode_strings{$mode};
	if ( $tstamp && $mode ne 'prorated' )
	{	carp "$me: Ignoring tstamp on non-prorated replenish " .
			"of proj $projname at ";
		$tstamp = undef;
	}
	if ( $mode eq 'prorated' )
	{	$tstamp = time unless defined $tstamp;
	}
	
	#This is definitely runtime, but serious enough still should croak
	croak "$me: Replenishment needs to run from within the HPC cluster at "
		unless $hpc->running_in_cluster;
	
	#Get project record and its type
	my $err;
	my $proj = $GHPProj->project_by_name($hpc, $projname);
	unless ( $proj )
	{	$err = "$me:Invalid project $projname, not found";
		carp "$err at ";
		return ( $err, [ $err ] );
	}
	unless ( ref($proj) )
	{	$err = "$me: Error looking up project $projname: $proj ";
		carp "$err at ";
		return ( $err, [ $err ] );
	}
	my $ptypeid = $proj->projtype;
	my $ptype = $GHPPType->projecttype_by_pkey($hpc, $ptypeid);
	my $ptname = $ptype->name;

	my $replenishes = $ptype->auto_replenishes;
	unless ( $replenishes )
	{	$err = "$me: Ignoring attempt to replenish proj $projname" .
			"(type=$ptname)";
		carp "$err at ";
		return ( 0, [ $err ], $ptname ); #Not really an error?
	}

	#Pass the buck to the appropriate methods for different project types
	my $msgs;
	my $method = "autoreplenish_${ptname}_proj";
	if ( $class->can($method) )
	{	my @results = $class->$method(
				hpc => $hpc,
				project => $projname,
				mode => $mode,
				tstamp => $tstamp,
				reason => $reason,
			);
		$GHPLog->add_log_entry($hpc,
			cluster => $hpc->db_clustername,
			action_code => "autoreplenish-$mode",
			project => $projname,
			tablename => 'project',
			reason => $reason,
			description => "$mode autoreplenish (type $ptname)",
		) ;

		return @results;
	} else
	{	carp "$me: Unable to find method $method for class $class at ";
		$err = "$me: Don't know how to replenish project $projname" .
			"(type=$ptname)";
		carp "$err at ";
		return ( $err, [ $err ], $ptname );
	}
}

sub autoreplenish_coop_proj($@)
#Autoreplenishes a coop type project.
#Agruments and return values same as for autoreplenish_project
{	my $class = shift;
	my %args = @_;
	my $me = 'autoreplenish_coop_proj';

	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $mode = delete $args{mode};
	my $tstamp = delete $args{tstamp};
	my $reason = delete $args{reason};

	#These are more 'compile' time errors, should die on
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parm 'project' at " unless $projname;
	croak "$me: Missing required parm 'mode' at " unless $mode;
	croak "$me: Missing required parm 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	$mode = lc $mode;
	croak "$me: Illegal value '$mode' for mode, expecting one of "
		. "[$valid_mode_strings] at " 
		unless exists $valid_mode_strings{$mode};
	if ( $tstamp && $mode ne 'prorated' )
	{	carp "$me: Ignoring tstamp on non-prorated replenish " .
			"of proj $projname at ";
		$tstamp = undef;
	}
	
	#Pass to scheduler specific code for mode/ptname
	my $method = "sched_${mode}_replenish_coop_proj";
	my @args = ( hpc => $hpc, project => $projname, reason => $reason);
	push @args, ( tstamp => $tstamp ) if $mode eq 'prorated';
		
	return $class->$method(@args) if $class->can($method);

	carp "$me: Unable to find method $method for class $class at ";
	my $err = "$me: Don't know how to do $mode replenish on proj "
			. "$projname (type=coop)";
	carp "$err at ";
	return ( $err, [ $err ], 'coop' );
}
	
sub autoreplenish_condo_proj($@)
#Autoreplenishes a condo type project.
#Agruments and return values same as for autoreplenish_project
#
#Really nothing to do, mode does not matter, we just report
#usage and then zero usage.
#
{	my $class = shift;
	my %args = @_;
	my $me = 'autoreplenish_condo_proj';

	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $mode = delete $args{mode};
	my $tstamp = delete $args{tstamp};
	my $reason = delete $args{reason};

	#These are more 'compile' time errors, should die on
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parm 'project' at " unless $projname;
	croak "$me: Missing required parm 'mode' at " unless $mode;
	croak "$me: Missing required parm 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	$mode = lc $mode;
	croak "$me: Illegal value '$mode' for mode, expecting one of "
		. "[$valid_mode_strings] at " 
		unless exists $valid_mode_strings{$mode};
	if ( $tstamp && $mode ne 'prorated' )
	{	carp "$me: Ignoring tstamp on non-prorated replenish " .
			"of proj $projname at ";
		$tstamp = undef;
	}
	
	#Pass to scheduler specific code for mode/ptname
	#my $method = "sched_${mode}_replenish_condo_proj";
	#Ignore mode 
	my $method = "sched_replenish_condo_proj";
	my @args = ( 	hpc => $hpc, 
			project => $projname, 
			reason => $reason,
			mode => $mode,
		);
	push @args, ( tstamp => $tstamp ) if $mode eq 'prorated';
		
	return $class->$method(@args) if $class->can($method);

	carp "$me: Unable to find method $method for class $class at ";
	my $err = "$me: Don't know how to do $mode replenish on proj "
			. "$projname (type=condo)";
	carp "$err at ";
	return ( $err, [ $err ], 'condo' );
}

1;
__END__

=head1 NAME

Glue::HPCC::ClusterUtil::Project - Scheduler independent base class for Projects

=head1 SYNOPSIS

  package Glue::HPCC::ClusterUtil::SCHEDULER::Project;
  use base qw(Glue::HPCC::ClusterUtil::Project);

  ...


=head1 DESCRIPTION

This is the base class for Glue::HPCC::ClusterUtil::XXX::Project classes,
where XXX is a scheduler type.  This contains assorted utilities for
managing projects within a cluster, or at least the scheduler independent
portions.  Most code should not be using this directly, but a scheduler
specific child class.

Most of the routines called are in the parent class, which then do
the scheduler independent pieces, then fork off to scheduler specific
routines in the child class.

There are no data members or instances of this class; the class is basically
only for namespace separation.  All methods have OO calling conventions (to
further reduce namespace pollution and allow for inheritance).  Unless 
otherwise specified, methods take key => value pairs.

Included methods are:

=over 4

=item B<verbose(;FLAG)>:

Gets/sets the verbosity level for package.

=item B<dryrun(;FLAG)>:

Gets/sets the dryrun flag for package.

=item B<user_class()>:

Returns the name of the Glue::HPCC::ClusterUtil:XXX:User sister class for
this class.

=item B<add_project_to_pdb>:

Adds a new project to the projects DB.  Does not do any scheduler-specific
tasks.  All of the project args given below are normally required.  In 
addition, the boolean B<FIXME> can be given; normally it is an error if
while creating the project it sees anything already existing; if FIXME is
set such errors will be ignored, and the code will continue, fixing any
problems it finds.

=item B<update_project_in_pdb>:

Updates an existing project in the projects DB.  See project args below 
for arguments.  hpc and project name are required; otherwise you should
only give the fields you wish to change.

=item B<pdb_add_or_update_project>:

Will add the specified project if it does not already exist.  If it does
exist, will update it.  Expects all the args in project args section below.

=item B<pdb_activate_project>:

Activates (sets active flag to true) the named project.  Project must
already exist.  Requires B<hpc>, B<project> and B<reason> arguments.  This ONLY
does the PDB side, not the scheduler dependent bits.

=item B<pdb_deactivate_project>:

Deactivates (sets active flag to false) the named project.  Project must
already exist.  Requires B<hpc>, B<project> and B<reason> arguments.  This ONLY
does the PDB side, not the scheduler dependent bits.

=item B<create_project>:

Creates the specified project.  Requires all of the arguments in the 
project args section below.  It does the scheduler independent bits, and
then calls B<sched_create_project> (defined in the scheduler specific child
class) to do the scheduler specific bits.

=item B<delete_project>:

Deletes the specified project.  Requires B<hpc>, B<project> and B<reason>
arguments.  You can also specify B<projtype>, which is only required if
the project does not still exist in the PDB.  The argument B<DELETE_PDB>
can also be given, and if true the project's rows will be deleted from the
PDB (normally it will just be set active to false).
It does the scheduler independent bits, and
then calls B<sched_create_project> (defined in the scheduler specific child
class) to do the scheduler specific bits.

=item B<update_project>:

Updates the specified project, in both PDB and scheduler.
Requires B<hpc>, B<project> and B<reason>
arguments.  Any other fields will cause be changed.  Scheduler specific
bits are passed off to B<sched_update_project>.

=item B<get_autoreplenish_projects>:

Only accepts B<hpc> argument.  Returns a list ref of names of projects
in the cluster (specified via B<hpc>) which are of a type which 
auto_replenishes.

=item B<autoreplenish_project>:

Replenishes the specified project, in both PDB and scheduler.
Requires B<hpc>, B<project> and B<reason> arguments.  Also
requires a B<mode> argument, which should be one of 'monthly',
'quarterly', or 'prorated' for the type of replenishment to do.
The optional parameter B<tstamp> can be given, in seconds from 1970, as
the date to prorate to for the 'prorated' mode.  It is ignored in 
other modes.  Basically just passes off to B<autoreplenish_coop_proj> or
B<autoreplenish_condo_proj> as appropriate.

=item B<autoreplenish_coop_proj>:

This has same arguments/return values as B<autoreplenish_project>, and
does the work of replenishing coop type projects.  Actually, for most
part just passes to B<sched_>I<MODE>B<_replenish_coop_proj> in the
scheduler specific child class.

=item B<autoreplenish_condo_proj>:

This has same arguments/return values as B<autoreplenish_project>, and
does the work of replenishing condo type projects.  Actually, for most
part just passes to B<sched_>I<MODE>B<_replenish_condo_proj> in the
scheduler specific child class.

=item B<is_project_in_any_sister_clusters>

This checks if a project with the same name exists in any sister clusters.
Takes key => value pairs.  The list ref of cluster names should be in the
key 'sisters' and is required, as is the name of the project in 'project'.
The key 'hpc' is optional; if given, that cluster will be omitted in the
results (even if in 'sisters').  The boolean 'noactive' is also optional,
defaults to false.  Normally, only active projects considered, if 'noactive'
is set, active and inactive projects are considered.  The method returns
true (the name of a sister cluster) is the project is found, false otherwise.

=item B<is_unixgrp_in_sister_cluster>

This takes key 'unixgroup', a string representing the name of an unixgroup,
and a list ref 'sisters' of clusters to check.  It then determines if
any projects in any of the 'sisters' has that 'unixgroup'.  If the optional
'hpc' is given, the cluster represented by 'hpc' is skipped even if in the
'sisters' list.  Normally, only active projects considered; if 'noactive' is
set, both active and inactive projects considered.  Returns true (the sister
clustername) if found, false otherwise.


=back

For the methods dealing with projects, the following arguments are commonly
accepted (exceptions noted in list above):

=over 4

=item B<hpc>: 

Always required, instance of B<Glue::HPCC::Cluster> used to determine
which cluster to act on.

=item B<project>:

Always required, the name of the project to act on.

=item B<desc>:

The description of the project.

=item B<unixgroup>:

The name of the unix group associated with the project.

=item B<parent>:

The name of the project which is this project's parent.  Give the
string B<NULL> to unset an existing parent.

=item B<dept>:

The dept code (typically 4 letter string) for department the project is
associated with.

=item B<projtype>:

The name for the type of project this is (e.g. coop, condo, aacgrant).

=item B<su>:

The value of the allocation in SUs (cpu-core hours)

=item B<active>:

Boolean.  Whether the project is active.  When setting to inactive, use 0
(using undef will cause methods to act as if the field was not given, and
either ignore or default it).

=item B<expiration>:

Expiration date of project.  Use YYYY-MM-DD optionally followed by time of day.

=item B<thrhld_pct_over>:

Threshold for warning if overuage.  Percentage.

=item B<thrhld_pct_under>:

Threshold for warning if underuage.  Percentage.

=item B<reason>:

Most actions will be logged and will require a reason.

=back



