#!/usr/local/bin/perl
#
#Does the heavy lifting for Cluster management
#project creation, user management, etc.
#For slurm based clusters

package Glue::HPCC::ClusterUtil::Slurm::Project;
use strict;
use warnings;
use base qw(Glue::HPCC::ClusterUtil::Project);
use Carp qw(carp croak);
use DBI;
use POSIX qw(floor);

use Glue::HPCC::Cluster;
use Glue::HPCC::ClusterUtil::Slurm::Common;

use Glue::HPCC::Project::Log;
use Glue::HPCC::Project::Project;
use Glue::HPCC::Project::ProjectType;

use Slurm::Sacctmgr;
use Slurm::Sacctmgr::Account;
use Slurm::Sacctmgr::Association;
use Slurm::Sacctmgr::Cluster;
use Slurm::Sacctmgr::Qos;
use Slurm::Sshare;

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $DEBUG;
$DEBUG=1;

#Cached sacctmgr and sshare instances for the class
my $_cached_sacctmgr;
my $_cached_sshare;

my $CUCommon = 'Glue::HPCC::ClusterUtil::Slurm::Common';

my $SAAcct = 'Slurm::Sacctmgr::Account';
my $SAAssoc = 'Slurm::Sacctmgr::Association';
my $SACluster = 'Slurm::Sacctmgr::Cluster';
my $SAQos = 'Slurm::Sacctmgr::Qos';

my $GHPLog = 'Glue::HPCC::Project::Log';
my $GHPProj = 'Glue::HPCC::Project::Project';
my $GHPPType = 'Glue::HPCC::Project::ProjectType';

#-------------------------------------------------------------------
#	Common Methods: sacctmgr and sshare
#-------------------------------------------------------------------

sub sacctmgr($)
#Returns a (possibly cached) Slurm::Sacctmgr instance.
#Creates if needed
{	my $class = shift;
	my $me = 'sacctmgr';
	return $_cached_sacctmgr 
		if $_cached_sacctmgr && ref($_cached_sacctmgr);

	my @args = ();
	my $slurm_version = $class->scheduler_version;
	push @args, slurm_version => $slurm_version if $slurm_version;
	my $slurm_rootdir = $class->scheduler_rootdir;
	my $sacctmgr_path;
	$sacctmgr_path = "${slurm_rootdir}/bin/sacctmgr" if $slurm_rootdir;
	push @args, sacctmgr => $sacctmgr_path if $sacctmgr_path;
	$_cached_sacctmgr = Slurm::Sacctmgr->new(@args);
	croak "$me: Unable to create Slurm::Sacctmgr: $_cached_sacctmgr at "
		unless $_cached_sacctmgr && ref($_cached_sacctmgr);

	#Propagate dryrun, verbose settings
	$_cached_sacctmgr->dryrun( $class->dryrun );
	$_cached_sacctmgr->verbose( $class->verbose );

	return $_cached_sacctmgr;
}

sub sshare($)
#Returns a (possibly cached) Slurm::Sshare instance.
#Creates if needed
{	my $class = shift;
	my $me = 'sshare';
	return $_cached_sshare 
		if $_cached_sshare && ref($_cached_sshare);

	$_cached_sshare = Slurm::Sshare->new;
	croak "$me: Unable to create Slurm::Sshare: $_cached_sshare at "
		unless $_cached_sshare && ref($_cached_sshare);

	#Propagate dryrun, verbose settings
	#no dryrun mode, sshare is view only
	#$_cached_sshare->dryrun( $class->dryrun );
	$_cached_sshare->verbose( $class->verbose );

	#Set slurm sshare path and version if given
	my $slurm_version = $class->scheduler_version;
	my $slurm_rootdir = $class->scheduler_rootdir;
	my $sshare_path;
	$sshare_path = "${slurm_rootdir}/bin/sshare" if $slurm_rootdir;
	if ( $sshare_path )
	{	#Have an sshare path, possibly with version
		$_cached_sshare->sshare($sshare_path, $slurm_version);
	} elsif ( $slurm_version)
	{	#Only have a version, not really a good combo
		$_cached_sshare->set_sshare_capabilities_by_version(
			$slurm_version);
	}


	return $_cached_sshare;
}

#-------------------------------------------------------------------
#	Common Methods: verbosity and dryrun mode
#-------------------------------------------------------------------

sub verbose($;$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		#Propagate to sacctmgr and sshare
		$_cached_sacctmgr->verbose($val) if $_cached_sacctmgr;
		$_cached_sshare->verbose($val) if $_cached_sshare;
	}

	#Parent class handles the rest
	return $class->SUPER::verbose($val);
}

sub dryrun($$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		#Propagate to sacctmgr and sshare
		$_cached_sacctmgr->verbose($val) if $_cached_sacctmgr;
                #sshare is view only, no dryrun mode
		#$_cached_sshare->verbose($val) if $_cached_sshare;
	}

	#Parent class handles the rest
	return $class->SUPER::dryrun($val);
}

#--------------------------------------------------------------------------
#		Cluster utilities 
#--------------------------------------------------------------------------

sub get_sacctmgr_cluster_hash($)
#Returns a hash of Slurm::Sacctmgr::Cluster objects, keyed on Cluster name
#
#Should we cache this???
{	my $class = shift;
	my $me = 'get_sacctmgr_cluster_hash';
	my $sacctmgr = $class->sacctmgr;

	my $Clusters = $SACluster->sacctmgr_list($sacctmgr);
	my %Clusters = map { $_->cluster => $_ } @$Clusters;
	return \%Clusters;
}

#-------------------------------------------------------------------
#	Project/account query
#-------------------------------------------------------------------

sub sacctmgr_list_accounts($@)
#Returns a list ref of names of accounts known to sacctmgr (accounts which
#have at least one association attached to them).
#Can be restricted by cluster and/or account name(s).
#	cluster => name of cluster to restrict to
#	account => either scalar name of account to restrict to
#		or a list ref of account names.
#Takes key=>value pairs
{	my $class = shift;
	my %args = @_;
	my $me = 'sacctmgr_list_accounts';

	#Validate parms
	my $cluster = delete $args{cluster};
	my $accounts = delete $args{account};
	$class->_no_extraneous_args($me, %args);

	my $sacctmgr = $class->sacctmgr;
	my @where = ();
	push @where, cluster => $cluster if $cluster;
	push @where, account => $accounts 
		if $accounts && ! ref($accounts);
	$accounts = [ $accounts ] unless ref($accounts);

	my $assocs = $SAAssoc->sacctmgr_list($sacctmgr, @where);
	my @accounts = ();
	if ( scalar(@$accounts) )
	{	foreach my $acct (@$accounts)
		{	my @tmp = grep { $_->account eq $acct } @$assocs;
			push @accounts, $acct if scalar(@tmp);
		}
	} else
	{	@accounts = map { $_->account } @$assocs;
	}

	my %temp = map { $_ => undef } @accounts;
	@accounts = sort ( keys %temp);
	return [ @accounts ];
}

sub sched_list_accounts($@)
#Wrapper for sacctmgr_list_accounts
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_list_accounts';

	#Validate parms
	my $cluster = delete $args{cluster};
	my $accounts = delete $args{account};
	$class->_no_extraneous_args($me, %args);

	return $class->sacctmgr_list_accounts(
		cluster => $cluster,
		account => $accounts );
}
	
		

#--------------------------------------------------------------------------
#		Account usage
#--------------------------------------------------------------------------

sub sshare_get_account_usage($@)
#Returns a hash ref describing to-date usage for a specific account in 
#a specific cluster, useful for providing in report to PoC when account
#replenished.
#
#Takes key=>value pairs for arguments, recognizing:
#	hpc => Glue::HPCC:Cluster instance, REQUIRED
#	account => account name, REQUIRED
#
#Returns a hash ref with the following keys:
#       account => name of the account  
#       cluster => name of cluster
#       limit_ksu => the limit on the account, in kSU
#       used_ksu  => the total usage on the account, in kSU
#       avail_ksu => the amount of time avail, in kSU
#       byuser => a hash ref containing usage stats by user, keyed on username
#               Values are the usage by that user for this account, in kSU
#NOTE: kSU values are NOT rounded/truncated (e.g. at full internal precision)
#
{	my $class = shift;
	my %args = @_;
	my $me = 'sshare_get_account_usage';

	#Validate parameters
	my $hpc = delete $args{hpc};
	my $account = delete $args{account};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required paramenter 'account' at " unless $account;
	$class->_no_extraneous_args($me, %args);
	my $cluster = $hpc->db_clustername;
	$CUCommon->_require_cluster_exists_in_sacctmgr($me, $cluster);

	my $sshare = $class->sshare;
	#Set users to undef to get ALL users associated with the account
	my $usagedata = $sshare->usage_for_account_in_cluster(
		cluster=>$cluster, account=>$account, users=>undef );
	unless ( $usagedata && ref($usagedata) eq 'ARRAY' )
	{	carp "$me: Error getting usage data from sshare for account " .
			"$account in cluster $cluster: $usagedata at ";
		#Try not to abort
		return;
	}
	my ( $used_cpusec, $limit_cpumin, $avail_cpumin, $rawuserdata) = 
		@$usagedata;
	my $userdata;
	if ( $rawuserdata && ref($rawuserdata) eq 'HASH' )
	{	foreach my $user (keys %$rawuserdata)
		{	my $value = $rawuserdata->{$user}; #Cpu-secs
			$value = $value/60/60/1000; #kSU
			$userdata = {} 
				unless $userdata && ref($userdata) eq 'HASH';
			$userdata->{$user} = $value;
		}
	}
	my $used_ksu = $used_cpusec/60/60/1000;
	my $limit_ksu = $limit_cpumin/60/1000;
	my $avail_ksu = $avail_cpumin/60/1000;
	my $usagehash =
	{	account => $account,
		cluster => $cluster,
		limit_ksu => $limit_ksu,
		used_ksu => $used_ksu,
		avail_ksu => $avail_ksu,
		byuser => $userdata,
	};

	return $usagehash;
}

sub sacctmgr_reset_account_balances($@)
#This routine will reset the limit and/or zero the usage on the specified
#account in specified cluster.
#Takes key => value pairs, recognizing:
#	account => name of account to reset
#	cluster => name of cluster account in on
#	tresmin => new TRESmins or GrpCPUmins limit for account.
#		If undef, we leave the limit alone.
#		If hash ref or string with '='. assumed to be a TRES hash
#		Else, should be numeric GrpCPUmins (i.e. a TRES { cpu=>$cpumin})
#		Either format supported for all versions of sacctmgr, but
#		pre v15.x.y only GrpCPUmins can/will be set.
#	zero_usage => Boolean.  If true, we zero usage on the account.
#
#One can use cpumin as an alias for tresmin; if both are given, only tresmin
#is used and a warning will be generated.
#
#Use $class->sacctmgr.
#
#Returns false on success; returns true (error) message on error.
{	my $class = shift;
	my %args = @_;
	my $me = 'sacctmgr_reset_account_balances';

	#Validate parameters
	my $account = delete $args{account};
	my $cluster = delete $args{cluster};
	my $zero_usage = delete $args{zero_usage};
	my $cpumin = delete $args{tresmin};
	if ( exists $args{cpumin} )
	{	if ( defined $cpumin )
		{	carp "$me: Both tresmin and cpumin given, ignoring cpumin at ";
			delete $args{cpumin};
		} else
		{	$cpumin  = delete $args{cpumin};
		}
	}
	croak "$me: Missing required parameter 'account' at " unless $account;
	croak "$me: Missing required parameter 'cluster' at " unless $cluster;
	$CUCommon->_require_cluster_exists_in_sacctmgr($me, $cluster);
	$class->_no_extraneous_args($me, %args);

	unless ( defined $cpumin || $zero_usage )
	{	#Nothing to do.  Warn, but not an error
		carp "$me: Called for $account/$cluster, but nothing to do at ";
		return;
	}

	my $error;
	my $sacctmgr = $class->sacctmgr;

	#To avoid possibly killing jobs, always reset usage before changing
	#cpumin

	if ( $zero_usage )
	{   $error = $SAAcct->zero_usage_on_account_cluster(
		$sacctmgr, $account, $cluster);
	    if ( $error )
	    {	$error = "$me: Error zeroing usage for $account on $cluster: "
			. "$error";
		carp "$error at ";
		return $error;
	    }
	}

	if ( defined $cpumin )
	{   #set_grptresmin_on_account_cluster will take care of converting
	    #cpumin to/from TRES, etc.
	    $error = $SAAcct->set_grptresmin_on_account_cluster(
		$sacctmgr, $account, $cluster, $cpumin);
	    if ( $error )
	    {	$error = "$me: Error setting GrpCPUMin on account $account, " .
			"cluster $cluster to $cpumin: $error";
		carp "$error at ";
		return $error;
	    }
	}

	return;
}

#--------------------------------------------------------------------------
#		Account/project creation
#--------------------------------------------------------------------------

sub sacctmgr_create_account_in_cluster($@)
#Creates specified account in specified cluster, from sacctmgr point of view
#Takes key => value pairs for arguments, recognizing:
#	account => Name of account, REQUIRED
#	project => Name of project(organization), REQUIRED
#	cluster => Name of cluster, REQUIRED
#	desc => Description of account, REQUIRED
#	tresmin => TRESmin or GrpCPUMins setting for account, 
#		not used for condos.
#		TRESmin must be hash ref or string of a=val1, b=val2.
#		Otherwise interpretted as GrpCPUMins and must be numeric
#		Either format can be used with any version of sacctmgr/slurm,
#		But Slurm < v15.x.y can/will only actually set GrpCPUMins 
#		(i.e. the "cpu" field of TRESmin).
#	qos => list ref of QoSes for this account, REQUIRED
#	defaultqos => Default QoS to use, optional, only for condos?
#	FIXME => if set, operate in FIXME mode.
#
#cpumins can be used as an alias for tresmin; if both given, warning generated
#and cpumins is ignored.
#
#Caller should log transaction.
#	
#Returns false (undef) on success.
#On error, either croaks or returns true (Error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'sacctmgr_create_account_in_cluster';

	#Verify paramenters
	my $account = delete $args{account};
	my $projname = delete $args{project};
	my $cluster = delete $args{cluster};
	my $desc = delete $args{desc};
	my $cpumins = delete $args{tresmin};
	if ( defined $args{cpumins} )
	{	if ( defined $cpumins)
		{	carp "$me: Both tresmin and cpumins given; cpumins " .
				"will be ignored at ";
			delete $args{cpumins};
		} else
		{	$cpumins = delete $args{cpumins};
		}
	}
	my $qos = delete $args{qos};
	my $defqos = delete $args{defaultqos};
	my $FIXME = delete $args{FIXME};
	croak"$me: Missing required arg 'account' at " unless $account;
	croak"$me: Missing required arg 'project' at " unless $projname;
	croak"$me: Missing required arg 'desc' at " unless $desc;
	#CPUmins is optional; e.g. for condo accounts
	#croak"$me: Missing required arg 'cpumins' at " unless defined $cpumins;
	croak"$me: Missing required arg 'qos' at " unless $qos;
	$class->_no_extraneous_args($me, %args);

	#Convert qos to a string
	$qos = [ $qos ] unless $qos && ref($qos) eq 'ARRAY';
	my $qosstr = join ',', @$qos;

	my $sacctmgr = $class->sacctmgr;
	my $error;

	my @fields_to_set =
	(	description => $desc,
		organization => $projname,
		qos => $qosstr,
	);
	push @fields_to_set, defaultqos => $defqos if defined $defqos;
	#We split into 2 operations; 1)  create the account, then
	# 2) set GrpCPUMins/TRESmins if desired.
	#This is because 1st step is not Slurm version dependent, but
	#second is.
	#push @fields_to_set, grpcpumins => $cpumins if defined $cpumins;

	#Check for assocs for this account/cluster
	#If there are none, we need to $SAAcct->sacctmgr_add instead of
	# $SAAcct->sacctmgr_modify
	my $assoclist = $SAAssoc->sacctmgr_list( $sacctmgr,
			account=>$account,
			cluster=>$cluster,
		);

	
	#--- Step 1: Create the account (w/out GrpCPUMin/TRESmin setting)

	if ( $assoclist && ref($assoclist) eq 'ARRAY' && scalar(@$assoclist) )
	{	#Association exists for this account/cluster

		#Abort unless in FIXME mode
		croak "$me: Failed to create $account/$cluster because " .
			"it already exists and not in fixmode at "
			unless $FIXME;

		my $where = {
			account => $account,
			cluster => $cluster, 
		};
		my $set = { @fields_to_set };

		$error = $SAAcct->sacctmgr_modify( $sacctmgr, $where, $set );
		if ( $error && ! ref($error) eq 'ARRAY' )
		{	die "$me: Error modifying acct $account/$cluster: "
				. "$error at ";
		}

	} else
	{	#No Association exists for this account/cluster
		
		$error = $SAAcct->sacctmgr_add( $sacctmgr,
			account => $account,
			cluster=> $cluster,
			@fields_to_set,
		);
		if ( $error && ! ref($error) eq 'ARRAY' )
		{	die "$me: Error adding acct $account/$cluster: "
				. "$error at ";
		}
	}

	#--- Step 2: Set GrpCPUMin/TRESmin if requested
	return unless defined $cpumins;

	$error = $SAAcct->set_grptresmin_on_account_cluster( $sacctmgr,
		$account, $cluster, $cpumins);
	if ( $error && ! ref($error) eq 'ARRAY' )
	{	die "$me: Error setting GrpCPUMins/TRESmin for " .
			" $account/$cluster: $error at ";
	}

	return;
}

sub sched_create_project($@)
#Creates named project in given cluster, creating the normal and/or
#high priority accounts needed for this type.
#Takes key=>value pairs as arguments, recognizing
#	hpc: Glue::HPCC:Cluster instance, REQUIRED
#	project: name of project, REQUIRED
#	type: type of project in this cluster, REQUIRED
#	su: Value of allocation in this cluster, in SU, Required for 
#		metered types
#	reason: for logging, REQUIRED
#	FIXME: operate in FIXME mode
#
#This only does the Slurm sacctmgr part.  Normally expect it to be
#called from Glue::HPCC::Cluster::create_project
#
#Returns false on success.
#On error, either croaks or returns true (error message).
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_create_project';

	#Validate parameters
	my $projname = delete $args{project};
	my $su = delete $args{su};
	my $type = delete $args{type};
	my $hpc = delete $args{hpc};
	my $reason = delete $args{reason};
	my $FIXME = delete $args{FIXME};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required arg project at " unless $projname;
	croak "$me: Missing required arg type at " unless $type;
	croak "$me: Missing required arg reason at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	my $cluster = $hpc->db_clustername;
	$CUCommon->_require_cluster_exists_in_sacctmgr($me, $cluster);

	
	my $typerec = $GHPPType->projecttype_by_name($hpc, $type);
	croak "$me: Invalid project type '$type' at " unless ( $typerec );
	croak "$me: Error looking up project type '$type': $typerec at " 
		unless ref($typerec);

	if ( $typerec->is_metered )
	{	croak "$me: Missing arg 'su' required for project " .
			"type $type  at " unless defined $su;
		croak "$me: Invalid su=$su.  Require su >= 0 at " 
			unless ( $su >= 0 );
	}

	#Create accounts in sacctmgr, with association for specified cluster

	my ($acctname, $org, $acctcpumins, $acctdesc);
	my $error;
	my $assoc;

	#PDB values are now AFTER TAX!!!!
	my $cpumins = 60*$su if defined $su;
	my ($acctobj, $assoclist);

	my $qos_tag = $typerec->qos_tag;
	my $qoses = $hpc->get_qos_list($qos_tag, $projname) if $qos_tag;

	if ($typerec->has_normal_priority )
	{	#Create normal priority account
		$acctname = $projname;
		$org = $projname;
		$acctdesc = "Normal priority account for $projname";
		$acctcpumins = $cpumins;
		$acctcpumins = undef unless $typerec->is_metered;

		if ( $type eq 'coop' )
		{	$acctcpumins = 2*$cpumins/3; #1/3 goes to high priority
		}

		$error = $class->sacctmgr_create_account_in_cluster(
			account => $acctname,
			project => $org,
			cluster => $cluster,
			desc => $acctdesc,
			cpumins => $acctcpumins,
			qos => $qoses,
			FIXME => $FIXME,
		);
		croak "$me: Error creating acct $acctname: $error at "
			if $error;
		$GHPLog->add_log_entry($hpc,
			action_code=>'create-sacctmgr-account',
			reason => $reason,
			project => $projname,
			cluster => $cluster,
			description => "created std pri account for project",
			account => $acctname,
		);
	}

	if ($typerec->has_high_priority )
	{	#Create normal priority account
		$acctname = "${projname}-hi";
		$org = $projname;
		$acctdesc = "High priority account for $projname";
		$acctcpumins = $acctcpumins;
		$acctcpumins = undef unless $typerec->is_metered;

		if ( $type eq 'coop' )
		{	$acctcpumins = $cpumins/3; #2/3 goes to normal pri
		}

		$error = $class->sacctmgr_create_account_in_cluster(
			account => $acctname,
			project => $org,
			cluster => $cluster,
			desc => $acctdesc,
			cpumins => $acctcpumins,
			qos => $qoses,
			FIXME => $FIXME,
		);
		croak "$me: Error creating acct $acctname: $error at "
			if $error;
		$GHPLog->add_log_entry($hpc,
			action_code=>'create-sacctmgr-account',
			reason => $reason,
			project => $projname,
			cluster => $cluster,
			description => "created hi pri account for project",
			account => $acctname,
		);
	}

	if ( $typerec->has_special_partition )
	{	#Condo account
		#die "Cannot handle condo accounts at this time. at ";

		#Create the condo Qoses if needed
		my $sacctmgr = $class->sacctmgr;

		CONDOQOS: foreach my $tmpqosname (@$qoses)
		{	my $tmpqos = $SAQos->new_from_sacctmgr_by_name(
				$sacctmgr, $tmpqosname);
			next CONDOQOS if $tmpqos && ref($tmpqos);
			croak "$me: Error retrieving QoS $tmpqosname: $tmpqos "
				. "at " if $tmpqos;

			#Does not exist, create it
			my $prio = 10;
			my $tmptxt = "standard";
			if ( $tmpqosname =~ /-hi$/ )
			{	$prio = 20;
				$tmptxt = "high";
			}
			my $desc = "$projname partition $tmptxt priority qos";
			$SAQos->sacctmgr_add($sacctmgr,
				name=>$tmpqosname,
				priority => $prio,
				description => $desc,
				usagefactor => 1.0,
				preempt => 'scavenger',
				preemptmode => 'cluster',
			);
			$GHPLog->add_log_entry($hpc,
				action_code=>'create-condo-qos',
				reason => $reason,
				project => $projname,
				cluster => $cluster,
				description => "created QoS $tmpqosname for condo project",
				account => $projname,
			);
		} # CONDOQOS

		my @tmp = grep !/-hi/, @$qoses;
		my $defqos = $tmp[0];

		$error = $class->sacctmgr_create_account_in_cluster(
			account => $projname,
			project => $projname,
			cluster => $cluster,
			desc => "$projname condo account",
			qos => $qoses,
			defaultqos => $defqos,
			FIXME => $FIXME,
		);
		
		croak "$me: Error creating condo acct $projname: $error at "
			if $error;
		$GHPLog->add_log_entry($hpc,
			action_code=>'create-sacctmgr-account',
			reason => $reason,
			project => $projname,
			cluster => $cluster,
			description => "created condo account for project",
			account => $projname,
		);
	}
	return;
}

sub sacctmgr_delete_account($@)
#Deletes named account in given cluster
#Takes key=>value pairs as arguments, recognizing
#	cluster: clustername, REQUIRED
#	account: name of account to delete, REQUIRED
#	reason: for logging, REQUIRED
#
#This only does the Slurm sacctmgr part.  Normally expect it to be
#called from Glue::HPCC::Cluster::create_project
#
#Returns false on success.
#On error, either croaks or returns true (error message).
{	my $class = shift;
	my %args = @_;
	my $me = 'sacctmgr_delete_account';

	#Validate parameters
	my $account = delete $args{account};
	my $cluster = delete $args{cluster};
	my $reason = delete $args{reason};
	croak "$me: Missing required arg cluster at " unless $cluster;
	croak "$me: Missing required arg account at " unless $account;
	croak "$me: Missing required arg reason at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	$CUCommon->_require_cluster_exists_in_sacctmgr($me, $cluster);

	#Verify no users with associations for this account
	my $usrclass = $class->user_class;
	my $users = $usrclass->sacctmgr_list_users( 
			cluster => $cluster, account => $account );
	if ( $users && ref($users) eq 'ARRAY' && scalar(@$users) )
	{	my $tmp = join ", ", @$users;
		croak "$me: Refusing to delete account $account in " .
			"cluster $cluster; users exist [$tmp] at ";
	}
	my $sacctmgr = $class->sacctmgr;

	#OK, safe to delete
	$SAAcct->sacctmgr_delete($sacctmgr,
		cluster => $cluster, account => $account );
	return;
}

sub sched_delete_project($@)
#Deletes named project in given cluster (on slurm side)
#Takes key=>value pairs as arguments, recognizing
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	project: name of project to delete, REQUIRED
#	type : name of project type, REQUIRED
#	reason: for logging, REQUIRED
#
#This only does the Slurm sacctmgr part.  Normally expect it to be
#called from Glue::HPCC::Cluster::delete_project
#
#Returns false on success.
#On error, either croaks or returns true (error message).
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_delete_project';

	#Validate parameters
	my $projname = delete $args{project};
	my $hpc = delete $args{hpc};
	my $reason = delete $args{reason};
	my $ptname = delete $args{type};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required arg project at " unless $projname;
	croak "$me: Missing required arg reason at " unless $reason;
	croak "$me: Missing required arg type at " unless $ptname;
	$class->_no_extraneous_args($me, %args);
	my $cluster = $hpc->db_clustername;
	$CUCommon->_require_cluster_exists_in_sacctmgr($me, $cluster);

	my $ptype = $GHPPType->projecttype_by_name($hpc, $ptname);
	croak "$me: Invalid type '$ptname', not in PDB at "
		unless $ptype;
	croak "$me: Error looking up type $ptname: $ptype at "
		unless ref($ptype);

	my @accounts = ();
	my $acctstd = $projname;
	my $accthi = "${projname}-hi";
	push @accounts, $acctstd 
		if $ptype->has_normal_priority
		|| $ptype->has_special_partition;
	push @accounts, $accthi if $ptype->has_high_priority;

	#Verify no users with associations for this account
	my ($account, $users);
	my $usrclass = $class->user_class;
	foreach $account (@accounts)
	{	$users = $usrclass->sacctmgr_list_users( 
			cluster => $cluster, account => $account );
		if ( $users && ref($users) eq 'ARRAY' && scalar(@$users) )
		{	my $tmp = join ", ", @$users;
			croak "$me: Refusing to delete project $projname in " .
				"cluster $cluster; users exist [$tmp] " .
				"for account $account at ";
		}
	}

	#OK, safe to delete
	my $sacctmgr = $class->sacctmgr;
	foreach $account (@accounts)
	{	$SAAcct->sacctmgr_delete($sacctmgr,
			cluster => $cluster, account => $account );
		$GHPLog->add_log_entry($hpc,
			action_code=>'delete-sacctmgr-account',
			reason => $reason,
			project => $projname,
			account => $account,
			cluster => $cluster,
			description => "deleted account $account in sacctmgr",
		);
	}
	return;
}

sub sched_update_project($@)
#Updates project (scheduler dependent pieces)
#Takes key => value pairs, recognizing:
#	hpc : Glue::HPCC::Cluster object, REQUIRED
#	project: name of project, REQUIRED,
#	ksu => new ksu
#	active => new active
#	type => new type
#	reason => for logging, REQUIRED
#
#Caller is expected to log.
#Returns false on success, croaks or returns error message on error
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_update_project';

	#Validate args
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $ksu = delete $args{ksu};
	my $active = delete $args{active};
	my $type = delete $args{type};
	my $reason = delete $args{reason};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at "
		unless $projname;
	croak "$me: Missing required parameter 'reason' at "
		unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $proj = $GHPProj->project_by_name($hpc, $projname);
	croak "$me: Invalid project $projname, not in PDB at "
		unless $proj;
	croak "$me: Error looking up project $projname: $proj at "
		unless ref($proj);
	my $ptypeid = $proj->projtype;
	my $projtype = $GHPPType->projecttype_by_pkey($hpc, $ptypeid);
	croak "$me: Unable to find ptype for proj $projname ($ptypeid) at "
		unless $projtype;
	croak "$me: Error looking up ptype: $projtype at "
		unless ref($projtype);
	my $cluster = $hpc->db_clustername;

	my $changed = 0;
	my $error;

	if ( $type && $type ne $projtype->name )
	{	#Change type of project
		#Cant handle yet
		$changed = 1; #To avoid stupid warnings
		carp "$me: Not yet able to change project types, skipping at ";
	}

	if ( $ksu && $ksu != $proj->ksu )
	{	$changed = 1; #To avoid stupid warnings
		carp "$me: Not changing ksu for project $projname in slurm, "
			. "not sure what to do, at ";
	}

	if ( defined $active )
	{	if ( $active && ! $proj->active )
		{	#Activate project
			#Basically, need to recreate the slurm bits
			$changed = 1;
			#Use new values of type, ksu if provided,
			#otherwise use what looked up in PDB
			$type = $projtype->name unless $type; 
			$ksu = $proj->ksu unless $ksu;
			my $su = 1000*$ksu;
			$error = $class->sched_create_project(
				hpc => $hpc,
				project => $projname,
				reason => $reason,
				type => $type,
				su => $su,
				FIXME => 1, #Just in case not deleted properly
			);
			croak "$me: Error re-creating project $projname: $error at "
				if $error;
			$GHPLog->add_log_entry($hpc,
				action_code=>'sched-activate-project',
				reason => $reason,
				project => $projname,
				cluster => $cluster,
				description => "recreated project $projname in slurm",
				account => $projname,
			);
		}
		if ( $proj->active && ! $active )
		{	#Deactivate project
			$changed = 1;
			$error = $class->sched_delete_project(
				hpc => $hpc,
				project => $projname,
				type => $type,
				reason => $reason,
			);
			croak "$me: Error deleting project $projname: $error at "
				if $error;
			$GHPLog->add_log_entry($hpc,
				action_code=>'sched-deactivate-project',
				reason => $reason,
				project => $projname,
				cluster => $cluster,
				description => "deleted project $projname in slurm",
				account => $projname,
			);
		}
	}

	unless ( $changed )
	{	carp "$me: No changes made to project $projname at ";
	}
	return;
}


#-------------------------------------------------------------------
#	Autoreplenishment methods
#
# All autoreplenishment methods take as arguments key=>value pairs, as follows:
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	project => name of project, REQUIRED
#	reason => for logging.  REQUIRED
# The 'mode' is specific to each routine.  The 'prorated' mode may also
# accept an argument 'tstamp'; if omitted defaults to current time.
#
# All autoreplenishment methods return as follows:
# 	( $errflag, $msgs, $usage, $usage2 ) 
# On success, 
#	$errflag is false
#	$msgs is a listref of text strings describing what happened.
#		Suitable for emailing to project owner.
#	$usage and $usage2 are account usages for normal and hi-pri
#	    accounts, resp. assoc with project just before replenishment
#	    For use in reports to PoC.
#	    If only one account, $usage2 will be undef.  Either may be undef
#	    These are hash refs with keys
#		account: name of account
#		cluster: name of cluster
#		limit_ksu: the limit on the account, kSU
#		used_ksu: the total usage on the account, kSU
#		avail_ksu: the amount of time available, kSU
#		byuser : Hash ref with per user stats.  
#			Keyed on username
#			Value is usage by this user, in kSU.
# On error, $errflag is true (error message).  $msgs should contain what
# was done up to error, as well as error message.  Rest is likely undef.
#
#-------------------------------------------------------------------

#---------------	coop projects -------------------------

sub sched_quarterly_replenish_coop_proj($@)
#Does quarterly autoreplenishment for 'coop' type projects
#See above for inputs/outputs.
#
#This code _tries_ not to abort on run-time like errors.
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_quarterly_replenish_coop_proj';

	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $reason = delete $args{reason};

	#These are more compilation time errors, abort
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	my $cluster = $hpc->db_clustername;

	#This is definitely runtime, but serious enough to abort
 	croak "$me: Account replenishments must be done from within HPCC"
		unless $hpc->running_in_cluster;
	
	my $msgs = [];
	my $err;

	#Find our project
	my $proj = $GHPProj->project_by_name($hpc, $projname);
	unless ( $proj )
	{	$err = "$me: Project $projname not found in PDB";
		carp "$err at ";
		return ($err, [$err]);
	}
	unless ( ref($proj) )
	{	$err = "$me: error querying PDB for project $projname: $proj";
		carp "$err at ";
		return ($err, [$err]);
	}

	#Make sure if project type coop
	my $ptypeid = $proj->projtype;
	my $ptype = $GHPPType->projecttype_by_pkey($hpc, $ptypeid);
	my $ptname = $ptype->name;
	unless ( $ptname eq 'coop' )
	{	$err = "$me: PType mismatch, expect coop, got '$ptname' for "
		. "project $projname";
		carp "$err at ";
		return ($err, [$err]);
	}

	my $su = $proj->su;
	#PDB values are now AFTER TAX
        #$su = 0.8 * $su; # 20% overhead to DivIT; TODO lookup tax in DB??
        my $cpumins = 60*$su;

	my $account = $projname;
	my $accounthi = "${projname}-hi";
	my $cpulimit = 2*$cpumins/3; #2/3 normal, 1/3 high
	my $cpulimithi = $cpumins/3;
	#Make sure are integral
	$cpulimit = floor($cpulimit + 0.5);
	$cpulimithi = floor($cpulimithi + 0.5);
	#Make sure are non-negative
	$cpulimit = 0 if $cpulimit < 0;
	$cpulimithi = 0 if $cpulimithi < 0;

	my $sacctmgr = $class->sacctmgr;
	my $error;

	#Get usage stats
	my $usage = $class->sshare_get_account_usage(
		hpc => $hpc, account => $account);
	my $usage2 = $class->sshare_get_account_usage(
		hpc => $hpc, account => $accounthi);

	#Replenish normal account
	$error = $class->sacctmgr_reset_account_balances(
		account => $account,
		cluster => $cluster,
		tresmin => { cpu => $cpulimit, },
		zero_usage => 1);
	if ( $error )
	{	$err = "$me: Error replenishing acct $account in " .
			"$cluster: $error";
		carp "$err at ";
		push @$msgs, $err;
		return ( $err, $msgs, $usage, $usage2);
	}
	push @$msgs, "Zeroed usage for account $account in cluster $cluster";
	push @$msgs, "Set CPU min limit on account $account in cluster $cluster"
		. " to $cpulimit CPU-min";
	$GHPLog->add_log_entry($hpc,
		action_code=>'sched-replenish-account',
		reason => $reason,
		project => $projname,
		account => $account,
		cluster => $cluster,
		description => "Slurm: quarterly reset coop std-pri account",
		tablename => 'sacctmgr',
		fieldname => 'grpcpumin',
		newval => $cpulimit,
		);

	#Replenish hipri account
	$error = $class->sacctmgr_reset_account_balances(
		account => $accounthi,
		cluster => $cluster,
		tresmin => { cpu => $cpulimithi, },
		zero_usage => 1);
	if ( $error )
	{	$err = "$me: Error replenishing acct $accounthi in " .
			"$cluster: $error";
		carp "$err at ";
		push @$msgs, $err;
		return ( $err, $msgs, $usage, $usage2);
	}
	$GHPLog->add_log_entry($hpc,
		action_code=>'sched-replenish-account',
		reason => $reason,
		project => $projname,
		account => $accounthi,
		cluster => $cluster,
		description => "Slurm: quarterly reset coop hi-pri account",
		tablename => 'sacctmgr',
		fieldname => 'grpcpumin',
		newval => $cpulimithi,
		);

	#Make friendly message
	my $ksu = sprintf "%.2f kSU", ($cpulimit/60/1000);
	my $ksuhi = sprintf "%.2f kSU", ($cpulimithi/60/1000);
	$msgs = [	
	"Reset limit for account $account in cluster $cluster to $ksu, usage 0",
	"Reset limit for account $accounthi in cluster $cluster to $ksuhi, usage 0",
	];

	return (0,$msgs, $usage, $usage2);
}

sub sacctmgr_null_coop_reset($@)
#This does a "null" reset on a coop account, not changing any balances
#but resetting so usage is zero and new limit is the old balance.
#Useful for cases when hipri already maxed out, or insufficient funds in
#normal priority
#
#Takes key => value pairs
#	account, REQUIRED
#	hpc, REQUIRED
#	reason, REQUIRED
#	norm_balance => Balance on normal account, in CPUMIN, REQUIRED
#	high_balance => Balance on hipri account, in CPUMIN, REQUIRED
#	msgs => list ref of any previous messages.  Defaults to []
#	me => masquerade as $me in error messages
#	usage, usage2 => usage hashs for norm, hipri resp.  REQUIRED
#
#Returns same as sched_monthly_replenish_coop_proj
#Mainly exists for error handling
{	my $class = shift;
	my %args = @_;
	my $realme = 'sacctmgr_null_coop_reset';

	#Validate arguments
	my $me = delete $args{me} || $realme;
	my $account = delete $args{account};
	my $hpc = delete $args{hpc};
	my $normbal = delete $args{norm_balance};
	my $hibal = delete $args{high_balance};
	my $msgs = delete $args{msgs} || [];
	my $usage = delete $args{usage};
	my $usage2 = delete $args{usage2};
	my $reason = delete $args{reason};

	#Make sure balances are integral
	$normbal = floor($normbal + 0.5);
	$hibal = floor($hibal + 0.5);
	#Make sure are non-negative --- slurm underflows if negative
	$normbal = 0 if $normbal < 0;
	$hibal = 0 if $hibal < 0;

	#These are compile time errors, croak on
	croak "$realme: Missing required parm 'account' at " unless $account;
	$class->_validate_hpc($realme, $hpc);
	croak "$realme: Missing required parm 'norm_balance' at " 
		unless defined $normbal;
	croak "$realme: Missing required parm 'usage' at " 
		unless defined $usage;
	croak "$realme: Missing required parm 'usage2' at " 
		unless defined $usage2;
	croak "$realme: Missing required parm 'high_balance' at " 
		unless defined $hibal;
	croak "$realme: Missing required parm 'reason' at " unless $reason;

	my $cluster = $hpc->db_clustername;
	my ($error, $err, @errs);
	my $accounthi = $account . '-hi';

	#Reset usage/limit on hipri so usage2 reports properly next mon
	$error = $class->sacctmgr_reset_account_balances(
			account => $accounthi,
			cluster => $cluster,
			tresmin => { cpu => $hibal},
			zero_usage => 1);
	if ( $error )
	{	@errs = (
			"Error resetting acct balances for $accounthi: $error",
			"Old balance = $hibal cpumin",
		);
		push @$msgs, @errs;
		foreach $err (@errs) { carp "$me: $err at " };
		return ( 1, $msgs, $usage, $usage2 );
	}
	$GHPLog->add_log_entry($hpc,
		action_code=>'sched-replenish-account',
		reason => $reason,
		project => $account,
		account => $accounthi,
		cluster => $cluster,
		description => "Slurm: null reset coop hi-pri account",
		tablename => 'sacctmgr',
		fieldname => 'balance',
		oldval => $hibal,
		newval => $hibal,
		);

	#And do same for normal account
	$error = $class->sacctmgr_reset_account_balances(
			account => $account,
			cluster => $cluster,
			tresmin => { cpu => $normbal},
			zero_usage => 1);
	if ( $error )
	{   	@errs = (
			"Error resetting acct balances for $account: $error",
			"Old balance = $normbal cpumin",
		      );
		push @$msgs, @errs;
		foreach $err (@errs) { carp "$me: $err at " };
		return ( 1, $msgs, $usage, $usage2 );
	}
	$GHPLog->add_log_entry($hpc,
		action_code=>'sched-replenish-account',
		reason => $reason,
		project => $account,
		account => $account,
		cluster => $cluster,
		description => "Slurm: null reset coop std-pri account",
		tablename => 'sacctmgr',
		fieldname => 'balance',
		oldval => $normbal,
		newval => $normbal,
		);

	my $high_ksu_bal = sprintf "%0.2f kSU", $hibal/60/1000;
	my $norm_ksu_bal = sprintf "%0.2f kSU", $normbal/60/1000;
	#OK, no errors
	$msgs = 
	[ 	"Account $account not changed (avail=$norm_ksu_bal)",
		"Account $accounthi not changed (avail=$high_ksu_bal)",
	];

	if ( $normbal )
	{	unshift @$msgs,
			"Account $accounthi ($cluster) already at maximum, "
                        . "no funds transferred.";
	} else
	{	unshift @$msgs,
			"No funds available in account $account (cluster " .
			"$cluster) for transfer to $accounthi";
	}

	return (0, $msgs, $usage, $usage2 );
}

sub sched_monthly_replenish_coop_proj($@)
#Does monthly autoreplenishment for projects of type 'coop', 
#does the sacctmgr bits
#Takes key => value pairs
#	hpc => Glue::HPCC:Cluster instance, REQUIRED
#	project => project name, REQUIRED
#	reason => for logging, REQUIRED
#
#Does monthly autoreplenishment of named project in cluster
#	1) Get usage and limits for normal and high priority accounts, and
#		computes balances, and monthly allotment (1/3 of after tax 
#		quarterly amount)
#	2) If normal balance <= 0, zero normal usage and limits.  Zero hipri
#		usage and set limit to current balance.
#	3) If hipri balance => monthly allotment, Zero hipri usage and set
#		limit to monthly allotment if not already there
#	4) Compute the amount of funds $xfer need to xfer to bring hipri 
#		account balance back to monthly allotment.
#	5) Reduce $xfer to available funds in normal balance if available
#		funds less than what was computed in 4
#	6) Zero usage on normal account, set limit to previous balance less
#		the amount transferred $xfer.
#	7) Zero usage on hipri account, set limit to prev balance + $xfer
#
#Returns ( $errflag, $msgs, $usage, $usage2 ) where $errflag is false on 
#success, true (error #message) on errors, and $msgs is a listref of text 
#strings describing what #was done (suitable for emailing to project owner).
# $usage and $usage2 are the account usages (from sshare_get_account_usage)
#for the normal and high-priority accounts associated with this project,
#taken just before the account replenished, for use in reports to the PoC
#
#This code _tries_ not to abort on run-time like errors.
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_monthly_replenish_coop_proj';

	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $reason = delete $args{reason};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	my $cluster = $hpc->db_clustername;

	#This is definitely runtime, but serious enough to abort
 	croak "$me: Account replenishments must be done from within HPCC"
		unless $hpc->running_in_cluster;

	my $msgs = [];
	my $err;

	#Find our project
	my $proj = $GHPProj->project_by_name($hpc, $projname);
	unless ( $proj )
	{	$err = "$me: Project $projname not found in PDB";
		carp "$err at ";
		return ($err, [$err]);
	}
	unless ( ref($proj) )
	{	$err = "$me: error querying PDB for project $projname: $proj";
		carp "$err at ";
		return ($err, [$err]);
	}

	#Make sure if project type coop
	my $ptypeid = $proj->projtype;
	my $ptype = $GHPPType->projecttype_by_pkey($hpc, $ptypeid);
	my $ptname = $ptype->name;
	unless ( $ptname eq 'coop' )
	{	$err = "$me: PType mismatch, expect coop, got '$ptname' for "
		. "project $projname";
		carp "$err at ";
		return ($err, [$err]);
	}

	my $su = $proj->su;
	#PDB values are now AFTER TAX
        #$su = 0.8 * $su; # 20% overhead to DivIT; TODO lookup tax in DB??
        my $cpumins = 60*$su;

	my $account = $projname;
	my $accounthi = "${projname}-hi";
	my $monthly_cpumins = $cpumins/3;
	my $monthly_ksu= sprintf "%0.2f kSU", $monthly_cpumins/60/1000;

	#Get balances
	my $sshare = $class->sshare;
	my $tmp = $sshare->usage_for_account_in_cluster(
		account=>$account, cluster=>$cluster);
	unless ( $tmp && ref($tmp) eq 'ARRAY' )
	{	$err = "$me: Error looking up balance for $account in " .
			"$cluster: $tmp";
		carp "$err at ";
		return ($err, [ $err ]);
	}
	my ($norm_cpusec_used, $norm_cpumin_limit, $norm_cpumin_bal) = @$tmp;
	#Get usage for reporting.  A bit wasteful since we already called sshare
	my $usage = $class->sshare_get_account_usage(
		hpc => $hpc, account => $account);

	$tmp = $sshare->usage_for_account_in_cluster(
		account=>$accounthi, cluster=>$cluster);
	unless ( $tmp && ref($tmp) eq 'ARRAY' )
	{	$err = "$me: Error looking up balance for $accounthi in " .
			"$cluster: $tmp";
		carp "$err at ";
		return ($err, [ $err ]);
	}
	my ($high_cpusec_used, $high_cpumin_limit, $high_cpumin_bal) = @$tmp;
	my $high_ksu_used = sprintf "%0.2f kSU", $high_cpusec_used/60/60/1000;
	my $high_ksu_limit = sprintf "%0.2f kSU", $high_cpumin_limit/60/1000;
	my $high_ksu_bal = sprintf "%0.2f kSU", $high_cpumin_bal/60/1000;
	#Get usage for reporting.  
	my $usage2 = $class->sshare_get_account_usage(
		hpc => $hpc, account => $accounthi);

	if ( $norm_cpumin_bal <= 0 )
	{	#No funds available to xfer
		$err = "No funds available in account $account ( cluster " .
			"$cluster) for transfer to $accounthi";
		push @$msgs, $err;
		return $class->sacctmgr_null_coop_reset(
			account => $account,
			hpc => $hpc,
			norm_balance => $norm_cpumin_bal,
			high_balance => $high_cpumin_bal,
			msgs => $msgs,
			me => $me,
			usage => $usage,
			usage2 => $usage2,
			reason => $reason,
		);
	}

	if ( $high_cpumin_bal >= $monthly_cpumins )
	{	#High priority already at maximum
		$err = "Account $accounthi already at maximum $monthly_ksu, "
				. "no funds transferred.";
		push @$msgs, $err;
		return $class->sacctmgr_null_coop_reset(
			account => $account,
			hpc => $hpc, 
			norm_balance => $norm_cpumin_bal,
			high_balance => $high_cpumin_bal,
			msgs => $msgs,
			me => $me,
			usage => $usage,
			usage2 => $usage2,
			reason => $reason,
		);
	}

	#How much do we need to xfer?
	my $xfer_mins = $monthly_cpumins - $high_cpumin_bal;
	if ( $xfer_mins > $norm_cpumin_bal )
	{	#Insufficient funds to do the full xfer
		$xfer_mins = $norm_cpumin_bal; #Xfer all that we have
		push @$msgs, "Insufficient funds in account $account (cluster "
			. "$cluster) to bring $accounthi up to $monthly_ksu";
	}
	my $xferksu = sprintf "%.2f", $xfer_mins/60/1000;
	my ($error, @errs);

	#Add funds to $accounthi
	my $newcpumin = $high_cpumin_bal + $xfer_mins;
	#Make sure is integral
	$newcpumin = floor($newcpumin + 0.5);
	#And non-negative (slurm will either drop limit or underflow)
	$newcpumin = 0 if $newcpumin < 0;
	my $newksu = sprintf "%.2f", $newcpumin/60/1000;
	my $new_hiksu = $newksu;
	$error = $class->sacctmgr_reset_account_balances(
			account => $accounthi,
			cluster => $cluster,
			tresmin => { cpu => $newcpumin, },
			zero_usage => 1);
	if ( $error )
	{   @errs = (
		"Error replenishing acct $accounthi on $cluster: $error",
		"Erroring adding $xferksu to bring up to $newksu",
	    );
	    push @$msgs, @errs;
	    foreach $err (@errs) { carp "$me: $err at "; };
	    return ($err, $msgs, $usage, $usage2);
	}
	$GHPLog->add_log_entry($hpc,
		action_code=>'sched-replenish-account',
		reason => $reason,
		project => $projname,
		account => $accounthi,
		cluster => $cluster,
		description => "Slurm: monthy reset coop hi-pri account",
		tablename => 'sacctmgr',
		fieldname => 'balance',
		oldval => $high_cpumin_bal,
		newval => $newcpumin,
		);

	#Remove funds from $account
	$newcpumin = $norm_cpumin_bal - $xfer_mins;
	#Make sure is integral
	$newcpumin = floor($newcpumin + 0.5);
	#And non-negative, need to avoid underflow
	$newcpumin = 0 if $newcpumin < 0;
	$newksu = sprintf "%.2f", $newcpumin/60/1000;
	my $new_normksu = $newksu;
	$error = $class->sacctmgr_reset_account_balances(
			account => $account,
			cluster => $cluster,
			tresmin => { cpu => $newcpumin, },
			zero_usage => 1);
	if ( $error )
	{   @errs = (
		"Error moving funds out of acct $account on $cluster: $error",
		"Erroring removing $xferksu to bring make it $newksu",
	    );
	    push @$msgs, @errs;
	    foreach $err (@errs) { carp "$me: $err at "; };
	    return ($err, $msgs, $usage, $usage2);
	}
	$GHPLog->add_log_entry($hpc,
		action_code=>'sched-replenish-account',
		reason => $reason,
		project => $projname,
		account => $account,
		cluster => $cluster,
		description => "Slurm: monthy reset coop std-pri account",
		tablename => 'sacctmgr',
		fieldname => 'balance',
		oldval => $norm_cpumin_bal,
		newval => $newcpumin,
		);
	
	#Success
	push @$msgs, "Transfered $xferksu from $account to $accounthi";
	push @$msgs, "$account: New balance=$new_normksu";
	push @$msgs, "$accounthi: New balance=$new_hiksu";
	return ( 0, $msgs, $usage, $usage2);
}

sub sched_prorated_replenish_coop_proj($@)
#Does prorated autoreplenishment for projects of type 'coop', 
#does the sacctmgr bits
#Takes key => value pairs
#	hpc => Glue::HPCC:Cluster instance, REQUIRED
#	project => project name, REQUIRED
#	tstamp => time to prorate to, defaults to now
#	reason => for logging, REQUIRED
#
#Does prorated autoreplenishment of named project in cluster
#Returns ( $errflag, $msgs, $usage, $usage2 ) where $errflag is false on 
#success, true (error #message) on errors, and $msgs is a listref of text 
#strings describing what #was done (suitable for emailing to project owner).
# $usage and $usage2 are the account usages (from sshare_get_account_usage)
#for the normal and high-priority accounts associated with this project,
#taken just before the account replenished, for use in reports to the PoC
#
#This code _tries_ not to abort on run-time like errors.
{	my $class = shift;
	my $me = 'sched_prorated_replenish_coop_proj';
	my $err =  "Prorated replenishment of coop proj under slurm not " .
		"yet implemented";
	carp "$me: $err at ";
	return ( 1, $err, [ $err ] );
	#This is definitely runtime, but serious enough to abort
 	#croak "$me: Account replenishments must be done from within HPCC"
	#	unless $hpc->running_in_cluster;

}

sub sched_replenish_condo_proj($@)
#Autoreplenishes projects of type 'condo', does the sacctmgr bits
#Takes key => value pairs
#	hpc => Glue::HPCC:Cluster instance, REQUIRED
#	cluster => cluster name, defaults from $hpc->db_clustername
#	project => project name, REQUIRED
#	mode => mode, one of quarterly, monthly, prorate, REQUIRED
#	tstamp => seconds from epoch for use with prorate mode.  Defaults to
#		now.
#	reason => for logging, REQUIRED
#mode and tstamp are for consistency with other project types.  
#Currently, all that
#the autoreplenish of condo projects does is zero the usage, and return usage
#stats.  This is the same for monthly, quarterly, and prorated modes.
#
#Returns ( $errflag, $msgs, $usage, $usage2 ) where $errflag is false on 
#success, true (error #message) on errors, and $msgs is a listref of text 
#strings describing what #was done (suitable for emailing to project owner).
# $usage is the account usages (from sshare_get_account_usage) for
#the account associated with this project.
#taken just before the account replenished, for use in reports to the PoC
# $usage2 is undef; only there for consistency with other project types.
#
#This code _tries_ not to abort on run-time like errors.
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_replenish_condo_proj';

	#Validate arguments.  These are more compile time args, so ok to croak
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $mode = delete $args{mode};
	my $reason = delete $args{reason};
	my $tstamp = delete $args{tstamp} || time;
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'mode' at " unless $mode;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	$mode = lc $mode;
	my $err;
	my $cluster = $hpc->db_clustername;

	#This is definitely runtime, but serious enough to abort
 	croak "$me: Account replenishments must be done from within HPCC"
		unless $hpc->running_in_cluster;

	#Find our project
	my $proj = $GHPProj->project_by_name($hpc, $projname);
	unless ( $proj )
	{	$err = "$me: Project $projname not found in PDB";
		carp "$err at ";
		return ($err, [$err]);
	}
	unless ( ref($proj) )
	{	$err = "$me: error querying PDB for project $projname: $proj";
		carp "$err at ";
		return ($err, [$err]);
	}

	#Make sure if project type coop
	my $ptypeid = $proj->projtype;
	my $ptype = $GHPPType->projecttype_by_pkey($hpc, $ptypeid);
	my $ptname = $ptype->name;
	unless ( $ptname eq 'condo' )
	{	$err = "$me: PType mismatch, expect condo, got '$ptname' for "
		. "project $projname";
		carp "$err at ";
		return ($err, [$err]);
	}

	#Get current usage stats
	my $msgs = [];
	my $account = $projname;
	my $usage = $class->sshare_get_account_usage(
		hpc => $hpc, account => $account);

	my $error = $class->sacctmgr_reset_account_balances(
			account => $account,
			cluster => $cluster,
			zero_usage => 1);
	if ( $error )
	{   $err = "Error zeroing usage on acct $account on $cluster: $error";
	    push @$msgs, $err;
	    carp "$me: $err at ";
	    return ($err, $msgs, $usage);
	}
	$GHPLog->add_log_entry($hpc,
		action_code=>'sched-replenish-account',
		reason => $reason,
		project => $projname,
		account => $account,
		cluster => $cluster,
		description => "Slurm: $mode reset coop account",
		tablename => 'sacctmgr',
		fieldname => 'usage',
		newval => 0,
		);

	push @$msgs, "Zeroed usage for account $account in cluster $cluster";
	return (0,$msgs, $usage);
}

	
1;
__END__

=head1 NAME

Glue::HPCC::ClusterUtil::Slurm::Project - Project management under Slurm

=head1 SYNOPSIS

  ...


=head1 DESCRIPTION

This provides various routines for managing projects in an HPC 
environment.  This package defines the Slurm specific stuff, but it
inherits the general, scheduler independent stuff from
B<Glue::HPCC::ClusterUtil::Project>, 
so it should be called by various scripts.
See also B<Glue::HPCC::ClusterUtil::User> for user management.

Typically, a script will get the B<Glue::HPCC::Cluster> object for the 
specific cluster (using CLUSTER in hostinfo to determine the cluster).
It then gets a B<Glue::HPCC::ClusterUtil> object from that using the
B<get_hpcUtil> method, which should be the version for the specified
scheduler.  The B<project> method on that object should return the class
name for this package.

All of the routines defined in this package are class routines.  A minimal
constructor is inherited if desired, but there is no instance data, etc.
Routines are packaged as an OO class to keep namespaces clean and allow
inheritence.

Significant routines are:

=over 4

=item B<verbose>

Gets/sets the verbosity level.  When setting, propagates to all classes
we depend on.

=item B<dryrun>

Gets/sets the dryrun level.  When setting, propagates to all classes
we depend on.

=item B<sacctmgr>

Returns a B<Slurm::Sacctmgr> object.  Will try to use a cached one if can,
but will create if need be.  The returned object has B<verbose> and B<dryrun>
synchronized with above methods.  By default, this will be configured for the
appropriate path and Slurm version as specified in the Cluster instance.

=item B<sshare>

Returns a B<Slurm::Sshare> object.  Will try to use a cached one if can,
but will create if need be.  The returned object has B<verbose> (no
dryrun is available since sshare is view only) synchronized with above method.  By default, this will be configured for the
appropriate path and Slurm version as specified in the Cluster instance.

=item B<get_sacctmgr_cluster_hash>

Returns a hash ref of B<Slurm::Sacctmgr::Cluster> objects, keyed on cluster
name.

=item B<sshare_get_account_usage>

Returns a hash ref containing to-date usage information for an account.  This
is based on B<sshare> command, so is since the last account reset, etc.

=item B<sched_quarterly_replenish_coop_proj>,B<sched_monthly_replenish_coop_proj>,B<sched_prorated_replenish_coop_proj>:

These do quartlerly, monthly, and prorated replenishments of 'coop' type
projects.  That is an inherently scheduler specific task; these handle it
for the Slurm scheduler.

=item B<sacctmgr_null_coop_reset>:

This does a "null" reset on a coop account, not changing any balances
but resetting so usage is zero and new limit is the old balance.

=item B<sched_replenish_condo_proj>:

This does replenishment of 'condo' type projects.  For this project type,
no distinction between various modes of replenishment.

=item B<sacctmgr_list_accounts>,  B<sched_list_accounts>:

Returns a list ref of names of accounts known to scheduler.  The second
form is just a wrapper to enable parent class to call w/out knowing the
actual scheduler.

=item B<sacctmgr_create_account_in_cluster>:

This does the Slurm specific bits for creating an B<account> in the
cluster.  Normally called from B<sched_create_project> for the various
accounts in a project.  It takes a list of key=>value pairs for arguments.
The arguments account (name of account), project (name of project), cluster
(name of cluster), desc (description of account), qos (list ref of valid
QOSes for this account) are all required.  The optional defaultqos sets
the default QoS to use (only used for condo projects).  FIXME is boolean,
and if set means do not complain if pieces of the account already exist but
assume there was a failed creation attempt and attempt to repair.

The optional field 'tresmin' will set the TRESmin limits for the 
allocation account,
and its value should either be a hash ref (e.g. { TRES1 => val1, TRES2=>val2})
or a comma delimited string a TRES assignments (e.g. TRES1=val1, TRES2=val2),
or a simple number which is interpretted as GrpCPUMins (e.g. 
{cpu => GrpCPUMins}).  

Slurm versions which do not support TRESes will still work with the TRES
values described above, but only the 'cpu' TRES will be set (i.e., only
GrpCPUmins will be set); any other TRESes will be silently ignored.

The key 'cpumins' is an alias for 'tresmin'; if both are provided a warning
is generated and 'cpumins' is ignored.

The account 'creation' is now normally done with two sacctmgr calls; the
first does the basic creation stuff which is _not_ dependent on Slurm version.
The second sets the TRESmin/GrpCPUmin limits, which is Slurm version dependent.

=item B<sched_create_project>:

This does the Slurm specific bits for creating a B<project>.  Calls
B<sacctmgr_create_account_in_cluster> with appropriate arguments to 
create the appropriate accounts.  Normally this is called from
B<create_project> in parent class so that both PDB and Slurm bits are
done properly.

=item B<sacctmgr_delete_account>:

This does the Slurm specific  bits for deleting an B<account>.  Normally
called from B<sched_delete_project>.

=item B<sched_delete_project>:

This does the Slurm specific bits for deleting a B<project>.  Calls
B<sacctmgr_delete_account> to delete the Slurm accounts associated with
the project.  Normally this is called from B<delete_project> in the
parent class so both PDB and Slurm bits get done.

=item B<sched_update_project>:

This does the Slurm specific parts of updating a project.  I.e., can
change active status, sus, or type.  Except cannot change type yet, and
working on sus.

=item B<sacctmgr_reset_account_balances>:

Resets the balance on a specified account in Slurm.  Takes a list of 
key => value pairs for arguments. Required arguments are
B<account> (name of account) and B<cluster> (name of cluster). The optional
argument B<tresmin> gives the new TRESmin limits for the account; if not 
defined the limits are left alone.  Set a limit to -1 to unlimit it.
The optional boolean argument B<zero_usage> 
determines whether the usage for the account is set to zero.

TRESmin can be a hash ref (e.g. { TRES1 => limit1, TRES2 => limit2 }), or a
comma delimited string of limit assignments (e.g. TRES1=limit1, TRES2=limit2),
or a number which is interpretted as GrpCPUmins (e.g. { cpu => GrpCPUmins }).
The key 'cpumin' can be used as an alias for the key 'tresmin'; if both are
given, a warning will be outputted and 'cpumins' will be ignored.

For Slurm versions not supporting TRESes, only the 'cpu' TRES field is 
considered and it will be used to set the GrpCPUmins limit for the allocation 
account.

=back






