#!/usr/local/bin/perl
#
#Does the heavy lifting for Cluster management
#project creation, user management, etc.
#For slurm based clusters

package Glue::HPCC::ClusterUtil::Slurm::User;
use strict;
use warnings;
use base qw(Glue::HPCC::ClusterUtil::User);
use Carp qw(carp croak);
use DBI;
use POSIX qw(floor);

use Slurm::Sacctmgr;
use Slurm::Sacctmgr::Association;
use Slurm::Sacctmgr::User;
use Slurm::Sshare;

use Glue::HPCC::Cluster;
use Glue::HPCC::ClusterUtil::Slurm::Common;

use Glue::HPCC::Project::Log;
use Glue::HPCC::Project::ProjectType;


#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $DEBUG;
$DEBUG=1;

#Cached Slurm::Sacctmgr and Slurm::Sshare instances
my $_cached_sacctmgr;
my $_cached_sshare;

#Convenience
my $GHPLog = 'Glue::HPCC::Project::Log';
my $GHPPType = 'Glue::HPCC::Project::ProjectType';
my $CUCommon = 'Glue::HPCC::ClusterUtil::Slurm::Common';

my $SAAssoc = 'Slurm::Sacctmgr::Association';
my $SAUser = 'Slurm::Sacctmgr::User';

#-------------------------------------------------------------------
#	Common Methods: sacctmgr and sshare
#-------------------------------------------------------------------

sub sacctmgr($)
#Returns a (possibly cached) Slurm::Sacctmgr instance.
#Creates if needed
{	my $class = shift;
	my $me = 'sacctmgr';
	return $_cached_sacctmgr 
		if $_cached_sacctmgr && ref($_cached_sacctmgr);

	my @args = ();
	my $slurm_version = $class->scheduler_version;
	push @args, slurm_version => $slurm_version if $slurm_version;
	my $slurm_rootdir = $class->scheduler_rootdir;
	my $sacctmgr_path;
	$sacctmgr_path = "${slurm_rootdir}/bin/sacctmgr" if $slurm_rootdir;
	push @args, sacctmgr => $sacctmgr_path;
	$_cached_sacctmgr = Slurm::Sacctmgr->new(@args);
	croak "$me: Unable to create Slurm::Sacctmgr: $_cached_sacctmgr at "
		unless $_cached_sacctmgr && ref($_cached_sacctmgr);

	#Propagate dryrun, verbose settings
	$_cached_sacctmgr->dryrun( $class->dryrun );
	$_cached_sacctmgr->verbose( $class->verbose );

	return $_cached_sacctmgr;
}

sub sshare($)
#Returns a (possibly cached) Slurm::Sshare instance.
#Creates if needed
{	my $class = shift;
	my $me = 'sshare';
	return $_cached_sshare 
		if $_cached_sshare && ref($_cached_sshare);

	$_cached_sshare = Slurm::Sshare->new;
	croak "$me: Unable to create Slurm::Sshare: $_cached_sshare at "
		unless $_cached_sshare && ref($_cached_sshare);

	#Propagate dryrun, verbose settings
	#no dryrun mode, sshare is view only
	#$_cached_sshare->dryrun( $class->dryrun );
	$_cached_sshare->verbose( $class->verbose );

	#Set slurm sshare path and version if given
	my $slurm_version = $class->scheduler_version;
	my $slurm_rootdir = $class->scheduler_rootdir;
	my $sshare_path;
	$sshare_path = "${slurm_rootdir}/bin/sshare" if $slurm_rootdir;
	if ( $sshare_path )
	{	#Have an sshare path, possibly with version
		$_cached_sshare->sshare($sshare_path, $slurm_version);
	} elsif ( $slurm_version)
	{	#Only have a version, not really a good combo
		$_cached_sshare->set_sshare_capabilities_by_version(
			$slurm_version);
	}

	return $_cached_sshare;
}

#-------------------------------------------------------------------
#	Common Methods: verbosity and dryrun mode
#-------------------------------------------------------------------

sub verbose($;$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		$GHPLog->verbose($val);
		#Propagate to sacctmgr and sshare
		$_cached_sacctmgr->verbose($val) if $_cached_sacctmgr;
		$_cached_sshare->verbose($val) if $_cached_sshare;
	}

	#Parent class handles the rest
	return $class->SUPER::verbose($val);
}

sub dryrun($$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		$GHPLog->dryrun($val);
		$_cached_sacctmgr->dryrun($val) if $_cached_sacctmgr;
		#sshare is view only, no dryrun mode
		#$_cached_sshare->dryrun($val) if $_cached_sshare;
	}

	#Parent class handles the rest
	return $class->SUPER::dryrun($val);
}

#-------------------------------------------------------------------
#	Get/set default account for user
#-------------------------------------------------------------------

sub get_default_account_for_user($@)
#Returns the default account associated with user account.
#Takes key=>value pairs:
#	user: username, REQUIRED
#	nocroak: if set, invalid username or lookup errors NOT fatal
#	nowarn: if set, non-fatal errors do not generate warning output
#	hpc: Glue::HPCC::Cluster instance, not required
#
#Returns true (account name) on success.
#Otherwise, either croaks (unless nocroak is set) or returns false
#(possibly with warning, unless nowarn is also set)
{	my $class = shift;
	my %args = @_;

	my $hpc = delete $args{hpc};
	my $username = delete $args{user};
	my $me = 'get_default_account_for_user';
	my $nocroak = delete $args{nocroak};
	my $nowarn = delete $args{nowarn};

	#
	#$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter username at "
		unless $username;
	$class->_no_extraneous_args($me, %args);

	my $sacctmgr = $class->sacctmgr;
	my $user = $SAUser->new_from_sacctmgr_by_name($sacctmgr, $username);
	unless ( $user )
	{ 	croak "$me: Invalid user $username at " unless $nocroak;
	 	carp "$me: Invalid user $username at " unless $nowarn;
		return;
	};
	unless ( ref($user) )
	{ 	croak "$me: Error looking up user $username in sacctmgr: $user at "
			unless $nocroak;
	 	carp "$me: Error looking up user $username in sacctmgr: $user at "
			unless $nowarn;
		return;
	};

	my $account = $user->defaultaccount;
	return $account;
}

sub set_default_account_for_user($@)
#Sets the default account associated with user
#Takes key => value pairs
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	user: username, REQUIRED
#	account: account name, REQUIRED
#	reason: for logging, REQUIRED
#	cluster: cluster to use in setting def acct.
#		Defaults to cluster from $hpc
#
#Returns false on success.
#Otherwise, either croaks or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'set_default_account_for_user';

	my $hpc = delete $args{hpc};
	my $username = delete $args{user};
	my $account = delete $args{account};
	my $reason = delete $args{reason};
	my $cluster = delete $args{cluster};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter username at "
		unless $username;
	croak "$me: Missing required parameter account at "
		unless $account;
	croak "$me: Missing required parameter reason at "
		unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $sacctmgr = $class->sacctmgr;
	$cluster = $hpc->db_clustername unless defined $cluster;
	my $where = { name => $username, cluster => $cluster };
	my $set = { defaultaccount => $account };

	my $error = $SAUser->sacctmgr_modify( $sacctmgr, $where, $set);
	$GHPLog->add_log_entry( $hpc,
		action_code => 'set-default-account',
		person => $username,
		account => $account,
		cluster => $hpc->db_clustername,
		description => "Set def account for user $username to $account",
		reason => $reason,
	) unless $error;
	return $error;
}

#-------------------------------------------------------------------
#	Generic add/remove user from sacctmgr account
#-------------------------------------------------------------------

sub sacctmgr_add_user_assoc_to_clus_acct_part($@)
#Creates association for specified user to use specified account
#to access specified partitions in specified cluster at specified qoses.
#
#Takes key => value pairs for arguments, recognizing:
#	user => username, REQUIRED
#	account => Name of account, REQUIRED
#	cluster => Name of cluster, REQUIRED
#	partition => Name of partition, REQUIRED
#	qos => list ref of QoSes for this association, REQUIRED
#	defaultqos => default QoS for user
#	FIXME => operate in repair mode.  Ignored, as always in repair mode
#	
#Returns false (undef) on success.
#On error, either croaks or returns true (Error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'sacctmgr_add_user_assoc_to_clus_acct_part';

	my $user = delete $args{user};
	my $account = delete $args{account};
	my $cluster = delete $args{cluster};
	my $partition = delete $args{partition};
	my $qos = delete $args{qos};
	my $defqos = delete $args{defaultqos};
	my $FIXME = delete $args{FIXME};

	croak "$me: Missing required arg 'user' at " unless $user;
	croak "$me: Missing required arg 'account' at " unless $account;
	croak "$me: Missing required arg 'cluster' at " unless $cluster;
	croak "$me: Missing required arg 'partition' at " unless $partition;
	croak "$me: Missing required arg 'qos' at " unless $qos;
	$class->_no_extraneous_args($me, %args);

	#Convert qos to a string
	$qos = [ $qos ] unless $qos && ref($qos) eq 'ARRAY';
	my $qosstr = join ',', @$qos;

	my $sacctmgr = $class->sacctmgr;
	my $error;

	my @fields_to_set =
	(	qos => $qosstr,
	);
	push @fields_to_set, defaultqos => $defqos if $defqos;

	my @where_clause = 
	(		account => $account,
			cluster => $cluster,
			user => $user,
			partition => $partition,
	);

	#Check for assocs for this account/cluster/user/part combo
	#If there are none, we need to $SAAcct->sacctmgr_add instead of
	# $SAAcct->sacctmgr_modify
	my $assoclist = $SAAssoc->sacctmgr_list( $sacctmgr, @where_clause);

	
	if ( $assoclist && ref($assoclist) eq 'ARRAY' && scalar(@$assoclist) )
	{	#Association exists for this account/cluster/user/partition

		my $where = { @where_clause };
		my $set = { @fields_to_set };

		$error = $SAUser->sacctmgr_modify( $sacctmgr, $where, $set );
		if ( $error && ! ref($error) eq 'ARRAY' )
		{	croak "$me: Error modifying assoc for " .
				"$cluster/$account/$user/$partition : "
				. "$error at ";
		}

	} else
	{	#No Association exists for this account/cluster
		$error = $SAUser->sacctmgr_add( $sacctmgr,
			@where_clause, @fields_to_set,);
		if ( $error && ! ref($error) eq 'ARRAY' )
		{	croak "$me: Error adding acct $account/$cluster: "
				. "$error at ";
		}
	}

	return;
}

sub sacctmgr_set_new_defacct_if_needed($@)
#When removing associations for given user/account, sacctmgr
#will fail if the account being removed from is the user's default
#account, and not all associations for user are being removed;
#i.e. the user will still exist in Slurm.  In such cases, we
#need to change the user's default account _before_ doing the
#removal of associations.
#
#This routine takes an username, account and cluster being deleted
#from, and if necessary will change the user's default account
#to allow the removal from account/cluster to succeed.  Does nothing
#unless has to.
#
#Basically, does nothing if account != user's default account,
#or user will still have access to the account after the deletion.  
#Or if that is the only account the user is associated with on that cluster.
#(users can have different defacct per cluster!)
#
#But if deleting user from his default account, and user has access
#to another account, it will set the user's default account to 
#one of the other accounts is associated with (arbitrarily/randomly
#if multiple such accounts).
#
#Normally called from sacctmgr_del_user_assoc_from_clus_acct_part
#and basically takes same arguments.
#
#Takes key => value pairs for arguments, recognizing:
#	user => username, REQUIRED
#	account => account you plan to delete user from, REQUIRED
#	cluster => cluster you plan to delete user from, REQUIRED
#	hpc => Glue::HPCC::Cluster instance, REQUIRED (for add_log)
#	partition => partition you plan to delete user from.
#		Defaults to all partitions for this user/account/cluster
#	qos => list ref of qoses you plan to delete user from.  Defaults to
#		all QoSes for this user/account/cluster/(partition)
#	silent => If set, no info message when changing default account.
{	my $class = shift;
	my %args = @_;
	my $me = 'sacctmgr_set_new_defacct_if_needed';

	my $user = delete $args{user};
	my $account = delete $args{account};
	my $cluster = delete $args{cluster};
	my $partition = delete $args{partition};
	my $qos = delete $args{qos};
	my $hpc = delete $args{hpc};
	my $silent = delete $args{silent};

	croak "$me: Missing required arg 'user' at " unless $user;
	croak "$me: Missing required arg 'account' at " unless $account;
	croak "$me: Missing required arg 'cluster' at " unless $cluster;
	$class->_validate_hpc($me, $hpc);

	$qos = [ $qos ] if ( $qos && ref($qos) ne 'ARRAY' );

	my $sacctmgr = $class->sacctmgr;
	my $error;
	

	my $olddef = $class->get_default_account_for_user(
		user => $user, nocroak=>1);

	#Don't need to do anything if cannot get default account
	return unless $olddef;

	#Don't need to do anything unless trying to delete from default account
	#print STDERR "[DEBUG]: user $user defacct = $olddef != $account\n" unless $account eq $olddef;
	return unless $account eq $olddef;

	#OK, trying to delete from default account.  Will there be any associations
	#for user with this account after we are done?

	#Find all associations for user on specified cluster (users can
	#have different defacct per cluster, and sacctmgr will not remove user
	#from assoc if will leave assocs on the cluster w/out defacct for user)
	my $assocs = Slurm::Sacctmgr::Association->sacctmgr_list($sacctmgr,
		user => $user, cluster => $cluster );
	unless ( $assocs && ref($assocs) eq 'ARRAY' )
	{	croak "$me: Error looking up associations for user=$user: $! at ";
	}

	#If user is already not associated with any accounts on cluster, nothing to do.  Just return
	#print STDERR "[DEBUG]: no assocs for user $user\n" unless (@$assocs);
	return unless ( @$assocs );

	#Get associations related to this account, and those related to other accts
	my @acctassocs = grep { $_->account eq $account } @$assocs;

	my @otherassocs = grep { $_->account && $_->account ne $account } @$assocs;
	
	#If user is not associated with any other accounts, just return
	#print STDERR "[DEBUG]: no assocs for user $user other than $account\n" unless (@otherassocs);
	return unless @otherassocs;

	#Get hash of qoses to delete (if any specified)
	my %qoses = map { $_ => undef } @$qos if $qos;
	
	#We assume if partition and/or QoS given, that other partitions/QoSes are
	#defined, so a record with no partition/QoS specified matches another partition
	#
	#Find any associations for this account not matching our partition/qos as applicable)
	@acctassocs = grep { !( (! $qos || ! $_->qos || exists $qoses{$_->qos} ) &&
		(! $partition || ! $_->partition || $partition eq $_->partition ) 
	       )
	    } @acctassocs;
	#If any such assocs exist, do NOT need to change default account
	#print STDERR "[DEBUG]: $user will have assoc for $account after deletion\n" if @acctassocs;
	#use Data::Dumper;
	#print STDERR "Assocs are:\n", Dumper(\@acctassocs), "\n";
	return if @acctassocs;

	#Need to change default account, pick first one found in @otherassocs
	my $newrec = $otherassocs[0];
	my $newacct = $newrec->account;
	my $newclus = $newrec->cluster;
	my $reason = "Removing user $user from previous defacct $account on $cluster";
	$class->set_default_account_for_user(
		hpc => $hpc,
		user => $user,
		account => $newacct,
		reason => $reason,
		cluster => $newclus,
	);
	print STDERR "Changed default account for user $user in cluster $newclus to $newacct\n"
		unless $silent;
}


sub sacctmgr_del_user_assoc_from_clus_acct_part($@)
#Deletes association for specified user to use specified account
#to access specified partitions in specified cluster at specified qoses.
#
#Takes key => value pairs for arguments, recognizing:
#	hpc => Glue::HPCC:Cluster instance, REQUIRED (for sacctmgr_set_new_defacct_if_needed)
#	user => username, REQUIRED
#	account => Name of account, REQUIRED
#	cluster => Name of cluster, REQUIRED
#	partition => Name of partition, optional.  If omitted,
#		will delete associations for all partitions for this
#		user/account/cluster combo
#	qos => list ref of QoSes for this association.  If omitted,
#		will delete for all QoSes.
#	adjust_balances => if true, account's limit will have deleted
#		user's usage deducted from it after deletion.  (Otherwise,
#		only the user's usage is deleted, so balance artificially increases).
#	
#Returns false (undef) on success.
#On error, either croaks or returns true (Error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'sacctmgr_del_user_assoc_from_clus_acct_part';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $account = delete $args{account};
	my $cluster = delete $args{cluster};
	my $partition = delete $args{partition};
	my $qoses = delete $args{qos};
	my $adj_bal = delete $args{adjust_balances};

	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required arg 'user' at " unless $user;
	croak "$me: Missing required arg 'account' at " unless $account;
	croak "$me: Missing required arg 'cluster' at " unless $cluster;

	my $sacctmgr = $class->sacctmgr;
	my $error;

	if ( $qoses && ref($qoses) eq 'ARRAY' && @$qoses && $adj_bal )
	{	return "Unable to adjust balance when QoSes specified.";
	}

	#Get usage for user
	my $sshare = $class->sshare;
	my $newlimit_cpumin;
	my $oldlimit_cpumin;
	my $cpumin_used_by_user;
	my $pclass = $class->project_class;
	if ( $adj_bal )
	{	my $temp = $sshare->usage_for_account_in_cluster(
			cluster => $cluster, account=> $account, users => [$user]);
		die "Unable to get usage for account $account in cluster $cluster in order to adjust balance on deletion on $user: $temp at "
			unless $temp && ref($temp) eq 'ARRAY';
		$oldlimit_cpumin = $temp->[1];
		my $userdata = $temp->[3] || {};
		my $cpusec_used_by_user = $userdata->{$user} || 0;
		$cpumin_used_by_user = floor($cpusec_used_by_user/60 + 0.5);
		if ( $cpumin_used_by_user > 0 )
		{	$newlimit_cpumin = $oldlimit_cpumin - $cpumin_used_by_user;
			if ( $newlimit_cpumin < 0 )
			{	warn "WARNING: account $account: adjusted limit=$newlimit_cpumin < 0; truncating to 0 at ";
				$newlimit_cpumin = 0;
			}
		}
	}

	my @where_clause = 
	(		account => $account,
			cluster => $cluster,
			user => $user,
	);
	push @where_clause, partition => $partition if $partition;

	if ( $qoses && ref($qoses) eq 'ARRAY' && @$qoses)
	{	#We need to change defacct for user if will conflict with
		#our deletion of associations
		$class->sacctmgr_set_new_defacct_if_needed(@where_clause,
			qos => $qoses, hpc => $hpc);
		foreach my $qos (@$qoses)
		{	$error = $SAUser->sacctmgr_delete( $sacctmgr,
				@where_clause, qos => $qos);
			if ( $error && ! ref($error) eq 'ARRAY' )
			{	die "$me: Error deleting qos $qos from  " .
				    "$user/$account/$cluster/$partition: "
					. "$error at ";
			}
		}
	} else
	{	#We need to change defacct for user if will conflict with
		#our deletion of associations
		$class->sacctmgr_set_new_defacct_if_needed(@where_clause,
			hpc => $hpc);
		$error = $SAUser->sacctmgr_delete( $sacctmgr, 
			@where_clause);
		if ( $error && ! ref($error) eq 'ARRAY' )
		{	die "$me: Error deleting user $user/$account/$cluster: "
				. "$error at ";
		}
	}

	#Adjust balances if requested/needed
	if ( $adj_bal && defined $newlimit_cpumin )
	{	$pclass->sacctmgr_reset_account_balances( account=>$account,
			cluster => $cluster, cpumin => $newlimit_cpumin,
			zero_usage => 0 );
	}
	return;
}

#-------------------------------------------------------------------
#	Generic add/remove user from sacctmgr accounts for a project
#-------------------------------------------------------------------

sub sacctmgr_add_user_to_project_in_cluster($@)
#Does the sacctmgr related work of adding specified user to
#specified project in specified cluster.
#
#Takes key=>value pairs as arguments, recognizing
#	hpc: Glue::HPCC:Cluster instance, REQUIRED
#	user: username of user to add, REQUIRED
#	project: name of project, REQUIRED
#	type: type of project in this cluster, REQUIRED
#	highpriority: flag, if true add to highpriority QOS for condo
#		Only valid for project type 'condo'
#	FIXME: operate in repair mode
#
#Returns false on success.
#On error, either croaks or returns true (error message).
{	my $class = shift;
	my %args = @_;
	my $me = 'sacctmgr_add_user_to_project_in_cluster';
	my $user = delete $args{user};
	my $project = delete $args{project};
	my $type = delete $args{type};
	my $hpc = delete $args{hpc};
	my $hipri = delete $args{highpriority};
	my $FIXME = delete $args{FIXME};

	#Validate parms
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter user at " unless $user;
	croak "$me: Missing required parameter project at " unless $project;
	croak "$me: Missing required parameter type at " unless $type;
	$class->_no_extraneous_args($me, %args);
	my $cluster = $hpc->db_clustername;
	$CUCommon->_require_cluster_exists_in_sacctmgr($me, $cluster);

	if ( defined $hipri && $type ne 'condo' )
	{	croak "$me: Error- highpriority flag only valid for project " .
			"type 'condo', not project type $type at ";
	}
	
	#Get project type
	my $typerec = $GHPPType->projecttype_by_name($hpc, $type);
	unless ( $typerec )
	{	croak "$me: Invalid project type '$type'.\n";
	}

	#Create user associations in sacctmgr as indicated by project type
	my ($acctname, $partname, $error);
	my $acctname_normal = $project;
	my $acctname_high = "${project}-hi";
	my $qos_tag = $typerec->qos_tag;
	my $qoses;
	my $defqos;
	if ( $qos_tag )
	{	$qoses = $hpc->get_qos_list($qos_tag, $project);
	}

	if ($typerec->has_normal_priority )
	{	#Add to normal priority account
		$acctname = $acctname_normal;
		$partname = 'standard';

		$error = $class->sacctmgr_add_user_assoc_to_clus_acct_part(
			user => $user,
			account => $acctname,
			partition => $partname,
			cluster => $cluster,
			qos => $qoses,
			FIXME => $FIXME,
		);
		
		croak "$me: Error adding $user to $acctname for $partname: "
			. "$error at " if $error;
	}

	if ($typerec->has_high_priority )
	{	#Add to high priority account
		$acctname = $acctname_high;
		$partname = 'high-priority';

		$error = $class->sacctmgr_add_user_assoc_to_clus_acct_part(
			user => $user,
			account => $acctname,
			partition => $partname,
			cluster => $cluster,
			qos => $qoses,
			FIXME => $FIXME,
		);
		
		croak "$me: Error adding $user to $acctname for $partname: "
			. "$error at " if $error;
	}

	if ( $typerec->has_special_partition )
	{	#Add to special condo partition
		$partname = $project;
		$acctname = $project;
		
		my @tmp;
		if ( $hipri )
		{	@tmp = grep /-hi/, @$qoses;
			$defqos = $tmp[0];
		} else
		{	@tmp = grep !/-hi/, @$qoses;
			$qoses = [ @tmp ];
			$defqos = $tmp[0];
		}

		$error = $class->sacctmgr_add_user_assoc_to_clus_acct_part(
			user => $user,
			account => $acctname,
			partition => $partname,
			cluster => $cluster,
			qos => $qoses,
			defaultqos => $defqos,
			FIXME => $FIXME,
		);
		
		croak "$me: Error adding $user to condo $acctname for " .
			"$partname: $error at " if $error;
	}

	if ($typerec->can_use_debug_q )
	{	#Grant access to debug partition
		$partname = 'debug';
		$qoses = $hpc->get_qos_list('debug', $project);

		if ( $typerec->has_normal_priority  ||
			$typerec->has_special_partition )
		{   $acctname = $acctname_normal;
		    $error = $class->sacctmgr_add_user_assoc_to_clus_acct_part(
			user => $user,
			account => $acctname,
			partition => $partname,
			cluster => $cluster,
			qos => $qoses,
			FIXME => $FIXME,
		    );
		
		    croak "$me: Error adding $user to $acctname for " .
			" $partname: $error at " if $error;
		}

		if ( $typerec->has_high_priority )
		{   $acctname = $acctname_high;
		    $error = $class->sacctmgr_add_user_assoc_to_clus_acct_part(
			user => $user,
			account => $acctname,
			partition => $partname,
			cluster => $cluster,
			qos => $qoses,
			FIXME => $FIXME,
		    );
		
		    croak "$me: Error adding $user to $acctname for " .
			" $partname: $error at " if $error;
		}
	}

	if ($typerec->can_use_scavenger_q )
	{	#Grant access to scavenger partition
		$partname = 'scavenger';
		$qoses = $hpc->get_qos_list('scavenger', $project);

		if ( $typerec->has_normal_priority  || 
			$typerec->has_special_partition )
		{   $acctname = $acctname_normal;
		    $error = $class->sacctmgr_add_user_assoc_to_clus_acct_part(
			user => $user,
			account => $acctname,
			partition => $partname,
			cluster => $cluster,
			qos => $qoses,
			FIXME => $FIXME,
		    );
		
		    croak "$me: Error adding $user to $acctname for " .
			"$partname: $error at " if $error;
		}

		if ( $typerec->has_high_priority )
		{   $acctname = $acctname_high;
		    $error = $class->sacctmgr_add_user_assoc_to_clus_acct_part(
			user => $user,
			account => $acctname,
			partition => $partname,
			cluster => $cluster,
			qos => $qoses,
			FIXME => $FIXME,
		    );
		
		    croak "$me: Error adding $user to $acctname for " .
			"$partname: $error at " if $error;
		}
	}

	return;
}

sub sacctmgr_del_user_from_project_in_cluster($@)
#Does the sacctmgr related work of removing specified user to
#specified project in specified cluster.
#
#Takes key=>value pairs as arguments, recognizing
#	hpc: Glue::HPCC:Cluster instance, REQUIRED
#	user: username of user to add, REQUIRED
#	project: name of project, REQUIRED
#	type: type of project in this cluster, REQUIRED
#	adjust_balances: if set, reduce account balances by users usage after deletion.
#
#Returns false on success.
#On error, either croaks or returns true (error message).
{	my $class = shift;
	my %args = @_;
	my $me = 'sacctmgr_add_user_to_project_in_cluster';

	my $user = delete $args{user};
	my $project = delete $args{project};
	my $type = delete $args{type};
	my $adj_bals = delete $args{adjust_balances};
	my $hpc = delete $args{hpc};

	#Validate parms
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter user at " unless $user;
	croak "$me: Missing required parameter project at " unless $project;
	croak "$me: Missing required parameter type at " unless $type;
	$class->_no_extraneous_args($me, %args);
	my $cluster = $hpc->db_clustername;
	$CUCommon->_require_cluster_exists_in_sacctmgr($me,$cluster);

	#Get project type
	my $typerec = $GHPPType->projecttype_by_name($hpc, $type);
	unless ( $typerec )
	{	croak "$me: Invalid project type '$type'.\n";
	}

	#Do NOT adjust balances unless project is metered
	unless ($typerec->is_metered)
	{	warn "Ignoring adjust_balances on unmetered proj $project at "
			if $adj_bals;
		$adj_bals = 0;
	}

	#Remove user associations in sacctmgr as indicated by project type

	my ($acctname, $error);
	my $acctname_normal = $project;
	my $acctname_high = "${project}-hi";

	if ($typerec->has_normal_priority )
	{	#delete normal priority account
		#We delete for all partitions
		$acctname = $acctname_normal;

		$error = $class->sacctmgr_del_user_assoc_from_clus_acct_part(
			user => $user,
			account => $acctname,
			cluster => $cluster,
			adjust_balances => $adj_bals,
			hpc => $hpc,
		);
		
		croak "$me: Error deleting $user from $acctname: " .
			"$error at " if $error;
	}

	if ($typerec->has_high_priority )
	{	#delete hi priority account
		$acctname = $acctname_high;

		$error = $class->sacctmgr_del_user_assoc_from_clus_acct_part(
			user => $user,
			account => $acctname,
			cluster => $cluster,
			adjust_balances => $adj_bals,
			hpc => $hpc,
		);
		
		croak "$me: Error deleting $user from $acctname: " .
			"$error at " if $error;
	}

	if ( $typerec->has_special_partition )
	{	#delete special condo account
		$acctname = $project;

		$error = $class->sacctmgr_del_user_assoc_from_clus_acct_part(
			user => $user,
			account => $acctname,
			cluster => $cluster,
			adjust_balances => $adj_bals,
			hpc => $hpc,
		);
		
		croak "$me: Error deleting $user from condo $acctname: " .
			"$error at " if $error;
	}
	return;
}

#-------------------------------------------------------------------
#	Add/remove user from cluster (scheduler tasks)
#-------------------------------------------------------------------

sub sched_add_user_to_hpc_cluster($@)
#Adds specified user from the specified HPC cluster.
#
#Nothing really to do in Slurm
#
#Mainly for compatibility with other schedulers
#
#Takes key=>value pairs, recognizing:
#       hpc => Glue::HPCC::Cluster instance, REQUIRED
#       user => username of user to remove from cluster. REQUIRED
#       reason => reason user being removed, REQUIRED
#	FIXME => operate in FIXME mode.
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_add_user_to_hpc_cluster';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $reason = delete $args{reason};
	my $FIXME = delete $args{FIXME};

	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'user' at " unless $user;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	return;
}

sub sched_remove_user_from_hpc_cluster($@)
#Removes specified user from the specified HPC cluster.
#
#Nothing really to do in Slurm, but we will barf if any associations found
#for the user (unless in dryrun mode, in which case just warn)
#
#Mainly for compatibility with other schedulers
#
#Takes key=>value pairs, recognizing:
#       hpc => Glue::HPCC::Cluster instance, REQUIRED
#       user => username of user to remove from cluster. REQUIRED
#       reason => reason user being removed, REQUIRED
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_remove_user_from_hpc_cluster';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $reason = delete $args{reason};

	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'user' at " unless $user;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $sacctmgr = $class->sacctmgr;
	my $cluster = $hpc->db_clustername;
	my @where = ( cluster => $cluster, user => $user );
	my $assocs = $SAAssoc->sacctmgr_list($sacctmgr, @where);
	croak "$me: Error listing associations for $user: $assocs at "
		if $assocs && ! ref($assocs);
	
	if ( scalar(@$assocs) )
	{	if ( $class->dryrun)
		{	warn "$me: User $user in cluster $cluster still has " .
				"associations in Slurm: assuming these would " .
				"have been removed but for DRYRUN at ";
		} else
		{ 	croak "$me: Unable to remove user $user from cluster " .
				"$cluster:  associations in Slurm still " .
				"exist.\n" 
		}
	}
	return;
}

#-------------------------------------------------------------------
#	Add/remove user from project (scheduler tasks)
#-------------------------------------------------------------------

sub sched_add_user_to_project_in_cluster($@)
#Does the slurm scheduler specific tasks for adding 
#specified user to specified project in specified cluster.
#
#Takes  key => value pairs,
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	user => username to add, REQUIRED
#	project => project to add to, REQUIRED
#	projtype => project type (name), REQUIRED e.g. coop, condo
#	reason => reason for logs, REQUIRED
#	highpriority => should user get access to condo's hi-prio QoS
#		Only valid on projects of type 'condo'
#	defaultaccount => if given, sets user default account to this
#	FIXME: if true, operate in repair mode.
#Cluster is specified via $hpc
#
#This handles the scheduler side of things, and is usually called from
#add_user_to_project_in_cluster from scheduler independent base class.
#
#Returns false on success.  On error, either croaks or returns true (error
#message)
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_add_user_to_project_in_cluster';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $project = delete $args{project};
	my $projtype = delete $args{projtype};
	my $reason = delete $args{reason};
	my $highpri = delete $args{highpriority};
	my $defacct = delete $args{defaultaccount};
	my $FIXME = delete $args{FIXME};
	my $cluster = $hpc->db_clustername;

	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'user' at " unless $user;
	croak "$me: Missing required parameter 'project' at " unless $project;
	croak "$me: Missing required parameter 'projtype' at " unless $projtype;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $error = $class->sacctmgr_add_user_to_project_in_cluster(
		hpc => $hpc,
		user => $user,
		project => $project,
		type => $projtype,
		highpriority => $highpri,
		FIXME => $FIXME,
	);
	croak "$me: Error adding $user to $project in $cluster: $error at "
		if $error;
	$GHPLog->add_log_entry( $hpc, 
		action_code => 'add-user-to-project',
		person => $user,
		project => $project,
		cluster => $cluster,
		description => "Sacctmgr stuff to add user $user to $project",
		reason => $reason,
	);

	$class->set_default_account_for_user(
		hpc => $hpc,
		user => $user,
		account => $defacct,
		reason => $reason,
	) if $defacct;

	return;
}

sub sched_del_user_from_project_in_cluster($@)
#Does the slurm scheduler specific tasks for adding 
#specified user to specified project in specified cluster.
#
#Takes  key => value pairs,
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	user => username to delete, REQUIRED
#	project => project to delete from, REQUIRED
#	projtype => project type (name), REQUIRED e.g. coop, condo
#	reason => reason for logs, REQUIRED
#	adjust_balances => adjust proj/account balances after delete.  See below
#Cluster is specified via $hpc
#
#This handles the scheduler side of things, and is usually called from
#delete_user_from_project_in_cluster from scheduler independent base class.
#
#adjust_balances: When user deleted from Slurm account, his usage against that account
#is also deleted, and account's balance goes up.  If adjust_balances is set, the user's
#usage against the project's accounts will be deducted after deletion.
#Returns false on success.  On error, either croaks or returns true (error
#message)
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_del_user_from_project_in_cluster';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $project = delete $args{project};
	my $projtype = delete $args{projtype};
	my $reason = delete $args{reason};
	my $adj_bals = delete $args{adjust_balances};
	my $cluster = $hpc->db_clustername;

	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'user' at " unless $user;
	croak "$me: Missing required parameter 'project' at " unless $project;
	croak "$me: Missing required parameter 'projtype' at " unless $projtype;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $error = $class->sacctmgr_del_user_from_project_in_cluster(
		hpc => $hpc,
		user => $user,
		project => $project,
		type => $projtype,
		adjust_balances => $adj_bals,
	);
	croak "$me: Error deleting $user from $project in $cluster: $error at "
		if $error;
	$GHPLog->add_log_entry( $hpc, 
		action_code => 'del-user-from-project',
		person => $user,
		project => $project,
		cluster => $cluster,
		description => "Sacctmgr stuff to delete user $user from $project",
		reason => $reason,
	);
	return;
}

#-------------------------------------------------------------------
#	Generic list users
#-------------------------------------------------------------------

sub sacctmgr_list_users($@)
#Returns list ref of names of users known to sacctmgr.  Possibly
#restricted to a cluster and/or account(s).
#Takes key=>value pairs
#	cluster => cluster name
#	account => either scalar string of account name, or
#		list ref of account names.
#
#Actually, we just look at associations.  Won't see an user w/o
#associations.  But an user w/o associations is pretty worthless
{	my $class = shift;
	my %args = @_;
	my $me = 'sacctmgr_list_users';

	#Validate parameters
	my $cluster = delete $args{cluster};
	my $accounts = delete $args{account};
	$class->_no_extraneous_args($me, %args);

	my @where = ();
	push @where, cluster => $cluster if $cluster;
	push @where, account => $accounts 
		if $accounts && ! ref($accounts);
	$accounts = [ $accounts ] unless 
		$accounts && ref($accounts) eq 'ARRAY';

	my $sacctmgr = $class->sacctmgr;
	my $assocs = $SAAssoc->sacctmgr_list($sacctmgr, @where);

	#Filter on accounts
	my @users = ();
	if ( scalar (@$accounts) )
	{	foreach my $acct (@$accounts)
		{	my @tmp = grep { $_->account eq $acct } @$assocs;
			@tmp = map { $_->user } @tmp;
			@tmp = grep { $_ } @tmp; #Drop user=undef lines
			push @users, @tmp;
		}
	} else
	{	@users = map { $_->user } @$assocs;
		@users = grep { $_ } @users;
	}

	my %temp = map { $_ => undef } @users;
	@users = sort ( keys %temp );

	return [ @users ];
}


sub sched_list_users($@)
#Returns list ref of names of users known to scheduler.  Possibly
#restricted to a cluster and/or account(s).
#Takes key=>value pairs
#
#Just a wrapper for sacctmgr_list_users
#	cluster => cluster name
#	account => either scalar string of account name
{	my $class = shift;
	my %args = @_;
	my $me = 'sched_list_users';

	#Validate parameters
	my $cluster = delete $args{cluster};
	my $accounts = delete $args{account};
	$class->_no_extraneous_args($me, %args);

	return $class->sacctmgr_list_users(
		cluster => $cluster,
		account => $accounts,
	);
}



1;
__END__

=head1 NAME

Glue::HPCC::ClusterUtil::Slurm::User - User management under Slurm

=head1 SYNOPSIS

  ...


=head1 DESCRIPTION

This provides various routines for managing users in an HPC 
environment.  This package defines the Slurm specific stuff, but it
inherits the general, scheduler independent stuff from
B<Glue::HPCC::ClusterUtil::User>, so it should be called by various scripts.
See also B<Glue::HPCC::ClusterUtil::Project> for project/account management.

Typically, a script will get the B<Glue::HPCC::Cluster> object for the 
specific cluster (using CLUSTER in hostinfo to determine the cluster).
It then gets a B<Glue::HPCC::ClusterUtil> object from that using the
B<get_hpcUtil> method, which should be the version for the specified
scheduler.  The B<user> method on that object should return the class
name for this package.

All of the routines defined in this package are class routines.  A minimal
constructor is inherited if desired, but there is no instance data, etc.
Routines are packaged as an OO class to keep namespaces clean and allow
inheritence.

Significant routines are:

=over 4

=item B<verbose>

Gets/sets the verbosity level.  When setting, propagates to all classes
we depend on.

=item B<dryrun>

Gets/sets the dryrun level.  When setting, propagates to all classes
we depend on.

=item B<get_default_account_for_user>

Returns the default account for the specified user.

=item B<set_default_account_for_user>

Sets the default account for the specified user.

=item B<sacctmgr_add_user_assoc_to_clus_acct_part>

Adds user associations in Slurm for given account, partition, cluster
and optionally QoSes.  Does NOT log.

=item B<sacctmgr_del_user_assoc_from_clus_acct_part>

Deletes user associations in Slurm for given account, partition and cluster.
Does NOT log.

=item B<sacctmgr_add_user_to_project_in_cluster>

Adds associations for user for accounts related to the specified project.
Based on flags in ProjectType record.  Does NOT log.

=item B<sacctmgr_del_user_from_project_in_cluster>

Removes associations for user for accounts related to the specified project.
Based on flags in ProjectType record.  Does NOT log.

=item B<sched_add_user_to_project_in_cluster>

Does the slurm scheduler specific tasks for adding 
specified user to specified project in specified cluster.
Basically just calls B<sacctmgr_add_useR_to_project_in_cluster> and logs.

=item B<sched_del_user_from_project_in_cluster>

Does the slurm scheduler specific tasks for deleting
specified user to specified project in specified cluster.
Basically just calls B<sacctmgr_del_user_from_project_in_cluster> and logs.

=back
