#!/usr/local/bin/perl
#
#Base class for Glue::HPCC::ClusterUtil::UserXXX
#where XXX is a scheduler type
#
#Does the heavy lifting for Cluster management,
#User management related stuff
#
#This handles the scheduler-independent stuff.
#
#All methods are class methods, 
#OO only to keep namespace clean and for inheritance

package Glue::HPCC::ClusterUtil::User;
use strict;
use warnings;
use base qw(Glue::HPCC::ClusterUtil::Base);
use Carp qw(carp croak);

use Glue::HPCC::Project::Log;
use Glue::HPCC::Project::Person;
use Glue::HPCC::Project::PersonProject;
use Glue::HPCC::Project::Project;
use Glue::HPCC::Project::ProjectType;

use Glue::HPCC::Utils::NetGroup qw(
	user_in_netgroup 
	add_user_to_netgroup
	remove_user_from_netgroup
);
use Glue::HPCC::Utils::ShellUtils qw(run_system_command);
use Glue::HPCC::Utils::UnixGroup qw(
	unix_group_exists
	user_in_unix_group
	add_user_to_unix_group
	remove_user_from_unix_group
	set_users_primary_unix_group
);
use Glue::HPCC::Utils::UserAccounts qw(	user_exists 
					uid_from_username
					username_from_uid
				);
use Glue::HPCC::Utils::FileUtils qw(files_in_dir_owned_by_uid);

#For mailing list management
use Glue::HPCC::Utils::MailUtils qw(send_email);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $DEBUG;
$DEBUG=1;
my $VERBOSE;
my $DRYRUN;

my $GHPLog = 'Glue::HPCC::Project::Log';
my $GHPPers = 'Glue::HPCC::Project::Person';
my $GHPPerPrj = 'Glue::HPCC::Project::PersonProject';
my $GHPProj = 'Glue::HPCC::Project::Project';
my $GHPPType = 'Glue::HPCC::Project::ProjectType';

my $MK_NEWHOME = '/local/etc/mk_newhome';



#-------------------------------------------------------------------
#	Common Methods: verbosity and dryrun mode
#-------------------------------------------------------------------

#Not needed for Glue::HPCC:Utils::NetGroup and
#Glue::HPCC::Utils::UnixGroup as those are controlled
#by ShellUtils::run_system_command

my @_subord_classes_for_dryrun = (
	"Glue::HPCC::Project::Log",
	"Glue::HPCC::Project::Person",
	"Glue::HPCC::Project::PersonProject",
	"Glue::HPCC::Project::Project",
	"Glue::HPCC::Project::ProjectType",
	"Glue::HPCC::Utils::ShellUtils",
	#Start with a ! means dryrun/verbose should not get called with
	#OO semantics
	"!Glue::HPCC::Utils::MailUtils",
);

sub verbose($;$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		my $sclass;
		foreach $sclass (@_subord_classes_for_dryrun)
		{	if ( $sclass =~ /^!/ )
			{	#Leading ! means don't call as OO method
				my $tmpcl = $sclass;
				$tmpcl =~ s/^!//;
				#OK, this is ugly
				no strict 'refs';
				&{$tmpcl . '::verbose'}($val);
			} else
			{	#Use standard OO semantics
				$sclass->verbose($val);
			}
		}
		$VERBOSE=$val;
	}
	return $VERBOSE;
}

sub dryrun($$)
#Sets verbosity on all of our subordinate modules
{	my $class = shift;
	my $val = shift;

	if ( defined $val )
	{	$val = $val?1:0;
		my $sclass;
		foreach $sclass (@_subord_classes_for_dryrun)
		{	if ( $sclass =~ /^!/ )
			{	#Leading ! means don't call as OO method
				my $tmpcl = $sclass;
				$tmpcl =~ s/^!//;
				#OK, this is ugly
				no strict 'refs';
				&{$tmpcl . '::dryrun'}($val);
			} else
			{	#Use standard OO semantics
				$sclass->dryrun($val);
			}
		}
		$DRYRUN=$val;
	}
	return $DRYRUN;
}

#-------------------------------------------------------------------
#	Method to get our project counterpart
#-------------------------------------------------------------------

sub project_class($)
{	my $class = shift;
	$class = ref($class) if ref($class);
	$class =~ s/User$/Project/;
	return $class;
}

#-------------------------------------------------------------------
#	Default user group/gid:  If primary group matches, change it
#-------------------------------------------------------------------

sub default_users_unix_group_name($)
{	my $class = shift;
	return 'users';
}

sub default_users_unix_group_gid($)
{	my $class = shift;
	return 100;
}

#-------------------------------------------------------------------
#	Common person PDB utilities 
#-------------------------------------------------------------------

sub _add_person_to_pdb($@)
#Does actual insert of specified person to person DB table
#Takes args:
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	username: username to add, REQUIRED
#	fullname: fullname of user, defaulted from hesiod pwd entry
#	uid: uid of user, defaulted from hesiod pwd entry
#	email: email of user, defaults to $username@umd.edu
#	active: boolean, true if active.  Must set to 0 (NOT undef) to turn off
#	me: method name for error messages
#
#Caller is expected to log the change
#Must not already exist in DB
#Returns false on success.  On error, croaks or returns true (error msg)
{	my $class = shift;
	my %args = @_;
	my $realme = '_add_person_to_pdb';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($realme, $hpc);
	my $me = delete $args{me} || $realme;

	my $username = delete $args{username};
	my $fullname = delete $args{fullname};
	my $uid = delete $args{uid};
	my $email = delete $args{email};
	my $active = delete $args{active};

	$class->_no_extraneous_args($realme, %args);

	croak "$realme: Missing required argument username at " 
		unless $username;
	my $pwent = user_exists($username);
	croak "$me: Invalid user $username, not found in hesiod, at "
		unless $pwent;
	my @pwent = split /:/, $pwent;

	$uid = $pwent[2] unless defined $uid;
	$email = "${username}\@umd.edu" unless $email;
	unless ( $fullname )
	{	$fullname = $pwent[4];
		$fullname =~ s/,.*$// if $fullname;
		$fullname = $username unless $fullname;
	}
	$active = 1 unless defined $active;

	my @user_fields = 
	(	fullname => $fullname,
		uid => $uid,
		email => $email,
		active => $active,
	);

	my $error = $GHPPers->pdb_insert($hpc, 
			username=>$username, @user_fields);
	croak "$me: Error inserting user $username: $error at " if $error;

	my $user = $GHPPers->person_by_username($hpc, $username);
	croak "$me: user $username not found after insert at "
		unless $user || $class->dryrun;
	croak "$me: Error looking up user $username after insert at "
		unless ref($user) || $class->dryrun;
	return;
}
		
sub _update_person_in_pdb($@)
#Does actual insert of specified person to person DB table
#Takes args:
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	username: username to update, REQUIRED
#	fullname: fullname of user
#	uid: uid of user
#	email: email of user
#	active: boolean, true if active.  Must set to 0 (NOT undef) to turn off
#	me: method name for error messages
#
#Caller is expected to log the change
#Must already exist in DB
#Returns false on success.  On error, croaks or returns true (error msg)
{	my $class = shift;
	my %args = @_;
	my $realme = '_update_person_in_pdb';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($realme, $hpc);
	my $me = delete $args{me} || $realme;

	my $username = delete $args{username};
	my $fullname = delete $args{fullname};
	my $uid = delete $args{uid};
	my $email = delete $args{email};
	my $active = delete $args{active};

	$class->_no_extraneous_args($realme, %args);

	croak "$realme: Missing required argument username at " 
		unless $username;

	my $user = $GHPPers->person_by_username($hpc, $username);
	croak "$me: No user $username found at " unless $user;
	croak "$me: Error looking user $user: $user at " unless $user;

	my @user_fields = ();
	push @user_fields, fullname => $fullname 
		if $fullname && $fullname ne $user->fullname;
	push @user_fields, uid => $uid if $uid && $uid != $user->uid;
	push @user_fields, email => $email 
		if $email && $email ne $user->email;
	my $tmpactdiff = 0;
	$tmpactdiff = 1 if $active && ! $user->active;
	$tmpactdiff = 1 if ! $active && $user->active;
	push @user_fields, active => $active if defined $active && $tmpactdiff;

	croak "$me: Nothing to update for user $username at "
		unless @user_fields;

	my $error = $user->pdb_update_me($hpc, @user_fields);
	croak "$me: Error updating user $username: $error at " if $error;
	#No need to check for existance, if doesn't exist update would have
	#failed

	return;
}
		
sub add_person_to_pdb($@)
#Adds specified person to the person database table
#Takes key => value pairs
#	username: username to add, REQUIRED
#	fullname: fullname of user, defaulted from hesiod pwd entry
#	uid: uid of user, defaulted from hesiod pwd entry
#	email: email of user, defaults to $username@umd.edu
#	active: boolean, default 1 (explicitly set to 0 for false)
#We also recognize:
#	hpc: Glue::HPCC::Cluster object, REQUIRED
#	FIXME: if set, do not abort if user exists in DB, just try
#		to fix it.
#	reason: reason text for logs, REQUIRED
#
#User record is always added as 'active'; if already exists and not 'active',
#in FIXME mode will always activate him
#
#Returns false on success.  On error, croaks or returns true (error msg)
{	my $class = shift;
	my %args = @_;
	my $me = 'add_person_to_pdb';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $username = delete $args{username};
	my $fullname = delete $args{fullname};
	my $uid = delete $args{uid};
	my $email = delete $args{email};
	my $FIXME = delete $args{FIXME};
	my $reason = delete $args{reason};
	my $active = delete $args{active};
	croak "$me: Missing required argument reason at " unless $reason;
	croak "$me: Missing required argument username at " unless $username;
	$class->_no_extraneous_args($me, %args);
	$active = 1 unless defined $active;

	my $error;
	my $tmp;
	my $user = $GHPPers->person_by_username($hpc, $username);
	croak "$me: Error looking up user $username: $user at "
		if $user && ! ref($user);
	if ( $user )
	{   	croak "$me: Failed, user $username already in table at "
			unless $FIXME;
	    	#FIXME mode, update him
	    	my @update = ();
	    	push @update, fullname => $fullname if $fullname;
		push @update, uid => $uid if $uid;
		push @update, email => $email if $email;
		$tmp = 0;
		$tmp = 1 if $active && ! $user->active;
		if ( $user->active && ! $active )
		{	carp "$me: Refusing to disable existing user $username"
				. " at ";
		}
		push @update, active=>$active if $tmp;

		return unless @update;
		$error = $class->update_person_in_pdb( hpc =>$hpc, 
			username => $username, reason=>$reason,
			@update);
		croak "$me: Error updating user $username: $error at " 
			if $error;
		return;
	}

	my $pwent = user_exists($username);
	croak "$me: Invalid user $username, not found in hesiod, at "
		unless $pwent;
	my @pwent = split /:/, $pwent;

	$uid = $pwent[2] unless defined $uid;
	$email = "${username}\@umd.edu" unless $email;
	unless ( $fullname )
	{	$fullname = $pwent[4];
		$fullname =~ s/,.*$// if $fullname;
		$fullname = $username unless $fullname;
	}

	my @user_fields = 
	(	fullname => $fullname,
		uid => $uid,
		email => $email,
		active => $active,
	);

	$error = $GHPPers->pdb_insert($hpc, 
		username=>$username, @user_fields);
	croak "$me: Error inserting user $username: $error at " if $error;

	$user = $GHPPers->person_by_username($hpc, $username);
	croak "$me: user $username not found after insert at "
		unless $user || $class->dryrun;
	croak "$me: Error looking up user $username after insert at "
		unless ref($user) || $class->dryrun;
		
	$error = $GHPLog->add_log_entry($hpc,
		action_code => "add-user-to-pdb",
		person => $username,
		reason => $reason,
		tablename => 'person',
		description => "add user $username to PDB",
	);
	return;
}

sub update_person_in_pdb($@)
#Adds specified person to the person database table
#Takes key => value pairs
#	username: username to add, REQUIRED
#	fullname: fullname of user
#	uid: uid of user
#	email: email of user
#	active: boolean, true if active.  Must set to 0 (NOT undef) to turn off
#We also recognize:
#	hpc: Glue::HPCC::Cluster object, REQUIRED
#	reason: reason text for logs, REQUIRED
#	QUIET: don't complain if nothing to update
#
#fullname, uid, and email updated only if provided AND differ from what
#is already in DB
#
#Returns false on success.  On error, croaks or returns true (error msg)
{	my $class = shift;
	my %args = @_;
	my $me = 'update_person_in_pdb';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $username = delete $args{username};
	my $fullname = delete $args{fullname};
	my $uid = delete $args{uid};
	my $email = delete $args{email};
	my $active = delete $args{active};
	my $reason = delete $args{reason};
	my $quiet = delete $args{QUIET};
	croak "$me: Missing required argument reason at " unless $reason;
	croak "$me: Missing required argument username at " unless $username;

	$class->_no_extraneous_args($me, %args);

	my $user = $GHPPers->person_by_username($hpc, $username);
	$user = $GHPPers->person_by_username($hpc, $username);
	croak "$me: Error looking up user $username: $user at "
		if $user && ! ref($user);
	croak "$me: Attempt to update non-existant user $username at "
		unless $user;

	my @user_fields = ();
	push @user_fields, fullname=>$fullname 
		if $fullname && $fullname ne $user->fullname;
	push @user_fields, email=>$email if $email && $email ne $user->email;
	push @user_fields, uid=>$uid if defined $uid && $uid != $user->uid;
	my $active_changed = 0;
	$active_changed = 1 if $active && ! $user->active;
	$active_changed = 1 if $user->active && ! $active;
	push @user_fields, active => $active if defined $active && $active_changed;

	unless (@user_fields)
	{	carp "$me: called with nothing to update at " unless $quiet;
		return;
	}

	my $error = $user->pdb_update_me($hpc, @user_fields);
	croak "$me: Error updating user $username: $error at "
		if $error;

	my %update = @user_fields;
	foreach my $key (keys %update)
	{	my $val = $update{$key};
		my $old = $user->$key;
		$error = $GHPLog->add_log_entry($hpc,
				action_code => "update-user-in-pdb",
				person => $username,
				reason => $reason,
				tablename => 'person',
				fieldname => $key,
				oldval => $old,
				newval => $val,
				description => "updating user $username in PDB",
			);
	}
	return;
}

sub add_or_update_person_to_pdb($@)
#Adds specified person to the person database table, if needed
#Takes key => value pairs
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	username: username to add, REQUIRED
#	fullname: fullname of user, defaulted from hesiod pwd entry
#	uid: uid of user, defaulted from hesiod pwd entry
#	email: email of user, defaults to $username@umd.edu
#	active: boolean, true if active.  Must set to 0 (NOT undef) to turn off
#	reason: reason text for logs, REQUIRED
#
#If user does not exist in PDB, we add, with all fields above, defaulting 
#as needed.
#If user does exist, we update fullname, uid, and/or email ONLY if 
#provided and different then what we already have
#
#Returns false on success.  On error, croaks or returns true (error msg)
{	my $class = shift;
	my %args = @_;
	my $me = 'add_or_update_person_to_pdb';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $username = delete $args{username};
	my $fullname = delete $args{fullname};
	my $uid = delete $args{uid};
	my $email = delete $args{email};
	my $active = delete $args{active};
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	croak "$me: Missing required argument username at " unless $username;
	$class->_no_extraneous_args($me, %args);

	my $user = $GHPPers->person_by_username($hpc, $username);
	croak "$me: Error looking up user $username: $user at "
		if $user && ! ref($user);

	my @args = ( 	hpc => $hpc,
			reason => $reason,
			username => $username,
			fullname=>$fullname, 
			uid=>$uid, 
			email=>$email,
			active => $active,
	);
	return $class->update_person_in_pdb(@args, QUIET=>1) if $user;
	return $class->add_person_to_pdb(@args);
}

sub activate_person_in_db($@)
#Activates user in PDB
#(i.e. sets active flag to true)
#User must already exist in PDB
#
#Takes key => value pairs, recognizing:
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	user: username to activate, REQUIRED
#	reason: for logging, REQUIRED
#
#Returns false on success.  On error, croaks or returns true (error msg)
{	my $class = shift;
	my %args = @_;
	my $me = "activate_person_in_db";

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $reason = delete $args{reason};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'user'" unless $user;
	croak "$me: Missing required parameter 'reason'" unless $reason;

	my $error = $class->update_person_in_pdb( 
		hpc => $hpc, 
		username => $user, 
		reason => $reason,
		active => 1 );
	return $error;
}

sub deactivate_person_in_db($@)
#Deactivates user in PDB
#(i.e. sets active flag to false)
#User must already exist in PDB
#
#Takes key => value pairs, recognizing:
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	user: username to activate, REQUIRED
#	reason: for logging, REQUIRED
#
#Returns false on success.  On error, croaks or returns true (error msg)
{	my $class = shift;
	my %args = @_;
	my $me = "deactivate_person_in_db";

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $reason = delete $args{reason};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'user'" unless $user;
	croak "$me: Missing required parameter 'reason'" unless $reason;

	my $error = $class->update_person_in_pdb( 
		hpc => $hpc, 
		username => $user, 
		reason => $reason,
		active => 0 );
	return $error;
}

	
sub _set_user_role_in_project($@)
#Takes username and project name, and role, and adds/deletes username to proj 
#in that role according to $mode.  But only in PDB
#Role should be one of 'member', 'poc', or 'coordinator'.  
#
#Args are given in key => value pairs, recognizing:
#	hpc:  REQUIRED
#	username: REQUIRED
#	project: REQUIRED
#	role: REQUIRED
#	mode: REQUIRED
#	me
#
#Calling routine is expected to log anything
#Returns false on success.  On error, aborts or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = delete $args{me} || '_set_user_role_in_project';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $username = delete $args{username};
	croak "$me: Missing required argument username at " unless $username;
	my $projname = delete $args{project};
	croak "$me: Missing required argument project at " unless $projname;
	my $role = delete $args{role};
	croak "$me: Missing required argument role at " unless $role;
	$role = lc $role;
	my $mode = delete $args{mode};
	croak "$me: Missing required argument mode at " unless defined $mode;
	$class->_no_extraneous_args($me, %args);

	my $verb = $mode?'set':'unset';

	my $user = $GHPPers->person_by_username($hpc, $username);
	if ( $class->dryrun && ! $user )
	{	#Avoid bogus errors in dryrun mode
		print STDERR "[DRYRUN] $verb user $username as $role in " .
			"$projname\n";
		return;
	}
	croak "$me: Invalid username $username, not found in PDB, at "
		unless $user;
	croak "$me: Error looking up user $username: $user at "
		unless ref($user);

	my $proj = $GHPProj->project_by_name($hpc, $projname);
	if ( $class->dryrun && ! $proj )
	{	#Avoid bogus errors in dryrun mode
		print STDERR "[DRYRUN] $verb user $username as $role in " .
			"$projname\n";
		return;
	}
	croak "$me: Invalid project $projname, not found in PDB at "
		unless $proj;
	croak "$me: Error looking up project $projname: $proj at "
		unless ref($proj);

	my @set_fields = ();
	if ( $role eq 'member' )
	{	push @set_fields, is_member => $mode;
	} elsif ( $role eq 'poc' )
	{	push @set_fields, is_poc => $mode;
	} elsif ( $role eq 'coordinator' )
	{	push @set_fields, is_coordinator => $mode;
	} else
	{	croak "$me: Invalid role $role at ";
	}

	my $error;

	my $prjperson = $GHPPerPrj->get_project_person_by_names($hpc,
		$projname, $username);
	if ( $prjperson )
	{	if (ref($prjperson) )
		{	$error = $prjperson->pdb_update_me($hpc,@set_fields);
			croak "$me: Error ${verb}ting user $username in " .
				"role $role for proj $projname: $error at "
				if $error;
		} else
		{	croak "$me: Error looking up proj $projname, user " .
				"$username: $error at ";
		}
	} else
	{	#Return if mode is false; no need to add a record with unset
		return unless $mode;
		$error = $GHPPerPrj->pdb_insert($hpc, 
			prj_pkey => $proj->prj_pkey,
			ppl_pkey => $user->ppl_pkey,
			@set_fields,
		);
		croak "$me: Error inserting row: $projname/$username as $role: "
			. "$error at " if $error;
	}
	return;
}

sub pdb_add_member_to_project($$$$)
#Adds user to project as member, in PDB only
#Takes key => value pairs as arguments
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	user => username, REQUIRED
#	project => project name, REQUIRED
#	reason => for logging purposes, REQUIRED
#	FIXME => operate in repair mode.  Ignored, always in repair mode
#
#Returns false on success.  On error, aborts or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'pdb_add_member_to_project';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $username = delete $args{user};
	croak "$me: Missing required argument user at " unless $username;
	my $project = delete $args{project};
	croak "$me: Missing required argument project at " unless $project;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	my $FIXME = delete $args{FIXME};
	$class->_no_extraneous_args($me, %args);

	my $role = 'member';
	my $cluster = $hpc->db_clustername;
	my $error = $class->_set_user_role_in_project(
		hpc => $hpc, username => $username,
		project => $project, role => $role,
		mode => 1, me => $me );
	croak "$me: Error adding user $username as $role of $project: " .
		"$error at " if $error;

	$error = $GHPLog->add_log_entry($hpc,
		action_code => "add-user-as-$role",
		person => $username,
		project => $project,
		tablename => 'project',
		cluster => $cluster,
		reason => $reason,
		description => "add user $username as $role of project $project in PDB",
	);
	croak "$me: error adding log entry add-user-as-$role, $username, $project, $reason: $error at "
		if $error;
	return;
}

sub pdb_add_poc_to_project($$$$)
#Adds user to project as poc, in PDB only
#Takes key => value pairs as arguments
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	user => username, REQUIRED
#	project => project name, REQUIRED
#	reason => for logging purposes, REQUIRED
#
#Returns false on success.  On error, aborts or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'add_poc_to_project';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $user = delete $args{user};
	croak "$me: Missing required argument user at " unless $user;
	my $project = delete $args{project};
	croak "$me: Missing required argument project at " unless $project;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $role = 'poc';
	my $cluster = $hpc->db_clustername;
	my $error = $class->_set_user_role_in_project(
		hpc => $hpc, username => $user,
		project => $project, role => $role,
		mode => 1, me => $me );
	croak "$me: Error adding user $user as $role of $project: " .
		"$error at " if $error;

	$error = $GHPLog->add_log_entry($hpc,
		action_code => "add-user-as-$role",
		person => $user,
		project => $project,
		tablename => 'project',
		cluster => $cluster,
		reason => $reason,
		description => "add user $user as $role of project $project in PDB",
	);
	croak "$me: error adding log entry add-user-as-$role, $user, $project, $reason: $error at "
		if $error;
	return;
}

sub pdb_remove_member_from_project($$$$)
#Removes user from project as member, in PDB only
#Takes key => value pairs as arguments
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	user => username, REQUIRED
#	project => project name, REQUIRED
#	reason => for logging purposes, REQUIRED
#
#Returns false on success.  On error, aborts or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'remove_member_from_project';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $user = delete $args{user};
	croak "$me: Missing required argument user at " unless $user;
	my $project = delete $args{project};
	croak "$me: Missing required argument project at " unless $project;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $role = 'member';
	my $cluster = $hpc->db_clustername;
	my $error = $class->_set_user_role_in_project(
		hpc => $hpc, username => $user,
		project => $project, role => $role,
		mode => 0, me => $me );
	croak "$me: Error removing user $user as $role of $project: " .
		"$error at " if $error;

	$error = $GHPLog->add_log_entry($hpc,
		action_code => "remove-user-as-$role",
		person => $user,
		project => $project,
		tablename => 'project',
		cluster => $cluster,
		reason => $reason,
		description => "remove user $user as $role of project $project in PDB",
	);
	croak "$me: error adding log entry remove-user-as-$role, $user, $project, $reason: $error at "
		if $error;
	return;
}

sub pdb_remove_poc_from_project($$$$)
#Removes user from project as poc, in PDB only
#Takes key => value pairs as arguments
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	user => username, REQUIRED
#	project => project name, REQUIRED
#	reason => for logging purposes, REQUIRED
#
#Returns false on success.  On error, aborts or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'remove_poc_from_project';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $user = delete $args{user};
	croak "$me: Missing required argument user at " unless $user;
	my $project = delete $args{project};
	croak "$me: Missing required argument project at " unless $project;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $role = 'poc';
	my $cluster = $hpc->db_clustername;
	my $error = $class->_set_user_role_in_project(
		hpc => $hpc, username => $user,
		project => $project, role => $role,
		mode => 0, me => $me );
	croak "$me: Error removing user $user as $role of $project: " .
		"$error at " if $error;

	$error = $GHPLog->add_log_entry($hpc,
		action_code => "remove-user-as-$role",
		person => $user,
		project => $project,
		tablename => 'project',
		cluster => $cluster,
		reason => $reason,
		description => "remove user $user as $role of project $project in PDB",
	);
	croak "$me: error adding log entry remove-user-as-$role, $user, $project, $reason: $error at "
		if $error;
	return;
}

sub _pdb_get_role_users_for_project($@)
#On success, returns a list ref of names of users who have role
#$role in specified project.
#On error, croaks or returns non-ref error
#text.
#Takes key=>value pairs:
#	hpc : Glue::HPCC::Cluster instance, REQUIRED
#	project: name of Project, REQUIRED
#	role: one of 'member', 'poc', 'coordinator', or 'all' REQUIRED
#	inactive: if set, will include inactive people
#	me: masquerade as this method for errors
{	my $class = shift;
	my %args= @_;
	my $realme = '_pdb_get_role_users_for_project';

	#Validate parameters
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $role = delete $args{role};
	my $inactive = delete $args{inactive};
	my $me = delete $args{me} || $realme;
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$realme: Missing required parameter 'role' at " unless $role;

	my $meth;
	if ( $role eq 'member' )
	{	$meth = 'get_members_of_project';
	} elsif ( $role eq 'coordinator' || $role eq 'coord' )
	{	$meth = 'get_coords_of_project';
	} elsif ( $role eq 'poc' )
	{	$meth = 'get_pocs_for_project';
	} elsif ( $role eq 'all' )
	{	$meth = 'get_project_persons_by_username';
	} else
	{	croak "$realme: Invalid parameter role='$role' at ";
	}

	my $list = $GHPPerPrj->$meth($hpc, $projname);
	my @ppl_pkeys = map { $_->ppl_pkey } @$list;
	my %temp = map { $_ => undef } @ppl_pkeys;
	@ppl_pkeys = keys %temp;

	my @users = map { $GHPPers->person_by_pkey($_) } @ppl_pkeys;
	unless ( $inactive )
	{	@users = grep { $_->active } @users;
	}

	return [ @users ];
}

sub pdb_get_members_for_project($@)
#On success, returns a list ref of names of users who are members 
#of the specified project.  On error, croaks or returns non-ref error
#text.
#Takes key=>value pairs:
#	hpc : Glue::HPCC::Cluster instance, REQUIRED
#	project: name of Project, REQUIRED
#	inactive: if set, will include inactive people
{	my $class = shift;
	my %args = @_;
	my $me = 'pdb_get_members_for_project';

	#Validate args
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $inactive = delete $args{inactive};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	$class->_no_extraneous_args($me, %args);
	my $role = 'member';

	return $class->_pdb_get_role_users_for_project(
		hpc => $hpc,
		project => $projname,
		inactive => $inactive,
		role => $role,
	);
}

sub pdb_get_pocs_for_project($@)
#On success, returns a list ref of names of users who are pocs
#of the specified project.  On error, croaks or returns non-ref error
#text.
#Takes key=>value pairs:
#	hpc : Glue::HPCC::Cluster instance, REQUIRED
#	project: name of Project, REQUIRED
#	inactive: if set, will include inactive people
{	my $class = shift;
	my %args = @_;
	my $me = 'pdb_get_pocs_for_project';

	#Validate args
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $inactive = delete $args{inactive};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	$class->_no_extraneous_args($me, %args);
	my $role = 'poc';

	return $class->_pdb_get_role_users_for_project(
		hpc => $hpc,
		project => $projname,
		inactive => $inactive,
		role => $role,
	);
}

sub pdb_get_coords_for_project($@)
#On success, returns a list ref of names of users who are coord
#of the specified project.  On error, croaks or returns non-ref error
#text.
#Takes key=>value pairs:
#	hpc : Glue::HPCC::Cluster instance, REQUIRED
#	project: name of Project, REQUIRED
#	inactive: if set, will include inactive people
{	my $class = shift;
	my %args = @_;
	my $me = 'pdb_get_coords_for_project';

	#Validate args
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $inactive = delete $args{inactive};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	$class->_no_extraneous_args($me, %args);
	my $role = 'coord';

	return $class->_pdb_get_role_users_for_project(
		hpc => $hpc,
		project => $projname,
		inactive => $inactive,
		role => $role,
	);
}


sub pdb_get_users_for_project($@)
#On success, returns a list ref of names of users who are member/poc/or coord
#of the specified project.  On error, croaks or returns non-ref error
#text.
#Takes key=>value pairs:
#	hpc : Glue::HPCC::Cluster instance, REQUIRED
#	project: name of Project, REQUIRED
#	inactive: if set, will include inactive people
{	my $class = shift;
	my %args = @_;
	my $me = 'pdb_get_users_for_project';

	#Validate args
	my $hpc = delete $args{hpc};
	my $projname = delete $args{project};
	my $inactive = delete $args{inactive};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'project' at " unless $projname;
	$class->_no_extraneous_args($me, %args);
	my $role = 'all';

	return $class->_pdb_get_role_users_for_project(
		hpc => $hpc,
		project => $projname,
		inactive => $inactive,
		role => $role,
	);
}

#-------------------------------------------------------------------
#	Handling shared resources between clusters 
#  Eg: hpcc_shared DB, shared mailing lists
#-------------------------------------------------------------------

sub is_user_in_any_sister_clusters($@)
#Determines if specified user is is any of the listed HPC clusters.
#Used to determine if user should be added/deleted from any resources
#shared between the clusters (e.g. mailing list, shared DB tables, etc).
#
#Returns true (the name of a sister cluster belongs to) if belongs to
#a sister cluster, and false (undef) otherwise.
#
#Returns true (a sister clustername user belongs to) if belongs to a sister
#cluster, and false otherwise.
#Takes key => value pairs for explicit arguments, recognizing:
#	hpc => Glue::HPCC::Cluster instance for cluster we are add/deleting
#		user from, used to skip matching cluster in $clusters.  Optional.
#	clusters => A list ref of sister clusternames to check, REQUIRED
#	user => name of user, REQUIRED
#	role => role for user to check for.  Default is 'ANY'.  If set
#		to 'DB', returns true if user exists in PDB for a sister cluster
#		even if no role (probably meaningless, as likely share PDB
#		with sister clusters).  If 'member', 'poc', or 'coord', requires
#		user to have that role for some project in the sister cluster.
#		If 'ANY', returns true if user has any role for any project
#		in sister cluster.
#	me => impersonate another routine
{	my $class = shift;
	my %args = @_;
	my $realme = __PACKAGE__ . '::is_user_in_any_sister_clusters';

	my $hpc = delete $args{hpc};
	my $clusters = delete $args{clusters};
	my $user = delete $args{user};
	my $role = delete $args{role} || 'ANY';
	my $me = delete $args{me} || $realme;

	$class->_no_extraneous_args($me, %args);
	$class->_validate_hpc($me, $hpc) if $hpc;
	croak "$me: Missing required argument user at " unless $user;
	croak "$me: Missing required argument clusters at " unless $clusters;
	croak "$me: Invalid value '$clusters' for clusters, expecting list ref,  at " 
		unless ref($clusters) eq 'ARRAY';

	my $this_cluster;
	$this_cluster = $hpc->clustername if $hpc;
	#If no sister clusters, then we return false always
	return unless $clusters && ref($clusters) eq 'ARRAY'
		&& scalar(@$clusters);

	SISTERS: foreach my $sister (@$clusters)
	{	next SISTERS if $this_cluster && $sister eq $this_cluster;
		my $sishpc = Glue::HPCC::Cluster->new_from_clustername($sister);
		unless ( $sishpc && ref($sishpc) )
		{	croak "$me: Unable to instantiate cluster for $sister: $sishpc at ";
		}
		#Skip if this is our cluster
		next SISTERS if $this_cluster && $sishpc->clustername eq $this_cluster;

		#First check if user is even in PDB for this cluster
		my $temp = Glue::HPCC::Project::Person->person_by_username(
			$sishpc, $user);
		next SISTERS unless $temp; #Not in PDB, so not in a project
		croak "$me: Error looking up user $user in PDB for cluster " .
			"$sister: $temp at " unless ref($temp); 

		$temp = Glue::HPCC::Project::PersonProject->get_project_persons_by_username($sishpc, $user);
		next SISTERS unless $temp; 
		croak "$me: Error looking up user $user in cluster $sister: " .
			"$temp at " unless $temp && ref($temp) eq 'ARRAY';
		next SISTERS unless scalar(@$temp);#No records, try next cluster

		return $sister if $role eq 'DB';  

		my @temp;
		if ( $role eq 'ANY' )
		{	@temp = grep { $_->is_member || $_->is_poc ||
				$_->is_coordinator } @$temp;
		} elsif ( $role eq 'member' )
		{	@temp = grep { $_->is_member } @$temp;
		} elsif ( $role eq 'poc' )
		{	@temp = grep { $_->is_poc } @$temp;
		} elsif ( $role eq 'coord' )
		{	@temp = grep { $_->is_coordinator } @$temp;
		}

		return $sister if scalar(@temp);
	}
	#We failed to find any sister cluster were user has specified role
	return;
}

sub is_user_in_maillist_sister_cluster($@)
#Determine if specified user currently has access to an HPC cluster
#we share a mailing list with.  E.g., no reason to add to the mailing
#list if already there (although should be safe/noop with listserv), and
#do NOT want to delete if still a member of allocation in other cluster.
#Returns true (a sister clustername user belongs to) if belongs to a sister
#cluster, and false otherwise.
#Takes key => value pairs for explicit arguments, recognizing:
#	hpc => Glue::HPCC::Cluster instance for cluster we are add/deleting
#		user from, REQUIRED
#	user => name of user, REQUIRED
#	role => role for user to check for.  Default is 'ANY'.  If set
#		to 'DB', returns true if user exists in PDB for a sister cluster
#		even if no role (probably meaningless, as likely share PDB
#		with sister clusters).  If 'member', 'poc', or 'coord', requires
#		user to have that role for some project in the sister cluster.
#		If 'ANY', returns true if user has any role for any project
#		in sister cluster.
{	my $class = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::is_user_in_maillist_sister_cluster';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $role = delete $args{role} || 'ANY';

	$class->_no_extraneous_args($me, %args);
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required argument user at " unless $user;

	my $sister_clusters = $hpc->clusters_sharing_mailinglist;
	#If no sister clusters, then we return false always
	return unless $sister_clusters && ref($sister_clusters) eq 'ARRAY'
		&& scalar(@$sister_clusters);

	return $class->is_user_in_any_sister_clusters(
		hpc => $hpc,
		user => $user,
		role => $role,
		clusters => $sister_clusters,
		me => $me,
	);
}

sub is_user_in_pdb_sister_cluster($@)
#Determine if specified user currently has access to an HPC cluster
#we share PDB tables with.  E.g., if removing user from one cluster, we
#need to know if has associations in any clusters sharing PDB tables before
#we de-activate user.
#Returns true (a sister clustername user belongs to) if belongs to a sister
#cluster, and false otherwise.
#Takes key => value pairs for explicit arguments, recognizing:
#	hpc => Glue::HPCC::Cluster instance for cluster we are add/deleting
#		user from, REQUIRED
#	user => name of user, REQUIRED
#	role => role for user to check for.  Default is 'ANY'.  If set
#		to 'DB', returns true if user exists in PDB for a sister cluster
#		even if no role (probably meaningless, as likely share PDB
#		with sister clusters).  If 'member', 'poc', or 'coord', requires
#		user to have that role for some project in the sister cluster.
#		If 'ANY', returns true if user has any role for any project
#		in sister cluster.
{	my $class = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::is_user_in_pdb_sister_cluster';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $role = delete $args{role} || 'ANY';

	$class->_no_extraneous_args($me, %args);
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required argument user at " unless $user;

	my $sister_clusters = $hpc->clusters_sharing_pdb_tables;
	#If no sister clusters, then we return false always
	return unless $sister_clusters && ref($sister_clusters) eq 'ARRAY'
		&& scalar(@$sister_clusters);

	return $class->is_user_in_any_sister_clusters(
		hpc => $hpc,
		user => $user,
		role => $role,
		clusters => $sister_clusters,
		me => $me,
	);
}

	
#-------------------------------------------------------------------
#	Common person utilities : netgroup, homedir, etc
#-------------------------------------------------------------------

sub add_user_to_cluster_netgroup($@)
#Adds user to netgroup for this cluster
#Takes key => value pairs for arguments
#	hpc => Glue::HPCC::Cluster instance, REQUIRED.
#	user => name of user to add, REQUIRED
#	reason => for logging, REQUIRED
#	FIXME => operate in repair mode.  Ignored for this method
#Cluster is from $hpc
{	my $class = shift;
	my %args = @_;
	my $me = 'add_user_to_cluster_netgroup';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $user = delete $args{user};
	croak "$me: Missing required argument user at " unless $user;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	my $FIXME = delete $args{FIXME};
	$class->_no_extraneous_args($me, %args);

	my $cluster = $hpc->db_clustername;
	my $netgrp = $hpc->netgroup;
	croak "$me: Invalid/missing netgroup for cluster $cluster at "
		unless $netgrp;

	my $ok = user_in_netgroup($user, $netgrp);
	return if $ok; #User already in netgroup, nothing to do

	my $error = add_user_to_netgroup(user=>$user, group=>$netgrp);
	croak "$me: Error running ng command to add $user to $netgrp: $? at "
		if $error;

	$error = $GHPLog->add_log_entry($hpc,
		action_code => "add-user-to-netgroup",
		person => $user,
		cluster => $cluster,
		reason => $reason,
		description => "add user $user to netgroup $netgrp",
	);
	return;
}

sub remove_user_from_cluster_netgroup($@)
#Removes user from netgroup for this cluster
#Takes key => value pairs for arguments
#	hpc => Glue::HPCC::Cluster instance, REQUIRED.
#	user => name of user to add, REQUIRED
#	reason => for logging
#Cluster is from $hpc
{	my $class = shift;
	my %args = @_;
	my $me = 'remove_user_from_cluster_netgroup';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $user = delete $args{user};
	croak "$me: Missing required argument user at " unless $user;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $cluster = $hpc->db_clustername;
	my $netgrp = $hpc->netgroup;
	croak "$me: Invalid/missing netgroup for cluster $cluster at "
		unless $netgrp;

	my $inng = user_in_netgroup($user, $netgrp);
	return unless $inng; #User already not in netgroup, nothing to do

	my $error = remove_user_from_netgroup(user=>$user, group=>$netgrp);
	croak "$me: Error running ng command to remove $user from $netgrp: $? at "
		if $error;

	$error = $GHPLog->add_log_entry($hpc,
		action_code => "del-user-from-netgroup",
		person => $user,
		cluster => $cluster,
		reason => $reason,
		description => "remove user $user from netgroup $netgrp",
	);
	return;
}

sub make_user_home_directory($@)
#Creates the home directory for a new user on the cluster, if needed.
#Takes key => value pairs for arguments, recognizing:
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	user: username to create homedir for, REQUIRED
#	reason: reason, for logging purposes, REQUIRED
#	admin: boolean, if true, create homedir in sysadmin area
#	uid: uid which should own dir, default is to lookup in hesiod.
#	gid: gid which should own dir, default is to lookup in hesiod.
#	FIXME: boolean, if true, operate in repair mode.
#		FIXME is actually ignored, always operate in that mode
#
#Must be run as root.  Cluster info obtained from $hpc
#Must be running on headnode of cluster
{	my $class = shift;
	my %args = @_;
	my $me = 'make_user_home_directory';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $user = delete $args{user};
	croak "$me: Missing required argument user at " unless $user;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	my $admin = delete $args{admin};
	my $FIXME = delete $args{FIXME};
	my $uid = delete $args{uid};
	my $gid = delete $args{gid};
	$class->_no_extraneous_args($me, %args);

	my $homeroot = $hpc->homeroot_user;
	$homeroot = $hpc->homeroot_admin if $admin;
	croak "$me: Unable to set homeroot for user $user, admin=$admin at "
		unless $homeroot;

	my $homedir = "$homeroot/$user";
	my $pwent = user_exists($user);
	croak "$me: Invalid user '$user', not in hesiod at " unless $pwent;
	my @pwent = split /\s*:\s*/, $pwent;
	$uid = $pwent[2] unless defined $uid;
	$gid = $pwent[3] unless defined $gid;

	my @stat = stat $homedir;
	if ( scalar(@stat) )
	{	#Directory already exists, verify ownership is correct
		my $tmpuid = $stat[4];
		unless ( $tmpuid == $uid )
		{	die <<EOF;
Home directory $homedir for user $user exists, but wrong ownership.
User $user has uid $uid.
Directory $homedir owned by uid $tmpuid.  
Mismatch.  Please correct this to continue. 
Aborting.
EOF
		}
		return; #Nothing to do
	} 
	
	#We need to create the home directory
	croak "$me: Must run as root to create home dir, not $< at " unless ( $< == 0 );
	#croak "$me: Must run on headnode to create home dir at "
	#	unless $hpc->running_on_headnode;

	my $cmd = "$MK_NEWHOME $homeroot $user $uid $gid";
	my $error = run_system_command($cmd);

	croak "$me: Error runnng '$cmd': " .
		"$? at " if $error;

	#And symlink std Glue homedirectory to ~/glue_home
	my $stdhomedir = "/glue_homes/$user";
	my $stdhdlink = "${homedir}/glue_home";
	$cmd = "ln -s $stdhomedir $stdhdlink";
	if ( $class->dryrun )
	{	print STDERR "[DRYRUN] $cmd\n";
	} else
	{	print STDERR "[VERBOSE] $cmd\n" if $class->verbose;
		my $ok = symlink($stdhomedir, $stdhdlink);
		croak "$me: Error symlinking std glue homedir to ~/glue_home "
			. "for $user: $! at " unless $ok;
	}

	$error = $GHPLog->add_log_entry($hpc,
		action_code => "make-homedir",
		person => $user,
		cluster => $hpc->db_clustername,
		reason => $reason,
		newval => $homedir, #Store path to home directory here
		description => "Created home directory for user $user",
	);
}

sub make_user_lustre_directories($@)
#Creates a lustre directory for a new user on the cluster, if needed.
#Takes key => value pairs for arguments, recognizing:
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	user: username to create homedir for, REQUIRED
#	reason: reason, for logging purposes, REQUIRED
#	uid: uid which should own dir, default is to lookup in hesiod.
#	gid: gid which should own dir, default is to lookup in hesiod.
#	FIXME: operate in repair mode.  Ignored as we always operate in
#		repair mode
#
#Uses $hpc->lustre_filesystems to get lustre filesystems for this cluster
#
#Must be run as root.  Cluster info obtained from $hpc
#Must be running on headnode of cluster
{	my $class = shift;
	my %args = @_;
	my $me = 'make_user_lustre_directories';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $user = delete $args{user};
	croak "$me: Missing required argument user at " unless $user;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	my $FIXME = delete $args{FIXME};
	my $uid = delete $args{uid};
	my $gid = delete $args{gid};
	$class->_no_extraneous_args($me, %args);

	my $pwent = user_exists($user);
	croak "$me: Invalid user '$user', not in hesiod at " unless $pwent;
	my @pwent = split /\s*:\s*/, $pwent;
	$uid = $pwent[2] unless defined $uid;
	$gid = $pwent[3] unless defined $gid;

	#create user owned directory in every lustre filesystem
	my $filesystems = $hpc->lustre_filesystems || [];
	my $fs;
	my $datadir_mode = 0700; #This SHOULD NOT BE QUOTED
	LUSTRE: foreach $fs (@$filesystems)
	{	next LUSTRE unless -d $fs;
		if ( -d "$fs/$user" )
		{	my @tmp = stat "$fs/$user";
			next LUSTRE if $tmp[4] == $uid;
			carp "$me: $fs/$user already exists, but owner is $tmp[4], not $uid at ";
			next LUSTRE;
		} elsif ( -e "$fs/$user" )
		{	carp "$me: $fs/$user exists but is not a directory at ";
			next LUSTRE;
		}

		if ( $class->dryrun )
		{	print STDERR "[DRYRUN] mkdir $fs/$user $datadir_mode\n";
			print STDERR "[DRYRUN] chown ${uid}:$gid $fs/$user\n";
			next LUSTRE;
		} 
		print STDERR "[VERBOSE] mkdir \"$fs/$user\", $datadir_mode\n" 
			if $class->verbose;
		mkdir "$fs/$user", $datadir_mode;
		print STDERR "[VERBOSE] chown $uid, $gid, \"$fs/$user\"\n" 
			if $class->verbose;
		chown $uid, $gid, "$fs/$user";

		my $error = $GHPLog->add_log_entry($hpc,
			action_code => "make-lustre-dir",
			person => $user,
			cluster => $hpc->db_clustername,
			reason => $reason,
			newval => "$fs/$user", #Store path to directory here
			description => "Created lustre directory for user $user",
		);
	}
}

sub make_user_directories($@)
#Creates the home and lustre directories for a new user on the cluster, as needed.
#Takes key => value pairs for arguments, recognizing:
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	user: username to create homedir for, REQUIRED
#	reason: reason, for logging purposes, REQUIRED
#	admin: boolean, if true, create homedir in sysadmin area
#	uid: uid which should own dir, default is to lookup in hesiod.
#	gid: gid which should own dir, default is to lookup in hesiod.
#	FIXME: operate in repair mode if true.
#
#Must be run as root.  Cluster info obtained from $hpc
#Must be running on headnode of cluster
{	my $class = shift;
	my %args = @_;
	my $me = 'make_user_directories';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $user = delete $args{user};
	croak "$me: Missing required argument user at " unless $user;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	my $admin = delete $args{admin};
	my $FIXME = delete $args{FIXME};
	my $uid = delete $args{uid};
	my $gid = delete $args{gid};
	$class->_no_extraneous_args($me, %args);

	my $error = $class->make_user_home_directory(
		hpc => $hpc, 
		user => $user,
		reason => $reason,
		admin => $admin,
		FIXME => $FIXME,
		uid => $uid,
		gid => $gid,
	);
	croak "$me: Error creating home directory for user $user: $error  "
		if ($error );

	$error = $class->make_user_lustre_directories(
		hpc => $hpc, 
		user => $user,
		reason => $reason,
		FIXME => $FIXME,
		uid => $uid,
		gid => $gid,
	);
	croak "$me: Error creating lustre directories for user $user: $error  "
		if ($error );

	return $error;
}

sub _find_user_dir_helper($$$$$)
#Helper function for find_user_directories
#
#For directory $root, return list ref of paths to all files/dirs
#directly under $root such that either:
#	file is owned by $uid, if $uid is given
#	file is named $username and owner not in /etc/passwd 
#		if $username is given and $uid is undef
#
#In latter case, pretty sure file is owned by user in question (or at least
#is orphaned and can be deleted anyway:)
#
#We return FULL PATH to the files
{	my $class = shift;
	my $type = shift;
	my $root = shift;
	my $username = shift;
	my $uid = shift;
	my $realme = '_find_user_dir_helper'; 
	my $me = 'find_user_directories'; #Lie about who we are

	die "$realme: Either username or uid must be defined at "
		unless defined $uid || $username;

	if ( defined $uid )
	{  	my $owns = files_in_dir_owned_by_uid($root, $uid);
		#This returns relative paths, fix
		my @temp = map { "$root/$_" } @$owns;
		return [ @temp ];
	}

	#No uid, look for dir named after username with unknown owner
	my $dir = "$root/$username";
	if ( -e $dir )
	{	my @stat = stat $dir;
		my $diruid = $stat[4]; #uid of directory
		my $dirowner = username_from_uid($diruid);
		if ( $dirowner && $dirowner eq $username )
		{	#Should this happen???
			#If can lookup diruid => username, should have gotten
			# uid <= diruid and did things in previous step
			# But just in case, do here as well
			return [ $dir ];
		} elsif ( $dirowner )
		{	#Has a real owner, not user in question
			carp "$me: Possible $type dir $dir exists, " .
				"but is owned by $dirowner.  Do not delete\n ";
			return [];
		} else
		{	#Owner is not in /etc/passwd, probably us
			carp "$me: $type dir $dir exists, owned by " .
				"$diruid not in /etc/passwd, " .
				"assuming is $username\n";
			return [ $dir ];
		}
	}
	#Should not reach here?
	return []; 
}
	
sub find_user_directories($@)
#Return a list ref of user directories found.
#Takes key => value pairs, recognizing
#	hpc: REQUIRED
#	user: username, REQUIRED
#List ref consist of list refs
#	[ type, path, ... ]
#Where type is 'home', 'lustre-data', or 'nfs-data'
#and path is path to the user owned directory
#
#Returns full paths to the dirs
{	my $class = shift;
	my %args = @_;
	my $me = 'find_user_directories';

	my $hpc = delete $args{hpc};
	my $username = delete $args{user};
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required argument user at " unless $username;
	$class->_no_extraneous_args($me, %args);

	my $user = Glue::HPCC::Project::Person->person_by_username(
		$hpc, $username);
	croak "$me: Error looking up user $username in PDB: $user at "
		if $user && ! ref($user);
	my $uid;
	if ( $user )
	{	$uid = $user->uid;
	} elsif ( $username =~ /^\d+$/ )
	{	carp "$me: treating $username as uid at ";
		$uid = $username;
		$username = undef;
	} elsif ( user_exists($username) )
	{	$uid = uid_from_username($username);
	} else
	{	carp "$me: no uid found for $username, trying to continue.\n";
	}	

	my $dirlist = [];
	my ($type, $root, $dir, $owns);
	my $filesystems;

	$type = 'lustre-data';
	$filesystems = $hpc->lustre_filesystems || [];
	foreach $root (@$filesystems)
	{	$owns = $class->_find_user_dir_helper(
			$type, $root, $username, $uid);
		foreach $dir (@$owns)
		{	push @$dirlist, [ $type, "$dir" ];
		}
	}

	$type = 'nfs-data';
	$filesystems = $hpc->nfs_data_filesystems || [];
	foreach $root (@$filesystems)
	{	$owns = $class->_find_user_dir_helper(
			$type, $root, $username, $uid);
		foreach $dir (@$owns)
		{	push @$dirlist, [ $type, "$dir" ];
		}
	}

	$type = 'home';
	$root = $hpc->homeroot_user;
	$dir = $hpc->homeroot_admin;
	$filesystems = [ $root ];
	push @$filesystems, $dir if $dir && $dir ne $root;
	foreach $root (@$filesystems)
	{	$owns = $class->_find_user_dir_helper(
			$type, $root, $username, $uid);
		foreach $dir (@$owns)
		{	push @$dirlist, [ $type, "$dir" ];
		}
	}
	
	return $dirlist;
}

#------	Mailing list utilities

sub OLD_is_user_in_maillist_sister_cluster($@)
#Determine if specified user currently has access to an HPC cluster
#we share a mailing list with.  E.g., no reason to add to the mailing
#list if already there (although should be safe/noop with listserv), and
#do NOT want to delete if still a member of allocation in other cluster.
#Returns true (a sister clustername user belongs to) if belongs to a sister
#cluster, and false otherwise.
#Takes key => value pairs for explicit arguments, recognizing:
#	hpc => Glue::HPCC::Cluster instance for cluster we are add/deleting
#		user from, REQUIRED
#	user => name of user, REQUIRED
#	role => role for user to check for.  Default is 'ANY'.  If set
#		to 'DB', returns true if user exists in PDB for a sister cluster
#		even if no role (probably meaningless, as likely share PDB
#		with sister clusters).  If 'member', 'poc', or 'coord', requires
#		user to have that role for some project in the sister cluster.
#		If 'ANY', returns true if user has any role for any project
#		in sister cluster.
{	my $class = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::is_user_in_maillist_sister_cluster';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $role = delete $args{role} || 'ANY';

	$class->_no_extraneous_args($me, %args);
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required argument user at " unless $user;

	my $this_cluster = $hpc->clustername;
	my $sister_clusters = $hpc->clusters_sharing_mailinglist;
	#If no sister clusters, then we return false always
	return unless $sister_clusters && ref($sister_clusters) eq 'ARRAY'
		&& scalar(@$sister_clusters);

	SISTERS: foreach my $sister (@$sister_clusters)
	{	next SISTERS if $sister eq $this_cluster;
		my $sishpc = Glue::HPCC::Cluster->new_from_clustername($sister);
		unless ( $sishpc && ref($sishpc) )
		{	croak "$me: Unable to instantiate cluster for $sister: $sishpc at ";
		}
		#Skip if this is our cluster
		next SISTERS if $sishpc->clustername eq $this_cluster;

		#First check if user is even in PDB for this cluster
		my $temp = Glue::HPCC::Project::Person->person_by_username(
			$sishpc, $user);
		next SISTERS unless $temp; #Not in PDB, so not in a project
		croak "$me: Error looking up user $user in PDB for cluster " .
			"$sister: $temp at " unless ref($temp); 

		$temp = Glue::HPCC::Project::PersonProject->get_project_persons_by_username($sishpc, $user);
		next SISTERS unless $temp; 
		croak "$me: Error looking up user $user in cluster $sister: " .
			"$temp at " unless $temp && ref($temp) eq 'ARRAY';
		next SISTERS unless scalar(@$temp);#No records, try next cluster

		return $sister if $role eq 'DB';  

		my @temp;
		if ( $role eq 'ANY' )
		{	@temp = grep { $_->is_member || $_->is_poc ||
				$_->is_coordinator } @$temp;
		} elsif ( $role eq 'member' )
		{	@temp = grep { $_->is_member } @$temp;
		} elsif ( $role eq 'poc' )
		{	@temp = grep { $_->is_poc } @$temp;
		} elsif ( $role eq 'coord' )
		{	@temp = grep { $_->is_coordinator } @$temp;
		}

		return $sister if scalar(@temp);
	}
	#We failed to find any sister cluster were user has specified role
	return;
}

sub add_user_to_cluster_mailinglist($@)
#This adds the specified user to the cluster mailing list.
#It takes key => value pairs for parameters, recognizing:
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#		This determines listserv email, listserv name, etc.
#	user => username of user to add, REQUIRED.
#	email => email address to add to list, REQUIRED.
#	fullname => full name of user to add to list, REQUIRED.
#	FORCE => boolean, default false.  Normally, we only add the
#		user if the user does not belong to a project in a sister
#		cluster sharing the same mailing list.  If set to true,
#		we always add the user (should be safe, listserv rejects
#		duplicates).
#
#Returns false on success.  On error, croaks or returns true (error message)
#Note that "success" basically means we managed to send an email to the listserv
#off --- we don't actually wait around for confirmation from listserv that
#the action was successful.
{	my $class = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::add_user_to_cluster_mailinglist';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $email = delete $args{email};
	my $fullname = delete $args{fullname};
	my $FORCE = delete $args{FORCE};
	$class->_no_extraneous_args($me, %args);
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required argument user at " unless $user;
	croak "$me: Missing required argument email at " unless $email;
	croak "$me: Missing required argument fullname at " unless $fullname;
	croak "$me: Invalid email '$email' at " unless $email &&
		$email =~ /[\w\.\-]+@[\w\.\-]+/;

	#Listserv requires fullname be at least two words
	unless ($fullname =~ /\S\s+\S/ )
	{	#Force it to be two words
		$fullname = "$fullname $user";
	}

	my $mlist = $hpc->mailinglist_name;
	unless ($mlist)
	{	print STDERR "[VERBOSE] No mailing list for cluster.\n";
		return;
	}
	my $domain = $hpc->domain || 'deepthought.umd.edu';
	my $from = "hpcc-auto-subscriber\@$domain";
	my $pwd = $hpc->mailinglist_password;
	my $toaddr = $hpc->mailinglist_listserv;
	my $subj = 'list management';
	
	my $tmp = $class->is_user_in_maillist_sister_cluster(
		hpc => $hpc, user => $user, role => 'ANY' );
	if ( $tmp )
	{	#User should already be in the list
		if ( $FORCE )
		{	print STDERR "[VERBOSE] User $user should already be " .
				"in $mlist from $tmp, but adding again due " .
				"to FORCE=$FORCE\n" if $class->verbose;
		} else
		{ 	print STDERR "[VERBOSE] User $user already should be " .
				"in $mlist from membership in $tmp, not " .
				"adding again\n" if $class->verbose;
			return;
		}
	}

	#We need to add the user
	my $header =
	{	To => $toaddr,
		From => $from,
		Subject => $subj,
	};

	my $useraddr = "$email $fullname";

	my $message = <<EOF;
QUIET ADD $mlist DD=XYZ IMPORT PW=$pwd
//XYZ DD *
$useraddr
/*
EOF

	send_email($header, $message);

	return;
}
	

sub remove_user_from_cluster_mailinglist($@)
#This removes the specified user from the cluster mailing list.
#It takes key => value pairs for parameters, recognizing:
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#		This determines listserv email, listserv name, etc.
#	user => username of user to remove, REQUIRED.
#	email => email address to remove to list, REQUIRED.
#	FORCE => boolean, default false.  Normally, we only remove the
#		user if the user does not belong to a project in a sister
#		cluster sharing the same mailing list.  If set to true,
#		we always remove the user.
#
#Returns false on success.  On error, croaks or returns true (error message)
#Note that "success" basically means we managed to send an email to the listserv
#off --- we don't actually wait around for confirmation from listserv that
#the action was successful.
{	my $class = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::remove_user_from_cluster_mailinglist';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $email = delete $args{email};
	my $FORCE = delete $args{FORCE};
	$class->_no_extraneous_args($me, %args);
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required argument user at " unless $user;
	croak "$me: Missing required argument email at " unless $email;
	croak "$me: Invalid email '$email' at " unless $email &&
		$email =~ /[\w\.\-]+@[\w\.\-]+/;

	my $mlist = $hpc->mailinglist_name;
	unless ($mlist)
	{	print STDERR "[VERBOSE] No mailing list for cluster.\n";
		return;
	}
	my $domain = $hpc->domain || 'deepthought.umd.edu';
	my $from = "hpcc-auto-subscriber\@$domain";
	my $pwd = $hpc->mailinglist_password;
	my $toaddr = $hpc->mailinglist_listserv;
	my $subj = 'list management';
	
	unless ( $FORCE )
	{	my $tmp = $class->is_user_in_maillist_sister_cluster(
			hpc => $hpc, user => $user, role => 'ANY' );
		if ( $tmp )
		{	#User should already be in the list
			print STDERR "[VERBOSE] Not removing user $user from " .
				"list $mlist because still has  membership " .
				"in $tmp\n" if $class->verbose;
			return;
		}
	}

	#We need to add the user
	my $header =
	{	To => $toaddr,
		From => $from,
		Subject => $subj,
	};

	my $useraddr = "$email";

	my $message = <<EOF;
QUIET DELETE $mlist DD=XYZ PW=$pwd
//XYZ DD *
$useraddr
/*
EOF

	send_email($header, $message);

	return;
}
	


	


#-------------------------------------------------------------------
#	Higher level stuff: add/remove user to/from cluster/project
#-------------------------------------------------------------------

sub add_user_to_hpc_cluster($@)
#Adds specified user to the specified HPC cluster
#Takes key=>value pairs, recognizing:
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	user => username of user to add to cluster. REQUIRED
#	fullname: fullname of user, to put in person DB
#	uid: uid of user, to put in person DB
#	gid: gid of user, to set as owner of directories.  If omitted, the
#		child routines should default to using users primary dir.
#		That will likely be WRONG for a new HPC user (will be users,
#		not dt-project of their new project), hence this option
#	email: email of user, to put in person DB
#	admin: boolean, true if this is a sysadmin account and should
#		go in special admin homespace
#	reason: reason user being added, REQUIRED
#	FIXME: repair mode, don't complain if partially done
#Cluster is the cluster represented by $hpc
#
#If user exists in PDB, fullname, uid, and email will be updated if
#given and different than what previously existed. And if disabled,
#the user will be enabled.  If the user does
#not exist in PDB, the values given will be used, or defaulted from hesiod.
#The user MUST exist in hesiod already.
#
#This does the stuff which is NOT project specific, i.e.:
#	adds to cluster netgroup, if needed
#	creates the home directory for user, if needed
#	adds to PDB person table, if needed
#	adds user to mailing list for HPC cluster, if needed
#
#Verifies user is valid before all of that
#Calls sched_add_user_to_hpc_cluster to do any (if any) scheduler specific
#tasks
#
#Returns false on success.  On error, croaks or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'add_user_to_hpc_cluster';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $user = delete $args{user};
	croak "$me: Missing required argument user at " unless $user;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	my $admin = delete $args{admin};
	my $full = delete $args{fullname};
	my $email = delete $args{email};
	my $uid = delete $args{uid};
	my $gid = delete $args{gid};
	my $FIXME = delete $args{FIXME};
	$class->_no_extraneous_args($me, %args);

	croak "$me: Invalid user '$user', not in hesiod at " 
		unless ( user_exists($user) );
	my $cname = $hpc->db_clustername;

	my $error;

	$error = $class->sched_add_user_to_hpc_cluster( 
		hpc=> $hpc,
		user => $user, 
		reason => $reason, 
		FIXME=>$FIXME );

	croak "$me: Error adding user $user to netgroup for $cname: " .
		"$error at " if $error;

	$error = $class->add_user_to_cluster_netgroup( hpc=> $hpc,
		user => $user, reason => $reason, FIXME=>$FIXME );
	croak "$me: Error adding user $user to netgroup for $cname: " .
		"$error at " if $error;

	$error = $class->make_user_directories( hpc=> $hpc, 
		user => $user, admin => $admin, reason=>$reason,
		uid => $uid, gid => $gid,
		FIXME => $FIXME );
	croak "$me: Error making homedir for $user for cluster $cname: " .
		"$error at " if $error;

	$error = $class->add_person_to_pdb(
		hpc => $hpc,
		username => $user,
		fullname => $full,
		email => $email, 
		uid => $uid,
		reason => $reason,
		FIXME=>1, #Always set, person might already be there 
			  #from another project
	);
	croak "$me: Error adding/updating user $user in PDB person table: " .
		"$error at " if $error;

	#We need email and fullname for adding to mailing list.  This
	#is most likely defaulted in add_person_to_pdb, so get the PDB entry
	my $userobj = Glue::HPCC::Project::Person->person_by_username(
		$hpc, $user);
	if ( $userobj && ref($userobj) )
	{	#Got a valid user object from PDB
		#Use it to set email and full if not already set
		$email = $userobj->email unless $email;
		$full = $userobj->fullname unless $full;
	} elsif ( $userobj )
	{	#Got an error looking up user in PDB, croak
		croak "$me: Error looking up newly created user $user in PDB: "
			. "$userobj at ";
	} else
	{	#No user object found.  OK if dryrun mode, otherwise bad
		if ( $class->dryrun)
		{	#OK, but fake email/full if needed
			$email = "${user}\@umd.edu" unless $email;
			$full = $user unless $full;
		} else
		{	croak "$me: Error, cannot find $user in PDB even after"
				. " adding user to PDB.  Huh? at ";
		}
	}

	$error = $class->add_user_to_cluster_mailinglist(
		hpc => $hpc,
		user => $user,
		fullname => $full,
		email => $email,
		#Use FORCE if in FIXME mode:  assume previous attempt failed
		#before adding to mailing list/don't trust PDB is completely
		#correct.  Should be safe; listserv will drop duplicates anyway
		FORCE => $FIXME, 
	);
	croak "$me: Error adding user $user to cluster mailing list: " .
		"$error at " if $error;

	return;
}

sub remove_user_from_hpc_cluster($@)
#Removes specified user from the specified HPC cluster
#Takes key=>value pairs, recognizing:
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	user => username of user to remove from cluster. REQUIRED
#	reason => reason user being removed, REQUIRED
#Cluster is the cluster represented by $hpc
#
#User must exists in PDB, but must not belong to any projects
#(but we make an exception in dryrun mode, as maybe is only in projects
#was supposed to be removed from)
#The user need not exist in hesiod, etc. (e.g. so can remove users
#who have been removed from Glue already)
#
#This does the stuff which is NOT project specific, i.e.:
#	removes from cluster netgroup, if needed
#
#We run sched_remove_user_from_hpc_cluster; this generally does NOT
#need to do anything, but should barf if it is NOT safe to remove
#the user (i.e. there are Slurm associations for the user).
#
#We do NOT delete home directory or lustre directories
#
#We DO deactivate the PDB entry in Person table IF the user DOES NOT BELONG to
#projects in another cluster sharing our PDB tables
#
#Returns false on success.  On error, croaks or returns true (error message)
{	my $class = shift;
	my %args = @_;
	my $me = 'remove_user_from_hpc_cluster';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $username = delete $args{user};
	croak "$me: Missing required argument user at " unless $username;
	my $reason = delete $args{reason};
	croak "$me: Missing required argument reason at " unless $reason;
	$class->_no_extraneous_args($me, %args);

	my $cluster = $hpc->db_clustername;
	my $user = $GHPPers->person_by_username($hpc, $username);
	croak "$me: Error looking up user $username in PDB: $user at "
		if $user && ! ref($user);

	if ( $user )
	{	#See if user belongs to any projects
		my $projpers = $GHPPerPrj->get_project_persons_by_username(
			$hpc, $username);
		croak "$me: Error looking up projects for $username: $projpers at"
			if $projpers && ! ref($projpers);
		PRPER: foreach my $prper (@$projpers)
		{	#See if member (PoC is OK, want to keep PoCs of disabled project)
			my $tmp = $prper->is_member ;
			if ( $tmp )
			{	#Can't delete if MEMBER of a project
				#Allow delete for PoC/coords, so can keep record of PoC
				#of disabled projects
				my $proj = $GHPProj->project_by_pkey($hpc, $prper->prj_pkey);
				my $projname = $proj->name;
				if ( $class->dryrun )
				{   warn "$me: User $username is member of " . 
					"$projname --- assuming would have " .
					"been removed except for DRYRUN mode " .
					"at ";
				} else
				{  croak "$me: Cannot delete user $username " .
					"from cluster $cluster: because is " .
					"member of $projname at ";
				}
			}

			#We have a row, but is not a member or anything.  Delete the row
			$GHPPerPrj->pdb_delete($hpc, prj_pkey=>$prper->prj_pkey,
				ppl_pkey => $prper->ppl_pkey);
		}
	}

	#OK, looks safe to delete now, not part of any projects

	#Do Scheduler specific tasks (do this first, because likely is only
	#a chance to barf if user looks active to scheduler)
	my $error = $class->sched_remove_user_from_hpc_cluster(
		hpc => $hpc, user => $username, reason => $reason );
	carp "$me: Error in scheduler portion of removing user from $cluster: "
		. "$error. Trying to continue at " if $error;

	#Remove netgroup (do this even if NOT in PDB)
	$error = $class->remove_user_from_cluster_netgroup(
		hpc => $hpc, user => $username, reason => $reason);
	carp "$me: Error removing user $username from NG for $cluster: $error."
		. "  Trying to continue at " if $error;

	my $email = $user->email;
	$error = $class->remove_user_from_cluster_mailinglist(
		hpc => $hpc,
		user => $username,
		email => $email,
		#Never FORCE this, might be in a sister cluster sharing the
		#same mailing list
		FORCE => 0,
	);
	croak "$me: Error removing user $user to cluster mailing list: " .
		"$error at " if $error;

	#Do NOT deactivate the person, UNLESS they are NOT in any other cluster we
	#share PDB tables with.
	my $in_sister = $class->is_user_in_pdb_sister_cluster(
		hpc => $hpc,
		user => $username, 
	);
	if ( $in_sister )
	{	print STDERR "[VERBOSE] Not deactivating user $user in PDB " .
			"because active in $in_sister\n" if $class->verbose;
	} else
	{	$class->deactivate_person_in_db( 
			hpc=>$hpc,
			user => $username, 
			reason => $reason);
	}
	return $error;
}

sub add_user_to_project_in_cluster($@)
#Adds the specified user to specified project in specified cluster.
#Args are in key => value pairs,
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	user => usaername to add, REQUIRED
#	project => project to add to, REQUIRED
#	reason => reason for logs, REQUIRED
#	highpriority => should user get access to condo's hi-prio QoS
#		Only valid on projects of type 'condo'
#	defaultaccount => default account for user.  Optional.
#	FIXME => operate in repair mode
#Cluster is specified via $hpc
#
#This does the scheduler independent AND the scheduler dependent stuff.
#	1) add user to unix group for project
#	2) set user's primary gid to unix group for project if it
#		is  currently $class->default_users_unix_group_gid
#	3) calls sched_add_user_to_project_in_cluster 
#		(which should be overridden by scheduler specific child class)
#		to do scheduler specific processing
#	4) adds user as member to project in PDB
#
#Updates the logs
#
#User and project must already exist, both on system and in PDB (and for this
#cluster).
#
#Returns false on success.  On error, either croaks or returns true (error
#message)
{	my $class = shift;
	my %args = @_;
	my $me = 'add_user_to_project_in_cluster';

	my $hpc = delete $args{hpc};
	my $user = delete $args{user};
	my $project = delete $args{project};
	my $reason = delete $args{reason};
	my $highpri = delete $args{highpriority};
	my $defacct = delete $args{defaultaccount};
	my $cluster = $hpc->db_clustername;
	my $FIXME = delete $args{FIXME};

	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'user' at " unless $user;
	croak "$me: Missing required parameter 'project' at " unless $project;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	
	my $pwent = user_exists($user);
	croak "$me: Invalid user '$user', not found in hesiod at "
		unless $pwent || $hpc->sheduler_type eq 'dummy';
	my @pwent = split /\s*:\s*/, $pwent;
	my $primary_gid = $pwent[3];

	#Get our project and type and unix group
	my $proj = Glue::HPCC::Project::Project->project_by_name($hpc, 
		$project);
	croak "$me: Cannot find proj $project in PDB at " unless $proj;
	croak "$me: Error looking up project $project: $proj at " 
		unless ref($proj);
	my $prjgroup = $proj->unixgroup;

	my $ptypeid = $proj->projtype;
	my $ptype = $GHPPType->projecttype_by_pkey( $hpc, $ptypeid);
	croak "$me: Cannot lookup project type for pkey=$ptypeid at "
		unless $ptype;
	croak "$me: Error looking up ptype for pkey=$ptypeid: $ptype at "
		unless ref($ptype);
	my $projtype = $ptype->name;
	if ( defined $highpri && $projtype ne 'condo' )
	{	carp "$me: Ignoring highpriority flag for project type " .
			"$projtype at ";
		$highpri = undef;
	}

	my $error;
	if ( $hpc->scheduler_type ne 'dummy' )
	{   unless ( user_in_unix_group($user, $prjgroup) )
	    {	$error = add_user_to_unix_group(
			user => $user, group => $prjgroup);
		croak "$me: Error adding user $user to unix group $prjgroup: $error at "
			if $error;
		$error = $GHPLog->add_log_entry($hpc,
			action_code => "add-user-to-unixgroup",
			person => $user,
			project => $project,
			cluster => $cluster,
			reason => $reason,
			description => "add user $user to unix group $prjgroup",
		);
	    }
	    if ( $primary_gid == $class->default_users_unix_group_gid)
	    {	$error = set_users_primary_unix_group(user => $user, group => $prjgroup);
		croak "$me: Error setting user ${user}'s to prim gid to $prjgroup: $error at "
			if $error;
		$error = $GHPLog->add_log_entry($hpc,
			action_code => "set-primarygroup",
			person => $user,
			project => $project,
			cluster => $cluster,
			reason => $reason,
			description => "setting primary group for  user $user to $prjgroup",
		);
	    }
	}

	$error = $class->sched_add_user_to_project_in_cluster(
		hpc => $hpc, 
		user => $user, 
		project => $project,
		projtype => $projtype,
		reason => $reason,
		highpriority => $highpri,
		defaultaccount => $defacct,
		FIXME => $FIXME,
	);
	croak "$me: Error in sched part of adding $user to $project in $cluster: $error at "
			if $error;

	$error = $class->pdb_add_member_to_project(
			hpc => $hpc,
			user => $user,
			project => $project,
			reason => $reason,
			FIXME => $FIXME,
			);
	croak "$me: Error adding user $user to proj $project in PDB: $error at "
		if $error;
	return;
}

sub delete_user_from_project_in_cluster($@)
#Removes the specified user to specified project in specified cluster.
#Args are in key => value pairs,
#	hpc => Glue::HPCC::Cluster instance, REQUIRED
#	user => usaername to remove, REQUIRED
#	project => project to remove from, REQUIRED
#	reason => reason for logs, REQUIRED
#	adjust_balanaces => adjust balances on Slurm accounts for project.  See below.
#Cluster is specified via $hpc
#
#This does the scheduler independent AND the scheduler dependent stuff.
#	1) calls sched_del_user_from_project_in_cluster 
#		(which should be overridden by scheduler specific child class)
#		to do scheduler specific processing
#	2) removes user as member to project in PDB
#	3) removes user from unix group for project
#	4) if user's primary group is the projects unix group,
#		set it to $class->default_users_unix_group_name
#		I was thinking of doing something more complicated, like
#		setting to another group is a member of, but could get
#		tricky if deleting from multiple projects due to delays
#		(e.g. if in foo and bar projects, primary gid=dt-bar, and
#		delete from foo, no change.  Then delete from bar before
#		remove from dt-foo propagates, and we set primary gid to dt-foo)
#
#adjust_balances: When user deleted from account in Slurm, the usage is also deleted,
#artificially raising the account's balance.  If set, the code will lower limit by 
#deleted user's usage after deleting user from project.
#Updates the logs
#
#Project must already exist (for real and in PDB).
#
#Returns false on success.  On error, either croaks or returns true (error
#message)
{	my $class = shift;
	my %args = @_;
	my $me = 'delete_user_from_project_in_cluster';

	my $hpc = delete $args{hpc};
	my $username = delete $args{user};
	my $projname = delete $args{project};
	my $reason = delete $args{reason};
	my $adj_bals = delete $args{adjust_balances};
	my $cluster = $hpc->db_clustername;

	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required parameter 'user' at " unless $username;
	croak "$me: Missing required parameter 'project' at " unless $projname;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$class->_no_extraneous_args($me, %args);
	
	#Lookup project in PDB
	my $proj = $GHPProj->project_by_name($hpc, $projname);
	croak "$me: Cannot find project $projname in PDB at " unless $proj;
	croak "$me: Error looking up project $projname in PDB: $proj at " unless ref($proj);
	my $prjgrp = $proj->unixgroup;
	my $ptypeid = $proj->projtype;
	my $ptype = $GHPPType->projecttype_by_pkey($hpc, $ptypeid);
	croak "$me: Cannot lookup project type for pkey=$ptypeid at "
		unless $ptype;
	croak "$me: Error looking up ptype for pkey=$ptypeid: $ptype at "
		unless ref($ptype);
	my $projtype = $ptype->name;

	my @errors = ();
	my $error = $class->sched_del_user_from_project_in_cluster(
		hpc => $hpc,
		user => $username,
		project => $projname,
		projtype => $projtype,
		reason => $reason,
		adjust_balances => $adj_bals,
	);
	carp "$me: Error doing sched side of deleting $username from "
		. "$projname: $error, but trying to continue at " 
		if $error;
	push @errors, "Error in sched_del_users_from_project_in_cluster: $error"
		if $error;

	#Lookup user in PDB
	my $user = $GHPPers->person_by_username($hpc, $username);

	$error = $class->pdb_remove_member_from_project(
		hpc => $hpc, 
		user => $username,
		project => $projname,
		reason => $reason,
	);
	carp "$me: Error deleting $username from $projname in PDB," .
		"but trying to continue anyway,  at " 
		if $error;
	push @errors, "Error in pd_remove_member_from_project: $error"
		if $error;

	my $pwent = user_exists($username);
	unless ( $pwent )
	{	carp "$me: User $username not in hesiod at ";
		return;
	}

	my @pwent = split /\s*:\s*/, $pwent;
	my $primary_gid = $pwent[3];

	my $grent = unix_group_exists($prjgrp);
	unless ( $grent )
	{	carp "$me: Unix group $prjgrp not in hesiod at ";
		return;
	}
	my @grent = split /\s*:\s*/, $grent;
	my $prjgid = $grent[2];

	if ( user_in_unix_group($username, $prjgrp) )
	{	$error = remove_user_from_unix_group(
			user => $username, group => $prjgrp);
		carp "$me: Error removing user $user to unix group $prjgrp: " .
			"$error, trying to continue at "
			if $error;
		push @errors, "Error in remove_user_from_unix_group: $error" if $error;
		$GHPLog->add_log_entry($hpc,
			action_code => "add-user-to-unixgroup",
			person => $username,
			project => $projname,
			cluster => $cluster,
			reason => $reason,
			description => "add user $username to unix group $prjgrp",
		) unless $error;
	}

	if ( $primary_gid == $prjgid )
	{	$error = set_users_primary_unix_group(user => $username, 
			group => $class->default_users_unix_group_name);
		carp "$me: Error setting user ${username}'s to prim gid to " .
			$class->default_users_unix_group_name .
			": $error at " if $error;
		push @errors, "Error in set_users_primary_group: $error" if $error;
		
		$GHPLog->add_log_entry($hpc,
			action_code => "set-primarygroup",
			person => $username,
			project => $projname,
			cluster => $cluster,
			reason => $reason,
			description => "setting primary group for  user $username to $prjgrp",
		) unless $error;
	}
	return;
}
1;
__END__

=head1 NAME

Glue::HPCC::ClusterUtil::User - Base class for Glue::HPCC::ClusterUtil::XXX:User classes

=head1 SYNOPSIS

  package Glue::HPCC::ClusterUtil::Slurm::User;
  use base qw(Glue::HPCC::ClusterUtil::User);

  ...


=head1 DESCRIPTION

This is the base class for Glue::HPCC::ClusterUtil::XXX::User classes.

This provides various routines for managing users in an HPC environment.
These routines deal with the scheduler independent portions, e.g. the 
projects/hpcc database (PDB), etc., and should be inherited by the
scheduler specific child classes, which should be called by the 
management scripts.

All of the routines are class routines.  A minimal constructor is
provided if desired.  The routines are packaged as an OO class to
reduce namespace pollution and allow inheritence.

Significant routines are:

=over 4

=item B<verbose>

Gets/sets the verbosity level.  When setting, propagates to all
classes we depend on.

=item B<dryrun>

Gets/sets the dryrun flag.  When setting, propagates to all
classes we depend on.

=item B<project_class>

Returns the name of our sister B<Glue::HPCC::ClusterUtil::XXX::Project>
class.

=item B<add_person_to_pdb>

Adds the specified person to the person DB table.

=item B<update_person_in_pdb>

Updates information about specified user in the person DB table.

=item B<add_or_update_person_to_pdb>

Adds the specified person to the person DB table if they are
not already present.  If they are already present, update the
record with the specified data if needed.

=item B<activate_user_in_pdb>

Sets the active flag for specified user to true (in PDB).

=item B<deactivate_user_in_pdb>

Sets the active flag for specified user to false (in PDB).

=item B<pdb_add_member_to_project>

Adds the specified user as a member of specified project, in
the PDB only.

=item B<pdb_add_poc_to_project>

Adds the specified user as a Point-of-Contact for specified project, in
the PDB only.

=item B<pdb_remove_member_from_project>

Removes the specified user as a member of specified project, in
the PDB only.

=item B<pdb_remove_poc_from_project>

Removes the specified user as a Point-of-Contact for specified project, in
the PDB only.

=item B<add_user_to_cluster_netgroup>

Adds the specified user to the netgroup controlling login access
to the login nodes for the HPC cluster specified by the 
Glue::HPCC::Cluster instance provided.

=item B<remove_user_from_cluster_netgroup>

Removes the specified user from the netgroup controlling login access
to the login nodes for the HPC cluster specified by the 
Glue::HPCC::Cluster instance provided.

=item B<make_user_home_directory>

Creates the home directory for the specified new user on the 
cluster specified by the Glue::HPCC::Cluster instance, if needed.

=item B<make_user_lustre_directories>

Creates the directories for the specified new user in any lustre
filesystems for the cluster, as needed.

=item B<make_user_directories>

Does B<make_user_home_directory> and B<make_user_lustre_directories>

=item B<add_user_to_cluster_mailinglist>

This sends email to the listserv management interface to add the specified
user to the cluster mailinglist.  It takes key => value argument pairs,
requiring hpc (for Glue::HPCC::Cluster instance), user (for the username),
email (the email address of user), and fullname (the fullname of the user).
The mailing list information is contained in $hpc.  Normally, no email is
sent to the listserv if the specified user is already in the PDB as belonging
to a project in one of the clusters that shares the mailing list with the
cluster in $hpc, but if FORCE is given and true, the user will be added 
regardless.  

=item B<remove_user_from_cluster_mailinglist>

This sends email to the listserv management interface to remove the specified
user from the cluster mailinglist.  It takes key => value argument pairs,
requiring hpc (for Glue::HPCC::Cluster instance), user (for the username), and
email (the email address of user).  The mailing list information is contained 
in $hpc.  Normally, no email is sent to the listserv if the specified user is 
still in the PDB as belonging to a project in one of the clusters that shares 
the mailing list with the cluster in $hpc (e.g. if user has access to both
DT and DT2, removing DT access should not remove from list as still has DT2
access), but if FORCE is given and true, the user will be removed regardless.  

=item B<add_user_to_hpc_cluster>

Sets up the specified new user for accessing the specified
HPC cluster (via Glue::HPCC::Cluster instance).  This adds the
user to the PDB if needed, adds them to the cluster login-access
netgroup, and creates the home directory for user if needed.
Does NOT do any project specific stuff.

=item B<delete_user_from_hpc_cluster>

Removes specified user from hpc cluster.  User need not exist
in hesiod, etc.  Expected to be in PDB, and must NOT be part
of any projects for this cluster.  We just remove login access
to the cluster (remove from netgroup).  Do not deactivate PDB
record (much less delete it) as might be active in another cluster.
We do NOT delete any files (homedir, lustre etc), nor do we do
anything Slurm (other to verify no associations)

=item B<add_user_to_project_in_cluster>

Adds the specified user to the specified project in the
specified cluster (via Glue::HPCC::Cluster instance).  This does
both the scheduler independent and the scheduler dependent
parts (the latter by calling B<sched_add_user_to_project_in_cluster>.

=item B<delete_user_from_project_in_cluster>

Removes the specified user from the specified project in the
specified cluster (via Glue::HPCC::Cluster instance).  This does
both the scheduler independent and the scheduler dependent
parts (the latter by calling B<sched_del_user_from_project_in_cluster>.

=back
