#!/usr/local/bin/perl
#
#College table in projects DB

package Glue::HPCC::Project::College;
use strict;
use warnings;
use base qw(Glue::HPCC::Project::RWBase);
use Carp qw(croak carp confess);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $GHPCollege = 'Glue::HPCC::Project::College';

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @ro_accessors = qw( coll_pkey );
my @rw_accessors = qw(
	code
	name
	description
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);

sub _ro_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_ro_fields;
	return [ @ro_accessors, @$parent ];
}

sub _rw_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_rw_fields;
	return [ @rw_accessors, @$parent ];
}

sub _required_fields($)
{	return [ qw( name ) ];
}

sub _db_schema_name($)
{       my $class = shift;
        return 'hpcc_shared';
}

sub _db_table_primary_keyname($)
#Name of primary key for our table.  Can return a scalar string or a list ref
#of scalar strings if multikey table
{	my $class = shift;
	return 'coll_pkey';
}

#--------------------------------------------------------------------
#	Special class methods
#--------------------------------------------------------------------

	
sub college_by_code($$$)
{	my $class = shift;
	my $hpc = shift;
	my $code = shift;
	my $me = 'college_by_code';
	$class->_validate_hpc($me, $hpc);

	my $list = $class->pdb_query($hpc, code => $code );
	croak "$me: Error querying code '$code': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 )
	{	my @tmp = map { $_->coll_pkey } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching code=$code, pkeys=[$tmp] at ";
	}
	return $list->[0];
}

sub college_by_name($$$)
{	my $class = shift;
	my $hpc = shift;
	my $name = shift;
	my $me = 'college_by_name';
	$class->_validate_hpc($me, $hpc);

	my $list = $class->pdb_query($hpc, name => $name );
	croak "$me: Error querying name '$name': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 )
	{	my @tmp = map { $_->coll_pkey } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching name=$name, pkeys=[$tmp] at ";
	}
	return $list->[0];
}

sub college_by_pkey($$$)
{	my $class = shift;
	my $hpc = shift;
	my $pkey = shift;
	my $me = 'college_by_pkey';
	$class->_validate_hpc($me, $hpc);

	my $list = $class->pdb_query($hpc, coll_pkey => $pkey );
	croak "$me: Error querying dept_pkey '$pkey': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 ) 
	{	my @tmp = map { $_->name } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching coll_pkey=$pkey, names=[$tmp] at ";
	}
	return $list->[0];
}


1;
__END__

=head1 NAME

Glue::HPCC::Project::College;

=head1 SYNOPSIS

  use Glue::HPCC::Project::College;
  use Glue::HPCC::Cluster;

  my $hpc = new Glue::HPCC::Cluster ...;

  my $cmns = Glue::HPCC::Project::ProjectType->college_by_code($hpc, 'cmns');

  ...


=head1 DESCRIPTION

This is the base class representing a record in the college table
in the Project database.   The college record has the following 
attributes:

=over 4

=item B<coll_pkey>: primary key

=item B<code>: This is the 4 letter or so code we use when referring to the college on the command line scripts, etc.

=item B<name>: This is the name which is shown on web usage stats page, etc.

=item B<description>: Longer, more descriptive name

=back

=head2 Special Functions

In all of the following, B<$hpc> is a B<Glue::HPCC::Cluster> object, providing
DB connection info, etc.

=over 4

=item B<college_by_code($hpc, $code)>

Returns the B<Glue::HPCC:Project::College> object for the college with
code $code, if it exists.  Returns undef if no college found.  Croaks or
returns a true, non-ref on errors.

=item B<college_by_pkey($hpc, $pkey)>

Returns the B<Glue::HPCC:Project::College> object for the college with
pkey $pkey, if it exists.  Returns undef if no college found.  Croaks or
returns a true, non-ref on errors.

=back
