#!/usr/local/bin/perl
#
#Department table in projects DB

package Glue::HPCC::Project::Department;
use strict;
use warnings;
use base qw(Glue::HPCC::Project::RWBase);
use Carp qw(croak carp confess);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $GHPDept = 'Glue::HPCC::Project::Department';

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @ro_accessors = qw( dept_pkey );
my @rw_accessors = qw(
	code
	name
	description
	college
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);

sub _ro_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_ro_fields;
	return [ @ro_accessors, @$parent ];
}

sub _rw_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_rw_fields;
	return [ @rw_accessors, @$parent ];
}

sub _required_fields($)
{	return [ qw( name college ) ];
}

sub _db_schema_name($)
{       my $class = shift;
        return 'hpcc_shared';
}

sub _db_table_primary_keyname($)
#Name of primary key for our table.  Can return a scalar string or a list ref
#of scalar strings if multikey table
{	my $class = shift;
	return 'dept_pkey';
}

#--------------------------------------------------------------------
#	Special class methods
#--------------------------------------------------------------------

	
sub dept_by_code($$$)
{	my $class = shift;
	my $hpc = shift;
	my $code = shift;
	my $me = 'dept_by_code';
	$class->_validate_hpc($me, $hpc);

	my $list = $class->pdb_query($hpc, code => $code );
	croak "$me: Error querying code '$code': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 )
	{	my @tmp = map { $_->dept_pkey } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching code=$code, pkeys=[$tmp] at ";
	}
	return $list->[0];
}

sub dept_by_name($$$)
{	my $class = shift;
	my $hpc = shift;
	my $name = shift;
	my $me = 'dept_by_name';
	$class->_validate_hpc($me, $hpc);

	my $list = $class->pdb_query($hpc, name => $name );
	croak "$me: Error querying name '$name': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 )
	{	my @tmp = map { $_->dept_pkey } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching name=$name, pkeys=[$tmp] at ";
	}
	return $list->[0];
}

sub dept_by_pkey($$$)
{	my $class = shift;
	my $hpc = shift;
	my $pkey = shift;
	my $me = 'dept_by_pkey';
	$class->_validate_hpc($me, $hpc);

	my $list = $class->pdb_query($hpc, dept_pkey => $pkey );
	croak "$me: Error querying dept_pkey '$pkey': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 ) 
	{	my @tmp = map { $_->name } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching dept_pkey=$pkey, names=[$tmp] at ";
	}
	return $list->[0];
}


1;
__END__

=head1 NAME

Glue::HPCC::Project::Department;

=head1 SYNOPSIS

  use Glue::HPCC::Project::Department;
  use Glue::HPCC::Cluster;

  my $hpc = new Glue::HPCC::Cluster ...;

  my $math = Glue::HPCC::Project::Department->dept_by_code($hpc,'math');

  ...


=head1 DESCRIPTION

This is the base class representing a record in the Department table
in the Project database.   The dept record has the following 
attributes:

=over 4

=item B<dept_pkey>: primary key

=item B<code>: Short name (~ 4 chars) we use on cmd line tools

=item B<name>: Longer name shown on web usage page, etc.

=item B<description>: Even longer, more descriptive name

=item B<college>: Reference to coll_pkey in College table, for the college to which the department belongs.

=back

=head2 Special Functions

In all of the following, B<$hpc> is a B<Glue::HPCC::Cluster> object, providing
DB connection info, etc.

=over 4

=item B<dept_by_code($hpc, $code)>

Returns the B<Glue::HPCC:Project::Department> object for the dept with
code $code, if it exists.  Returns undef if no department found.  Croaks or
returns a true, non-ref on errors.

=item B<dept_by_pkey($hpc, $pkey)>

Returns the B<Glue::HPCC:Project::Department> object for the dept with
pkey $pkey, if it exists.  Returns undef if no department found.  Croaks or
returns a true, non-ref on errors.

=back
