#!/usr/local/bin/perl
#
#last_email table in projects DB

package Glue::HPCC::Project::LastEmail;
use strict;
use warnings;
use base qw(Glue::HPCC::Project::RWBase);

use Carp qw(croak carp confess);
use POSIX qw(strftime);
use SQL::Interpolate qw(sql_interp);
use Time::ParseDate;

use Glue::HPCC::Utils::MailUtils qw(send_template_email);
use Glue::HPCC::Project::Person;

#Do slightly less than full days to ensure not skipped if cron running a bit late
my $DEFAULT_MIN_HRS_BTWN_ALERTS = 4*24 - 0.5;

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @ro_accessors = qw( lmail_pkey );
my @rw_accessors = qw(
	username
	cluster
	project
	account
	mailcode
	time
	description
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);

sub _ro_fields($)
{	my $class = shift;
	#my $parent = $class->SUPER::_ro_fields;
	#We don't want parent fields because we don't have timestamps
	#return [ @ro_accessors, @$parent ];
	return [ @ro_accessors ];
}

sub _rw_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_rw_fields;
	return [ @rw_accessors, @$parent ];
}

sub _required_fields($)
{	return [ qw( username time cluster project account mailcode) ];
}

sub _db_schema_name($)
{	my $class = shift;
	return 'hpcc_shared';
}

sub _db_table_name($)
{	my $class = shift;
	return 'last_email';
}

sub _db_table_primary_keyname($)
#Name of primary key for our table.  Can return a scalar string or a list ref
#of scalar strings if multikey table
{	my $class = shift;
	return 'lmail_pkey';
}

sub _autoset_lastupdated($)
{	return 0;
}

sub _autoset_created($)
{	return 0;
}

#-------------------------------------------------------------------
#	Special class methods 
#-------------------------------------------------------------------

sub get_last_email_record($@)
#Given a list of key => value pairs identifying what we want, returns
#the last email record from table, or undef if no such record.
#Returns non-ref error string on non-fatal errors
{	my $class = shift;
	my %args = @_;
	my $me = delete $args{me} || 'get_last_email_record';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $cluster = $hpc->db_clustername;

	my $uname = delete $args{username};
	my $project = delete $args{project};
	my $account = delete $args{account};
	my $mailcode = delete $args{mailcode};

	#These represent mistakes in coding, so are fatal
	unless ( $uname )
	{	croak "$me: Missing required argument 'username' at ";
	}
	#unless ( $cluster )
	#{	croak "$me: Missing required argument 'cluster' at ";
	#}
	unless ( $project )
	{	croak "$me: Missing required argument 'project' at ";
	}
	unless ( $account )
	{	croak "$me: Missing required argument 'account' at ";
	}
	unless ( $mailcode )
	{	croak "$me: Missing required argument 'mailcode' at ";
	}
	if ( scalar(%args) )
	{	my $tmp = join ', ', (keys %args);
		croak "$me: Extraneous arguments [ $tmp ], aborting at ";
	}
	
	my $tmpparms = "$uname|$cluster|$project|$account|$mailcode";

        my $list = $class->pdb_query($hpc, 
		username => $uname,
		cluster => $cluster,
		project => $project,
		account => $account, 
		mailcode => $mailcode,
	);


        croak "$me: Error querying last_email record for $tmpparms; $list at "
                if $list && ref($list) ne 'ARRAY';
        return unless scalar(@$list);

        if ( scalar(@$list) > 1 )
        {       my @tmp = map { $_->time } @$list;
                my $tmp = join ', ', @tmp;
                croak "$me: Multiple rows matching $tmpparms, times=[$tmp] at ";
        }
        return $list->[0];
}

sub time_of_last_email($@)
#Takes same args as get_last_email_record (and actually invokes it)
#Returns undef if never emailed, or time of last email
{	my $class = shift;
	my @args = @_;
	my $me = 'time_of_last_email';

	my $lmail = $class->get_last_email_record(@args, me => $me);
	return unless defined $lmail;

	unless ( $lmail && ref($lmail) )
	{	my $tmp = join ', ', @args;
		croak "$me: Error looking up last email time [$tmp]: $lmail at ";
	}

	my $time = $lmail->time;
	return $time;
}

sub tstamp_of_last_email($@)
#Takes same args as get_last_email_record (and actually invokes it)
#Returns undef if never emailed, or tstamp (seconds form 1970) of last email
{	my $class = shift;
	my @args = @_;
	my $me = 'time_of_last_email';

	my $lmail = $class->get_last_email_record(@args, me => $me);
	return unless defined $lmail;

	unless ( $lmail && ref($lmail) )
	{	my $tmp = join ', ', @args;
		croak "$me: Error looking up last email time [$tmp]: $lmail at ";
	}

	my $time = $lmail->time;
	return unless defined $time;

	my $tstamp = parsedate($time,
		WHOLE=>1, 
		DATE_REQUIRED=>1,
		NO_RELATIVE=>1,
		VALIDATE=>1,
	);
	croak "$me: Invalid time '$time' returned, cannot parse at "
		unless defined $tstamp;
	return $tstamp;
}

sub set_last_email_record($@)
#Given a list of key => value , sets the corresponding last email time.
#Recognizes parameters
#	hpc: Glue::HPCC::Cluster instance
#	username : user email sent to (REQUIRED)
#	#cluster: cluster email about (REQUIRED)  This is set from $hpc
#	project: project email about (REQUIRED)
#	account: account email about (REQUIRED)
#	mailcode: mailcode for email (REQUIRED)
#	time: time to set to.  Defaults to now
#	description: description to use, default is undef
#
#If a record already exists, the time and description will be set to what
#is specified (undef description will nullify any previous value).
#If no existing record, adds a new record
#
#Returns false on success, true (error message) on non-fatal error.
{	my $class = shift;
	my %args = @_;
	my $me = 'set_last_email_record';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $cluster = $hpc->db_clustername;

	my $uname = delete $args{username};
	my $project = delete $args{project};
	my $account = delete $args{account};
	my $mailcode = delete $args{mailcode};
	my $time = delete $args{time} || localtime;
	my $desc = delete $args{description};

	#These represent mistakes in coding, so are fatal
	unless ( $uname )
	{	croak "$me: Missing required argument 'username' at ";
	}
	#unless ( $cluster )
	#{	croak "$me: Missing required argument 'cluster' at ";
	#}
	unless ( $project )
	{	croak "$me: Missing required argument 'project' at ";
	}
	unless ( $account )
	{	croak "$me: Missing required argument 'account' at ";
	}
	unless ( $mailcode )
	{	croak "$me: Missing required argument 'mailcode' at ";
	}
	if ( scalar(%args) )
	{	my $tmp = join ', ', (keys %args);
		croak "$me: Extraneous arguments [ $tmp ], aborting at ";
	}
	
	my $tmpparms = "$uname|$cluster|$project|$account|$mailcode";
	my $err;

        my $list = $class->pdb_query($hpc, 
		username => $uname,
		cluster => $cluster,
		project => $project,
		account => $account, 
		mailcode => $mailcode,
	);

	if ( $list && ref($list) && scalar(@$list) )
	{	#Previous record exists
		warn "$me: Multiple records for $tmpparms, only updating first\n"
			if scalar(@$list) > 1;
		$list = $list->[0];
		my $pkey = $list->lmail_pkey;
		$err = $class->pdb_update($hpc,
			{ lmail_pkey => $pkey }, #where
			#Values to set to (most should already be set to these anyway)
			{ username => $uname,
			  cluster => $cluster,
			  project => $project,
			  account => $account,
			  mailcode => $mailcode,
			  time => $time,
			  description => $desc,
			}
		);
		$hpc->db_commit unless $err;
		return $err;
	}
	return "$me: Error checking for previous record [$tmpparms]: $list" 
		if $list & ! ref($list) eq 'ARRAY';

	#No previous record, create new
	$err = $class->pdb_insert($hpc,
			username => $uname,
			  cluster => $cluster,
			  project => $project,
			  account => $account,
			  mailcode => $mailcode,
			  time => $time,
			  description => $desc,
	);
	$hpc->db_commit unless $err;
	return $err;
}

sub nospam_send_template_email($@)
#This is a wrapper around Glue::HPCC::Utils::MailUtils::send_template_email
#The "nospam" is because it checks when the last email of this sort sent
#to the recipient, and won't send too frequently.
#
#It takes the following key => value pairs
#	hpc: Glue:HPCC::Cluster instance. REQUIRED
#	username: username of user being emailed.  REQUIRED
#	project: project being emailed about.  REQUIRED
#	account: account being emailed about.  REQUIRED
#	mailcode: mailcode for last email table.  REQUIRED
#	time: time to use for this email, defaults to current time
#	description: description to add to the last mail log.  Defaults to null
#	template: the template of email to send, see below.  REQUIRED.
#	macros: hash ref of macros to expand in template, defaults to {}
#Template can be given as any of the following:
#       scalar: filename containing template, relative to TEMPLATE_ROOT (see TemplateUtils)
#       scalar ref: a text string containing the template
#       array ref: a list of text strings, one per line, forming the template
#
#Username MUST be in the Person database, will be looked up, and min_hrs_btwn_alerts
#and email will be recorded.  (Defaulting to 0 and $username@umd.edu)
#The last matching email sent will also be found.
#If the last email was sent too recently, just return.
#Otherwise, we add the following values to the macro hash:
#	USERNAME: $username
#	FULLNAME: fullname of user (from Person DB)
#	TO_EMAIL: email address of user (from Person DB)
#	DB_CLUSTERNAME: from $hpc
#	CLUSTERNAME: from $hpc
#	ACCOUNT: $account
#	PROJECT: $project
#	BLOCK_UNTIL: human readable time of $time + $min_hrs_btwn_alerts
#
#The template will be expanded with these and any macros in $macros and sent as email
#(all this via send_template_email)
#
#NOTE: It is up to the template to actually have a header line causing email to
#be delivered to TO_EMAIL, this routine just sets the template variable, and
#updates the time in last_mail table (as we assume it was sent).
#
#Returns false on success (undef if no mail sent, 0 if mail sent), 
#or true (error message) on non-fatal errors
{	my $class = shift;
	my %args = @_;
	my $me = 'nospam_send_template_email';

	my $hpc = delete $args{hpc};
	$class->_validate_hpc($me, $hpc);
	my $cluster = $hpc->db_clustername;

	my $uname = delete $args{username};
	my $project = delete $args{project};
	my $account = delete $args{account};
	my $mailcode = delete $args{mailcode};
	my $time = delete $args{time} || localtime;
	my $desc = delete $args{description};
	my $template = delete $args{template};
	my $macros = delete $args{macros} || {};

	croak "$me: Missing required parameter 'username' at " unless ($uname);
	croak "$me: Missing required parameter 'project' at " unless ($project);
	croak "$me: Missing required parameter 'account' at " unless ($account);
	croak "$me: Missing required parameter 'mailcode' at " unless ($mailcode);
	croak "$me: Missing required parameter 'template' at " unless ($template);

	if ( scalar(%args) )
	{	my $tmp = join ', ', (keys %args);
		croak "$me: Extraneous arguments [ $tmp ], aborting at ";
	}

	#Get tstamp for current time
	my $tstamp = parsedate( $time,
		WHOLE => 1,
		VALIDATE => 1,
		DATE_REQUIRED => 1,
		NO_RELATIVE => 1,
	);
	return "$me: Invalid time '$time' given, cannot parse" unless defined $tstamp;


	my $person = Glue::HPCC::Project::Person->person_by_username($hpc, $uname);
	my ($fullname, $email, $min_hrs_btwn_alerts);
	my $err;

	if ( $person &&  ref($person) )
	{	#We got a live one
		$fullname = $person->fullname || $uname;
		$email = $person->email || "${uname}\@umd.edu";
		$min_hrs_btwn_alerts = $person->min_hrs_btwn_alerts || $DEFAULT_MIN_HRS_BTWN_ALERTS;
	} elsif ( $person )
	{	#We got an error looking up person
		return "$me: Error looking up recipient '$uname': $person";
	} else
	{	#No user in system with that name
		#Should this continue or not?
		$fullname = $uname;
		$email = "${uname}\@umd.edu";
		$min_hrs_btwn_alerts = $DEFAULT_MIN_HRS_BTWN_ALERTS;
		return "$me: No user $uname found in PDB";
	}

	my $tstamp_last = $class->tstamp_of_last_email(
		hpc => $hpc,
		username => $uname,
		project => $project,
		account => $account,
		mailcode => $mailcode,
	);
	$tstamp_last = 0 unless defined $tstamp_last; #Default to start of epoch

	if ( $tstamp < $tstamp_last + 60*60*$min_hrs_btwn_alerts )
	{	#We emailed them too recently, so don't email again
		print STDERR "skipping email to $email ($uname, $fullname), emailed too recently...\n";
		return;
	}

	#OK, we should be sending email
	#Generate macros
	$macros->{USERNAME} = $uname;
	$macros->{FULLNAME} = $fullname;
	$macros->{TO_EMAIL} = $email;
	$macros->{DB_CLUSTERNAME} = $hpc->db_clustername;
	$macros->{CLUSTERNAME} = $hpc->clustername;
	$macros->{PROJECT} = $project;
	$macros->{ACCOUNT} = $account;
	my $tstamp_block = $tstamp + 60*60*$min_hrs_btwn_alerts;
	my $time_block = strftime("%Y %b %d %H:%M", localtime($tstamp_block));
	$macros->{BLOCK_UNTIL} = $time_block;

	#Send mail
	$err = send_template_email($template, $macros);
	#send_template_email does not do great error handling/return value, should fix
	$err = 0;
	return "$me: Error sending mail to $email ($uname): $err" if $err;

	#Update last email timestamp
	$err = $class->set_last_email_record(
		hpc => $hpc,
		username => $uname,
		project => $project,
		account => $account,
		mailcode => $mailcode,
		time => $time,
		description => $desc,
	);
	return "$me: Error setting last email tstamp for $uname to $time: $err" if $err;
	return 0; #Mail sent, so return 0
}

1;
__END__

=head1 NAME

Glue::HPCC::Project::LastMail

=head1 SYNOPSIS

  use Glue::HPCC::Project::LastMail;
  use Glue::HPCC::Cluster;

  my $hpc = new Glue::HPCC::Cluster ...;

  my $error = Glue::HPCC::Project::LastMail->nospam_send_template_email(
	username => 'kevin',
	project => 'tptest',
	account => 'tptest-hi',
	mailcode => 'usage-warning',
	description => 'tptest-hi has excessive consumption rate',
	template => "dt2_excessive_usage_coop.tmplate"
	macros => 
		{	variable => value,
		}
  );
  ...


=head1 DESCRIPTION


This is the base class representing a record in the last_email table
in the Project database.  The following fields are supported:

=over 4

=item B<lmail_pkey>: The primary key

=item B<username>: The username of person mail was sent to

=item B<cluster>: db_clustername of cluster email was about

=item B<project>: name of project email was about

=item B<account>: Name of account affected.

=item B<mailcode>: A terse string code type of email (e.g. usage-warning)

=item B<time>: Time of last email sent

=item B<description>: Short description of email

=back

All of the above are text strings, so can contain either text or numeric
values.  No foreign keys, so do not need to delete anything after a project/
account/user is deleted.  Although B<lmail_pkey> is listed as the primary
key, there is an unique constraint on B<username, cluster, project, account,>
and B<mailcode>.

=head2 Special Class Methods

The following special methods are provided for interacting with the
last_email table.  They are all class methods, and other than the class name
all variables are given as a list of key => value pairs.

The following parameters are common (and REQUIRED) in all meethods below
unless otherwise stated:

=over 4

=item B<hpc>: A Glue::HPCC::Cluster instance.  The B<cluster> field is obtained from this.

=item B<username>: the username value to match/set.

=item B<project>: the project name value to match/set.

=item B<account>: the account name value to match/set.

=item B<mailcode>: the mailcode value to match/set.

=back

The method B<get_last_email_record> will return the Glue::HPCC::Project::LastMail
record matching the username, project, account, mailcode, and cluster (from hpc),
or undef if no such record found.  The methods B<time_of_last_email> and
B<tstamp_of_last_email> are similar, but instead of returning the LastMail instance,
they return the time string or timestamp (seconds from 1970), respectively, if the
record was found (and undef if not found).

The method B<set_last_email_record> will also take a B<time> and B<description>
parameter, in addition to the required parameters above.  It will either update
the time and description of the matching existing record, or insert a new record.
B<time> will default to the current time, and B<description> to null (which will
overwrite any existing description).

But the most useful and commonly used method is B<nospam_send_template_email>.  
This is a wrapper around B<Glue::HPCC::Utils::MailUtils::send_template_email>, with
the "nospam" signifying that it will look up the recipient in Person and LastMail
databases, and only send the email if the person was not sent a similar mail within
the last B<min_hrs_btwn_alerts> (obtained from the Person DB) hours.  It takes the
required parameters above, as well as:

=over 4

=item B<time>: The time to use, mainly for debugging.  Defaults to current time.

=item B<description>: The description to use when updating last_email table.  Defaults to NULL.

=item B<template>: The template for the email to be sent.  It accepts any of the formats acceptable to Glue::HPCC::Utils::TemplateUtils::expand_template.  REQUIRED.

=item B<macros>: A hash ref of macros to use for template ( macro_name => value).  Default is {}

=back

The username MUST be in the person database, and mail will be sent to the email
address recorded for the user (or $username@umd.edu).  If matching email was
sent to this user within the past $min_hrs_btwn_alerts (from Person DB), the
method will just return (undef).  Otherwise, we add some macros to the $macros
list, call B<send_template_email>, update the last_email timestamp, and return 0.

The method always returns false (undef or 0, depending on whether mail actually sent)
on success, and true (an error message) on non-fatal errors.

The following macros are added to the macro list in hash ref $macros:

=over 4

=item USERNAME: the username provided

=item FULLNAME: the fullname corresponding to username, from Person DB.

=item TO_EMAIL: the email address corresponding to username, from Person DB

=item DB_CLUSTERNAME: from $hpc

=item CLUSTERNAME: from $hpc

=item PROJECT: the project provided

=item ACCOUNT: the account provided

=item BLOCK_UNTIL: human readable time before which no more emails to be sent

=back

B<NOTE:> The method actually only expands the template and sends it to sendmail.
The template must provide both HEADER and BODY of the email, and in particular
set the To: header to TO_EMAIL in order for the email to be sent to the user.
This is assumed by the method, and the last_email will actually be set even if
that is not the case.

