#!/usr/local/bin/perl
#
#Log table in projects DB

package Glue::HPCC::Project::Log;
use strict;
use warnings;
use base qw(Glue::HPCC::Project::RWBase);
use Carp qw(croak carp confess);
use POSIX qw(strftime);
use File::Basename qw(basename);
use SQL::Interpolate qw(sql_interp);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $_ptype_list;

my $_ptype_by_name;

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @ro_accessors = qw( log_pkey );
my @rw_accessors = qw(
	username
	program
	hostname
	time
	reason
	action_code
	project
	account
	cluster
	person
	tablename
	fieldname
	oldval
	newval
	description
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);

sub _ro_fields($)
{	my $class = shift;
	#my $parent = $class->SUPER::_ro_fields;
	#We don't want parent fields because we don't have timestamps
	#return [ @ro_accessors, @$parent ];
	return [ @ro_accessors ];
}

sub _rw_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_rw_fields;
	return [ @rw_accessors, @$parent ];
}

sub _required_fields($)
{	return [ qw( username program hostname time action_code) ];
}

sub _db_schema_name($)
{	my $class = shift;
	return 'hpcc_shared';
}

sub _db_table_primary_keyname($)
#Name of primary key for our table.  Can return a scalar string or a list ref
#of scalar strings if multikey table
{	my $class = shift;
	return 'log_pkey';
}

sub _autoset_lastupdated($)
{	return 0;
}

sub _autoset_created($)
{	return 0;
}

#-------------------------------------------------------------------
#	Special class methods 
#-------------------------------------------------------------------

sub add_log_entry($$@)
#Takes Glue::HPCC::Cluster instance and list of key => value pairs.
#Defaults fields it can, and adds a log entry to DB
{	my $class = shift;
	my $hpc = shift;
	my %args = @_;
	my $me = 'add_log_entry';
	$class->_validate_hpc($me, $hpc);

	#my $user = delete $args{username} || $ENV{USER};
	my $user = delete $args{username};
	unless ($user)
	{	$user = $ENV{USER};
		$user = $ENV{SUDO_USER}
			if ( $user eq 'root' && $ENV{SUDO_USER} );
	}
	my $program = delete $args{program} || (basename $0);
	my $hostname = delete $args{hostname} || $ENV{HOSTNAME};
	my $time = delete $args{time} || time;
	my $cluster = delete $args{cluster} || $hpc->db_clustername;

	my $reason = delete $args{reason};
	my $action_code = delete $args{action_code};
	my $description = delete $args{description};

	unless ( $reason )
	{	croak "$me: Missing required argument 'reason' at ";
	}
	unless ( $action_code )
	{	croak "$me: Missing required argument 'action_code' at ";
	}
	unless ( $description )
	{	croak "$me: Missing required argument 'description' at ";
	}

	#Fix time to mysql like string
	my $tstring = $time;
	if ( $tstring =~ /^\s*\d+\s*$/ )
	{	#Looks like a timestamp
		$tstring = strftime("%Y-%m-%d %H:%M:%S", localtime($time) );
	}

	my @args =
	(	action_code => $action_code,
		reason => $reason,
		description => $description,

		time => $tstring,
		username => $user,
		program => $program,
		hostname => $hostname,
		cluster => $cluster,
	);

	#Add any optional fields
	my @opt_fields = qw( project person tablename account
		fieldname oldval newval);
	KEY: foreach my $key (@opt_fields)
	{	my $val = delete $args{$key};
		next KEY unless defined $val;
		push @args, $key => $val;
	}

	if ( scalar(keys %args) )
	{	my @tmp = keys %args;
		my $tmp = join ", ", @tmp;
		croak "$me: Unrecognized arguments [ $tmp ] given at ";
	}

	return $class->pdb_insert($hpc, @args);
}
	
sub pdb_query($$@)
#Override SUPER pdb_query to allow for time to be a range.
#If $where{time} is not defined or a scalar, same as SUPER pdb_query.
#IF $where{time} is a 2-element list ref [ $start, $end ], treat as
#	time >= $start and time <= $end
#IF start is undef, default to 1970-01-01
#IF end is undef (or 1-element list ref), default end to 2270-01-01
#If $where{time} is anything else, croak
{	my $class = shift;
	my $hpc = shift;
	my %where = @_;
	my $me = 'pdb_query';
	$class->_validate_hpc($me, $hpc);

	my $dbh = $hpc->get_projects_dbh;

	my $time = delete $where{time};
	return $class->SUPER::pdb_query($hpc, %where) 
		unless $time && ref($time) eq 'ARRAY' && scalar(@$time);
	croak "$me: Too many elements for 'time'; expecting [start, end] at "
		if scalar(@$time) > 2;
	my $start = $time->[0] || '1970-01-01';
	my $end   = $time->[1] || '2270-01-01';

	my $table = $class->_db_fq_table_name;
	print STDERR "$me: DEBUG: start=$start, end=$end\n";
	my ($sql, @bind);
	if ( scalar(keys %where) )
	{	($sql, @bind) = sql_interp "SELECT * FROM $table WHERE", 
			\%where, 'AND time > ', \$start, ' AND time < ',
			\$end;
	} else
	{	($sql, @bind) = sql_interp 
			"SELECT * FROM $table WHERE time >", \$start,
			' AND time < ', \$end;
	}

	my $sth = $dbh->prepare_cached($sql);
	croak "$me: Error preparing SQL '$sql': " . $dbh->errstr . " at "
		unless $sth && ref($sth);

	if ( $hpc->verbose || $class->verbose )
	{	$class->verbose_display_sql($sql, @bind);
	}
	my $ok = $sth->execute(@bind);
	croak "$me: Error executing '$sql', " . $sth->errstr . " at "
		unless $ok;

	my @results = ();
	while (my $row = $sth->fetchrow_hashref )
	{	my $rowobj = $class->new_from_dbi_hashref($row);
		croak "$me: error creating from row: $rowobj at "
			unless $rowobj && ref($rowobj);
		push @results, $rowobj;
	}
	return [ @results ];
}

1;
__END__

=head1 NAME

Glue::HPCC::Project::Log

=head1 SYNOPSIS

  use Glue::HPCC::Project::Log
  use Glue::HPCC::Cluster;

  my $hpc = new Glue::HPCC::Cluster ...;

  my $error = Glue::HPCC::Project::Log->add_log_entry( $hpc,
	project => 'tptest',
	person => 'kevin',
	action_code => 'user-added-to-project',
	description => 'add kevin to tptest',
	reason => 'He gave me a tortilla chip',
  );
  if ($error )
  {	die "Error adding kevin to tptest: $error\n";
  ...


=head1 DESCRIPTION

This is the base class representing a record in the log table
in the Project database.  The following fields are supported:

=over 4

=item B<log_pkey>: The primary key

=item B<username>: Username of person making the change

=item B<program>: Name of program making the change

=item B<time>: Time of entry/change

=item B<reason>: Text string describing why the change was made.  Typically an ITSC number or similar.

=item B<action_code>: A terse string code describing type of change.

=item B<project>: Name of project affected.

=item B<hostname>: Hostname of system change made from.

=item B<cluster>: Name of cluster affected.

=item B<account>: Name of account affected.

=item B<person>: Username of person affected.

=item B<tablename>: Name of table affected.

=item B<fieldname>: Name of field affected.

=item B<oldval>: Old value of B<tablename.fieldname>

=item B<newval>: New value of B<tablename.fieldname>

=item B<description>: Textual description of change.

=back

All of the above are text strings, so can contain either text or numeric
values.  No foreign keys, so do not need to delete anything after a project/
account/user is deleted.

=head2 Special Functions

The primary method people are expect to use is the class method
B<add_log_entry>.  It takes a B<Glue::HPCC::Cluster> instance (to provide
DB connection information), followed by a list of key => value pairs.
Recognized keys are:

=over 4

=item B<user>: Username to log changed under.  Defaults to $ENV{USER}

=item B<program>: Name of program making change.  Defaults to basename $0

=item B<hostname>: Hostname of system change made from.  Defaults to system running on.

=item B<time>: Time of entry/change.  Defaults to current time.

=item B<reason>: Reason for the change.  REQUIRED.

=item B<action_code>: Terse string describing type of change.  REQUIRED.

=item B<project>: Name of project affected.  Default is NULL.

=item B<cluster>: Name of cluster affected.  Default is NULL.

=item B<account>: Name of account affected.  Default is NULL.

=item B<person>: Username of person affected.  Default is NULL.

=item B<tablename>: Name of table affected.  Default is NULL.

=item B<fieldname>: Name of field affected.  Default is NULL.

=item B<oldval>: Old value of B<tablename.fieldname>  Default is NULL.

=item B<newval>: New value of B<tablename.fieldname>  Default is NULL.

=item B<description>: Text describing change.  REQUIRED.

=back

It adds a new log entry as described.
