#!/usr/local/bin/perl
#
#Person table in projects DB

package Glue::HPCC::Project::Person;
use strict;
use warnings;
use base qw(Glue::HPCC::Project::RWBase);
use Carp qw(croak carp confess);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $GHPPerson = 'Glue::HPCC::Project::Person';

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @ro_accessors = qw( ppl_pkey );
my @rw_accessors = qw(
	username 
	fullname
	email
	uid
	active
	min_hrs_btwn_alerts
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);

sub _ro_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_ro_fields;
	return [ @ro_accessors, @$parent ];
}

sub _rw_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_rw_fields;
	return [ @rw_accessors, @$parent ];
}

sub _required_fields($)
{	return [ qw( username email uid ) ];
}

sub _db_schema_name($)
{       my $class = shift;
        return 'hpcc_shared';
}

sub _db_table_primary_keyname($)
#Name of primary key for our table.  Can return a scalar string or a list ref
#of scalar strings if multikey table
{	my $class = shift;
	return 'ppl_pkey';
}

#--------------------------------------------------------------------
#	Special class methods
#--------------------------------------------------------------------

	
sub person_by_username($$$)
{	my $class = shift;
	my $hpc = shift;
	my $username = shift;
	my $me = 'person_by_username';
	$class->_validate_hpc($me, $hpc);

	my $list = $class->pdb_query($hpc, username => $username );
	croak "$me: Error querying username '$username': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 )
	{	my @tmp = map { $_->ppl_pkey } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching username=$username, pkeys=[$tmp] at ";
	}
	return $list->[0];
}

sub person_by_uid($$$)
{	my $class = shift;
	my $hpc = shift;
	my $uid = shift;
	my $me = 'person_by_uid';
	$class->_validate_hpc($me, $hpc);

	my $list = $class->pdb_query($hpc, uid => $uid );
	croak "$me: Error querying uid '$uid': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 ) 
	{	my @tmp = map { $_->ppl_pkey } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching uid=$uid, pkeys=[$tmp] at ";
	}
	return $list->[0];
}

sub person_by_pkey($$$)
{	my $class = shift;
	my $hpc = shift;
	my $pkey = shift;
	my $me = 'person_by_pkey';
	$class->_validate_hpc($me, $hpc);

	my $list = $class->pdb_query($hpc, ppl_pkey => $pkey );
	croak "$me: Error querying pkey '$pkey': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 ) 
	{	my @tmp = map { $_->username } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching pkey=$pkey, usernames=[$tmp] at ";
	}
	return $list->[0];
}

1;
__END__

=head1 NAME

Glue::HPCC::Project::Person;

=head1 SYNOPSIS

  use Glue::HPCC::Project::Person;
  use Glue::HPCC::Cluster;

  my $hpc = new Glue::HPCC::Cluster ...;

  my $kevin = Glue::HPCC::Project::ProjectType->person_by_username(
	$hpc, 'kevin');

  ...


=head1 DESCRIPTION

This is the base class representing a record in the person table
in the Project database.   The person record has the following 
attributes:

=over 4

=item B<ppl_pkey>: primary key

=item B<username>: Glue username.  REQUIRED

=item B<fullname>: Fullname, human readable

=item B<email>: Email address.  Defaults to B<username>@umd.edu. REQUIRED.

=item B<uid>: Glue unix userid.  REQUIRED

=back

=head2 Special Functions

In all of the following, B<$hpc> is a B<Glue::HPCC::Cluster> object, providing
DB connection info, etc.

=over 4

=item B<person_by_username($hpc, $username)>

Returns the B<Glue::HPCC:Project::Person> object of the user with specified
username, if it exists.  Returns undef if no person found.  Croaks or
returns a true, non-ref on errors.

=item B<person_by_uid($hpc, $uid)>

Returns the B<Glue::HPCC:Project::Person> object of the user with specified
uid, if it exists.  Returns undef if no person found.  Croaks or
returns a true, non-ref on errors.


=back
