#!/usr/local/bin/perl
#
#person_project_m2m table in projects DB

package Glue::HPCC::Project::PersonProject;
use strict;
use warnings;
use base qw(Glue::HPCC::Project::RWBase);
use Carp qw(croak carp confess);

use Glue::HPCC::Project::Person;
use Glue::HPCC::Project::Project;

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @ro_accessors = qw( prj_pkey ppl_pkey );
my @rw_accessors = qw(
	is_member
	is_poc
	is_coordinator
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);

sub _ro_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_ro_fields;
	return [ @ro_accessors, @$parent ];
}

sub _rw_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_rw_fields;
	return [ @rw_accessors, @$parent ];
}

sub _required_fields($)
{	return [ ];
}


sub _db_table_name($)
#Default name of table we are in.  Subclasses can override, but a good
#default
{	my $class = shift;
	return 'person_project_m2m';
}

sub _db_table_primary_keyname($)
#Name of primary key for our table.  Can return a scalar string or a list ref
#of scalar strings if multikey table
{	my $class = shift;
	return [ 'prj_pkey', 'ppl_pkey' ];
}

#-------------------------------------------------------------------
#	Special class methods 
#-------------------------------------------------------------------

sub get_project_person_by_names($$$$)
#Takes project and username, and returns person_project_m2m object
#On error, croaks or returns non-ref true value (error message)
{	my $class = shift;
	my $hpc = shift;
	my $projname = shift;
	my $username = shift;

	my $me = 'get_project_person_by_names';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing project name at " unless $projname;
	croak "$me: Missing username at " unless $username;

	my $user = Glue::HPCC::Project::Person->person_by_username(
		$hpc,$username);
	return "$me: No such username '$username' at " 
		unless $user && ref($user);
	my $proj = Glue::HPCC::Project::Project->project_by_name(
		$hpc,$projname);
	return "$me: Invalid project name '$projname' at " 
		unless $proj && ref($proj);

	my $ppl_pkey = $user->ppl_pkey;
	my $prj_pkey = $proj->prj_pkey;

	my $list = $class->pdb_query($hpc, 
		ppl_pkey=>$ppl_pkey, prj_pkey=>$prj_pkey);
	croak "$me: Error querying db for prj_pkey=$prj_pkey, ppl_pkey=$ppl_pkey: $list at "
		unless $list && ref($list) eq 'ARRAY';
	return unless scalar(@$list);
	if  ( scalar(@$list) > 1 )
	{	my $tmp = scalar(@$list);
		croak "$me: Multiple ($tmp) rows returned for prj_pkey=$prj_pkey, ppl_pkey=$ppl_pkey at ";
	}
	return $list->[0];
}

sub get_project_persons_by_projectname($$$)
#Takes a project name, and returns person_project_m2m rows 
#matching this project name
#
#Returns list ref of rows matching projectname on success.
#On error, croaks or returns non-ref true value (error message)
{	my $class = shift;
	my $hpc = shift;
	my $projname = shift;
	my $me = 'get_project_persons_by_projectname';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing project name at " unless $projname;

	my $proj = Glue::HPCC::Project::Project->project_by_name(
		$hpc,$projname);
	unless ($proj)
	{	#Not fatal, allow for dryrun mode
		carp "$me: Invalid project '$projname', not in PDB, at ";
		return [];
	}
	return "$me: Invalid project name '$projname' at " 
		unless ref($proj);
	my $prj_pkey = $proj->prj_pkey;

	my $list = $class->pdb_query($hpc, prj_pkey=>$prj_pkey);
	croak "$me: Error querying db for prj_pkey=$prj_pkey: $list at "
		unless $list && ref($list) eq 'ARRAY';
	return $list;
}

sub get_project_persons_by_username($$$)
#Takes a username, and returns person_project_m2m rows 
#matching this username
#
#Returns list ref of rows matching username on success.
#On error, croaks or returns non-ref true value (error message)
{	my $class = shift;
	my $hpc = shift;
	my $username = shift;
	my $me = 'get_project_persons_by_username';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing username at " unless $username;

	my $user = Glue::HPCC::Project::Person->person_by_username(
		$hpc,$username);
	return "$me: Invalid username '$username' at " 
		unless $user && ref($user);
	my $ppl_pkey = $user->ppl_pkey;

	my $list = $class->pdb_query($hpc, ppl_pkey=>$ppl_pkey);
	croak "$me: Error querying db for ppl_pkey=$ppl_pkey: $list at "
		unless $list && ref($list) eq 'ARRAY';
	return $list;
}

sub get_members_of_project($$$)
#Gets members of project according to PDB
#REturns list ref of usernames in project on success
#On error, croaks or returns non-ref true value (error message)
{	my $class = shift;
	my $hpc = shift;
	my $projname = shift;
	my $me = 'get_members_of_project';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing projname at " unless $projname;

	my $list = $class->get_project_persons_by_projectname(
		$hpc,$projname);
	return [] unless $list;
	croak "$me: Error getting proj/person rows for $projname: $list at"
		unless ref($list) eq 'ARRAY';

	my @members = grep { $_->is_member } @$list;
	@members = map { Glue::HPCC::Project::Person->person_by_pkey(
				$hpc, $_->ppl_pkey) } @members;
	@members = map { $_->username } @members;
	return \@members;
}

sub get_coords_of_project($$$)
#Gets coordinators of project according to PDB
#REturns list ref of usernames in project on success
#On error, croaks or returns non-ref true value (error message)
{	my $class = shift;
	my $hpc = shift;
	my $projname = shift;
	my $me = 'get_coords_of_project';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing projname at " unless $projname;

	my $list = $class->get_project_persons_by_projectname(
		$hpc,$projname);
	return [] unless $list;
	croak "$me: Error getting proj/person rows for $projname: $list at"
		unless ref($list) eq 'ARRAY';

	my @members = grep { $_->is_coordinator } @$list;
	@members = map { Glue::HPCC::Project::Person->person_by_pkey(
				$hpc, $_->ppl_pkey) } @members;
	@members = map { $_->username } @members;
	return \@members;
}

sub get_pocs_for_project($$$)
#Gets pocs for project according to PDB
#REturns list ref of usernames in project on success
#On error, croaks or returns non-ref true value (error message)
{	my $class = shift;
	my $hpc = shift;
	my $projname = shift;
	my $me = 'get_pocs_for_project';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing projname at " unless $projname;

	my $list = $class->get_project_persons_by_projectname(
		$hpc,$projname);
	return [] unless $list;
	croak "$me: Error getting proj/person rows for $projname: $list at"
		unless ref($list) eq 'ARRAY';

	my @members = grep { $_->is_poc } @$list;
	@members = map { Glue::HPCC::Project::Person->person_by_pkey(
				$hpc, $_->ppl_pkey) } @members;
	@members = map { $_->username } @members;
	return \@members;
}

sub get_poc_emails_for_project($$$)
#Gets email addresses for pocs for project according to PDB
#REturns list ref of email addresses for PoCs of project on success
#On error, croaks or returns non-ref true value (error message)
{	my $class = shift;
	my $hpc = shift;
	my $projname = shift;
	my $me = 'get_poc_emails_for_project';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing projname at " unless $projname;

	my $list = $class->get_project_persons_by_projectname(
		$hpc,$projname);
	return [] unless $list;
	croak "$me: Error getting proj/person rows for $projname: $list at"
		unless ref($list) eq 'ARRAY';

	my @members = grep { $_->is_poc } @$list;
	@members = map { Glue::HPCC::Project::Person->person_by_pkey(
				$hpc, $_->ppl_pkey) } @members;
	@members = map { $_->email } @members;
	return \@members;
}

sub get_projects_user_is_attached_to($$$;$)
#Gets projects which given user is associated with, according to PDB
#REturns list ref of project names meeting criteria.
#Wants $hpc, $username, and optionally a $role, which should be one
#of 'member', 'poc', or 'coord', of 'ANY'.  If $role is given, returned list
#is limited to those projects for which user has specified role; if
#role is 'ANY', will return projects for which user has one or more of the
#roles 'member', 'poc' or 'coord'.  If $role is omitted, will return any
#projects for which a PersonProject record exists for that user, even if
#no role flags are set.
#On error, croaks or returns non-ref true value (error message)
{	my $class = shift;
	my $hpc = shift;
	my $username = shift;
	my $role = shift || 'any';
	my $me = 'get_projects_user_is_attached_to';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing username at " unless $username;
	$role = lc $role if $role;

	my $list = $class->get_project_persons_by_username(
		$hpc,$username);
	return [] unless $list;
	croak "$me: Error getting proj/person rows for $username: $list at"
		unless ref($list) eq 'ARRAY';

	my @recs = @$list;

	#Filter by role if requested
	if ( $role eq 'any' )
	{	@recs = grep { $_->is_member || $_->is_poc || 
			$_->is_coordinator } @recs;
	} elsif ( $role eq 'member' )
	{	@recs = grep { $_->is_member } @recs;
	} elsif ( $role eq 'poc' )
	{	@recs = grep { $_->is_poc } @recs;
	} elsif ( $role eq 'coord' || $role eq 'coordinator' )
	{	@recs = grep { $_->is_coordinator } @recs;
	} elsif ( $role )
	{	croak "$me: Invalid value '$role' for role, expecting " .
			"one of [ ANY, member, poc, coord] at ";
	}

	my @projs = map { Glue::HPCC::Project::Project->project_by_pkey(
		$hpc, $_->prj_pkey) } @recs;

	@projs = map { $_->name } @projs;
	return \@projs;
}

sub get_projects_user_is_member_of($$$)
#Returns a list ref of project names of which specified user is a member.
{	my $self = shift;
	my $hpc = shift;
	my $user = shift;
	my $role = 'member';
	return $self->get_projects_user_is_attached_to($hpc, $user, $role);
}

sub get_projects_user_is_coord_of($$$)
#Returns a list ref of project names of which specified user is a coordinator.
{	my $self = shift;
	my $hpc = shift;
	my $user = shift;
	my $role = 'coord';
	return $self->get_projects_user_is_attached_to($hpc, $user, $role);
}

sub get_projects_user_is_poc_of($$$)
#Returns a list ref of project names of which specified user is a PoC
{	my $self = shift;
	my $hpc = shift;
	my $user = shift;
	my $role = 'poc';
	return $self->get_projects_user_is_attached_to($hpc, $user, $role);
}



1;
__END__

=head1 NAME

Glue::HPCC::Project::PersonProject

=head1 SYNOPSIS

  use Glue::HPCC::Project::PersonProject;
  use Glue::HPCC::Cluster;

  my $hpc = new Glue::HPCC::Cluster ...;

  my $project_name = 'important-project';
  my $members = Glue::HPCC::Project::PersonProject->get_members_of_project(
	$hpc, $project_name);
  print "Members of $project_name are:\n\t", (join "\n\t", @$members), "\n";

  ...


=head1 DESCRIPTION

This is the base class representing a record in the person_project_m2m table
in the Project database.  This is the many-to-many table representing 
the association between users and projects.

=head2 Accessors

Instances of B<Glue::HPCC::PersonProject> have the following accessors:

=over 4

=item B<prj_pkey>: The primary key of the project the record is associated with

=item B<ppl_pkey>: The primary key of the person the record is associated with

=item B<is_member>: Boolean, true if user is a member of the project

=item B<is_poc>: Boolean, true if user is a point-of-contact for the project

=item B<is_coordinator>: Boolean, true if user is a coordinator of the project

=back

Note that it is permissible to have records with all of the boolean flags
unset --- this probably means that the user had a role at some time in the
past but was then removed from the role.  

=head2 Special Functions

=over 4

=item B<get_project_person_by_names($hpc, $project_name, $username)>

This looks up the record in the person_project_m2m table corresponding
to the given project and user names.  On success, it returns the corresponding
PersonProject instance.  If no record for that user and project exists in the
table, it returns undef.  Otherwise it returns a non-ref error message (e.g.
if there is no user or project by the given name).

=item B<get_project_persons_by_projectname($hpc, $project_name)>

This looks up records in the person_project_m2m table corresponding to the
named project.  On success, it returns a (possibly empty) list ref of matching
PersonProject instances.  On error, returns a non-ref error text.

=item B<get_project_persons_by_username($hpc, $username)>

This looks up records in the person_project_m2m table corresponding to the
named user.  On success, it returns a (possibly empty) list ref of matching
PersonProject instances.  On error, returns a non-ref error text.

=item B<get_members_of_project($hpc, $project_name)>

This returns a list ref of usernames for users with the role "member" in
the specified project.  On error, returns a non-ref error text.

=item B<get_coords_of_project($hpc, $project_name)>

This returns a list ref of usernames for users with the role "coord" in
the specified project.  On error, returns a non-ref error text.

=item B<get_pocs_for_project($hpc, $project_name)>

This returns a list ref of usernames for users with the role "poc" in
the specified project.  On error, returns a non-ref error text.

=item B<get_poc_emails_for_project($hpc, $project_name)>

Like B<get_pocs_for_project>, but on success returns a list ref
containing the email addresses of the points-of-contact for the named project.

=item B<get_projects_user_is_attached_to($hpc, $username, $role)>

On success, this returns a list ref of project names of the projects
which the specified user is attached to.  If B<$role> is given, it should
be (case insensitive) one of 'ANY', 'member', 'coord', or 'poc'.  If role
is set to 'member', 'coord', or 'poc', only projects for which the user has
the specified role are returned.  If role is 'ANY', only projects for which
the user has one of those roles (i.e. one of the role flags is set).  If
omitted or undef, any project for which user has a PersonProject record
will be returned, even if none of the role flags are set.  On non-fatal
errors, returns a non-ref error string.

=item B<get_projects_user_is_member_of($hpc, $username)>

A convenience function, returning B<get_projects_user_is_attached_to> with
B<$role> = 'member'.

=item B<get_projects_user_is_poc_of($hpc, $username)>

A convenience function, returning B<get_projects_user_is_attached_to> with
B<$role> = 'poc'.

=item B<get_projects_user_is_coord_of($hpc, $username)>

A convenience function, returning B<get_projects_user_is_attached_to> with
B<$role> = 'coord'.



=back
