#!/usr/local/bin/perl
#
#Probject table in projects DB

package Glue::HPCC::Project::Project;
use strict;
use warnings;
use base qw(Glue::HPCC::Project::RWBase);
use Carp qw(croak carp confess);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $GHPPerson = 'Glue::HPCC::Project::Project';

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @ro_accessors = qw( prj_pkey );
my @rw_accessors = qw(
	name 
	description
	unixgroup
	parent
	department

	projtype
	su
	active

	expiration

	thrhld_pct_over
	thrhld_pct_under
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);

sub _ro_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_ro_fields;
	return [ @ro_accessors, @$parent ];
}

sub _rw_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_rw_fields;
	return [ @rw_accessors, @$parent ];
}

sub _required_fields($)
{	return [ qw( name expiration ) ];
}


sub _db_table_primary_keyname($)
#Name of primary key for our table.  Can return a scalar string or a list ref
#of scalar strings if multikey table
{	my $class = shift;
	return 'prj_pkey';
}

#--------------------------------------------------------------------
#	Special class methods
#--------------------------------------------------------------------

sub project_by_pkey($$$)
{	my $class = shift;
	my $hpc = shift;
	my $pkey = shift;
	my $me = 'project_by_pkey';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing pkey at " unless $pkey;

	my $list = $class->pdb_query($hpc, prj_pkey => $pkey );

	croak "$me: Error querying pkey '$pkey': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 ) 
	{	my @tmp = map { $_->name } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching pkey=$pkey, names=[$tmp] at ";
	}
	return $list->[0];
}

sub project_by_name($$$)
{	my $class = shift;
	my $hpc = shift;
	my $name = shift;
	my $me = 'project_by_name';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing name at " unless $name;

	my $list = $class->pdb_query($hpc, name => $name );

	croak "$me: Error querying name '$name': $list at "
		if $list && ref($list) ne 'ARRAY';
	return unless scalar(@$list);

	if ( scalar(@$list) > 1 ) 
	{	my @tmp = map { $_->prj_pkey } @$list;
		my $tmp = join ', ', @tmp;
		croak "$me: Multiple rows matching name=$name, pkeys=[$tmp] at ";
	}
	return $list->[0];
}

sub children_of_project_by_pkey($$;$)
#Returns all projects whose parent matches $pkey; or without a parent if
#$pkey is omitted/undef.
#Returns a (possibly empty) list ref of Glue::HPCC::Project::Project objects
#on success.  
#Croaks or returns non-ref error message on error
{	my $class = shift;
	my $hpc = shift;
	my $pkey = shift;
	my $me = 'children_of_project_by_pkey';
	$class->_validate_hpc($me, $hpc);

	#pkey => undef is converted to IS NULL by SQL::Interpolate.  Nice!
	my $list = $class->pdb_query($hpc, parent => $pkey );

	croak "$me: Error querying children of $pkey: $list at "
		if $list && ref($list) ne 'ARRAY';
	return $list;
}

sub descendents_of_project_by_pkey($$$)
#Returns all projects whose parent matches $pkey, or whose parent's parent
#matches $pkey, or whose parent's parent's parent matches, and so on.
#$pkey is REQUIRED. (otherwise, just would return all projects inefficiently)
#Returns a (possibly empty) list ref of Glue::HPCC::Project::Project objects
#on success.  
#Croaks or returns non-ref error message on error
{	my $class = shift;
	my $hpc = shift;
	my $pkey = shift;
	my $me = 'descendents_of_project_by_pkey';
	$class->_validate_hpc($me, $hpc);

	return "$me: Missing required parameter pkey" unless defined $pkey;

	my $proj = $class->project_by_pkey($hpc, $pkey);
	return "$me: Error looking up proj pkey=$pkey: $proj"
		if $proj && ! ref($proj);
	return [] unless $proj; #No proj with pkey=$pkey found

	my %descendents = ();
	my @list = ( $proj );
	PROJECT_LOOP: while ( scalar(@list) )
	{	$proj = shift @list;
		unless ( $proj && ref($proj) )
		{	warn "$me: Invalid proj $proj found, skipping ...";
			next PROJECT_LOOP;
		}
		my $curpkey = $proj->prj_pkey;

		my $children = $class->children_of_project_by_pkey(
			$hpc, $curpkey);
		unless ( $children && ref($children) eq 'ARRAY' )
		{	warn "$me: Error looking up children of $curpkey: $children";
			next PROJECT_LOOP;
		}

		CHILDREN_LOOP: foreach my $prj (@$children)
		{	$curpkey = $prj->prj_pkey;
			unless ( exists $descendents{$curpkey} )
			{	$descendents{$curpkey} = $prj;
				push @list, $prj;
			}
		}
	}

	return [ values %descendents ];
}

1;
__END__

=head1 NAME

Glue::HPCC::Project::Project;

=head1 SYNOPSIS

  use Glue::HPCC::Project::Project;
  use Glue::HPCC::Cluster;

  my $hpc = new Glue::HPCC::Cluster ...;

  my $kevin = Glue::HPCC::Project::ProjectType->project_by_name(
	$hpc, 'secret-project');

  ...


=head1 DESCRIPTION

This is the base class representing a record in the project table
in the Project database.   The project record has the following 
attributes:

=over 4

=item B<prj_pkey>: primary key

=item B<name>: Name of the project.  REQUIRED

=item B<description>: Short (1 line) text description of project.

=item B<unixgroup>: Name of the Unix group for the project

=item B<parent>: Prj_pkey of parent project, if any

=item B<department>: dept_pkey of department for project, REQUIRED.

=item B<projtype>: ptype key to project type, REQUIRED.

=item B<su>: SUs for the project.  Exact meaning depends on project type.  Typically is after tax quarterly SUs for auto-replenishing projects, and one shot after tax SUs for others.

=item B<active>: Boolean flag.  If true, this is an active project; inactive otherwise.

=item B<expiration>: When the project expires

=item B<thrhld_pct_over>, B<thrhld_pct_under>: threshholds for automated warnings if usage is over/under. 

=back

=head2 Special Functions

In all of the following, B<$hpc> is a B<Glue::HPCC::Cluster> object, providing
DB connection info, etc.  The implicit class B<Glue::HPCC::Project::Project>
argument is not explicity shown below.

=over 4

=item B<project_by_pkey($hpc, $pkey)>

Returns the B<Glue::HPCC:Project::Project> object with the specified 
B<prj_pkey>, if it exists.  Returns undef if no such project found.  Croaks or
returns a true, non-ref on errors.

=item B<project_by_name($hpc, $name)>

Returns the B<Glue::HPCC:Project::Project> object with the specified 
B<name>, if it exists.  Returns undef if no such project found.  Croaks or
returns a true, non-ref on errors.

=item B<children_of_project_by_pkey($hpc, $pkey)>

Returns a list ref of B<Glue::HPCC::Project::Project> objects which have
have the specified parent object (given by B<pkey>).  If B<pkey> is omitted
or undef, returns a list ref of projects which have no parent.  List may
be empty.  On non-fatal errors, returns non-ref error message.

=item B<descendents_of_project_by_pkey($hpc, $pkey)>

Similar to B<children_of_project_by_pkey>, but the former only returns the
immediate children of the project.  This will return the grandchildren, etc.,
i.e. all projects whose parent matches B<pkey>, or whose parent's parent
matches B<pkey>, or whose parent's parent's parent ... etc.  NOTE: the 
B<pkey> = undef mode is NOT supported (as that would simply return ALL
projects very inefficiently).

=back
