#!/usr/local/bin/perl
#
#Project_type table in projects DB

package Glue::HPCC::Project::ProjectType;
use strict;
use warnings;
use base qw(Glue::HPCC::Project::RWBase);
use Carp qw(croak carp confess);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $_ptype_list;

my $_ptype_by_name;
my $_ptype_by_pkey;

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @ro_accessors = qw( pkey );
my @rw_accessors = qw(
	name 
	description 
	auto_replenishes 
	has_normal_priority
	has_high_priority
	has_special_partition
	can_use_scavenger_q
	can_use_debug_q
	default_expiry_days
	requires_unix_group
	can_have_members
	qos_tag
	is_metered
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);

sub _ro_fields($)
{	my $class = shift;
	#my $parent = $class->SUPER::_ro_fields;
	#We don't want parent fields because we don't have timestamps
	#return [ @ro_accessors, @$parent ];
	return [ @ro_accessors ];
}

sub _rw_fields($)
{	my $class = shift;
	my $parent = $class->SUPER::_rw_fields;
	return [ @rw_accessors, @$parent ];
}

sub _required_fields($)
{	return [ qw( name default_expiry_days ) ];
}

sub _db_schema_name($$)
#Takes a Glue::HPCC::Cluster instance as well, so can figure out what cluster
#we are in
{       my $class = shift;
        my $hpc = shift;
        return "hpcc_shared";
}


sub _db_table_name($)
#Default name of table we are in.  Subclasses can override, but a good
#default
{	my $class = shift;
	return 'project_type';
}

sub _db_table_primary_keyname($)
#Name of primary key for our table.  Can return a scalar string or a list ref
#of scalar strings if multikey table
{	my $class = shift;
	return 'pkey';
}

sub _autoset_lastupdated($)
{	return 0;
}

sub _autoset_created($)
{	return 0;
}

#-------------------------------------------------------------------
#	Special class methods 
#-------------------------------------------------------------------

sub _get_ptype_list($$;$)
{	my $class = shift;
	my $hpc = shift;
	my $me = shift || '_make_ptype_by_name_hash';

	$_ptype_list = Glue::HPCC::Project::ProjectType->pdb_query($hpc);
	return if ( $_ptype_list && ref($_ptype_list) eq 'ARRAY' );

	my $error = $_ptype_list || '<unknown error>';
	$_ptype_list = undef;
	croak "$me: Error querying project type: $error at ";
}

sub _make_ptype_by_name_hash($$;$)
{	my $class = shift;
	my $hpc = shift;
	my $me = shift || '_make_ptype_by_name_hash';

	unless ( $_ptype_list && ref($_ptype_list) eq 'ARRAY' )
	{	my $err =  $class->_get_ptype_list($hpc, $me);
		croak "$me: Error getting project type list: $err at " if $err;
	}

	my %hash = map { $_->name => $_ } @$_ptype_list;
	$_ptype_by_name = \%hash;
	return;
}
	
sub _make_ptype_by_pkey_hash($$;$)
{	my $class = shift;
	my $hpc = shift;
	my $me = shift || '_make_ptype_by_pkey_hash';

	unless ( $_ptype_list && ref($_ptype_list) eq 'ARRAY' )
	{	my $err =  $class->_get_ptype_list($hpc, $me);
		croak "$me: Error getting project type list: $err at " if $err;
	}

	my %hash = map { $_->pkey => $_ } @$_ptype_list;
	$_ptype_by_pkey = \%hash;
	return;
}
	
sub projecttype_by_name($$$)
{	my $class = shift;
	my $hpc = shift;
	my $name = shift;
	my $me = 'projecttype_by_name';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required param name at "
		unless defined $name;

	unless (  $_ptype_by_name && ref($_ptype_by_name) eq 'HASH' )
	{	my $err = $class->_make_ptype_by_name_hash($hpc, $me);
		croak "$me: Cant make ptype_by_name hash: $err at " if $err;
	}

	return $_ptype_by_name->{$name};
}

sub projecttype_by_pkey($$$)
{	my $class = shift;
	my $hpc = shift;
	my $pkey = shift;

	my $me = 'projecttype_by_pkey';
	$class->_validate_hpc($me, $hpc);
	croak "$me: Missing required param pkey at "
		unless defined $pkey;
	croak "$me: Invalid param pkey '$pkey' at " unless $pkey =~ /^\d+$/;

	unless (  $_ptype_by_pkey && ref($_ptype_by_pkey) eq 'HASH' )
	{	my $err = $class->_make_ptype_by_pkey_hash($hpc, $me);
		croak "$me: Cant make ptype_by_pkey hash: $err at " if $err;
	}

	return $_ptype_by_pkey->{$pkey};
}

sub project_types($$)
#Returns a list ref of project type names
{	my $class = shift;
	my $hpc = shift;
	my $me = 'project_types';
	$class->_validate_hpc($me, $hpc);

	unless (  $_ptype_by_name && ref($_ptype_by_name) eq 'HASH' )
	{	my $err = $class->_make_ptype_by_name_hash($hpc, $me);
		croak "$me: Cant make ptype_by_name hash: $err at " if $err;
	}

	my @keys = keys %$_ptype_by_name;
	return [ sort @keys ];
}

sub autoreplenishing_project_types($$)
#Returns a list ref of project types that autoreplenish
#
{	my $class = shift;
	my $hpc = shift;
	my $me = 'autoreplenishing_project_types';
	$class->_validate_hpc($me, $hpc);

	unless ( $_ptype_list && ref($_ptype_list) eq 'ARRAY' )
	{	my $err =  $class->_get_ptype_list($hpc, $me);
		croak "$me: Error getting project type list: $err at " if $err;
	}

	my @ptypes = grep { $_->auto_replenishes } @$_ptype_list;
	return \@ptypes;
}

sub autoreplenishing_project_type_names($$)
#Returns a list ref of project type names that autoreplenish
#
{	my $class = shift;
	my $hpc = shift;
	my $me = 'autoreplenishing_project_type_names';
	$class->_validate_hpc($me, $hpc);

	unless ( $_ptype_list && ref($_ptype_list) eq 'ARRAY' )
	{	my $err =  $class->_get_ptype_list($hpc, $me);
		croak "$me: Error getting project type list: $err at " if $err;
	}

	my @ptypes = grep { $_->auto_replenishes } @$_ptype_list;
	@ptypes = map { $_->name } @ptypes;
	return \@ptypes;
}


1;
__END__

=head1 NAME

Glue::HPCC::Project::ProjectType

=head1 SYNOPSIS

  use Glue::HPCC::Project::ProjectType;
  use Glue::HPCC::Cluster;

  my $hpc = new Glue::HPCC::Cluster ...;

  my $prjtype = Glue::HPCC::Project::ProjectType->projecttype_by_name(
	$hpc, 'coop');
  print $prjtype->description;

  ...


=head1 DESCRIPTION

This is the base class representing a record in the project_type table
in the Project database.

=head2 Special Functions

=over 4

=item B<projecttype_by_name($hpc, $name)>

Will return the object representing the project_type by that name, or 
undef if no such type.  Because project_type is a very static table, this
B<class> method will download the whole table the first time it is invoked,
and just use that cached download for all subsequent calls.

=back
