#!/usr/local/bin/perl
#
#Base class for read-only tables/views in Projects DB


package Glue::HPCC::Project::ROBase;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);
use DBI;
use SQL::Interpolate qw(sql_interp);
use POSIX qw(strftime);


use Data::Dumper;
#-------------------------------------------------------------------
#	Globals and Class methods
#-------------------------------------------------------------------

my $VERBOSE;
my $DRYRUN;

sub verbose($;$)
#This is shared by all child classes
{	my $class = shift;
	my $new = shift;

	$VERBOSE=$new?1:0 if defined $new;
	return $VERBOSE;
}

sub dryrun($;$)
#This is shared by all child classes
{	my $class = shift;
	my $new = shift;

	$DRYRUN=$new?1:0 if defined $new;
	return $DRYRUN;
}

sub _output_sql($$$@)
{	my $class = shift;
	my $sql = shift;
	my @bind = @_;
	print STDERR "sql: $sql\n";
	if ( @bind )
	{ 	print STDERR "with bind values:\n";
		my $i = 1;
		foreach my $val (@bind)
		{	my $tmpval = $val;
			$tmpval = '<NULL>' unless defined $tmpval;
			print STDERR "   $i) $tmpval\n";
			$i++;
		}
	 	print STDERR "End of bind values.\n";
	} else
	{	print STDERR "with no bind values.\n";
	}
}
	
sub verbose_display_sql($$@)
{	my $class = shift;
	my $sql = shift;
	my @bind = @_;
	print STDERR "[VERBOSE] About to execute  the sql\n";
	$class->_output_sql($sql, @bind);
}

sub dryrun_display_sql($$@)
{	my $class = shift;
	my $sql = shift;
	my @bind = @_;
	print STDERR "[DRYRUN] About to execute the sql \n";
	$class->_output_sql($sql, @bind);
}

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @ro_accessors = qw(created lastupdated);
my @rw_accessors = qw();

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);

sub _ro_fields($)
{	return [ @ro_accessors ];
}

sub _rw_fields($)
{	return [@rw_accessors];
}

sub _required_fields($)
{	return [];
}

sub _db_fields($)
{	my $class = shift;
	my $ro_fields = $class->_ro_fields || [];
	my $rw_fields = $class->_rw_fields || [];

	my @all_fields = ( @$ro_fields, @$rw_fields);
	return \@all_fields;
}

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	my @args = @_;
	$class = ref($class) if ref($class);

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my $accessors = $obj->_rw_fields;
	my ($arg, $meth, $val);
	RWARG: foreach $arg (@$accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	$accessors = $obj->_ro_fields;
	ROARG: foreach $arg (@$accessors)
	{	next ROARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next ROARG unless defined $val;
		$meth = $arg;
		$obj->set($meth,$val);
	}


	#Warn about unknown arguments
	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;

	return;
}

sub _init($)
{	my $obj = shift;

	my ($fld, $meth, $val);
	my $req_parms = $obj->_required_fields;
	foreach $fld (@$req_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld";
		}
	}

}

#-------------------------------------------------------------------
#	Special constructor to generate from a DB entry
#-------------------------------------------------------------------

sub new_from_dbi_hashref($$)
#Generates a new instance from a hash ref as obtained from DBI fetchrow_hashref
#and similar
{	my $class = shift;
	my $record = shift;
	my $me = 'new_from_dbi_hashref';
	unless ( $record && ref($record) eq 'HASH' )
	{	croak "$me: Illegal record $record, expecting hashref at ";
	}

	my @newargs = ();
	my ($fld, $val);
	foreach $fld (keys %$record)
	{	$val = $record->{$fld};

		push @newargs, $fld, $val;
	}

	my $obj = $class->new(@newargs);
	return $obj;
}

#-------------------------------------------------------------------
#	Misc utilities
#-------------------------------------------------------------------

sub _force_instance_method($$)
#Croaks if called as a class instead of instance method
{	my $self = shift;
	my $me = shift;

	croak "$me: Must be called as an instance method at "
		unless $self && ref($self);
}

sub _validate_hpc($$$)
#Validates hpc argument for method
{	my $class = shift;
	my $me = shift;
	my $hpc = shift;

	croak "$me: Missing Glue::HPCC::Cluster instance at "
		unless $hpc;
	croak "$me: Invalid value '$hpc' for Glue::HPCC::Cluster instance at "
		unless $hpc && ref($hpc);
}

#-------------------------------------------------------------------
#	Low level table definitions.   Might need overriding
#-------------------------------------------------------------------

sub _db_schema_name($$)
#Takes a Glue::HPCC::Cluster instance as well, so can figure out what cluster
#we are in
{	my $class = shift;
	my $hpc = shift;
	my $cluster = $hpc->clustershort;

	return "hpcc_${cluster}";
}

sub _db_table_name($)
#Default name of table we are in.  Subclasses can override, but a good
#default
{	my $class = shift;
	$class = ref($class) if ref($class);

	my $base = $class;
	$base =~ s/^.*:://;
	$base = lc $base;
	return $base;
}

sub _db_fq_table_name($$)
#Concatenate schema and table names
{	my $class = shift;
	my $hpc = shift;

	my $schema = $class->_db_schema_name($hpc);
	my $table  = $class->_db_table_name;
	return "${schema}.${table}";
}

sub _db_table_primary_keyname($)
#Name of primary key for our table.  Can return a scalar string or a list ref
#of scalar strings if multikey table
{	my $class = shift;
	return 'pkey';
}


sub _my_where_clause($)
#This might need to be overloaded.
#Returns a where clause hash that should return the current entity.
{	my $obj = shift;
	croak "Must be called as an instance method at "
		unless $obj && ref($obj);

	#my $table = $obj->_db_table_name;
	my $keys = $obj->_db_table_primary_keyname;
	$keys = [ $keys ] unless ref($keys) eq 'ARRAY';

	my %where = ();
	my ($key, $val);
	foreach $key (@$keys)
	{	$val = $obj->$key;
		$where{$key} = $val;
	}
	return \%where;
}


#-------------------------------------------------------------------
#	Database interface routines
#-------------------------------------------------------------------

sub pdb_query($$@)
#Takes a Glue::HPCC::Cluster object, and a bunch of key => value
#pairs specifying a where clause, and returns a list ref containing
#Glue::HPCC::Project::XXX objects corresponding to the rows returned.
{	my $class = shift;
	my $hpc = shift;
	my %where = @_;
	my $me = 'pdb_query';
	$class->_validate_hpc($me, $hpc);

	my $dbh = $hpc->get_projects_dbh;

	my $table = $class->_db_fq_table_name($hpc);
	my ($sql, @bind);
	if ( scalar(keys %where) )
	{	($sql, @bind) = sql_interp "SELECT * FROM $table WHERE", \%where;
	} else
	{	$sql = "SELECT * FROM $table";
	}

	my $sth = $dbh->prepare_cached($sql);
	croak "$me: Error preparing SQL '$sql': " . $dbh->errstr . " at "
		unless $sth && ref($sth);

	if ( $hpc->verbose || $class->verbose )
	{	$class->verbose_display_sql($sql, @bind);
	}
	my $ok = $sth->execute(@bind);
	croak "$me: Error executing '$sql', " . $sth->errstr . " at "
		unless $ok;

	my @results = ();
	while (my $row = $sth->fetchrow_hashref )
	{	my $rowobj = $class->new_from_dbi_hashref($row);
		croak "$me: error creating from row: $rowobj at "
			unless $rowobj && ref($rowobj);
		push @results, $rowobj;
	}
	return [ @results ];
}

sub pdb_query_me($$)
#Calls pdb_query with where clause set to get the row in DB representing
#this instance.
#Returns undef if no matches (I don't exist)
#Returns a new instance with that info
#On error, returns a non-ref error string
{	my $obj = shift;
	my $hpc = shift;
	my $me = 'pdb_query_me';
	$obj->_force_instance_method($me);
	$obj->_validate_hpc($me, $hpc);

	my $where = $obj->_my_where_clause;
	my $list = $obj->pdb_query($hpc, %$where);
	return $list unless $list && ref($list) eq 'ARRAY'; #Error
	return unless scalar(@$list); #No matches, I don't exist

	if ( scalar(@$list) > 1 )
	{	my @tmp = map { "$_='" .  $where->{$_} . "'" } (keys %$where);
		my $tmp = join ", ", @tmp;
		my $class = ref($obj);
		return "Multiple  objects of type $class found with [$tmp]";
	}

	my $obj2 = $list->[0];
	return $obj2;
}

sub compare($$)
#Compare two instances, field by field (using
#Returns a list ref of triplets [ fieldname, value1, value2 ] for every
#field that differs.  If no differences, returns undef.
#value1 is the value for the invocant, value2 is the value of the field
#of the explicit argument.
#
#Compares fieldsd from _sacctmgr_fields_in_order;
{	my $obj1 = shift;
	my $obj2 = shift;
	my $me = 'compare';
	$obj1->_force_instance_method($me);
	croak "Bad invalid argument to $me: $obj2"
		unless $obj2 && ref($obj2) eq ref($obj1);

	my $fields = $obj1->_db_fields;

	my @diffs = ();
	foreach my $fld (@$fields)
	{	my $meth = $fld;
		my $val1 = $obj1->$meth;
		my $val2 = $obj2->$meth;
	
		#print STDERR "$me: meth=$meth, val1='$val1', val2='$val2'\n";

		if ( defined $val1 )
		{	if ( defined $val2 )
			{	push @diffs, [ $fld, $val1, $val2 ]
					unless $val1 eq $val2;
			} else
			{	push @diffs, [ $fld, $val1, undef ];
			}
		} elsif ( defined $val2 )
		{	push @diffs, [ $fld, undef, $val2 ];
		}
	}

	return unless @diffs;
	return [@diffs];
}

	
1;
__END__

=head1 NAME

Glue::HPCC::Project::ROBase

=head1 SYNOPSIS

  package Glue::HPCC::Project::Project;
  use base qw(Glue::HPCC::Project::ROBase);

  sub _ro_fields($)
  {	my $class = shift;
	return qw( prj_pkey ); #Make primary key read-only
  }

  sub _rw_fields($)
  {	my $class = shift;
	return qw( name description unixgroup expiration); 
  }

  sub _db_table_primary_keyname($)
  {	return 'prj_pkey';
  }
  ...


=head1 DESCRIPTION

This is the base class representing a record in one of the tables
or views in the Project database.  This base class is for
tables which only support SELECT (although RWBase extends this
to support INSERT, UPDATE, and DELETE as well).
It provides common constructors, and routines for querying 
the database table, as well as some common stuff that is useful
when extend to RWBase.  Each table should provide its own
class inheriting from either ROBase or RWBase class as appropriate.

=head2 Methods to Overload

Child classes should overload the following methods to customize
behavior:

=over 4

=item B<_ro_fields> 

Should return a list ref of the names of the
read-only fields for this entity.  The Base class
version should be included in this list, and has
the common timestamp fields B<created>, B<lastupdated>.

=item B<_rw_fields>

Should return a list ref of the names of the
read-write fields for this entity.  Default is empty.

=item B<_db_fields>

Should be a list ref of all fields in the DB.  The default
just concatenates B<_ro_fields> and B<_rw_fields>, which
usually suffices.  However if the subclasses B<_ro_fields>
does not include B<SUPER::_ro_fields>, or similar for rw,
then you should add in B<SUPER::_db_fields>.

=item B<_required_fields>

A list of required fields for the constructor to make the object.
Default is empty.  Should be the NOT NULL fields in the table definition.

=item B<_db_schema_name($hpc)>

The schema the table is in.  Takes a Glue::HPCC::Cluster object so
can determine what cluster we are looking at.  Returns hpcc_dt or
hpcc_dt2, etc. depending on the cluster.

=item B<_db_table_name>

The name of the table to use.  Default is taken from the class name.
The schema name still needs to be prepended to this.

=item B<_db_fq_table_name($hpc)>

The fully qualified table name, including schema if needs it.
Default is to concatenate B<_db_schema_name($hpc)> and B<_db_table_name>,
with a '.' in between.

=item B<_db_table_primary_keyname>

This returns the name(s) of the primary key(s) for the table.  Can
return a simple scalar string, or a list ref of string names if the
table has a multicolumned key.  The default is 'pkey', so you probably
need to overload this.

=item B<_my_where_clause>

Returns a hash ref which via B<SQL::Interpolate> becomes a where clause
identifying us in the table.  Default is just to use 
B<_db_table_primary_keyname>.

=back

All of the above except for B<_my_where_clause> 
are class methods; the last requires an actual instance to operate on.

=head2 Useful methods

This package provides some useful methods for querying data
in the various tables of the Projects database.  This allows much code
sharing just by overloading the methods above.  In the following, 
B<$hpc> is an instance of B<Glue::HPCC::Cluster>, needed to get the database
connection information.

The following class methods wrap the basic DBI commands:

=over 4

=item B<new_from_dbi_hashref($hashref)>

This takes a hash ref, e.g. from DBI's B<fetchrow_hashref>, and converts
it into an object of this type.

=item B<pdb_query>($hpc, [ fld1=> val1, [ fld2 => val2 ... ]) 

This does a generic query on the table, with where clause specified by
the fld1 => val1, etc. field=>value pairs.  On success, it returns a 
(possibly empty) list ref of Glue::HPCC::Project::XXX objects matching
the query.  On error, it most likely aborts, but might return a non-ref
error message.

=back

There is also a similar query method operating on the
invocant instance of the class (actually only requires the primary
key be set).  The basic instance methods are:

=over 4
 
=item B<pdb_query_me($hpc)>

This looks up in the database for the object matching the invocant instance,
based on the primary key of invocant.

=item B<compare($obj2)> 

This compares the invocant to B<$obj2> field by field, (doing textual
comparisons, and using fields as given by B<_db_fields)).  If there is
no difference, returns undef.  Otherwise returns a list ref of triplets
[ fieldname, val1, val2 ] where val1 is the invocant value for the field,
and val2 is the value in B<$obj2>.

=back

