#!/usr/local/bin/perl
#
#Base class for rw tables in Projects DB

package Glue::HPCC::Project::RWBase;
use strict;
use warnings;
use base qw(Glue::HPCC::Project::ROBase);
use Carp qw(carp croak);
use DBI;
use SQL::Interpolate qw(sql_interp);
use POSIX qw(strftime);


use Data::Dumper;

#-------------------------------------------------------------------
#	Low level table definitions.   Might need overriding
#-------------------------------------------------------------------

sub _autoset_lastupdated($)
{	return 1;
}

sub _autoset_created($)
{	return 1;
}

#-------------------------------------------------------------------
#	Miscellaneous
#-------------------------------------------------------------------

sub mysql_timestamp2datetime(;$)
#Returns the specified timestamp (seconds from epoch) to mysql DATETIME
#timestamp defaults to now
{	my $tstamp = shift;
	$tstamp = time unless defined $tstamp;

	my @ltime = localtime($tstamp);
	my $datetime = strftime("%Y-%m-%d %H:%M:%S", @ltime);
	return $datetime;
}

#-------------------------------------------------------------------
#	Database interface routines
#-------------------------------------------------------------------

sub pdb_insert($$@)
#Inserts a record of this class into the projects DB.  Values to insert 
#given as key => value pairs
{	my $class = shift;
	my $hpc = shift;
	my %fields = @_;
	my $me = 'pdb_insert';
	$class->_validate_hpc($me, $hpc);

	my $dbh = $hpc->get_projects_dbh;

	#Set creation/lastupdated timestamps if table supports them
	#(and not already set, ensure insert_me can set them)
	my $now = mysql_timestamp2datetime;
	my $aset_created = $class->_autoset_created;
	my $aset_last = $class->_autoset_lastupdated;
	$fields{created} = $now 
		if $aset_created && ! defined $fields{created};
	$fields{lastupdated} = $now 
		if $aset_last && ! defined $fields{lastupdated};

	my $table = $class->_db_fq_table_name($hpc);
	my ($sql, @bind) = sql_interp "INSERT INTO $table", \%fields;

	my $sth = $dbh->prepare_cached($sql);
	croak "$me: Error preparing SQL '$sql': " . $dbh->errstr . " at "
		unless $sth && ref($sth);

	if ( $hpc->dryrun || $class->dryrun )
	{	$class->dryrun_display_sql($sql, @bind);
		return;
	}
	if ( $hpc->verbose || $class->verbose )
	{	$class->verbose_display_sql($sql, @bind);
	}
	my $ok = $sth->execute(@bind);
	croak "$me: Error executing '$sql', " . $sth->errstr . " at "
		unless $ok;

	return;
}

sub pdb_insert_me($$)
#Inserts this record into the projects DB
{	my $obj = shift;
	my $hpc = shift;
	my $me = 'pdb_insert_me';
        $obj->_force_instance_method($me);
	$obj->_validate_hpc($me, $hpc);

	my $myvals = $obj->_my_insert_clause;

	#Set the created and lastupdated timestamps if table supports them
	my $now = mysql_timestamp2datetime;
	if ( $obj->_autoset_created )
	{	$myvals->{created} = $now;
		$obj->set('created', $now);
	}
	if ( $obj->_autoset_lastupdated )
	{	$myvals->{lastupdated} = $now;
		$obj->set('lastupdated', $now);
	}

	return  $obj->pdb_insert($hpc, %$myvals);
}

sub pdb_update($$$$)
#Does SQL update on one or more rows of this table, as specified
#by $where hash ref (as key=>value pairs) to values in $update hash ref
#(again key=>value pairs)
{	my $class = shift;
	my $hpc = shift;
	my $where = shift || {};
	my $update = shift;

	my $me = 'pdb_update';
	$class->_validate_hpc($me, $hpc);
	croak "Invalid where hash ($where) passed to $me at "
		if $where && ref($where) ne 'HASH';
	croak "No/invalid update hash passed to $me at " 
		unless $update && ref($update) eq 'HASH' && 
		scalar(keys %$update);

	#Set the lastupdated timestamp if table supports and not already set
	my $now = mysql_timestamp2datetime;
	my $aset_last = $class->_autoset_lastupdated;
	$update->{lastupdated} = $now 
		if $aset_last && ! defined $update->{lastupdated};

	my $dbh = $hpc->get_projects_dbh;
	my $table = $class->_db_fq_table_name($hpc);
	my @cmd = ( "UPDATE $table SET", $update );
	push @cmd, 'WHERE', $where if $where && ref($where) eq 'HASH';
	my ($sql, @bind) = sql_interp @cmd;

	my $sth = $dbh->prepare_cached($sql);
	croak "$me: Error preparing SQL '$sql': " . $dbh->errstr . " at "
		unless $sth && ref($sth);

	if ( $hpc->dryrun || $class->dryrun )
	{	$class->dryrun_display_sql($sql, @bind);
		return;
	}
	if ( $hpc->verbose || $class->verbose )
	{	$class->verbose_display_sql($sql, @bind);
	}
	my $ok = $sth->execute(@bind);
	croak "$me: Error executing '$sql', " . $sth->errstr . " at "
		unless $ok;

	return;
}

sub pdb_update_me($$@)
#Does SQL update of the row corresponding to this Perl object, using
#the key => value pairs of %update as values to set.
{	my $obj = shift;
	my $hpc = shift;
	my %update = @_;

	my $me = 'pdb_update_me';
	croak "$me must be called as an instance method at "
		unless $obj && ref($obj);
	croak "No/invalid Glue::HPC::Cluster object passed to $me at "
		unless $hpc && ref($hpc);

	#Set the lastupdated timestamp if table supports
	my $now = mysql_timestamp2datetime;
	if ( $obj->_autoset_lastupdated )
	{	$update{lastupdated} = $now;
		$obj->set('lastupdated', $now);
	}

	my $where = $obj->_my_where_clause;
	return $obj->pdb_update($hpc, $where, \%update);
}

sub pdb_save_me($$)
{	my $obj = shift;
	my $hpc = shift;
	my $me = 'pdb_save_me';
        $obj->_force_instance_method($me);
	$obj->_validate_hpc($me, $hpc);

	my $current = $obj->pdb_query_me($hpc);
	unless ( defined $current )
	{	#No current row matching me, so just do an insert
		return $obj->insert_me($hpc);
	}
	croak "Error querying row in $me : $current at"
		unless ref($current);

	my $diffs = $obj->compare($current);
	return unless ( $diffs ); #Nothing to do

	my %updates = ();
	foreach my $rec (@$diffs)
	{	my ($fld, $val1, $val2) = @$rec;
		#$val1 = '' unless defined $val1; #undefs => NULL, right?
		$updates{$fld} = $val1;
	}

	return $obj->pdb_update_me($hpc, %updates);
}

sub pdb_delete($$@)
#Does SQL delete to delete all of the entities of this type
#matching specified criteria
#
#We REQUIRE at least one criterion for safety
{	my $class = shift;
	my $hpc = shift;
	my %where = @_;

	my $me = 'pdb_delete';
	$class->_validate_hpc($me, $hpc);
	croak "$me refusing to issue SQL delete w/out where clause at "
		unless scalar(keys %where);

	my $dbh = $hpc->get_projects_dbh;
	my $table = $class->_db_fq_table_name($hpc);
	my ($sql, @bind) = sql_interp "DELETE FROM $table WHERE", \%where;

	my $sth = $dbh->prepare_cached($sql);
	croak "$me: Error preparing SQL '$sql': " . $dbh->errstr . " at "
		unless $sth && ref($sth);

	if ( $hpc->dryrun || $class->dryrun )
	{	$class->dryrun_display_sql($sql, @bind);
		return;
	}
	if ( $hpc->verbose || $class->verbose )
	{	$class->verbose_display_sql($sql, @bind);
	}
	my $ok = $sth->execute(@bind);
	croak "$me: Error executing '$sql', " . $sth->errstr . " at "
		unless $ok;
	return;
}

sub pdb_delete_me($$)
#Does SQL delete to the delete the DB row corresponding to this Perl
#object instance
{	my $obj = shift;
	my $hpc = shift;
	my $me = 'pdb_delete_me';
        $obj->_force_instance_method($me);
	$obj->_validate_hpc($me, $hpc);

	my $where = $obj->_my_where_clause;
	return $obj->pdb_delete($hpc, %$where);
}
	
1;
__END__

=head1 NAME

Glue::HPCC::Project::RWBase

=head1 SYNOPSIS

  package Glue::HPCC::Project::Project;
  use base qw(Glue::HPCC::Project::Base);

  sub _ro_fields($)
  {	my $class = shift;
	return qw( prj_pkey ); #Make primary key read-only
  }

  sub _rw_fields($)
  {	my $class = shift;
	return qw( name description unixgroup expiration); 
  }

  sub _db_table_primary_keyname($)
  {	return 'prj_pkey';
  }
  ...


=head1 DESCRIPTION

This is the base class representing a record in one of the tables
in the Project database.  It extends the similar ROBase class
to handle tables which can be modified; e.g. it adds methods
corresponding to SQL INSERT, UPDATE, and DELETE operations.

=head2 Methods to Overload

Child classes should overload the following methods to customize
behavior (see also the similar section in ROBase):

=over 4

=item B<_autoset_lastupdated>

If true (the default), the B<lastupdated> field will be automatically set
on all updates to the DB.  This should only be unset on the rare tables
w/out the field.

=item B<_autoset_created>

If true (the default), the B<created> timestamp field will be automatically 
when record is created in the DB.  This should only be unset on the 
rare tables w/out the field.

=item B<_my_insert_clause>

Returns a hash ref which via B<SQL::Interpolate> becomes the fields and
values parts of an SQL insert command to insert a record representing
this object.

=back

All of the above except for B<_my_insert_clause>
are class methods; the last requires an actual instance to operate on.

=head2 Useful methods

This package provides some useful methods for modifying data
in the various tables of the Projects database using standard SQL
INSERT, UPDATE, and DELETE operations.  SELECT functionality
is inherited from ROBase.  This allows much code
sharing just by overloading the methods above.  In the following, 
B<$hpc> is an instance of B<Glue::HPCC::Cluster>, needed to get the database
connection information.

The following class methods follow the basic DB query and data 
manipulation commands:

=over 4

=item B<pdb_insert>($hpc, [ fld1 => val1, [ fld2 => val2 ... ]) 

This does a generic insert into the table, with values for the record
being inserted given by the field=>value pairs.
It returns false on success, and either croaks or returns true (an error
message) on errors.  The fields B<created> and B<lastupdated> will
automatically be set to the current time unless B<_autoset_created> and/or
B<_autoset_lastupdated> are false (default is both true).

=item B<pdb_update>($hpc, { where1=>wval1, ... }, { set1 => sval1, ... }) 

This does a generic SQL update on the table. The where clause is given
by a hash B<reference>, and the values to set by another hash B<reference>.
It returns false on success, and either croaks or returns true (an error
message) on errors.  The field B<lastupdated> will be automatically 
set to the current time unless B<_autoset_lastupdated> is false (default is
true).

=item B<pdb_delete>($hpc, fld1=>val1, [ fld2=>val2 ... ])

This does a generic SQL delete into the table, with where clause specified
by the key => value pairs.  At least one key and value must be specified
(we refuse to do a delete without a where clause).
It returns false on success, and either croaks or returns true (an error
message) on errors.

=back

For each class method above, there is a similar one operating on the
invocant instance of the class:

=over 4
 
=item B<pdb_insert_me($hpc)>

This inserts into the database a row matching the invocant instance.  The
fields B<created> and B<lastupdated>, for both in the invocant and the DB,
will be set to the current time (assuming B<_autoset_created> and/or
B<_autoset_lastupdated> not changed).

=item B<pdb_update_me>($hpc, fld1 => val1, [ .. fld2 => val2 ] )

This updates the DB record and the invocant with the specified fields.
The B<lastupdated> field will be automatically set to the current time
(assuming B<_autoset_lastupdated> not changed).

=item B<pdb_delete_me($hpc)>

This deletes the DB record corresponding to the invocant.  The invocant
object is NOT deleted.

=item B<pdb_save_me($hpc)>

This does a B<query_me> on the invocant, and if it does not exist in the
database already, continues with an B<insert_me>.  If it does exist, it
compares the DB object to the invocant, and does an B<update_me> to make
the DB record match the invocant.  The B<lastupdated> time will be set
to now.

=back

