#!/usr/local/bin/perl
#
# Common routines for checking usage of a project
# Shared by Glue::HPCC::Project and Glue::HPCC::UserPDB routines

package Glue::HPCC::ProjectUsage;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);
use Slurm::Sshare;

use Glue::HPCC::Cluster;
use Glue::HPCC::Project::Project;
use Glue::HPCC::Project::ProjectType;

use Glue::HPCC::UserPDB::Project;

use Glue::HPCC::Utils::TimeUtils qw(fraction_of_quarter);
use Time::ParseDate; #For grant-type projects

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $VERBOSE;
my $DRYRUN;
my $DEBUG;
$DEBUG=1;

#Percentage points
my $DEFAULT_OVER_THRHLD = 15;
my $DEFAULT_UNDER_THRHLD = 100; #Don;t warn on under

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @rw_accessors = qw(
	name
	ptname
	description

	timeperiod
	tstamp_now
	frac_tperiod

	ksu_total
	ksu_used
	ksu_avail
	frac_used

	over_thr
	under_thr
	replenish_thr

	overusage_flag
	underusage_flag
	replenish_alarm
);

my @ro_accessors = qw( 
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);


my @required_parms = qw(
);

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my ($arg, $meth, $val);
	RWARG: foreach $arg (@rw_accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	ROARG: foreach $arg (@ro_accessors)
	{	next ROARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next ROARG unless defined $val;
		$meth = $arg;
		$obj->set($meth,$val);
	}


	#Warn about unknown arguments
	if ( scalar(%args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;

	return;
}

sub _init($)
{	my $obj = shift;

	my ($fld, $meth, $val);
	foreach $fld (@required_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld at ";
		}
	}

}

#-------------------------------------------------------------------
#	Class methods: convert PDB projects into UserPDB projects
#-------------------------------------------------------------------

sub pdb2udb_project($$$;$)
#Takes Hlue::HPCC instance and either PDB or UserPDB project
#Returns corresponding UserPDB project, creating if needed
{	my $class = shift;
	my $hpc = shift;
	my $proj = shift;
	my $realme = __PACKAGE__ . '::pdb2udb_project';
	my $me = shift || $realme;

	if ( UNIVERSAL::isa($proj, 'Glue::HPCC::UserPDB::Project') )
	{	#Already is a UserPDB::Project instance, nothing to do
		return $proj;
	}


	unless ( UNIVERSAL::isa($proj, 'Glue::HPCC::Project::Project') )
	{	croak "$me: Invalid project argument; expecting either " .
			"Glue::HPCC::Project::Project or " .
			"UserPDB::Project, got $proj at ";
	}

        unless ( $hpc && ref($hpc) )
        {       croak "$me: Missing/invalid required param hpc (required for Glue::HPCC::Project::Project instances) at ";
        }

	my $udbproj = Glue::HPCC::UserPDB::Project->server_new_from_pdb_project(
		$hpc, $proj);

	unless ( $udbproj && ref($udbproj) )
	{	croak "$me: Error converting PDB to UserPDB project: $udbproj at ";
	}

	return $udbproj;
}

#-------------------------------------------------------------------
#	Class methods: Abstraction layer methods for usage thresholds
#
# These methods provide an abstraction layer between the thresholds stored
# in the databases and their use in the check usage methods.  In particular,
# some of these thresholds are currently used in several places; we use these
# methods to break this overusage into distinct use cases, so that if we ever
# need to refactor the usage routines can be left alone.
#
# These methods also allow for consistent defaulting, and we have multiple versions
# for different project types to allow different defaults by project type.
#
# These methods either take an Glue::HPCC::UserPDB::Project instance (in which case
# it will return either the appropriate DB threshold value if set or the default),
# or no argument (in which case the default is always returned).
#
#-------------------------------------------------------------------

sub coop_total_overusage_threshold($;$)
#For coop accounts; overusage threshold on combined (std+hi) for the quarter
{	my $class = shift;
	my $proj = shift;
	my $default = $DEFAULT_OVER_THRHLD;

	return $default unless $proj && ref($proj);
	my $tmp = $proj->thrhld_pct_over;
	return $tmp if defined $tmp;
	return $default;
}

sub coop_total_underusage_threshold($;$)
#For coop accounts; underusage threshold on combined (std+hi) for the quarter
{	my $class = shift;
	my $proj = shift;
	my $default = $DEFAULT_UNDER_THRHLD;

	return $default unless $proj && ref($proj);
	my $tmp = $proj->thrhld_pct_under;
	return $tmp if defined $tmp;
	return $default;
}

sub coop_high_overusage_threshold($;$)
#For coop accounts; overusage threshold on high-priority for the month
{	my $class = shift;
	my $proj = shift;
	my $default = $DEFAULT_OVER_THRHLD;

	return $default unless $proj && ref($proj);
	my $tmp = $proj->thrhld_pct_over;
	return $tmp if defined $tmp;
	return $default;
}

sub coop_high_underusage_threshold($;$)
#For coop accounts; underusage threshold on high-priority for the month
{	my $class = shift;
	my $proj = shift;
	my $default = $DEFAULT_UNDER_THRHLD;

	return $default unless $proj && ref($proj);
	my $tmp = $proj->thrhld_pct_under;
	return $tmp if defined $tmp;
	return $default;
}

sub grant_overusage_threshold($;$)
#For aacgrant/engrgrant accounts; overusage threshold over 1 year
{	my $class = shift;
	my $proj = shift;
	my $default = $DEFAULT_OVER_THRHLD;

	return $default unless $proj && ref($proj);
	my $tmp = $proj->thrhld_pct_over;
	return $tmp if defined $tmp;
	return $default;
}

sub grant_underusage_threshold($;$)
#For aacgrant/engrgrant accounts; underusage threshold over 1 year
{	my $class = shift;
	my $proj = shift;
	my $default = $DEFAULT_UNDER_THRHLD;

	return $default unless $proj && ref($proj);
	my $tmp = $proj->thrhld_pct_under;
	return $tmp if defined $tmp;
	return $default;
}

sub coop_replenish_threshold($;$)
#For coop accounts; ensure sufficient funds in std to replenish high for rest of quarter
{	my $class = shift;
	my $proj = shift;
	my $default = $DEFAULT_OVER_THRHLD;

	return $default unless $proj && ref($proj);
	my $tmp = $proj->thrhld_pct_over;
	return $tmp if defined $tmp;
	return $default;
}


#-------------------------------------------------------------------
#	Class methods: Do project usage calculations
#
# These all take either project => Glue::HPCC::UserPDB::Project instance or
#	project => Glue::HPCC::Project::Project and hpc => Glue::HPCC::Cluster instances
#-------------------------------------------------------------------

sub check_coop_project_usage($@)
#Returns a list ref of Glue::HPCC::ProjectUsage objects giving data about usage
#of project.
#Takes a list of key => value pairs as arguments
#	project: Either an Glue::HPCC::Project::Project or Glue::HPCC::UserPDB::Project
#		instance. REQUIRED
#	cluster: cluster name. REQUIRED
#	??? hpc: Glue::HPCC::Cluster instance.  Required if project is Glue::HPCC::Project::Project
#	timestamp: Time to do calculation at; defaults to current time
#	over_thr: threshold for excessive use, as percent
#	under_thr: threshold for under use, as percent
#
#Returns non-ref error message on non-fatal errors
#Returns [ high_priority_usage, total_usage ] where 
#high_priority_usage, total_usage are Glue::HPCC::ProjectUsage objects for high priority
#account and total allocation, respectively, on success.
{	my $class = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::check_coop_project_usage';

	my $project = delete $args{project};
	my $cluster = delete $args{cluster};
	my $hpc = delete $args{hpc};
	my $tstamp = delete $args{timestamp} || time;
	my $over_thr = delete $args{over_thr};
	my $under_thr = delete $args{under_thr};

	croak "$me: Missing required parameter 'project' at " unless $project;
	croak "$me: Missing required parameter 'cluster' at " unless $cluster;
	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}
	croak "$me: Invalid non-ref value '$project' for project at " unless ref($project);

	my ( $pname, $ptname);
	my $uproj = $class->pdb2udb_project($hpc, $project, $me);

	$pname = $uproj->name;
	$ptname = $uproj->projtype;

	my ( $thr_over_total, $thr_over_high, $thr_under_total, $thr_under_high, $thr_replenish);
	$thr_over_total = $thr_over_high = $thr_replenish = $over_thr if ( defined $over_thr );
	$thr_under_total = $thr_under_high = $under_thr if ( defined $under_thr );

	$thr_over_total = $class->coop_total_overusage_threshold($uproj) unless defined $thr_over_total;
	$thr_over_high = $class->coop_high_overusage_threshold($uproj) unless defined $thr_over_high;
	$thr_under_total = $class->coop_total_underusage_threshold($uproj) unless defined $thr_under_total;
	$thr_under_high = $class->coop_high_underusage_threshold($uproj) unless defined $thr_under_high;
	$thr_replenish = $class->coop_replenish_threshold($uproj) unless defined $thr_replenish;

	croak "$me: Called on project $pname of type $ptname at " unless $ptname eq 'coop';

	#Convert to fraction (0-1)
	my $over_thr_total = $thr_over_total / 100;
	my $over_thr_high = $thr_over_high / 100;
	my $under_thr_total = $thr_under_total / 100;
	my $under_thr_high = $thr_under_high / 100;
	my $replenish_thr = $thr_replenish / 100;


	my ($mon_quarter, $frac_month, $frac_quarter) = fraction_of_quarter($tstamp);
	my $stdacc = "${pname}";
	my $hiacc = "${pname}-hi";

	#Get usage for std and hi
	my $rec = Slurm::Sshare->usage_for_account_in_cluster(
		cluster => $cluster, account=>$stdacc);
	return "$me: Unable to get std pri balance for $pname: $rec" 
		unless $rec && ref($rec) eq 'ARRAY';
	#Userdata not useful because reset each month, and we are working on quarters
	my ($cpusec_used, $cpumin_limit, $cpumin_avail, $userdata) = @$rec;
	my $std_ksu_avail = $cpumin_avail/60/1000;

	$rec = Slurm::Sshare->usage_for_account_in_cluster(
		cluster => $cluster, account=>$hiacc);
	return "$me: Unable to get hi pri balance for $pname: $rec" 
		unless $rec && ref($rec) eq 'ARRAY';
	($cpusec_used, $cpumin_limit, $cpumin_avail, $userdata) = @$rec;
	my $hi_ksu_avail = $cpumin_avail/60/1000;
	#This will show under use if hi not fully replenished at start of month
	#my $hi_ksu_used  = $cpusec_used/60/60/1000;

	my $qu_su_tot = $uproj->su;
	my $qu_ksu_tot = $qu_su_tot/1000;
	#Avoid div-by-zero errors
	$qu_ksu_tot = 1 if $qu_ksu_tot <= 0;
	my $qu_ksu_used = $qu_ksu_tot - $std_ksu_avail - $hi_ksu_avail;
	$qu_ksu_used = 0 if $qu_ksu_used < 0;

	my $qu_pused = $qu_ksu_used/$qu_ksu_tot;
	my $hi_ksu_tot = $qu_ksu_tot/3;
	my $hi_ksu_used  = $hi_ksu_tot - $hi_ksu_avail;
	my $hi_pused = $hi_ksu_used/$hi_ksu_tot;

	my $over_qu = 0;
	my $over_hi = 0;
	my $under_qu = 0;
	my $under_hi = 0;
	my $repl_alarm = 0;

	$over_qu = 1 if $qu_pused > $frac_quarter + $over_thr_total;
	$over_hi = 1 if $hi_pused > $frac_month + $over_thr_high;
	$under_qu = 1 if $qu_pused < $frac_quarter - $under_thr_total;
	$under_hi = 1 if $hi_pused < $frac_month - $under_thr_high;

	#Check for sufficient SUs in std to replenish high for remaining months in quarter
	my $refreshes_left = 2 - $mon_quarter;
	$refreshes_left = 0  if $refreshes_left < 0;
	my $ksu_needed2replenish = $refreshes_left * $hi_ksu_tot;
	my $replenish_delta = $ksu_needed2replenish - $std_ksu_avail;
	$replenish_delta = 0 if $replenish_delta < 0;
	my $replenish_pdelta = $replenish_delta/ $hi_ksu_tot;
	$repl_alarm = 1 if $replenish_pdelta > $replenish_thr;

	my $total_usage = $class->new(
		name => $pname,
		description => 'Total',
		ptname => 'coop',

		timeperiod => 'quarter',
		tstamp_now => $tstamp,
		frac_tperiod => $frac_quarter,

		ksu_total => $qu_ksu_tot,
		ksu_used  => $qu_ksu_used,
		ksu_avail => $std_ksu_avail + $hi_ksu_avail,
		frac_used => $qu_pused,

		over_thr => $over_thr_total,
		under_thr => $under_thr_total,
		replenish_thr => $replenish_thr,

		overusage_flag => $over_qu,
		underusage_flag => $under_qu,
		replenish_alarm => $repl_alarm,
	);
		
	my $high_usage = $class->new(
		name => $pname,
		description => 'HiPriority',
		ptname => 'coop',

		timeperiod => 'month',
		tstamp_now => $tstamp,
		frac_tperiod => $frac_month,

		ksu_total => $hi_ksu_tot,
		ksu_used  => $hi_ksu_used,
		ksu_avail => $hi_ksu_avail,
		frac_used => $hi_pused,

		over_thr => $over_thr_high,
		under_thr => $under_thr_high,

		overusage_flag => $over_hi,
		underusage_flag => $under_hi,
	);
		
	return [ $high_usage, $total_usage];
}

sub _check_grant_project_usage($@)
#Does the work for aacgrant and engrgrant type projects.  I.e., projects
#which are non-renewing, and have a one-time grant of funds for a year.
#
#Takes a list of key => value pairs as arguments
#	project: Either an Glue::HPCC::Project::Project or Glue::HPCC::UserPDB::Project
#		instance. REQUIRED
#	cluster: cluster name. REQUIRED
#	??? hpc: Glue::HPCC::Cluster instance.  Required if project is Glue::HPCC::Project::Project
#	timestamp: Time to do calculation at; defaults to current time
#	over_thr: threshold for excessive use, as percent
#	under_thr: threshold for under use, as percent
#	daysvalid: length of time account is supposed to be valid for, in days
#		Defaults to 365.25 (1 year)
#	tperiod: human readable time period.  Will default to "$daysvalid days"
#		(with some minor fixes, e.g. year for 365.25 days).
#		If daysvalid and tperiod are both given, no checks made that
#		they are consistent.
#	me: Method name to use when reporting errors
#
#Returns non-ref error message on non-fatal errors
#Returns a listref with a single element, a  Glue::HPCC::ProjectUsage 
#object with total usage.
{	my $class = shift;
	my %args = @_;
	my $realme = __PACKAGE__ . '::_check_grant_project_usage';

	my $me = delete $args{me} || $realme;
	my $project = delete $args{project};
	my $cluster = delete $args{cluster};
	my $hpc = delete $args{hpc};
	my $tstamp = delete $args{timestamp} || time;
	my $over_thr = delete $args{over_thr};
	my $under_thr = delete $args{under_thr};
	my $daysvalid = delete $args{days_valid};
	my $tperiod = delete $args{tperiod};

	croak "$me: Missing required parameter 'project' at " unless $project;
	croak "$me: Missing required parameter 'cluster' at " unless $cluster;
	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}
	croak "$me: Invalid non-ref value '$project' for project at " unless ref($project);
	my $uproj = $class->pdb2udb_project($hpc, $project, $me);

	#Default daysvalid and/or tperiod
	unless ( $daysvalid )
	{	$daysvalid = 365.25;
		$tperiod = 'year';
	}

	unless ( $tperiod )
	{	$tperiod = "$daysvalid days";
		#We are going to be approximate
		if ( $daysvalid >= 360 && $daysvalid <= 366 )
		{	$tperiod = 'year';
		} elsif ( $daysvalid >= 90 && $daysvalid <= 92 )
		{	$tperiod = 'quarter';
		} elsif ( $daysvalid >= 30 && $daysvalid <= 31 )
		{	$tperiod = 'month';
		}
	}

	my ( $pname, $ptname, $thr_over, $thr_under);
	my ($expiration);


	$pname = $uproj->name;
	$ptname = $uproj->projtype;
	$thr_over = $uproj->thrhld_pct_over;
	$thr_under = $uproj->thrhld_pct_under;
	$expiration = $uproj->expiration;

	croak "$realme: Called on project $pname of type $ptname at " unless $ptname eq 'aacgrant' || $ptname eq 'engrgrant';

	$over_thr = $thr_over unless defined $over_thr;
	$under_thr = $thr_under unless defined $under_thr;
	$over_thr = $DEFAULT_OVER_THRHLD unless defined $over_thr;
	$under_thr = $DEFAULT_UNDER_THRHLD unless defined $under_thr;
	#Convert to fraction (0-1)
	$over_thr /= 100;
	$under_thr /= 100;

	my $ts_end = parsedate($expiration,
		WHOLE=>1, 
		DATE_REQUIRED=>1, 
		NO_RELATIVE => 1,
	);
	return "$me: Unable to parse expiration date '$expiration'"
		unless defined $ts_end;
	my $secsvalid = $daysvalid * 24 * 60 * 60;
	$secsvalid = 1 unless $secsvalid; #Avoid div by zero
	my $ts_start = $ts_end - $secsvalid;
	my $time_fraction = ($tstamp - $ts_start) / $secsvalid;
	$time_fraction = 1 if $time_fraction > 1; #Since we don't always expire

	#Get usage
	my $rec = Slurm::Sshare->usage_for_account_in_cluster(
		cluster => $cluster, account=>$pname);
	return "$me: Unable to get balance for $pname: $rec" 
		unless $rec && ref($rec) eq 'ARRAY';
	my ($cpusec_used, $cpumin_limit, $cpumin_avail, $userdata) = @$rec;
	my $ksu_avail = $cpumin_avail/60/1000;

	my $su_tot = $uproj->su;
	my $ksu_tot = $su_tot/1000;
	#Avoid div-by-zero errors
	$ksu_tot = 1 if $ksu_tot <= 0;
	my $ksu_used = $ksu_tot - $ksu_avail;
	$ksu_used = 0 if $ksu_used < 0;

	my $pused = $ksu_used/$ksu_tot;

	my $over_used = 0;
	my $under_used = 0;

	$over_used = 1 if $pused > $time_fraction + $over_thr;
	$under_used = 1 if $pused < $time_fraction - $under_thr;

	my $total_usage = $class->new(
		name => $pname,
		description => 'Total',
		ptname => $ptname,

		timeperiod => $tperiod,
		tstamp_now => $tstamp,
		frac_tperiod => $time_fraction,

		ksu_total => $ksu_tot,
		ksu_used  => $ksu_used,
		ksu_avail => $ksu_avail,
		frac_used => $pused,

		over_thr => $over_thr,
		under_thr => $under_thr,

		overusage_flag => $over_used,
		underusage_flag => $under_used,
	);
		
	return [ $total_usage ];
}

sub check_aacgrant_project_usage($@)
#Returns a Glue::HPCC::ProjectUsage object giving data about usage of an
#aacgrant project.
#Takes a list of key => value pairs as arguments
#	project: Either an Glue::HPCC::Project::Project or Glue::HPCC::UserPDB::Project
#		instance. REQUIRED
#	cluster: cluster name. REQUIRED
#	??? hpc: Glue::HPCC::Cluster instance.  Required if project is Glue::HPCC::Project::Project
#	timestamp: Time to do calculation at; defaults to current time
#	over_thr: threshold for excessive use, as percent
#	under_thr: threshold for under use, as percent
#	daysvalid:  the number of days the project is valid.  Default is 1 year
#	tperiod: a human readable text string which should equal daysvalid.
#		Will be overwritten if daysvalid not given.  If both daysvalid
#		and tperiod are given, no checks made that are consistent.
#		Default is more or less "$daysvalid days" (some special cases
#		like 'year' are handled, approximately).
#
#Returns non-ref error message on non-fatal errors
#Returns Glue::HPCC::ProjectUsage objects for total usage on success
{	my $self = shift;
	my @args = @_;
	my $me = __PACKAGE__ . '::check_aacgrant_project_usage';

	return $self->_check_grant_project_usage(@args);
}

sub check_engrgrant_project_usage($@)
#As check_aacgrant_project_usage, but for engrgrants.
#Only separate entry point in case the two project types diverge in future.
{	my $self = shift;
	my @args = @_;
	my $me = __PACKAGE__ . '::check_engrgrant_project_usage';

	return $self->_check_grant_project_usage(@args);
}


sub check_project_usage($@)
#Returns a list ref of Glue::HPCC::ProjectUsage objects giving data about usage
#of project.
#Takes a list of key => value pairs as arguments
#	project: Either an Glue::HPCC::Project::Project or Glue::HPCC::UserPDB::Project
#		instance. REQUIRED
#	cluster: cluster name. REQUIRED
#	??? hpc: Glue::HPCC::Cluster instance.  Required if project is Glue::HPCC::Project::Project
#	timestamp: Time to do calculation at; defaults to current time
#	over_thr: threshold for excessive use, as percent
#	under_thr: threshold for under use, as percent
#
#Returns non-ref error message on non-fatal errors
#Heavy lifting is done by project-type specific routines.
#Most return a list ref with a single Glue::HPCC::ProjectUsage object, although
#coop projects return a list ref with two such [ hi-pri-usage, total-usage ]
{	my $class = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::check_project_usage';

	my $project = delete $args{project};
	my $cluster = delete $args{cluster};
	my $hpc = delete $args{hpc};
	my $tstamp = delete $args{timestamp} || time;
	my $over_thr = delete $args{over_thr};
	my $under_thr = delete $args{under_thr};

	croak "$me: Missing required parameter 'project' at " unless $project;
	croak "$me: Missing required parameter 'cluster' at " unless $cluster;
	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}

	my ( $pname, $ptname);
	croak "$me: Invalid non-ref value '$project' for project at " unless ref($project);
	my $uproj = $class->pdb2udb_project($hpc, $project, $me);

	$pname = $project->name;
	$ptname = $project->projtype;

	my $retval;
	if ( $ptname eq 'coop' )
	{	$retval = $class->check_coop_project_usage(
			project => $uproj,
			cluster => $cluster,
			#hpc => $hpc, 
			timestamp => $tstamp,
			over_thr => $over_thr,
			under_thr => $under_thr,
		);
	} elsif ( $ptname eq 'aacgrant' )
	{	$retval = $class->check_aacgrant_project_usage(
			project => $uproj,
			cluster => $cluster,
			#hpc => $hpc, 
			timestamp => $tstamp,
			over_thr => $over_thr,
			under_thr => $under_thr,
		);
	} elsif ( $ptname eq 'engrgrant' )
	{	$retval = $class->check_engrgrant_project_usage(
			project => $uproj,
			cluster => $cluster,
			#hpc => $hpc, 
			timestamp => $tstamp,
			over_thr => $over_thr,
			under_thr => $under_thr,
		);
	} else {
		return "$me: Do not know how to handle projects of type $ptname for project $pname";
	}

	return $retval;
}


1;
__END__

=head1 NAME

Glue::HPCC::ProjectUsage - Information about usage of a project

=head1 SYNOPSIS

  package Glue::HPCC::ProjectUsage;
  use Glue::HPCC::UserPDB::Client;
  use Glue::HPCC::UserPDB::Project;

  my $updb = Glue::HPCC::UserPDB::Client->new;
  my $myprojects = Glue::HPCC::UserPDB::Project->client_get_my_projects(
		$updb, $filter, $impersonate);

  my $cluster = 'dt2';
  my $proj = $myprojects->[0];
  my $coop_usage = Glue::HPCC::ProjectUsage->check_coop_project_usage(
	project => $proj, cluster=>$cluster, );



=head1 DESCRIPTION

This class encapsulates information about the current usage of a project in
a cluster.  It is used for user commands so that users can query how
fast/slow funds for a project are being consumed for the current time
period (e.g. the current quarter for a coop account), as well as for
system cron jobs reporting to PoCs if project is being consumed too
rapidly or slowly.


It provides some common routines and default values.  


The accessors are:

=over 4

=item B<name>: name of the project

=item B<ptname>: name of the project type

=item B<description>: Description of the ProjectUsage object.  E.g., to distinguish accounts in coop projects

=item B<timeperiod>: The timeperiod for this PUobj, which in turn depends on the type of project.

=item B<tstamp_now>: The timestamp (seconds from 1970) used as 'now' in the calculations

=item B<frac_tperiod>: The fraction [0-1] of the time period that has past

=item B<ksu_total>: Total kSU for the account in question for the entire timeperiod

=item B<ksu_used>: The kSU for the account used during the timeperiod

=item B<ksu_avail>: The kSU available for the account.

=item B<frac_used>: ksu_used/ksu_total

=item B<over_thr>: The threshold (0-1) used for the overusage_flag

=item B<under_thr>: The threshold (0-1) used for the underusage flag

=item B<replenish_thr>: The threshold (0-1) used for the replenish alarm

=item B<overusage_flag>: Boolean.  True if allocation being consumed too quickly.

=item B<underusage_flag>: Boolean.  True if allocation being consumed too slowly.

=item B<replenish_alarm>: Boolean.  True if usage of std priority account on coop project type might impact replenishment of high priority account in successive months in the quarter.

=back


Common methods now described.


B<check_coop_project_usage(key-value pairs)> checks the usage on a coop
type project.  Parameters are key=>value pairs, recognizing:

=over 4

=item project: Either a Glue::HPCC::Project::Project or Glue::HPCC:UserPDB::Project object, REQUIRED

=item hpc: A Glue::HPCC::Cluster object, required only if project is Glue::HPCC::Project::Project object.

=item cluster:  cluster name, REQUIRED

=item timestamp: time to do calculation at, defaults to now.

=item over_thr: threshold for excessive use, as percent.

=item under_thr: threshold for under use, as percent.

=back

On success, returns a list ref of two Glue::HPCC::ProjectUsage objects, the first
for high priority ussage, the second for total usage.  On non-fatal errors returns
a non-ref error message.

B<check_aacgrant_project_usage(key-value pairs)> checks the usage on a accgrant
type project.  Input parameters are the same as for B<check_coop_project_usage>,
except we expect a B<aacgrant> type project.  Returns a single ProjectUsage
object (for total usage).

B<check_engrgrant_project_usage(key-value pairs)> does the same as 
B<check_aacgrant_project_usage> but for engrgrant projects.  (Actually, the
two are the same, but different names in case they behave differently in the
future).



