#!/usr/local/bin/perl
#
#File utilities

package Glue::HPCC::Utils::FileUtils;
use strict;
use warnings;

use Carp qw(carp croak);
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw( 
	dryrun
	verbose
	files_in_dir_owned_by_uid
	archive_home_dir
	archive_data_dir
	du_directory
	du_subdirs_of_directory
	du_subdirs_of_dir_by_uid
	df_disk_containing_file
	parse_archive_metainfo_file
	write_archive_metainfo_file
	make_archive_metainfo_file
	delete_archived_home_dir
	delete_archived_data_dir
);
use File::Basename qw(basename dirname);
use Filesys::Df;
use File::Copy qw(move);
use File::Find;
use Fcntl ':mode';
use Glue::HPCC::Utils::ShellUtils qw(
		run_system_command
		run_backticked_command_always
	);
use POSIX qw(strftime); #For defaulting expiraton in make_archive_metainfo_file
use Time::ParseDate; #For checking expiration/exptstamp agree, make_archive_metainfo_file
use Glue::HPCC::Project::Log; #For logging archiving of files
use Glue::HPCC::Project::PersonProject; #For getting PoCs for projects in delete_archived_data

#-------------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------------

my $VERBOSE=0;
$VERBOSE=1;
my $DRYRUN=0;
my $PENDING_DIRNAME="pending_deletion";
my $DELETED_DIRNAME="DELETED";
my $METASUFF=".readme";

my $EXPIRATION_TIME_FORMAT = "%e %b %Y %H:%M";
my $DEFAULT_GRACEDAYS = 35;

#Metadata fields for archive files

#Structure of values is
#	[ fieldname, type, required ] 
#	(or undef, causing space in text output)
#	type is one of (scalar, boolean, listref, csvlist)
my @KNOWN_METADATA_FIELDS = 
(	[ 'DirectoryArchived', 'scalar', 1 ],
	[ 'DirectoryType', 'scalar', 1 ],
	undef,
	[ 'ArchivalDate', 'scalar', 0 ],
	[ 'ArchivalHost', 'scalar', 0 ],
	[ 'ArchivedBy', 'scalar', 0 ],
	undef,
	[ 'DirectoryOwnerUsername', 'scalar', 0 ],
	[ 'DirectoryOwnerUID', 'scalar', 1 ],
	[ 'DirectoryOwnerFullname', 'scalar', 0 ],
	undef,
	[ 'DirectoryOwnerProjects', 'csvlist', 1 ],
	[ 'ExcludePoCs', 'csvlist', 0 ],
	[ 'ReasonExcludePoCs', 'listref', 0 ],
	undef,
	[ 'ExpirationTstamp', 'scalar', 1 ],
	[ 'Expiration', 'scalar', 0 ],
	[ 'DoNotDelete', 'boolean', 0 ],
	undef,
	[ 'ReasonForArchiving', 'listref', 1 ],
	undef,
	[ 'ls-ld_Output', 'scalar', 0 ],
	[ 'ls-nd_Output', 'scalar', 0 ],
	undef,
	[ 'DeletionDate', 'scalar', 0 ],
	[ 'DeletedBy', 'scalar', 0 ],
	[ 'DeletionReason', 'listref', 0 ],
);

my %KNOWN_METADATA_FIELDS = map { (lc $_->[0] ) => $_ } 
	( grep { defined $_ } @KNOWN_METADATA_FIELDS);

#-------------------------------------------------------------------------
#	Class methods, for debugging, etc
#-------------------------------------------------------------------------

sub verbose(;$)
{	my $new = shift;
	if ( defined $new )
	{	$VERBOSE=($new)?1:0;
	}
	return $VERBOSE;
}

sub dryrun(;$)
{	my $new = shift;
	if ( defined $new )
	{	$DRYRUN=($new)?1:0;
	}
	return $DRYRUN;
}

#-------------------------------------------------------------------------
#	Filesystem queries (mostly wrappers)
#-------------------------------------------------------------------------

sub files_in_dir_owned_by_uid($$)
#Returns list ref of names of all files directly beneath dir $dir owned
#by uid $uid.  $uid must be numeric uid
{	my $dir = shift;
	my $uid = shift;
	my $me = 'files_in_dir_owned_by_uid';

	my $DIR;
	opendir($DIR,$dir) or croak "$me: Error opening directory $dir: $! at ";
	my @files = readdir($DIR);
	@files = grep { $_ ne '.' && $_ ne '..' } @files;
	@files = grep { (stat("$dir/$_"))[4] == $uid } @files;
	return [ @files ];
}

#----------------------------------------------------------
#		Disk usage (du) utilities
#----------------------------------------------------------

my $_wanted_sum;
my $_wanted_options;

sub old_wanted_sum_directory($)
{	#return if $_ eq '.' || $_ eq '..';
	my ($size, $lastmod);
	my $wanted_lastmod_before = $_wanted_options->{last_modified_before};
	if ( -l $_ )
	{	#Want the symlink, not what is pointing to
		my @tmp = lstat $_;
		$size = $tmp[7];
		$lastmod = $tmp[9];
	} else
	{	my @tmp = stat $_;
		$size = $tmp[7];
		$lastmod = $tmp[9];
	}

	if ( defined $wanted_lastmod_before )
	{	return if $lastmod > $wanted_lastmod_before;
	}

	$_wanted_sum += $size if defined $size;
}

sub wanted_sum_directory($)
#Does the work for du_directory.
#New version: for non-directories files, if num links > 1, divide size by num links
#This should do much better wrt not overcounting hard links
{	#return if $_ eq '.' || $_ eq '..';
	my ($size, $lastmod);
	my ( $nlinks, $isregfile );
	my $wanted_lastmod_before = $_wanted_options->{last_modified_before};

	if ( -l $_ )
	{	#Want the symlink, not what is pointing to
		my @tmp = lstat $_;
		$size = $tmp[7];
		$lastmod = $tmp[9];
		$isregfile = 0;
	} else
	{	my @tmp = stat $_;
		$size = $tmp[7];
		$lastmod = $tmp[9];
		$nlinks = $tmp[3];
		my $mode = $tmp[2];
		$isregfile = S_ISREG($mode);
	}

	if ( defined $wanted_lastmod_before )
	{	return if $lastmod > $wanted_lastmod_before;
	}

	#For regular files with nlinks > 1, split size over each hard link
	$size /= $nlinks if ( $isregfile && $nlinks > 1 );

	$_wanted_sum += $size if defined $size;
}

sub du_directory($;$)
#Given a directory, compute total size in bytes of everything in it.
#Options is an optional hash specifying criteria for including file in
#the sum.  Recognized keys are:
#	last_modified_before: tstamp, only files last modified before this
#		tstamp are included
#If $age is given, only include in sum files older than $age
#
#The help function (wanted_sum_directory) has been updated to divide usage for regular
#files by the number of links.  This should prevent double counting of hard links.
#Not perfect, but not unreasonable even when directory tree does not include all of
#the symlinks (though in this case, we WILL differ from du, but will be under so users
#wont be complaining), and give much better answers when all hard links are within
#the directory tree.
{	my $dpath = shift;
	my $options = shift || {};
	my $me = __PACKAGE__ . '::du_directory';

	croak "$me: Invalid options '$options', expecting hash ref, at "
		unless $options && ref($options) eq 'HASH';

	#Validate options keys (now, before call wanted)
	my @options = keys %$options;
	my %tmp = map { $_ => undef } @options;
	#Validate timestamp options
	my $tmpval;
	OPTLIST: foreach my $tmpopt ( qw(last_modified_before) )
	{	delete $tmp{$tmpopt};
		$tmpval = $options->{$tmpopt};
		next OPTLIST unless $tmpval;
		croak "$me: Invalid value for option $tmpopt: '$tmpval', " .
			"expecting timestamp.  Aborting at "
			unless $tmpval =~ /^\d+$/;
	}
	#Complain about unknown options
	$tmpval = join ", ", (keys %tmp);
	croak "$me: Unrecognized options [$tmpval] at " if ( %tmp );
	
	$_wanted_sum = 0;
	$_wanted_options = $options;
	find( {	wanted=>\&wanted_sum_directory,
		follow=>0,
		no_chdir=>1, #Avoids taint issues w/out untainting
		}, $dpath);
	return $_wanted_sum;
}

sub du_subdirs_of_directory(@)
#Returns the du in bytes for each top-level subdir of a directory.
#Takes key => value pairs:
#	rootdir: REQUIRED, full path to root directory
#	exclude: list ref of names of subdirs of top level to exclude.  
#		Default is []
#	options: hash ref of options for filtering.  See du_directory.
#		Default is {} or no filtering.
#Returns list ref of hash refs with keys:
#	fullpath: full path to the subdirectory
#	dirname: name of the subdirectory (e.g. path relative to rootdir)
#	sizeBytes: usage of that subdirectory, in bytes
#	uid: uid of owner of the directory
#	gid: gid of directory
{	my %args = @_;
	my $realme = __PACKAGE__ . '::du_subdirs_of_directory';
	
	my $rootdir = delete $args{rootdir};
	my $exclude = delete $args{exclude} || [];
	my $options = delete $args{options};
	my $me = delete $args{me} || $realme;

	croak "$me: Missing required argument 'rootdir' at " unless $rootdir;
	croak "$me: Rootdir '$rootdir' not found at " unless -e $rootdir;
	croak "$me: Rootdir '$rootdir' is not a directory at "
			unless -d $rootdir;

	if (%args)
	{	my @tmp = keys %args;
		my $tmp = join ", ", @tmp;
		croak "$me: Unrecognized arguments [$tmp], aborting, at ";
	}

	my %ehash = map { $_ => undef } @$exclude;
	my $dubydir = [];

	opendir(DIR, $rootdir) || 
		croak "$me: Unable to open dir $rootdir for read: $!, at ";
	my @dirs = readdir(DIR);
	closedir(DIR);

	#Exclude non-dirs, ., .., and anything in %ehash
	@dirs = grep { $_ ne '.' && $_ ne '..' } @dirs;
	@dirs = grep { -d "$rootdir/$_" } @dirs;
	@dirs = grep { ! exists $ehash{$_} } @dirs;
	#And ignore any symlinks at top level
	@dirs = grep { ! -l "$rootdir/$_" } @dirs;

	foreach my $dir (@dirs)
	{	my $fdir = "$rootdir/$dir";
		my @stat = stat "$fdir";
		my $owner = $stat[4];
		my $group = $stat[5];
		my $size  = du_directory($fdir, $options);
		
		my $rec = { 	fullpath => $fdir,
				dirname  => $dir,
				sizeBytes=> $size,
				uid      => $owner,
				gid    => $group,
			};
		push @$dubydir, $rec;
	}

	return $dubydir;
}
	
sub du_subdirs_of_dir_by_uid(@)
#Does du_subidrs_of_directory and then consolidates entries for the
#same uid.  Input is same as di_subdirs_of_directory: e.g key=>value pairs:
#	rootdir: REQUIRED, full path to root directory
#	exclude: list ref of names of subdirs of top level to exclude.  
#		Default is []
#	options: hash ref of options for filtering.  See du_directory.
#		Default is {} or no filtering.
#Returns hash ref keyed on uid of hash refs with keys:
#	uid: uid of owner of the directory
#	gid: gid of directory (if multiple, takes from the first dir)
#	sizeBytes: combined usage of subdirs with same owner, in bytes
#	fullpaths: list ref of full paths to the subdirectories
#	dirnames: list ref of names of the subdirectories
{	my %args = @_;
	my $me = __PACKAGE__ . '::du_subdirs_of_dir_by_uid';
	
	my $rootdir = delete $args{rootdir};
	my $exclude = delete $args{exclude} || [];
	my $options = delete $args{options};

	croak "$me: Missing required argument 'rootdir' at " unless $rootdir;
	croak "$me: Rootdir '$rootdir' not found at " unless -e $rootdir;
	croak "$me: Rootdir '$rootdir' is not a directory at "
			unless -d $rootdir;

	if (%args)
	{	my @tmp = keys %args;
		my $tmp = join ", ", @tmp;
		croak "$me: Unrecognized arguments [$tmp], aborting, at ";
	}

	my $dubydirs = du_subdirs_of_directory( me => $me,
		rootdir => $rootdir, exclude=>$exclude, options=>$options);

	my $dubyuid = {};

	foreach my $rec (@$dubydirs)
	{	my $fpath = $rec->{fullpath};
		my $dir   = $rec->{dirname};
		my $size  = $rec->{sizeBytes};
		my $uid   = $rec->{uid} || 0;
		my $gid   = $rec->{gid} || 0; 
		#Force numeric evaluation
		$uid = 0 + $uid;
		$gid = 0 + $gid;

		if ( exists $dubyuid->{$uid} )
		{	my $rec0 = $dubyuid->{$uid};
			my $size0 = $rec0->{sizeBytes};
			my $fpaths = $rec0->{fullpaths};
			my $dirs = $rec0->{dirnames};

			$size0 += $size;
			push @$fpaths, $fpath;
			push @$dirs, $dir;

			$rec0->{sizeBytes} = $size0;
			$rec0->{fullpaths} = $fpaths;
			$rec0->{dirnames} = $dirs;

		} else
		{	my $tmprec = {
				uid => $uid,
				gid => $gid,
				sizeBytes => $size,
				fullpaths => [ $fpath ],
				dirnames => [ $dir ],
			};
			$dubyuid->{$uid} = $tmprec;
		}

	}
	return $dubyuid;
}

sub df_disk_containing_file($)
#Get disk usage stats for disk contain filepath
#Returns ($totalsize, $pctfree, $pctused)
#where totalsize is in Bytes
{	my $filepath = shift;
	my $me = __PACKAGE__ . '::get_disk_stats';
	
	my $ref = df($filepath);
	my $blocks = $ref->{blocks} || 0;
	my $pctused = $ref->{per} || 0;
	my $total = 1024 * $blocks;
	my $pctfree = 100 - $pctused;

	return ( $total , $pctfree, $pctused);
}

#-------------------------------------------------------------------------
#	Archival system: misc low-level utilities
#-------------------------------------------------------------------------

sub get_pending_dirname($)
{	my $dir = shift;
	my $rootdir = dirname($dir);
	my $pending_dir = "$rootdir/$PENDING_DIRNAME";

	my @stat = stat($pending_dir);
	unless ( @stat )
	{	print STDERR "Creating $pending_dir ...\n" if $VERBOSE;
		if ( $DRYRUN )
		{	print STDERR "[DRYRUN] mkdir $pending_dir, 0700";
			#Fake out stat
			@stat = ( undef, undef, 0700, undef, 0, 0);
		} else
		{	mkdir $pending_dir, 0700;
			@stat = stat($pending_dir);
		}
	}

	my $uid = $stat[4];
	my $gid = $stat[5];
	my $mode = S_IMODE($stat[2]);

	unless ( $uid == 0 && $gid == 0 )
	{	print STDERR "Setting ownership of $pending_dir to 0:0 from ${uid}:${gid}\n"
			if $VERBOSE;
		if ( $DRYRUN )
		{	print STDERR "[DRYRUN] chown 0:0 $pending_dir";
		} else
		{	chown 0, 0,  $pending_dir;
		}
	}

	unless ( $mode == 0700 )
	{	printf STDERR "Setting mode of $pending_dir to 0700 from %04o",
			$mode if $VERBOSE;
		if ( $DRYRUN )
		{	print STDERR "[DRYRUN] chmod 0700 $pending_dir";
		} else
		{	chmod 0700, $pending_dir;
		}
	}

	return $pending_dir;
}

sub _find_archived_directory($;$)
#This takes an absolute path to one of:
#	the "archived" directory (i.e. relocated under pending_deletion)
#	the "archived" tarball (also under pending_deletion)
#	the metadata file for one of the above
#	the original, unarchived path
#Optional parameter $me is used for impersonating another routine
#in error messages.
#
#It determines what was given, and returns a hash ref with the
#following fields
#	archive: the path to the "archived" dir or tarball
#	archive_type: 'dir' or 'tarball'
#	metafile: the path to the corresponding metadata information file 
#	origpath: the original path
#	metadata: The hash from reading metafile
#
#On non-fatal error, returns non-ref error message.
{	my $rawpath = shift;
	my $me = shift || '_find_archived_directory';

	#Make sure got an absolute path
	if ( $rawpath !~ '^/' )
	{	return "$me: archive must be an absolute path, not '$rawpath'";
	}

	#For archived dir or tarball, or metadata, the path must exist
	#and be in pending_deletion directory.  Anything else presumably
	#must be an original directory path

	my $archive; #The actual "archived" data, either dir or tarball
	my $metadata; #The corresponding metadata file
	my $origpath; #The original location of the data

	if ( ( ! -e $rawpath ) || 
	     ( -e $rawpath && $rawpath !~ /\/$PENDING_DIRNAME\// ) )
	{	#Must have given an original, unarchived path
		my $rootdir = dirname($rawpath);
		my $pending_dir = "$rootdir/$PENDING_DIRNAME";
		unless ( -d $pending_dir )
		{	#Pending deletion dir must already exist, or there is a problem
			return "$me: unable to find archive from orig path $rawpath"
				. "($pending_dir doesnt exist)";
		}

		$origpath = $rawpath;
		my $basename = basename($rawpath);
		$archive = "${pending_dir}/${basename}";
		$metadata = "${archive}${METASUFF}";
		unless ( -f $metadata )
		{	return "$me: unable to find archive from orig path $rawpath"
				. "($metadata doesnt exist)";
		}

		unless ( -e $archive )
		{	#Try a tarball
			my $tmp = "${archive}.tar.gz";
			unless ( -e $tmp )
			{	return "$me: unable to find archive from orig path $rawpath"
					. "($archive, $tmp dont exist)";
			}
			$archive = $tmp;
		}
	} elsif ( $rawpath =~ /$METASUFF$/ )
	{	#Looks like we got a metadata file (already know it exists and in pending_deletion dir)
		$metadata = $rawpath;
		$archive = $metadata;
		$archive =~ s/$METASUFF$//;
		unless ( -e $archive )
		{	#OK, try a tarball
			my $tmp = "${archive}.tar.gz";
			unless ( -e $tmp )
			{	return "$me: unable to find archive from meta $metadata"
					. "($archive, $tmp dont exist)";
			}
			$archive = $tmp;
		}
		
	} else
	{	#Presumably an archive (exists and in pending dir)
		$archive = $rawpath;
		if ( -d $archive )
		{	#Is a directory,
			$metadata = "${archive}$METASUFF";
		} else
		{	#Is a file
			my $tmp = $archive;
			$tmp =~ s/\.tar(\.gz)?$//;
			$metadata = "${tmp}$METASUFF";
		}
		unless ( -f $metadata )
		{	return "$me: unable to find metadata for $archive";
		}
	}

	#archive, metadata should be set.  origpath maybe.
	my $type = 'dir';
	$type = 'tarball' if -f $archive;

	my $minfo = parse_archive_metainfo_file($metadata);
	unless ( $minfo && ref($minfo) eq 'HASH' )
	{	return "$me: unable to parse metadata $metadata: $minfo";
	}

	if ( $origpath )
	{	#Orig path given, make sure metadata agrees
		unless ( $minfo->{DirectoryArchived} eq $origpath )
		{	return "$me: archive dir mismatch, given $origpath, "
				. "but metadata says " . $minfo->{DirectoryArchived};
		}
	} else
	{	$origpath = $minfo->{DirectoryArchived};
	}

	my $res = 
	{	archive => $archive,
		archive_type => $type,
		metafile => $metadata,
		metadata => $minfo,
		origpath => $origpath,
	};

	return $res;
}

#-------------------------------------------------------------------------
#	Archival system: metadata information file stuff
#-------------------------------------------------------------------------
	
sub parse_archive_metainfo_file($;$)
#Parses an archive metainfo file/text.  On success returns
#a hash ref with keys:
#	DirectoryArchived: (string) directory archived*
#	DirectoryType: 'home', 'nfs-data', 'lustre-data',
#	ArchivalDate: (string) date archive made
#	ArchivalHost: (string) host archive made on
#	ArchivedBy: (string) username of person who archived the directory
#	DirectoryOwnerUsername: (string) username that owned directory
#	DirectoryOwnerUID: (int) UID that owned directory*
#	DirectoryOwnerFullname: (string) fullname of owner
#	DirectoryOwnerProjects: (list ref) list of project names
#		owner belonged to*
#	DoNotDelete: (bool), if set, directory should not be deleted by
#		archival deletion routines.  Default is false
#	ExpirationTstamp: (int) tstamp of expiration of archive*
#	Expiration: (string) human readable time of expiration of archive*
#	ReasonForArchiving: (list ref) strings giving reason for archival*
#	ls-ld_Output: Output of ls -ld command on archive 
#	ls-nd_Output: Output of ls -nd command on archive 
#	ExcludePoCs: (list ref) list of PoC usernames that do not want email
#		about this directory anymore
#	ReasonExcludePoCs: (list ref) reason for adding names to ExcludePoCs
#	ExtraLines: A list ref containing any lines that do not have a colon
#
#The following additional ones are valid after archive deleted
#	DeletionDate: date deleted
#	DeletedBy: person who deleted
#	DeletionReason: reason for deletion (list ref)
#
#The following keys may be set, but are not part of the Metadata record:
#	MetadataSourceType: 'file' if given filename
#	MetadataSource: filename for type 'file'
#
#On non-fatal errors, a non-ref error string will be returned.
#
#The $metainfo parameter can be any of the following forms:
#	a string representing path to a metainfo file
#	a string reference representing the contents (as a string) of
#		a metainfo file
#	an array reference, representing the contents (one string 
#		element per line) of a metainfo file
#	 a file handle reference (e.g. IO::Handle or sub-class): metainfo
#		will be read from that file handle 
#	a GLOB (e.g. \*STDIN): metainfo	will be read from that file handle
#
#If $verify is given and true, will generate warnings if important fields
#are missing (the important fields are starred above).  If greater than 1, 
#will generate warnings if any fields are missing.  If $verify is true, 
#warnings will be generated if any ExtraLines as well.
#
#If one of Expiration or ExpirationTstamp missing, will default the missing
#one from the one given.  If both given and $verify is true, it will complain
#if they do not agree (within $GRACESECS seconds).  If neither given, it will
#complain if $verify true and default from ArchivalDate if present.  If neither
#found and no ArchivalDate, will return an error string.
#
#File format is 
#field: value
#Fields returning list refs can be repeated, and values will be appended to
#the list ref.
#The special field '-' is a shortcut for the previous field name.
#Lines starting with # are ignored (comments), as are blank lines
#Leading/trailing spaces around field and value deleted.
#Lines without a ':' are silently ignored unless $verify.
#If $verify > 1, will cause non-fatal error, otherwise warning if $verify
#and ignored.
#
# *NOTE*: If you make any changes to output format, please make 
# corresponding changes to write_archive_metainfo_file
{	my $metainfo = shift;
	my $verify = shift;
	my $me = __PACKAGE__ . '::parse_archive_metainfo_file';
	$verify = 0 unless $verify;
	$verify = 1 if $verify && $verify !~ /^\s*\d+\s*$/;

	croak "$me: Missing required parameter 'metainfo' at " unless $metainfo;
	my @lines;
	my $fname = "unknownFile";

	my $retval = {};
	my $extraLines = [];

	#Read the file, etc.
	if ( ref($metainfo) eq 'ARRAY' )
	{	#Should be a listref of strings with actual metainfo
		@lines = @$metainfo;
		$retval->{MetadataSourceType} = 'array';
		$retval->{MetadataSource} = $metainfo;
	} elsif ( ref($metainfo) eq 'SCALAR' )
	{	#Contains string with actual metainfo
		@lines = split($$metainfo,/\n/);
		$retval->{MetadataSourceType} = 'text';
		$retval->{MetadataSource} = $$metainfo;
	} elsif ( ! ref($metainfo) )
	{	#Non-ref, is path to a file
		return "$me: Unable to find metainfo file '$metainfo'"
			unless -f $metainfo;
		$fname = $metainfo;
		$retval->{MetadataSourceType} = 'file';
		$retval->{MetadataSource} = $metainfo;
		my $MIFILE;
		open($MIFILE, "<$metainfo") or 
			return "$me: Error opening metainfo file '$metainfo'" .
				" for reading: $!";
		@lines = <$MIFILE>;
		close $MIFILE;
		chomp @lines;
	} else
	{	@lines = <$metainfo>;
		chomp @lines;
		$retval->{MetadataSourceType} = 'filehandle';
		$retval->{MetadataSource} = $metainfo;
	}

	#Now parse it

	my $linenum = 0;
	my $lastkey;
	my $tmp;
	LINE: foreach my $line (@lines)
	{	chomp $line;
		$linenum++;
		my $fnln = $fname . ':' . $linenum;
		next LINE unless $line;
		$line =~ s/^\s*//; $line =~ s/\s*$//;
		next LINE if $line =~ /^#/;
		$fname = $metainfo;
		unless ( $line =~ /:/ )
		{	#
			#warn "$me: Bad line $fnln '$line', no colon, added to ExtraLines and skipping"
			#	if $verify;
			push @$extraLines, $line;
			next LINE;
		}

		my ($key, $val ) = split /\s*:\s*/, $line, 2; #Anything after first : is part of value
		if ( $key eq '-' )
		{	unless ( $lastkey )
			{	return "$me: Bad line $fnln '$line', using '-' for field" .
					" but no prior fieldname.";
				#next LINE;
			}
			$key = $lastkey;
		}
		unless ( $key )
		{	return "$me: Missing fieldname at $fnln, '$line'";
		}

		$lastkey = $key;
		my $lckey = lc $key;
		unless ( exists $KNOWN_METADATA_FIELDS{$lckey} )
		{	warn "$me: Unrecognized field '$key' at $fnln, skipping";
			next LINE;
		}

		unless ( defined $val )
		{	warn "$me: Missing value for key $key at $fnln"
				if $verify;
			return "$me: Missing value for key $key at $fnln"
				if $verify && $verify > 1;
			next LINE;
		}

		my $rec = $KNOWN_METADATA_FIELDS{$lckey};
		my ( $fld, $type, $req ) = @$rec;

		if ( $type eq 'scalar' )
		{	$tmp = $retval->{$key};
			if ( defined $tmp )
			{	warn "$me: Duplicate entry for non-listref field $fld at " .
					"$fnln; previous value is '$tmp'."
					if $verify;
				return "$me: Duplicate entry for non-listref field $fld " .
					"at $fnln; previous value is '$tmp'"
					if ( $verify && $verify > 1 );
			}
			$retval->{$key} = $val;
			next LINE;
		} 
		
		if ( $type eq 'boolean' )
		{	$tmp = $retval->{$key};
			if ( defined $tmp )
			{	warn "$me: Duplicate entry for non-listref field $fld at " .
					"$fnln; previous value is '$tmp'."
					if $verify;
				return "$me: Duplicate entry for non-listref field $fld " .
					"at $fnln; previous value is '$tmp'"
					if ( $verify && $verify > 1 );
			}

			#Strip leading/trailing spaces, and make lowercase
			$val =~ s/^\s*//; $val =~ s/\*$//;
			$val = lc $val;

			if ( $val eq '1' || $val eq 'on' ||
				$val eq 'y' || $val eq 'yes' ||
				$val eq 't' || $val eq 'true' )
			{	$retval->{$key} = 1;
				next LINE;
			}

			if ( $val eq '0' || $val eq 'off' || $val eq '' ||
				$val eq 'n' || $val eq 'no' ||
				$val eq 'f' || $val eq 'false' )
			{	$retval->{$key} = 0;
				next LINE;
			}

			#Not clearly true or false, will default to true
			#but warn
			warn "$me: In $fnln, non-standard boolean value " .
				"'$val' at " if $verify;
			return "$me: In $fnln, non-std bool value '$val'"
				if $verify && $verify > 1;
			#Default to true
			$retval->{$key} = 1;
			next LINE;
		} 
		
		if ( $type eq 'csvlist' )
		{	$tmp = $retval->{$key};
			$tmp = [] unless defined $tmp;
			my @temp = split /\s*,\s*/, $val;
			push @$tmp, @temp;
			$retval->{$key} = $tmp;
			next LINE;
		}

		if ( $type eq 'listref' )
		{	$tmp = $retval->{$key};
			$tmp = [] unless defined $tmp;
			push @$tmp, $val;
			$retval->{$key} = $tmp;
			next LINE;
		}
			
		croak "$me: Illegal field type '$type' for field $fld at ";
	}

	#Do any defaulting required
	my $expts = $retval->{ExpirationTstamp};
	my $expstr = $retval->{Expiration};

	my $tempts;
	$tempts = parsedate($expstr,
		WHOLE => 1, DATE_REQUIRED => 1, NO_RELATIVE => 1,
		PREFER_FUTURE => 1, VALIDATE => 1)
		if $expstr;
	my $EXPIRATION_TIME_FORMAT = "%e %b %Y %H:%M";

	if ( defined $expts )
	{	#We have an expiration tstamp
		my $GRACESECS = 60*60*24; #1 day in seconds
		my $tempstr = strftime($EXPIRATION_TIME_FORMAT, localtime($expts) );
		
		if ( $expstr )
		{	#Both expiration string and tstamp given
			#Make sure agree (within $GRACESECS)
			unless ( $tempts - $GRACESECS < $expts &&
				$tempts + $GRACESECS > $expts )
			{	warn "$me: Expiration tstamp ($tempstr) and string ($expstr) " .
					"disagree at ";
				return "$me: Exp tstamp ($tempstr) and string ($expstr) disagree";
				
			}
		} else
		{	warn "$me: Using exp tstamp to set exp string at " if $verify;
			$expstr = $tempstr;
			$retval->{Expiration} = $expstr;
		}
	} elsif ( $expstr )
	{	#Expiration string exists but not tstamp
		warn "$me: Using exp string to set exp tstamp at " if $verify;
		$expts = $tempts;
		$retval->{ExpirationTstamp} = $expts;
	} else
	{	#Neither expiration string/tstamp set
		my $tempstr = $retval->{ArchivalDate};
		$tempts = undef;
		if ( $tempstr )
		{ 	$tempts =  parsedate($expstr,
				WHOLE => 1, DATE_REQUIRED => 1, NO_RELATIVE => 1,
				PREFER_FUTURE => 1, VALIDATE => 1);
		}
	
		if ( defined $tempts )
		{	#Got archival date, add 6 months to default expiration
			$tempts += 60*60*24*31*6; #6 months in seconds
			$tempstr = strftime($EXPIRATION_TIME_FORMAT, localtime($tempts) );
			$expts = $tempts;
			$expstr = $tempstr;
			$retval->{ExpirationTstamp} = $expts;
			$retval->{Expiration} = $expstr;
			warn "$me: Defaulting expiration (string/tstamp) from archival date at "
				if $verify;
		} else
		{	warn "$me: No expiration string/tstamp (and no archival date) in $fname at ";
		}
	}


	#Make sure all required fields are present
	FIELD: foreach my $rec ( grep { defined $_ } @KNOWN_METADATA_FIELDS )
	{ 	my ( $fld, $type, $req ) = @$rec;
		my $val = $retval->{$fld};
		
		if ( defined $val )
		{	#Defined, make sure is of correct type
			if ( $type eq 'scalar' || $type eq 'boolean' )
			{	warn "$me: In $fname, illegal value '$val' for scalar field '$fld', " .
					"expecting scalar at "
					if ref($val);
			} elsif ( $type eq 'csvlist' || $type eq 'listref' )
			{	warn "$me: In $fname, illegal value '$val' for $type field '$fld', " .
					"expecting listref at "
					unless ref($val) eq 'ARRAY';
				#Make sure not an empty list if required/verify>1
				unless ( scalar(@$val) )
				{	if ( $req )
					{	warn "$me: In $fname, empty list for required $type " .
							"field $fld at "
							if $verify;
					} elsif ( $verify && $verify > 1 )
					{	warn "$me: In $fname, empty list for $type field $fld at";
					}
				}
			} else
			{	#Code error, not runtime error
				die "$me: In $fname, Illegal type '$type' for field '$fld' at ";
			}

			next FIELD;
		}

		#Missing field, is it required
		if ( $req )
		{	warn "$me: In $fname, missing required field '$fld' at "
				if $verify;
		} elsif ( $verify && $verify > 1 )
		{	warn "$me: In $fname, missing field '$fld' at ";
		}
	}

	#Complain if any ExtraLines in $verify mode
	if ( $extraLines && ref($extraLines) eq 'ARRAY' && scalar(@$extraLines) )
	{	$retval->{ExtraLines} = $extraLines;
		if ( $verify )
		{	my $temptxt = join "\n:   ", @$extraLines;
			warn "$me: In $fname, the following lines found without a colon\n:   $temptxt\nat ";
			die "Aborting due to lines w/out colon at " if $verify > 1;
		}
	}

	return $retval;
}

sub write_archive_metainfo_file($$;$)
#Writes the specified metadata to the specified metadata info file
#Optional options hash recognizes keys:
#	force: allows clobbering of existing file
#
#Returns false on success, true (error message) on non-fatal errors.
#
#Pseudo-fields MetadataSourceType, MetadataSource are ignored
#
# *NOTE*: If you make any changes to output format, please make 
# corresponding changes to parse_archive_metainfo_file
{	my $mfile = shift;
	my $metadata = shift;
	my $options = shift || {};
	my $me = 'write_archive_metainfo_file';

	croak "$me: Missing required parameter 'mfile' at "
		unless $mfile;
	croak "$me: Missing required parameter 'metadata' at "
		unless $metadata;
	croak "$me: Invalid value '$metadata' for metadata, expect hash ref, at "
		unless ref($metadata) eq 'HASH';
	
	my $force = delete $options->{force};
	if ( %$options )
	{	my @tmp = keys %$options;
		my $tmp = join ", ", @tmp;
		croak "$me: Unrecognized options [ $tmp ] at ";
	}

	if ( -f $mfile )
	{	#Don;t clobber unless force
		return "$me: Refusing to clobber $mfile w/out force mode"
			unless $force;
	}

	my $MFILE;
	if ($DRYRUN)
	{	$MFILE = \*STDERR;
		print STDERR "[DRYRUN] Would be outputing to $metadata: \n";
	} else
	{	open($MFILE, ">$mfile") 
			or die "Error opening $mfile for write: $!";
	}
	print $MFILE "#Metainfo for archival of directory\n";
	my $prog = $0 || '<unknown>';
	my $user = $ENV{USER} || '<unknown>';
	my $when = strftime("%Y-%m-%eT%H:%M:%S", localtime);
	print $MFILE "#Written by program $prog, user $user, $when\n";

	#Track what metadata keys have processed
	my %tmpmd = map { $_ => undef } (keys %$metadata);

	#We ignore pseudo-fields MetadataSourceType, MetadataSource
	delete $tmpmd{MetadataSourceType};
	delete $tmpmd{MetadataSource};

	FLD: foreach my $frec (@KNOWN_METADATA_FIELDS)
	{	#Non-array ref records mean print a blank line
		unless ( $frec && ref($frec) eq 'ARRAY' )
		{	print $MFILE "\n";
			next FLD;
		}

		my ( $fname, $ftype, $freq ) = @$frec;
		my $val = $metadata->{$fname};
			delete $tmpmd{$fname};

			unless ( defined $val )
			{	print $MFILE "#$fname: \n";
				next FLD;
			}

			if ( $ftype eq 'scalar' )
			{	print $MFILE "$fname: $val\n";
				next FLD;
			}

			if ( $ftype eq 'boolean' )
			{	#Canonicalize booleans
				$val = $val?'true':'false';
				print $MFILE "$fname: $val\n";
				next FLD;
			}

			if ( $ftype eq 'listref' )
			{	return "$me: Invalid value '$val' for listref field $fname,"
					. " expecting listref"
					unless ( ref($val) eq 'ARRAY' );
				unless ( scalar(@$val) )
				{	print $MFILE "#$fname: \n";
					next FLD;
				}
				my $tmpstr = join "\n-:\t", @$val;
				print $MFILE "$fname:	$tmpstr\n";
				next FLD;
			}

			if ( $ftype eq 'csvlist' )
			{	return "$me: Invalid value '$val' for csvlist field $fname,"
					. " expecting listref"
					unless ref($val) eq 'ARRAY';
				unless ( scalar(@$val) )
			{	print $MFILE "#$fname: \n";
				next FLD;
			}
			my $tmpstr = join ", ", @$val;
			print $MFILE "$fname:	$tmpstr\n";
			next FLD;
		}

	}

	my $val = $metadata->{ExtraLines};
	delete $tmpmd{ExtraLines};
	if ( $val && ref($val) eq 'ARRAY' && scalar(@$val) )
	{	print $MFILE "\n#---------------------------------------------\n";
		print $MFILE "#		Extra Lines\n";
		print $MFILE "#---------------------------------------------\n";
		foreach my $tmp (@$val)
		{	print $MFILE "$tmp\n";
		}
		print $MFILE "#---------------------------------------------\n\n";
	}

	if ( %tmpmd )
	{	#We were given keys that did not process
		my @tmp = keys %tmpmd;
		my $tmp = join ", ", @tmp;
		return "$me: Unrecognized keys [ $tmp ] in metadata hash, aborting write";
	}

	unless ( $DRYRUN )
	{	close($MFILE) or die "Error closing $metadata after write: $!";
	}
	return;
}

sub make_archive_metainfo_file(@)
#Create a metadata file for the archive
#Takes key => value pairs for arguments, recognizing
#	METAINFO_FILE: full path to metainfo file to write. Will default
#		based on directory name, etc.  Usually want to let default.
#	me: method to masquerade as
#	hpc: Glue::HPCC::Cluster instance, needed for getting
#		info re fileowner, groups, etc.  Required??? If omitted,
#		much metainfo missing
#	force: Boolean, are we in force mode.  If set, allows clobbering
#		of metadata, etc.
#	logreason: reason for inclusion in PDB log table.  Default is
#		to concatenate $reason
#
#	directory: directory to archive, REQUIRED
#	reason: reason for archiving (for metainfo file).  Can be string 
#		or list ref of strings.	REQUIRED
#	uid: UID of owner of directory.  If omitted, get from filesystem
#	owner: Owner of directory (username).  If omitted, try to get
#		from uid and passwd file and/or hpc
#	fullname: Fullname of owner.  Will get from owner and hpc if hpc
#		given and user in PDB
#	projects: Projects owner belongs to (who should get notified of
#		archival status, pending deletion).  Will try to get from
#		owner and hpc if hpc given and user in PDB, otherwise []
#	expiration: When archive expires and should be deleted (or at least
#		when warn users is being deleted).  Human readable format
#	exptstamp: When the archive expires and should be deleted, in
#		seconds from 1970.
#	fstype: 'home', 'nfs-data', 'lustre-data'.  
#	dontdelete: Boolean, if set, add DoNotDelete: 1 line
#	excludePocs: list ref: list of PoCs to exclude from sending email
#		Defaults to empty list.
#	exclPocsReason: list ref, list of reasons for Excluding PoCs.
#		Defaults to empty list, must not be empty if excludePocs is
#		non-empty.
#	deletionDate: Human readable date of when archive was deleted. Optional.
#		Only relevant if archive has been deleted.  This
#		typically is only used by the delete_archive method.
#		Default is to omit this field.
#	deletionReason: list ref.  Reason for deletion.  Optional, unless
#		deletionDate given, in which case required (and must be non-empty).
#		Not valid unless deletionDate set.
#	deletedBy: Name of user who deleted directory.  Optional, unless
#		deletionDate given, in which case required.
#		Not valid unless deletionDate set.
#
#If only one of expiration and exptstamp given, default the missing one
#from the one given.  If both given, they must agree (within $GRACESECS
#seconds, see code) or else a warning will be generated.  If both given,
#exptstamp is definitive, and expiration will be set from that.  If neither
#given, defaults to 6 months from now.
#
#Returns false (undef) on success, true (error message) on non-fatal errors.
#
# *NOTE*: If you make any changes to output format, please make 
# corresponding changes to parse_archive_metainfo_file
#
{	my %args = @_;
	my $me = delete $args{me} || 
		( __PACKAGE__ . '::make_archive_metainfo_file' );

	my $metadata = delete $args{METAINFO_FILE};
	my $logreason = delete $args{logreason};
	my $force = delete $args{force};
	my $hpc = delete $args{hpc};

	my $dir = delete $args{directory};
	my $reason = delete $args{reason};
	my $uid = delete $args{uid};
	my $owner = delete $args{owner};
	my $fullname = delete $args{fullname};
	my $projects = delete $args{projects};
	my $expiration = delete $args{expiration};
	my $exptstamp = delete $args{exptstamp};
	my $fstype = delete $args{fstype};
	my $dontdelete = delete $args{dontdelete};
	my $exclpocs = delete $args{excludePocs};
	my $exclpoc_reasons = delete $args{exclPocsReason};
	my $deletionDate = delete $args{deletionDate};
	my $deletionReason = delete $args{deletionReason};
	my $deletedBy = delete $args{deletedBy};

	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}
	croak "$me: Missing required parameter 'directory' at " unless $dir;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$reason = [ $reason ] unless $reason && ref($reason) eq 'ARRAY';
	croak "$me: Missing req parm 'reason', got empty list, at " unless scalar(@$reason);

	#Do defaults

	#Default metadata file name
	my $pending = get_pending_dirname($dir);
	my $basename = basename($dir);
	$metadata = "${pending}/${basename}${METASUFF}" unless ( $metadata );

	if ( -e $metadata )
	{	if ( $force )
		{	print STDERR "Clobbering old $metadata file\n"
				if $VERBOSE;
		} else
		{	return "Refusing to clobbering existing file $metadata";
		}
	}


	unless ( defined $uid )
	{	#Get uid of owner from filesystem
		my @stat = stat $dir;
		$uid = $stat[4] if @stat;
	}

	$logreason = join "; ",  @$logreason if $logreason && ref($logreason) eq 'ARRAY';
	$logreason = join "; ",  @$reason unless $logreason;

	#Default fstype if can
	unless ( defined $fstype )
	{	$fstype = 'nfs-data' if $dir =~ /data/;
		$fstype = 'lustre-data' if $dir =~ /lustre/;
		$fstype = 'home' if $dir =~ /home/;
	}
	croak "$me: Missing required parameter 'fstype' and unable to default at " unless $fstype;
	$fstype = lc $fstype;
	croak "$me: Invalid value '$fstype' for fstype; expecting one of " .
		"'home', 'nfs-data', or 'lustre-data' at "
		unless $fstype eq 'home' || $fstype eq 'nfs-data' || $fstype eq 'lustre-data';


	my $pdbuser;
	$pdbuser = Glue::HPCC::Project::Person->person_by_uid($hpc, $uid)
		if $hpc && ref($hpc) && defined $uid;
	my @pwuser;
	@pwuser = getpwuid($uid);
	
	unless ( $owner )
	{	#Try PDB first
		$owner = $pdbuser->username if ( $pdbuser && ref($pdbuser) );
		#and fallback to password file
		$owner = $pwuser[0] unless $owner;
	}
	unless ( $fullname )
	{	#Try PDB first
		$fullname = $pdbuser->fullname if  ($pdbuser && ref($pdbuser));
		#and fallback to password file
		unless ( $fullname )
		{	$fullname = $pwuser[6] unless $owner;
			#Strip anything after 1st comma
			$fullname =~ s/,.*$//;
		}
	}
	unless ( $projects )
	{	if ( $hpc && ref($hpc) )
		{	$projects = Glue::HPCC::Project::PersonProject->get_projects_user_is_attached_to(
				$hpc, $owner, 'ANY');
		}
		$projects = [] unless $projects;
	}
	unless ( ref($projects) eq 'ARRAY' )
	{	$projects = [ split /\s*,\s*/, $projects ];
	}
	unless ( ref($reason) eq 'ARRAY' )
	{	$reason = [ split /\s*;\s*/, $reason ];
	}

	my $expformat = "%e %b %Y %H:%M";
	my $defexptstamp = time + 6*31*24*60*60; #6 months from now, in seconds
	my $defexpstring = "6 months from now"; #Make sure agrees with above

	if ( $exptstamp )
	{	#tstamp formatted expiration given
		my $temptime = strftime($EXPIRATION_TIME_FORMAT, localtime($exptstamp) );
		if ( $expiration )
		{	#gave both expiration and exptstamp, make sure agree
			my $GRACESECS = 60*60*24; #Grace period, 1 day in seconds
			my $tempts = parsedate($expiration, 
				WHOLE => 1, DATE_REQUIRED => 1, NO_RELATIVE => 1,
				PREFER_FUTURE => 1, VALIDATE => 1);
			if ( defined $tempts )
			{	#Make sure agree to within $GRACESECS
				unless ( $tempts - $GRACESECS < $exptstamp &&
					$tempts + $GRACESECS > $exptstamp )
				{   warn "$me: Expiration string ($expiration) and tstamp ($temptime) "
						. "do NOT agree; using timestamp at ";
				}
			} else
			{	#Unable to parse $expiration
				warn "$me: Unable to parse expiration string ($expiration), " .
					"using $temptime from timestamp instead, at ";
			}
		}
		#If both given, we use the timestamp value (silently if withing $GRACESECS 
		#of each other).  
		#	--- OR ---
		#string format was not given, in which case we still default from timestamp
		$expiration = $temptime;
	} elsif ( $expiration )
	{	#tstamp not given, but string was given
		#Default timestamp from string
		$exptstamp = parsedate($expiration, 
			WHOLE => 1, DATE_REQUIRED => 1, NO_RELATIVE => 1,
			PREFER_FUTURE => 1, VALIDATE => 1);
		unless ( defined $exptstamp )
		{	warn "$me: Unable to parse expiration string ($expiration), " .
				"defaulting to $defexpstring at ";
			$exptstamp = $defexptstamp;
			$expiration = strftime($EXPIRATION_TIME_FORMAT, localtime($exptstamp) );
		}
	} else
	{	#Neither given, default to $defexptstamp
		$exptstamp = $defexptstamp;
		$expiration = strftime($EXPIRATION_TIME_FORMAT, localtime($exptstamp) );
	}
				

	my $date = strftime("%d %b %Y %H:%M", localtime);
	my $archiver = $ENV{USER} || 'unknown';
	my $host = $ENV{HOST} || $ENV{HOSTNAME} || 'unknown';

	my $lsld = run_backticked_command_always("ls -ld $dir");
	my $lsnd = run_backticked_command_always("ls -nd $dir");
	chomp ($lsld);
	chomp ($lsnd);

	$dontdelete = 0 unless defined $dontdelete;
	#Strip leading/trailing spaces, set to lowercase
	$dontdelete =~ s/^\s*//; 
	$dontdelete =~ s/\s*$//;  
	$dontdelete = lc $dontdelete;
	$dontdelete = 1 if $dontdelete eq 'on' || 
		$dontdelete eq 'y' || $dontdelete eq 'yes' ||
		$dontdelete eq 't' || $dontdelete eq 'true';
	$dontdelete = 0 if $dontdelete eq 'off' || $dontdelete eq '' ||
		$dontdelete eq 'n' || $dontdelete eq 'no' ||
		$dontdelete eq 'f' || $dontdelete eq 'false';
	unless ( $dontdelete eq '0' || $dontdelete eq '1' )
	{	my $tmp = $dontdelete?1:0;
		warn "$me: Treating nonstandard value '$dontdelete' for dontdelete as $tmp at ";
		$dontdelete = $tmp;
	}

	$exclpocs = [] unless defined $exclpocs;
	$exclpocs = [ $exclpocs ] unless ref($exclpocs) eq 'ARRAY';
	
	$exclpoc_reasons = [] unless defined $exclpoc_reasons;
	$exclpoc_reasons = [ $exclpoc_reasons ] unless ref($exclpoc_reasons) eq 'ARRAY';

	if ( scalar(@$exclpocs) )
	{	unless ( scalar(@$exclpoc_reasons) )
		{	return "$me: ExcludePocs non-empty, but ExclPocReasons empty";
		}
	}
	
	if ( $deletionDate )
	{	#Verify other required fields provided
		unless ( $deletedBy )
		{	return "$me: deletedBy not set, required when deletionDate set";
		}
		$deletionReason = [] unless $deletionReason;
		$deletionReason = [ $deletionReason ] unless ref($deletionReason) eq 'ARRAY';
		unless ( scalar(@$deletionReason) )
		{	return "$me: deletionReason required when deletionDate set";
		}
	} else
	{	#Verify other deletion fields NOT set
		return "$me: ERROR- deletedBy set but deletionDate is not set"
			if $deletedBy;
		return "$me: ERROR- deletionReason set but deletionDate is not set"
			if ( ($deletionReason && ref($deletionReason) ne 'ARRAY' )
			     || ( $deletionReason && ref($deletionReason) eq 'ARRAY' 
					&& scalar(@$deletionReason) ) );
	}
		

	my $minfo = 
	{	DirectoryArchived => $dir,
		DirectoryType => $fstype,
		ArchivalDate => $date,
		ArchivalHost => $host,
		ArchivedBy => $archiver,
		DirectoryOwnerUsername => $owner,
		DirectoryOwnerUID => $uid,
		DirectoryOwnerFullname => $fullname,
		DirectoryOwnerProjects => $projects,
		ExcludePoCs => $exclpocs,
		ReasonExcludePoCs => $exclpoc_reasons,
		ExpirationTstamp => $exptstamp,
		Expiration => $expiration,
		DoNotDelete => $dontdelete,
		ReasonForArchiving => $reason,
		'ls-ld_Output' => $lsld,
		'ls-nd_Output' => $lsnd,
		DeletionDate => $deletionDate,
		DeletedBy => $deletedBy,
		DeletionReason => $deletionReason,
	};
		
	my $options = { force => $force };

	return write_archive_metainfo_file($metadata, $minfo, $options);
}

sub update_archive_metainfo_file(@)
#Update an existing metainfo file with new information.
#The existing metainfo file will be read (an error returned if it cannot
#be) using parse_archive_metainfo_file, then the metadata arguments provided
#to this routine will add to/override the information read, and the file
#then written out.
#Takes key => value pairs for arguments, recognizing the same arguments
#that make_archive_metainfo_file does, with following additions/changes:
#
#	METAINFO_FILE: this is the full path of the existing metainfo file 
#		to read from.. Will default based on directory name, etc.
#	NEW_METAINFO_FILE: this is the full path of the updated metainfo
#		file to WRITE to.  It defaults to METAINFO_FILE (in which
#		case one needs to set force)
#	hpc: Glue::HPCC::Cluster instance, probably not needed as most
#		defaults should already be set in the original file,
#		but used to set additional defaults (e.g. if uid/owner changes).
#	force: If set, it allows for overwriting existing metadata files.
#		It MUST be set if NEW_METAINFO_FILE not given (as otherwise
#		overwriting existing METAINFO_FILE will fail).
#	directory: This option is NOT ALLOWED.  You cannot change the
#		archive directory name.
#	exptstamp, expiration: Same as with make_archive_metainfo_file, but
#		if either is given, will default the other to same value
#		if not also given.
#	uid, owner: Same as with make_archive_metinfo_file, but if either
#		given, will default the other plus fullname if not also given.
#	extrareason: list ref.  If given, will be appended to existing reason.
#	extraprojects: list ref.  If given, will be appended to existing projects.
#	archivalDate, archivedBy, archivalHost: can be given to set archivalDate,
#		archivedBy, archivalHost fields, resp.
#
#Except where noted above, any metadata not provided in arguments to this
#method, the value from the old file will remain unchanged.
#
#This method actually basically just does a parse_archive_metainfo_file, 
#updates the fields provided, and then does a make_archive_metainfo_file
#
#Returns undef on success, true (error message) on error.
{	my %args = @_;
	my $me = delete $args{me} || 
		( __PACKAGE__ . '::update_archive_metainfo_file' );

	my $metadata = delete $args{METAINFO_FILE};
	my $new_metadata = delete $args{NEW_METAINFO_FILE};

	my $force = delete $args{force};
	my $hpc = delete $args{hpc};
	my $logreason = delete $args{logreason};

	#my $dir = delete $args{directory};
	my $reason = delete $args{reason};
	my $uid = delete $args{uid};
	my $owner = delete $args{owner};
	my $fullname = delete $args{fullname};
	my $projects = delete $args{projects};
	my $expiration = delete $args{expiration};
	my $exptstamp = delete $args{exptstamp};
	my $fstype = delete $args{fstype};
	my $dontdelete = delete $args{dontdelete};
	my $exclpocs = delete $args{excludePocs};
	my $exclpoc_reasons = delete $args{exclPocsReason};
	my $deletionDate = delete $args{deletionDate};
	my $deletionReason = delete $args{deletionReason};
	my $deletedBy = delete $args{deletedBy};

	my $extrareason = delete $args{extrareason};
	my $extraprojects = delete $args{extraprojects};

	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}
	croak "$me: Missing required parameter 'METAINFO_FILE' at " unless $metadata;
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$reason = [ $reason ] unless $reason && ref($reason) eq 'ARRAY';
	croak "$me: Missing req parm 'reason', got empty list, at " unless scalar(@$reason);

	unless ( -r $metadata )
	{	return "$me: metadata file to read from not found/readable: $metadata";
	}

	$new_metadata = $metadata unless ( $new_metadata );

	if ( $metadata eq $new_metadata )
	{	#Input/output file the same
		unless ( $force )
		{	return "$me: Unable to use same input/output metadata file w/out setting force";
		}
	}

	#Read old data

	my $metainfo = parse_archive_metainfo_file($metadata);
	unless ( $metainfo && ref($metainfo) eq 'HASH' )
	{	return "$me: Error reading old metadata: $metainfo";
	}

	#update fields as directed
	
	$metainfo->{DirectoryType} = $fstype if ( $fstype );
	$metainfo->{DoNotDelete} = $dontdelete if defined $dontdelete;
	$metainfo->{DeletionDate} = $deletionDate if $deletionDate;
	$metainfo->{DeletedBy} = $deletedBy if $deletedBy;
	if ( $deletionReason )
	{	$deletionReason = [ $deletionReason ] unless ref($deletionReason) eq 'ARRAY';
		$metainfo->{DeletionReason} = $deletionReason;
	}

	if ( $exclpocs )
	{	$exclpocs = [ $exclpocs ] unless ref($exclpocs) eq 'ARRAY';
		$metainfo->{ExcludePoCs} = $exclpocs;
	}
	if ( $exclpoc_reasons )
	{	$exclpoc_reasons = [ $exclpoc_reasons ] unless ref($exclpoc_reasons) eq 'ARRAY';
		$metainfo->{ReasonExcludePoCs} = $exclpoc_reasons;
	}

	if ( $reason )
	{	$reason = [ $reason ] unless ref($reason) eq 'ARRAY';
		$metainfo->{ReasonForArchiving} = $reason;
	}
	if ( $extrareason )
	{	$extrareason = [ $extrareason ] unless ref($extrareason) eq 'ARRAY';
		my $tmp = $metainfo->{ReasonForArchiving};
		push @$tmp, @$extrareason;
		$metainfo->{ReasonForArchiving} = $tmp;
	}

	if ( $projects )
	{	$projects = [ $projects ] unless ref($projects) eq 'ARRAY';
		$metainfo->{DirectoryOwnerProjects} = $projects;
	}
	if ( $extraprojects )
	{	$extraprojects = [ $extraprojects ] unless ref($extraprojects) eq 'ARRAY';
		my $tmp = $metainfo->{DirectoryOwnerProjects};
		push @$tmp, @$extraprojects;
		$metainfo->{DirectoryOwnerProjects} = $tmp;
	}

	$metainfo->{DirectoryOwnerUID} = $uid if defined $uid;
	$metainfo->{DirectoryOwnerUsername} = $owner if $owner;
	$metainfo->{DirectoryOwnerFullname} = $fullname if $fullname;
	my $pdbuser;
	my @pwuser;
	if ( defined $uid )
	{	$pdbuser = Glue::HPCC::Project::Person->person_by_uid(
			$hpc, $uid) if $hpc && ref($hpc);
		@pwuser = getpwuid($uid);
		#Default owner if not explicitly given
		unless ( $owner )
		{	my $tmp;
			$tmp = $pdbuser->username if ( $pdbuser && ref($pdbuser) );
			$tmp = $pwuser[0] unless $tmp;
			$metainfo->{DirectoryOwnerUsername} = $tmp if $tmp;
		}
		#Default fullname if not explicitly given
		unless ( $fullname )
		{	my $tmp;
			$tmp = $pdbuser->fullname if ( $pdbuser && ref($pdbuser) );
			$tmp = $pwuser[6] unless $tmp;
			$metainfo->{DirectoryOwnerFullname} = $tmp if $tmp;
		}
	}
	if ( $owner )
	{	$pdbuser = Glue::HPCC::Project::Person->person_by_username(
			$hpc, $owner) if $hpc && ref($hpc);
		@pwuser = getpwnam($owner);
		#Default uid if not explicitly given
		unless ( defined $uid )
		{	my $tmp = $pdbuser->uid if ( $pdbuser && ref($pdbuser) );
			$tmp = $pwuser[0] unless defined $tmp;
			$metainfo->{DirectoryOwnerUID} = $tmp if defined $tmp;
		}
		unless ( $fullname )
		{	my $tmp;
			$tmp = $pdbuser->fullname if ( $pdbuser && ref($pdbuser) );
			$tmp = $pwuser[6] unless $tmp;
			$metainfo->{DirectoryOwnerFullname} = $tmp if $tmp;
		}
	}

	$metainfo->{Expiration} = $expiration if $expiration;
	$metainfo->{ExpirationTstamp} = $exptstamp if defined $exptstamp;
	if ( defined $exptstamp )
	{	#Default expiration if not given
		unless ( $expiration )
		{ 	my $temptime = strftime(
				$EXPIRATION_TIME_FORMAT, localtime($exptstamp) );
			$metainfo->{Expiration} = $temptime if $temptime;
		}
	}
	if ( $expiration )
	{	#Default exptstamp if not given
		unless ( defined $exptstamp )
		{	my $tempts = parsedate($expiration, 
				WHOLE => 1, DATE_REQUIRED => 1, NO_RELATIVE => 1,
				PREFER_FUTURE => 1, VALIDATE => 1);
			$metainfo->{ExpirationTstamp} = $tempts
				if defined $tempts;
		}
	}

	my $options = { force => $force };
	return write_archive_metainfo_file($new_metadata, $metainfo, $options);
}
	
#-------------------------------------------------------------------------
#	Archival system: methods for archiving, etc
#-------------------------------------------------------------------------
	
sub _generic_archive_dir(@)
#"Archives" a directory
#This basically means take directory $dir, and move it to pending_deletion
#directory in dirname($dir); pending_deletion will be created if needed, and
#have owner set to root and readable only by root.
#Also sticks a metainfo directory with some brief notes.
#
#Takes key => value pairs for arguments, recognizing:
#	directory: directory to archive, REQUIRED
#	hpc: Glue::HPCC::Cluster instance, needed for getting
#		info re fileowner, groups, etc.  REQUIRED
#		(we must log archiving)
#	reason: reason for archiving.  Can be string or list ref of strings
#		This is reason placed in the metainfo file.  REQUIRED
#	logreason: reason for archiving, for insertion into PDB logs.
#		Defaults to concat of @$reason.
#	force: Boolean, are we in force mode.  If set, allows clobbering
#		of metadata, etc.
#	me: method to masquerade as
#	fstype: Type of filesystem.  REQUIRED
#	uid: UID of owner of directory.  If omitted, get from filesystem
#	owner: Owner of directory (username).  If omitted, try to get
#		from uid and passwd file and/or hpc
#	fullname: Fullname of owner.  Will get from owner and hpc if hpc
#		given and user in PDB
#	projects: Projects owner belongs to (who should get notified of
#		archival status, pending deletion).  Will try to get from
#		owner and hpc if hpc given and user in PDB, otherwise []
#		Probably unwise to default, as likely user already removed
#		from all projects.
#	exptstamp: When archive expires and should be deleted, in seconds
#		from 1970.
#	expiration: When archive expires, human readable
#
#owner, fullname, projects, expiration, exptstamp just passed to 
#make_archive_metainfo_file which actually handles any defaulting.
#
#Returns false on success, true (error message) on error
{	my %args = @_;
	my $me = delete $args{me} || ( __PACKAGE__ . "::_generic_archive_dir");

	my $dir = delete $args{directory};
	my $reason = delete $args{reason};
	my $logreason = delete $args{logreason};
	my $force = delete $args{force};
	my $fstype = delete $args{fstype};
	my $uid = delete $args{uid};
	my $hpc = delete $args{hpc};
	my $owner = delete $args{owner};
	my $fullname = delete $args{fullname};
	my $projects = delete $args{projects};
	my $expiration = delete $args{expiration};
	my $exptstamp = delete $args{exptstamp};

	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}
	croak "$me: Missing required parameter 'directory' at " unless $dir;
	croak "$me: Missing required parameter 'hpc' at " unless $hpc;
	croak "$me: Invalid value '$hpc' for 'hpc', expecting Glue::HPCC::Cluster ref at "
		unless $hpc && ref($hpc);
	croak "$me: Missing required parameter 'reason' at " unless $reason;
	$reason = [ $reason ] unless $reason && ref($reason) eq 'ARRAY';
	croak "$me: Missing req parm 'reason'; given an empty list at "
		unless scalar(@$reason);
	$logreason = join "; ", @$reason unless $logreason;

	#Default fstype if can
	unless ( defined $fstype )
	{	$fstype = 'nfs-data' if $dir =~ /data/;
		$fstype = 'lustre-data' if $dir =~ /lustre/;
		$fstype = 'home' if $dir =~ /home/;
	}
	croak "$me: Missing required parameter 'fstype' and unable to default at " unless $fstype;
	$fstype = lc $fstype;
	croak "$me: Invalid value '$fstype' for 'fstype'; expecting one of 'home', 'nfs-data', or 'lustre-data' at " 
		unless $fstype eq 'home' || $fstype eq 'nfs-data' || $fstype eq 'lustre-data';
	
	unless ( -d $dir )
	{	croak "$me: $dir exists but is not a directory at "
			if -e $dir;
		croak "$me: Directory $dir not found at ";
	}

	unless ( defined $uid )
	{	#Get uid of owner from filesystem
		my @stat = stat $dir;
		$uid = $stat[4];
	}

	my $pending = get_pending_dirname($dir);
	my $basename = basename($dir);

	my $newdir = "$pending/$basename";
	if ( -e $newdir )
	{	croak "$me: Refusing to clobber existing dir $newdir at ";
	}
	my $res = make_archive_metainfo_file(
		directory => $dir, 
		reason => $reason, 
		uid => $uid,
		hpc => $hpc,
		owner => $owner,
		fullname => $fullname,
		projects => $projects,
		expiration => $expiration,
		exptstamp => $exptstamp,
		logreason => $logreason,
		fstype => $fstype,
		force => $force);
	if ( $res )
	{	print STDERR "Error making metafile for $dir: $res\n"
			if $VERBOSE;
		return "Error making metadata file: $res";
	}
	
	if ( $DRYRUN )
	{	print STDERR "[DRYRUN] mv $dir $newdir\n";
	} else
	{ 	$res = move $dir, $newdir;
		unless ( $res )
		{	print STDERR "***ERROR*** renaming $dir to $newdir\n";
			print STDERR "Error: $!\n";
			return "Error: $!";
		}
	}
	

	return;
}

sub archive_data_dir(@)
#Archives a data directory.
#See _generic_archive_dir for args accepted (key => value pairs)
{	my %args = @_;
	my $me = __PACKAGE__ . '::archive_data_dir';

	my $dir = delete $args{directory};
	my $reason = delete $args{reason};
	my $force = delete $args{force};
	my $fstype = delete $args{fstype};
	my $hpc = delete $args{hpc};
	my $uid = delete $args{uid};
	my $owner = delete $args{owner};
	my $fullname = delete $args{fullname};
	my $projects = delete $args{projects};
	my $expiration = delete $args{expiration};
	my $exptstamp = delete $args{exptstamp};
	my $logreason = delete $args{logreason};

	croak "$me: Missing required parameter 'dir' at " unless $dir;
	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}

	unless ( $fstype )
	{	$fstype = 'nfs-data' if $dir =~ /data/;
		$fstype = 'lustre-data' if $dir =~ /lustre/;
	}

	return _generic_archive_dir(
		directory => $dir,
		reason => $reason,
		force => $force,
		me => $me,
		fstype => $fstype,
		hpc => $hpc,
		uid => $uid,
		owner => $owner,
		fullname => $fullname,
		projects => $projects,
		expiration => $expiration,
		exptstamp => $exptstamp,
		fstype => $fstype,
		logreason => $logreason,
	);

}

sub archive_home_dir(@)
#Archives a home directory.
#See _generic_archive_dir for args accepted (key => value pairs)
{	my %args = @_;
	my $me = __PACKAGE__ . '::archive_home_dir';

	my $dir = delete $args{directory};
	my $reason = delete $args{reason};
	my $force = delete $args{force};
	my $fstype = delete $args{fstype};
	my $hpc = delete $args{hpc};
	my $uid = delete $args{uid};
	my $owner = delete $args{owner};
	my $fullname = delete $args{fullname};
	my $projects = delete $args{projects};
	my $expiration = delete $args{expiration};
	my $exptstamp = delete $args{exptstamp};
	my $logreason = delete $args{logreason};

	croak "$me: Missing required parameter 'dir' at " unless $dir;
	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}

	#Default fstype
	$fstype = 'home' unless $fstype;

	return _generic_archive_dir(
		directory => $dir,
		reason => $reason,
		force => $force,
		me => $me,
		fstype => $fstype,
		hpc => $hpc,
		uid => $uid,
		owner => $owner,
		fullname => $fullname,
		projects => $projects,
		expiration => $expiration,
		exptstamp => $exptstamp,
		fstype => $fstype,
		logreason => $logreason,
	);

}

sub _generic_delete_archived_dir(@)
#Does the work for delete_archived_home_dir/delete_archived_data_dir
#Basically, deletes the specified archive, adds a couple lines
#to the metadata for the archive (DeletionDate, etc) and moves it
#to the DELETED folder of the pending_deletion directory.
#
#Returns false (undef) on success, and true (error message) on
#non-fatal errors.
#
#Takes key=>value pairs for arguments, recognizing:
#	archive: The archive to delete. REQUIRED. This should be the full,
#		absolute path to one of:
#			the archived directory (in pending_deletion folder)
#			the metadata ( *.readme in pending_deletion folder)
#			the original path to archived directory
#	reason: The reason file is being deleted.  Or actually, a list ref
#		of reasons.  This is used to set DeletionReason in metadata.
#		If the archive has expired (including grace period), this 
#		will be defaulted to "Archive Expired" 
#		If the archive has no non-excluded PoCs remaining, will
#		be defaulted to something to that effect.
#		Otherwise, this is REQUIRED.
#	logreason: Reason for deletion, for logging to PDB logs.  Defaults
#		from reason.
#	hpc: A Glue::HPCC::Cluster instance, REQUIRED (so can log actions)
#		Also needed so we can determine if any non-excluded PoCs left
#	deletedby: who deleted.  Defaults to $ENV{USER}
#	deletiondate: when deleted. Defaults to now
#	gracedays: length of grace period in days. Default is 35.  Must be non-negative
#	force: Boolean.  Force flag must be set is archive has not expired
#		and still has at least one non-excluded PoC.
#	me: Method we should impersonate in error messages, etc.
{	my %args = @_;

	my $me = delete $args{me} || '_generic_delete_archived_dir';
	my $rawarchive = delete $args{archive};
	my $reason = delete $args{reason};
	my $deletedby = delete $args{deletedby};
	my $deletiondate = delete $args{deletiondate};
	my $gracedays = delete $args{gracedays};
	my $force = delete $args{force};
	my $hpc = delete $args{hpc};
	my $logreason = delete $args{logreason};

	croak "$me: Missing required parameter 'archive' at " unless $rawarchive;
	croak "$me: Missing required parameter 'hpc' at " unless $hpc;
	croak "$me: Invalid value '$hpc' for hpc at " unless $hpc && ref($hpc);
	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}

	#Find our archive and metadata
	my $ahash = _find_archived_directory($rawarchive, $me);
	return "ERROR: $ahash" unless $ahash && ref($ahash) eq 'HASH';

	my $archive = $ahash->{archive};
	my $atype = $ahash->{archive_type};
	my $metafile = $ahash->{metafile};
	my $metadata = $ahash->{metadata};
	my $origpath = $ahash->{origpath};

	my $deleted_md_dir = dirname($metafile);
	$deleted_md_dir .= "/$DELETED_DIRNAME";
	my $deleted_mdfile = "${deleted_md_dir}/" . basename($metafile);

	#Default values
	my $now = time;
	$deletedby = $ENV{USER} unless $deletedby;
	$deletiondate = strftime($EXPIRATION_TIME_FORMAT, localtime($now) )
		unless $deletiondate;
	$gracedays = $DEFAULT_GRACEDAYS unless defined $gracedays;
	if ( $gracedays < 0 )
	{	return "$me: GRACEDAYS=$gracedays <0 is not allowed";
	}
	$reason = [] unless defined $reason;
	$reason = [ $reason ] unless ref($reason) eq 'ARRAY';

	#Is DoNotDelete flag set
	my $dontdelete = $metadata->{DoNotDelete};
	if ( $dontdelete )
	{	#Never delete if DoNotDelete set, even in force mode
		return "$me: Refusing to delete with DoNotDelete flag, even in force mode"
			if $force;
		return "$me: Refusing to delete with DoNotDelete flag";
	}

	#Is archive expired
	my $gracesecs = $gracedays * 24*60*60;
	my $expts = $metadata->{ExpirationTstamp};
	my $effexp = $expts + $gracesecs;
	my $exp = $metadata->{Expiration};

	#Get projects for archive
	my $projects = $metadata->{DirectoryOwnerProjects} || [];

	#See if any unexcluded PoCs
	my $exclpocs = $metadata->{ExcludePoCs} || [];
	my %exclpocs = map { $_ => undef } @$exclpocs;
	my $exclpoc_str = join ", ", @$exclpocs;
	my @allpocs = ();
	PROJ: foreach my $tmpproj (@$projects)
	{	my $tmppocs = Glue::HPCC::Project::PersonProject->get_pocs_for_project(
			$hpc, $tmpproj);
		unless ( $tmppocs && ref($tmppocs) eq 'ARRAY' )
		{	carp "$me: Unable to get PoCs for $tmpproj (for archive $archive): $tmppocs,  at";
			#Dont delete archive unless force is set
			return "$me: Refusing to delete archive $archive, issue getting PoCs for $tmpproj"
				. " ($tmppocs)" unless $force;
			next PROJ;
		}
		push @allpocs, @$tmppocs;
	}
	my @unexclpocs = grep { ! exists $exclpocs{$_} } @allpocs;

	push @$reason, "Archive expired [$exp, grace=$gracedays days]"  if ( $effexp < $now );
	push @$reason, "No projects associated with archive directory owner"
		unless scalar(@$projects);
	#Don't give no unexcluded PoCs message if no projects
	push @$reason, "No unexcluded PoCs assoc with archive (PoCs: $exclpoc_str excluded)"
		if scalar(@$projects) && ! scalar(@unexclpocs);

	return "$me: No reason given and unable to default" unless scalar(@$reason);

	if ( scalar(@unexclpocs) && $effexp >= $now )
	{	return "$me: Refusing to delete archive w/out force set: unexpired, unexcl pocs"
			unless $force;
	}

	unless ( $logreason )
	{	$logreason = $reason;
	}
	if ( ref($logreason) eq 'ARRAY' )
	{	$logreason = join "; ", @$logreason;
	}

	#Update metadata fields
	$metadata->{DeletedBy} = $deletedby;
	$metadata->{DeletionDate} = $deletiondate;
	$metadata->{DeletionReason} = $reason;

	
	#Write new metadata file (in DELETED directory)
	my $options = { force => 1 };
	my $err = write_archive_metainfo_file($deleted_mdfile, $metadata);
	if ( $err )
	{	return "$me: Error writing updating metadata in deleted dir: $err";
	}

	#Make sure in dryrun mode if either ShellUtils or FileUtils in dryrun mode
	my $shut_dryrun = Glue::HPCC::Utils::ShellUtils::dryrun;
	Glue::HPCC::Utils::ShellUtils::dryrun($DRYRUN || $shut_dryrun);
	my $shut_verbose = Glue::HPCC::Utils::ShellUtils::verbose;
	Glue::HPCC::Utils::ShellUtils::verbose($VERBOSE || $shut_verbose);

	#Delete the archive
	my $cmd;
	if ( $atype eq 'dir' )
	{	$cmd = "/bin/rm -rf $archive";
	} else
	{	$cmd = "/bin/rm $archive";
	}
	$err = run_system_command($cmd);
	if ( $err )
	{	return "$me: Error running '$cmd': system exited with $err, \$?=$?";
	}

	#Delete the old metadata file
	$cmd = "/bin/rm $metafile";
	$err = run_system_command($cmd);
	if ( $err )
	{	return "$me: Error running '$cmd': system exited with $err, \$?=$?";
	}
	
	#Restore dryrun/verbose on ShellUtils
	Glue::HPCC::Utils::ShellUtils::dryrun($shut_dryrun);
	Glue::HPCC::Utils::ShellUtils::verbose($shut_verbose);
	
	unless ( $DRYRUN )
	{	my $dtype = $metadata->{DirectoryType} || '';
		Glue::HPCC::Project::Log->add_log_entry($hpc,
			reason => $logreason,
			action_code => "delete-$dtype-archive",
			description => "deleting $dtype archive $archive",
			person => $metadata->{DirectoryOwnerUsername},
		);
	}

	return;
}

sub delete_archived_home_dir(@)
#Deletes a previously archived home directory.
#See _generic_delete_archived_dir for args accepted (key => value pairs)
{	my %args = @_;
	my $archive = delete $args{archive};
	my $reason = delete $args{reason};
	my $logreason = delete $args{logreason};
	my $hpc = delete $args{hpc};
	my $deletedBy = delete $args{deletedby};
	my $deletionDate = delete $args{deletiondate};
	my $gracedays = delete $args{gracedays};
	my $force = delete $args{force};

	my $me = "delete_archived_home_dir";

	croak "$me: Missing required parameter 'archive' at " unless $archive;
	croak "$me: Missing required parameter 'hpc' at " unless $hpc;
	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}

	return _generic_delete_archived_dir(
		archive => $archive,
		reason => $reason,
		logreason => $logreason,
		force => $force,
		me => $me,
		hpc => $hpc,
		deletedby => $deletedBy,
		deletiondate => $deletionDate,
		gracedays => $gracedays,
	);
}

sub delete_archived_data_dir(@)
#Deletes a previously archived data directory.
#See _generic_delete_archived_dir for args accepted (key => value pairs)
{	my %args = @_;
	my $archive = delete $args{archive};
	my $reason = delete $args{reason};
	my $logreason = delete $args{logreason};
	my $hpc = delete $args{hpc};
	my $deletedBy = delete $args{deletedby};
	my $deletionDate = delete $args{deletiondate};
	my $gracedays = delete $args{gracedays};
	my $force = delete $args{force};

	my $me = "delete_archived_data_dir";

	croak "$me: Missing required parameter 'archive' at " unless $archive;
	croak "$me: Missing required parameter 'hpc' at " unless $hpc;
	if ( %args )
	{	my @temp = keys %args;
		my $temp = join ", ", @temp;
		croak "$me: Unrecognized arguments [ $temp ], aborting at ";
	}

	return _generic_delete_archived_dir(
		archive => $archive,
		reason => $reason,
		logreason => $logreason,
		force => $force,
		me => $me,
		hpc => $hpc,
		deletedby => $deletedBy,
		deletiondate => $deletionDate,
		gracedays => $gracedays,
	);
}

1;


__END__

=head1 NAME

Glue::HPCC::Utils::FileUtils- Utils for dealing with files, directories, etc

=head1 SYNOPSIS

  use Glue::HPCC::Utils::FileUtils qw( files_in_dir_owned_by_uid );

  my $res = files_in_dir_owned_by_uid($dir, $uid);
  ...


=head1 DESCRIPTION

Assorted utilities for dealing with files and/or directories

=head2 Debugging type options

The class level accessor/mutator B<verbose> gets/sets the verbosity level
for the package.

The class level accessor/mutator B<dryrun> gets/sets the dryrun flag for
the package.  In dryrun mode, methods which normally make changes will
instead just print what they would have done.  (Methods which just run
queries without making changes run as usual).

To use as mutators, the above take a defined boolean argument.  Normal
perl truth semantics apply, except that undef is NOT allowed (passing an
undef argument to the above routines cause them to be interpretted as
accessors, not mutators).

=head2 Basic Lookup/disk usage Utilities 

The routine B<files_in_dir_owned_by_uid($dir, $uid)>
returns a list ref of the names of all files directly underneath directory
$dir which are owned by uid $uid.  $uid MUST be numeric uid, not username.

The routine B<du_directory($root, [$options])>
effectively does a du of directory B<$root>, returning the total size
(in bytes) of files in B<$root> and its subdirectories.  If B<$options> is
given, it should be a hash reference which filters which files are included
in the du sum.  The following options are supported:

=over 4

=item B<last_modified_before>: only files whose last modification timestamp is before this value are included.

=back

The routine B<du_subdirs_of_directory([key=>value pairs])>
returns disk usage stats for all the toplevel subdirectories of
a directory.  It takes key=>value pairs for arguments, recognizing:

=over 4

=item B<rootdir>: REQUIRED, full path to root directory.

=item B<exclude>: list ref of names of subdirs of rootdir to exclude. Default is none.

=item B<options>: A hash ref giving filtering options for the du, as described under B<du_directory>.  Default is empty hash/no options/all files processed.

=back

The return value is a list ref of hash refs with keys:

=over 4

=item B<fullpath>: full path to the subdirectory

=item B<dirname>: name of the subdirectory (e.g. path relative to rootdir)

=item B<sizeBytes>: usage of that subdirectory, in bytes

=item B<uid>: uid of owner of the directory

=item B<gid>: gid of directory

=back

The routine B<du_subdirs_of_dir_by_uid([key=>value pairs])>
does a B<du_subdirs_of_directory>, then consolidates
the results by uid.  It takes the same key=>value pairs
for arguments as B<du_subdirs_of_directory>.  It returns
a hash ref, keyed on uid, with the value being a hash
ref with keys B<uid>, B<gid>, B<sizeBytes>, B<fullpaths>,
and B<dirnames>, corresponding to the return fields of
B<du_subdirs_of_directory>.  B<uid> and B<gid> are 
the UID and GID of the owner of the directores (if different 
directories have different GIDs, then the GID of the first dir 
is returned).  B<sizeBytes> is the sum of the sizes of each
directory, and B<fullpaths> and B<dirnames> are list refs
with the individual B<fullapth> and B<dirname> values.

The routine B<df_disk_containing_file($file)>
is a wrapper around Filesys::Df.  Returns listref (totalsize,
pctfree, pctused) for disk containing $fire.  Totalsize
is in bytes.

=head2 Utilities related to archiving of directories

This package also provides a set of routines for the standardized of
"archiving" data as part of the user decommissioning process.  We do not
really "archive" data, but rename into an access restricted "pending_deletion"
directory.  This will cause any jobs/processes/etc using the files to fail,
with the thinking it is better they fail early, while the data is easily 
recoverable, then after data deleted.  The archival process also writes
some "metadata" associated with the archival event, which is used to track
when the data is slated for deletion, why it was archived, and especially
to allow for emails to be sent to the PoCs of allocations the user had belonged
to so they are aware of the pending deletion.

So e.g., when archiving /lustre/bigboote, the directory /lustre/bigboote
will get moved to /lustre/pending_deletion/bigboote (which is on same 
filesystem, so fast operation, but does NOT reduce disk usage) and metadata
regarding the action is written to /lustre/bigboote.readme.  Cron jobs
(like /cell_root/software/sbin/cron_notify_pocs_of_pending_data_deletions)
will use the metadata to send emails to PoCs about pending data expirations,
and to system staff re expired data (and/or delete the data).  If/when 
something like B<delete_archived_directory> is invoked, the metadata has
a few extra fields appended and is moved to /lustre/pending_deletion/DELETED
and the directory is permanently deleted.

The metadata for archived files includes a number of fields describing
how/why/when was archived (including reason(s) user access was revoked).
It includes an expiration date (when archive is able to be deleted, used
in warnings sent to PoCs), and a list of project names to which the user
was attached just before access was revoked.  There also is a list of PoC
usernames listing PoCs who have OKed the deletion of the data (along with
a field for explaining such).  Archives should not be deleted until:

=over 4

=item 1) The expiration date has past (and normally, past by at least GRACEDAYS days)

=item B<OR>

=item 2) All PoCs for all projects the user was attached to have agreed the data can be deleted.

=back


The metadata info files use a simple Field: Value text format, so they
can be editted manually. 

We start with the higher level routines (which unless otherwise noted
will make calls as needed to generate the metadata files).

=head3 Archiving utilities

The method B<archive_home_dir(key value pairs)> archives a home directory.  
The directory is moved to a B<pending_deletion> 
directory and a metainfo file (which is what most of the arguments are for)
is created with details about what/why/when/etc.
Arguments should be given as key value pairs, with the following recognized:

=over 4

=item B<directory>: path of directory to archive, REQUIRED

=item B<reason>: list ref of reasons for archiving, REQUIRED

=item B<fstype>:  type of filesystem, Unused?

=item B<force>:  Boolean, whether to clobber files, etc. Default is false

=item B<hpc>: Glue::HPCC::Cluster instance, optional, will use to help with defaulting other values if given 

=item B<uid>: UID owning directory, will default from filesystem 

=item B<owner>: username owning dir, will default from PDB if $hpc given and/or password file 

=item B<fullname> fullname of owner, defaults from PDB or password file 

=item B<projects> list ref of project names owner belongs to, defaults from PDB or to [].  Generally, do NOT want to default because typically user already deleted from all projects in PDB by time this is called.

=item B<expiration>:  when archive expires, human readable, defaults to 6 months from now.  

=back

The method B<archive_data_dir(key value pairs)> is basically the same as
B<archive_home_dir>, except with customizations for data directories.  The
allowed options are same as for B<archive_home_dir>.  Indeed, currently
there is no difference between the two methods, but the two methods are 
maintained in case differences are added in the future.

The methods B<delete_archived_home_dir(key value pairs)> and 
B<delete_archived_data_dir(key value pairs)> will delete the specified archive 
directory.  On success, they return false.  On non-fatal errors, they return
true (an error string).  Recognized arguments are:

=over 4

=item B<archive>:

The full path to the archive directory to delete.  If it does not contain
a B<pending_deletion> path component, it will be assumed that this is the
original path which was archived, and will attempt to find the archived path
from that.  There MUST be a corresponding metadata file or the routine will
fail with an error.

=item B<hpc>:

An instance of Glue::HPCC::Cluster.  REQUIRED.  Needed so that we can
log a message in PDB.  Also needed to determine the list of PoCs for the
listed projects (so can see if any not on excluded list).

=item B<reason>:

This is a list ref of strings, one per line, giving the reasons for 
deleting the directory.  It is optional if the expiration (plus grace period)
for the archive has already elapsed, in which case it defaults to something
like "Archive Expired".  It is also optional if there are no DirectoryOwnerProjects,
or if the B<all> of the PoCs for B<all> such projects are on the ExcludePoCs list,
in which cases an appropriate default is generated.  It is required in all
other cases.

=item B<logreason>:

A string or list ref of strings giving reason(s) for deletion, for use
in the PDB logs.  Defaults based on B<reason>.

=item B<deletedby>:

Who did the deletion.  Default is from $ENV{USER}.

=item B<deletiondate>:

Human readable date file deleted.  Default is now.

=item B<gracedays>:

The number of days to be used as a "grace" period.  Default is 30.
If the expiration time + the grace period is in the future, the
deletion WILL NOT OCCUR unless force mode is set.

=item B<force>:

Boolean.  Normally, an archive will not be deleted unless either
the expiration date (plus grace period) has passed, or there are
no projects listed, or all the PoCs for all the projects listed are
on the excluded list.  If the force flag is set, the archive can be 
deleted even if the above criteria do not hold.  Default is unset.

=back

=head3 Metadata utilities

The following methods are provided for handling the metadata files containing
information about archived directories.

The method B<parse_archive_metainfo_file($metainfo,[$verify])> 
parses the metainfo files generated by the B<archive_home_dir> and 
B<archive_data_dir> methods.  $metainfo is used to specify the metadata to
be processed, and can be any of the following:

=over 4

=item *) a scalar string giving path to metainfo file

=item *) a scalar reference containing metainfo content as a single string 

=item *) an array reference containing the  metainfo content line by line 

=item *) a filehandle which is read for the metainfo

=back

The return value is a hash ref with metainfo on success, or non-ref error 
string on non-fatal errors.  If $verify is true, warnings will be generated 
to stderr for missing fields and other irregularities; this can be made even 
more agressive by setting to something greater than 1, in which case many
issues become non-fatal errors.

The method B<write_archive_metainfo_file($file, $metadata, $options)>
essentially does the reverse of B<parse_archive_metainfo_file>.  It
takes a filename to write to, a metadata hash (as returned by 
B<parse_archive_metainfo_file>) and an optional options hash, and writes
the metadata hash into a properly formated metadata info file.  Options
hash currently has only one allowed option, 'force', which if sets allows
the output file to clobber an existing file.  It returns false (undef) on
success, or true (an error message) on non-fatal error.

The method B<make_archive_metainfo_file> is used to create metadata info
files.  On success, it returns
false, on non-fatal errors it returns true (an error message).  
It takes key => value pairs for arguments.  Arguments fall into two classes,
options for the method, and metadata.  The following options controlling
the behavior of the method are recognized:

=over 4

=item B<METAINFO_FILE> 

if given, this is the full path to the metainfo file to write.  Normally, it
should not be given, and it will default based on the name of the archived
directory, etc.

=item B<hpc>

A B<Glue::HPCC::Cluster> instance, which is used for defaulting any metadata
parameters that were omitted.  Not required, but strongly recommended, as w/out
it a lot of metadata info will be missing if not given explicitly.

=item B<logreason>

The reason the directory is being archived, for inclusion in the PDB log
table.  If omitted, it is defaulted from B<reason> metadata option
(concatenating the list of strings if a list).

=item B<force>

Boolean.  If set, it allows for the clobbering of existing metadata files, etc.
By default is false, and if the metadata file already exists, the method will
return an error message.

=item B<me>

If given, the method will masquerade as another method (i.e. warnings/errors
will list this method name).

=back

The following metadata parameters are recognized as arguments:

=over 4

=item B<directory>

The path to the directory to archive, REQUIRED.

=item B<reason>

The reason the directory is being archived.  This can be a string or a 
list ref of strings.  REQUIRED.  This gets included in the metadata file.

=item B<uid>

UID of the owner of the directory.  If omitted, it will default to the uid of 
the owner of the directory path, as reported by the filesystem.

=item B<owner>

The username of the owner of the directory.  If omitted, will try to default
from B<uid> and either B<hpc>/PDB or the password file.

=item B<fullname>

The full (human readable) name of the owner of the directory.  If omitted,
will try to default from B<uid> and either B<hpc>/PDB or the password file.

=item B<projects> 

A list ref of the names of projects the owner of the directory belongs to.
This is used to determine which PoCs should be notified of the pending deletion
of the archive.  Typically we want the last projects the user belonged to just
before was decommissioned.  Although this will try to default from the B<owner>
and B<hpc>/PDB, you normally DO NOT want to let this default, as typically
by the time this routine is called, the user is no longer a member of any
projects according to the PDB.  SO DO NOT DEFAULT THIS.

=item B<expiration>

When the archive expires and should be deleted (or at least when warn people
it can be deleted).  Human readable format.

=item B<exptstamp>

As B<expiration>, but as seconds since 1970.

=item B<fstype>

The type of data.  One of 'home', 'nfs-data', 'lustre-data'.

=back

If only one of expiration and exptstamp given, we will try to default the 
missing one from the one given.  If both given, they must agree or else a 
warning will be generated.  If both given, exptstamp is definitive, and 
expiration will be set from that.  If neither given, defaults to 6 months 
from now.

The method B<update_archive_metainfo_file> is used to edit the metadata in
an existing metadata file.  On success, it returns
false, on non-fatal errors it returns true (an error message).  
It takes key => value pairs for arguments.  Arguments fall into two classes,
options for the method, and metadata.  

The metadata arguments are basically the
same as for B<make_archive_metainfo_file>; basically the routine reads the
existing metadata file, updates the resulting data in memory with the metadata
arguments given, and then writes the metadata file back out.  Any metadata
values not given as arguments are left unchanged (from what was in the 
original metadata file), although might get defaulted if no value in original
metadata file either.  However, we make the following exceptions:

=over 4

=item 1) If either (but not both) of B<exptstamp> or B<expiration> is given as a metadata argument, the other will be set from what was given.

=item 2) If either of B<uid> or B<owner> is given, then the other and B<fullname> will be defaulted from that if not also given.

=item 3) If B<extrareason> is given (either a string or list ref), it will be appended to the existing B<reason> list.

=item 4) If B<extraprojects> is given (either string or list ref), it will be appended to the existing B<projects> list.

=item 5) The field B<directory> is not allowed to be changed.

=back

The following options controlling the behavior of the method are recognized:
=over 4

=item B<METAINFO_FILE> 

if given, this is the full path to the existing metainfo file (to read from).
If omitted, it will default based on the name of the archived directory, etc.
If this file does not exist, 

=item B<NEW_METAINFO_FILE>

if given, this is the full path to which the updated metainfo should be 
written out to.  It defaults to B<METAINFO_FILE>.  Note that if this parameter
is not set, one MUST set B<force> (as we will otherwise refuse to write back
to the same file we read from).

=item B<hpc>

A B<Glue::HPCC::Cluster> instance, which is used for defaulting any metadata
parameters that were omitted.  In theory, this should not be needed, as any
defaulting should already have been done.

=item B<logreason>

The reason the directory is being archived, for inclusion in the PDB log
table.  If omitted, it is defaulted from B<reason> metadata option
(concatenating the list of strings if a list).

=item B<force>

Boolean.  If set, it allows for the clobbering of existing metadata files, etc.
By default is false, and if the metadata file already exists, the method will
return an error message.

=item B<me>

If given, the method will masquerade as another method (i.e. warnings/errors
will list this method name).

=back


