#!/usr/local/bin/perl
#
#Unix group processing utilities

package Glue::HPCC::Utils::MailUtils;
use strict;
use warnings;

use Carp qw(carp croak);
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw( 
	send_email
	send_template_email
);

use Glue::HPCC::Utils::TemplateUtils qw(expand_template);

my $DRYRUN;
my $VERBOSE;
my $SENDMAIL = '/usr/lib/sendmail';

sub dryrun(;$)
{	my $new = shift;
	my $me = __PACKAGE__ . '::dryrun';
	if ( defined $new )
	{	#CAtch if someone called us in OO fashion, rather than
		#just setting dryrun to TRUE
		$new = ref($new) if ref($new);
		if ( $new eq __PACKAGE__ )
		{	carp "$me: Dropping 1st arg because looks like called in OO fashion at ";
			$new = shift;
		}
	}
	if (defined $new)
	{	$DRYRUN=$new?1:0;
		#carp "Set DRYRUN to $DRYRUN at ";
	}
	return $DRYRUN;
}

sub verbose(;$)
{	my $new = shift;
	my $me = __PACKAGE__ . '::verbose';
	if ( defined $new )
	{	#CAtch if someone called us in OO fashion, rather than
		#just setting dryrun to TRUE
		$new = ref($new) if ref($new);
		if ( $new eq __PACKAGE__ )
		{	carp "$me: Dropping 1st arg because looks like called in OO fashion at ";
			$new = shift;
		}
	}
	if (defined $new)
	{	$VERBOSE=$new?1:0;
	}
	return $VERBOSE;
}

sub split_header_and_body($)
#Given a list ref of strings representing a header plus message body of
#a mail message, returns two list refs representing header and body.
#Can also take a single string;
{	my $all = shift;
	my $me = __PACKAGE__ . '::split_header_and_body';

	croak "$me: Must supply text string at " unless $all;
	unless ( ref($all) )
	{	#Got a scalar string, split into an array
		$all = [ split /\n/, $all ];
	}
	croak "$me: Invalid argument all='$all', expecting list ref at"
		unless $all && ref($all) eq 'ARRAY';

	my $header = [];
	my $body = [ @$all ];

	my ($line);
	$line = shift @$body;
	$line =~ s/^\s*// if $line; 
	$line =~ s/\s*$// if $line;

	while ( $line )
	{	push @$header, $line;
		$line = shift @$body;
		$line =~ s/^\s*// if $line; 
		$line =~ s/\s*$// if $line;
	}

	return ($header, $body);
}

sub send_email($$)
#Takes header and message variables and sends email.
#The header MUST have a To: line, as we do not specify a destination on cmdline
#
#Header can be a hash ref or array ref, or scalar.  If scalar, it is treated
#the same as a single element array ref.  Arrays are just printed out, one
#element per line. Hash ref get printed out with key as the header field and
#value as header value.
#
#Message can be array ref or scalar.  Array ref is treated as lines of text,
#and scalar simply as straightforward message.
#
#If $dryrun set, don't actually send out mail
#
#Returns false (undef) on success, true (error message) on non-fatal errors.
{	my $header = shift;
	my $message = shift;
	my $me = __PACKAGE__ . '::send_email';

	croak "$me: You must supply a header at " unless $header;

	my $text = "";
	my $tmp;
	$header = [ $header ] unless ref($header);
	my $has_to = 0;
	if ( ref($header) eq 'ARRAY' )
	{	my @tmp = grep /^To: /i, @$header;
		$has_to = 1 if @tmp;
		#We want to handle whether they include CRs or not
		#So chomp first , then add
		chomp @$header;
		$text .= join "\n", @$header;
	} elsif ( ref($header) eq 'HASH' )
	{	my @tmp = grep /^To$/i, keys %$header;
		$has_to = 1 if @tmp;
		foreach my $key (keys %$header)
		{	$tmp = $header->{$key};
			chomp $tmp;
			$text .= "$key: $tmp\n";
		}
	} else
	{	croak "$me: Illegal value '$header' for header at ";
	}

	$text .= "\n\n"; #First CR is for last header, next for blank line
	if ( ref($message) eq 'ARRAY' )
	{	chomp @$message;
		$message = join "\n", @$message;
	}
	$text .= $message;

	if ( dryrun() )
	{	print STDERR "[DRYRUN] Would be mailing:\n-------------------------\n";
		print STDERR "$text\n";
		print STDERR "-------------------------------------\n";
		return;
	}

	if ( verbose() )
	{	print STDERR "[VERBOSE] $SENDMAIL -oi -t\n";
	}
	open(MAIL, "| $SENDMAIL -oi -t") or 
		croak "$me: Unable to open pipe to sendmail, $?,  at ";

	print MAIL "$text\n";
	close(MAIL) or return "$me: Error closing pipe to sendmail: $?";
	return;
	
}

sub send_template_email($$)
#Takes a template (as array ref of strings, one per line) and a 
#macro hash (keys are macron names, values are expansion),
#and sends out mail.
#
#Template MUST include the full mail headers, including the To: line.
#
#If $dryrun flag is set, only pretend to do it.
#
#Returns false (undef) on success, true (error message) on non-fatal errors.
{	my $template = shift;
	my $macros = shift;
	my $me = __PACKAGE__ . '::send_template_mail';

	my $text = expand_template($template,$macros);
	return "$me: Error, expand_template returned false" unless $text;

	my ($header, $body) = split_header_and_body($text);
	my $res = send_email($header, $body);
	return "$me: Error from send_email: $res" if $res;
	return;
}


1;
__END__

=head1 NAME

Glue::HPCC::Utils::UnixGroup- Utils for Unix Groups

=head1 SYNOPSIS

  use Glue::HPCC::Utils::UnixGroup qw( group_exists );

  my $res = group_exists("dt-fooproject");
  ...


=head1 DESCRIPTION

Assorted utilities for dealing with Unix groups.
Most (all?) commands go through Glue::HPCC:Utils::ShellUtils and so honor
its DRYRUN setting.

=over 4

=item B<unix_group_exists($name)>

Returns true if group named B<$name> exists in Hesiod, false otherwise.

=item B<add_user_to_unix_group( user=>USER, group=>GROUP, [ runas=>RUNAS], [force=>1])>

Adds USER to unix group GROUP.  If force set, adds --force flag to the grp
command.  Runs grp command as RUNAS if provided.

=item B<set_users_primary_unix_group( user=>USER, group=>GROUP, [ runas=>RUNAS], [force=>1])>

Sets the primary unix group for  USER to  GROUP.  If force set, adds 
--force flag to the grp
command.  Runs grp command as RUNAS if provided.

=back






