#!/usr/local/bin/perl
#
#Net group processing utilities

package Glue::HPCC::Utils::NetGroup;
use strict;
use warnings;

use Carp qw(carp croak);
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw( 
	add_user_to_netgroup
	ngbyuser
	user_in_netgroup
	remove_user_from_netgroup
);

use Glue::HPCC::Utils::ShellUtils qw(run_system_command);
use Net::Hesiod;

my $NGCMD = "/usr/local/scripts/ng";
my $hesiod = new Net::Hesiod;

sub ngbyuser($)
#Returns list of netgroups user belongs to. 
{	my $user = shift;
	my @results = $hesiod->query($user, "ngbyuser");
	return @results if wantarray;
	my $tmp = join ", ", @results;
	return $tmp;
}

sub user_in_netgroup($$)
#Returns true if specified user is in specified netgroup
{	my $user = shift;
	my $ngrp = shift;

	my @ngs = ngbyuser($user);
	my @temp = grep /^$ngrp$/, @ngs;
	return scalar(@temp);
}

sub add_user_to_netgroup(@)
#Takes key=>value pairs for arguments.  Recognized arguments are:
#	user: User to add to $group.  REQUIRED.
#	group: Group to to add $user to.  REQUIRED.
#	runas: User to runas
#	force: Whether to use --force flag to grp
#
#Honors dryrun (via run_system_command)
{	my %args = @_;

	my $user = delete $args{user};
	my $group = delete $args{group};
	my $runas = delete $args{runas};
	my $force = delete $args{force};

	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unexpected args [ $tmp ], aborting at ";
	}

	unless ( $user )
	{	croak "Missing required argument 'user' at ";
	}
	unless ( $group )
	{	croak "Missing required argument 'group' at ";
	}


	my $force_flag = '';
	$force_flag = '--force' if $force;
	my $cmd = "$NGCMD $force_flag add -user $user -group $group";
	run_system_command($cmd, $runas);
}

sub remove_user_from_netgroup(@)
#Takes key=>value pairs for arguments.  Recognized arguments are:
#	user: User to add to $group.  REQUIRED.
#	group: Group to to add $user to.  REQUIRED.
#	runas: User to runas
#	force: Whether to use --force flag to grp
#
#Honors dryrun (via run_system_command)
{	my %args = @_;

	my $user = delete $args{user};
	my $group = delete $args{group};
	my $runas = delete $args{runas};
	my $force = delete $args{force};

	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unexpected args [ $tmp ], aborting at ";
	}

	unless ( $user )
	{	croak "Missing required argument 'user' at ";
	}
	unless ( $group )
	{	croak "Missing required argument 'group' at ";
	}


	my $force_flag = '';
	$force_flag = '--force' if $force;
	my $cmd = "$NGCMD $force_flag remove -user $user -group $group";
	run_system_command($cmd, $runas);
}

1;
__END__

=head1 NAME

Glue::HPCC::Utils::NetGroup- Utils for NetGroups

=head1 SYNOPSIS

  use Glue::HPCC::Utils::NetGroup qw( user_in_netgroup add_user_to_netgroup );

  unless ( user_in_netgroup($user, $netgroup) )
  {	add_user_to_netgroup(user=>$user, netgroup=>$netgroup);
  }
  ...


=head1 DESCRIPTION

Assorted utilities for dealing with NetGroups.
Most (all?) commands go through Glue::HPCC:Utils::ShellUtils and so honor
its DRYRUN setting.

=over 4

=item B<nbbyuser($user)>

Returns list of all netgroups user B<$user> belongs.  Returns as list in
list context, or comma separated string in scalar context.

=item B<user_in_netgroup( $user, $group)

Returns true if B<$user> is in netgroup B<$group>, and false otherwise.

=item B<add_user_to_netgroup( user=>USER, group=>GROUP, [ runas=>RUNAS], [force=>1])>

Adds user USER to  netgroup GROUP.  If force set, adds 
--force flag to the ng command.  Runs grp command as RUNAS if provided.

=back

