#!/usr/local/bin/perl
#
#Utilities for dealing with shell commands

package Glue::HPCC::Utils::ShellUtils;
use strict;
use warnings;

use Carp qw(carp croak);
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw( 
	dryrun 
	verbose
	run_system_command
	run_backticked_command
	run_backticked_command_always
);

our $DRYRUN;
our $VERBOSE;
my $SU = "/bin/su";

sub dryrun(;$)
#Gets/sets the package dryrun flag
{	my $new = shift;
	if ( defined $new )
	{	$DRYRUN = $new;
	}
	return $DRYRUN;
}

sub verbose(;$)
#Gets/sets the package verbose flag
{	my $new = shift;
	if ( defined $new )
	{	$VERBOSE = $new;
	}
	return $VERBOSE;
}

sub _suify_cmd($$)
#Given cmd and RUNAS user, return the su-ified cmd
{	my $origcmd = shift;
	my $runas = shift;

	my $newcmd = $origcmd;
	if ( $runas )
	{	if ( $< )
		{	#We are not root!
			carp "WARNING: not running as root, ignoring runas at ";
		} else
		{	$newcmd = "$SU -c '$origcmd' $runas ";
		}
	}
	return $newcmd;
}

sub run_system_command($;$)
#Runs a system command, honoring $DRYRUN, and optionally as user $runas
#
#Returns the return value of system, which is the exit status of
#the command as per wait syscall (might want to shift 8 bits to right)
#BAsically, 0 if all went well, non-zero otherwise
{	my $cmd = shift;
	my $runas = shift;

	my $realcmd = _suify_cmd($cmd, $runas);

	if ( dryrun )
	{	print STDERR "[DRYRUN] $realcmd\n";
		return;
	}
	if ( verbose )
	{	print STDERR "[VERBOSE] $realcmd\n";
	}

	system($cmd);
}

sub _run_backticked_command_guts($)
#Runs a command in backticks style, returning the output
{	my $cmd = shift;

	my $PIPE;
	open($PIPE, "$cmd 2>&1 |") or die "Error opening pipe to $cmd: $! at ";

	my @results = <$PIPE>;
	my $results = join '', @results;
	return $results;
}

sub run_backticked_command($;$)
#Runs a shell command, honoring $DRYRUN, and optionally as user $runas
{	my $cmd = shift;
	my $runas = shift;

	my $realcmd = _suify_cmd($cmd, $runas);

	if ( dryrun )
	{	print STDERR "[DRYRUN] $realcmd\n";
		return;
	}
	if ( verbose )
	{	print STDERR "[VERBOSE] $realcmd\n";
	}

	_run_backticked_command_guts($cmd);
}

sub run_backticked_command_always($;$)
#Runs a shell command, and optionally as user $runas
#This version does NOT honor dryrun, it runs always.
#Should only be used on commands which collect information but do NOT
#change anything.
{	my $cmd = shift;
	my $runas = shift;

	my $realcmd = _suify_cmd($cmd, $runas);
	if ( verbose )
	{	print STDERR "[VERBOSE] $realcmd\n";
	}
	_run_backticked_command_guts($cmd);
}

1;
__END__

=head1 NAME

Glue::HPCC::Utils::ShellUtils- Utils for dealing with shell commands

=head1 SYNOPSIS

  use Glue::HPCC::Utils::ShellUtils qw( dryrun run_shell_command);

  dryrun(1); #Set us to DRYRUN mode
  run_shell_command("grp add user foo");
  ...


=head1 DESCRIPTION

Assorted utilities for dealing with shell commands

=over 4

=item B<dryrun([VAL])>

Gets/sets the dryrun setting for this module.

=item B<run_system_command(cmd, [user])>

Runs system command B<$cmd>, running as B<$user> if specified.  Honors
DRYRUN flag.

=item B<run_backticked_command(cmd, [user])>

Runs command B<$cmd> in a PIPE, returning combined STDOUT, STDERR as 
a string.  Runs as B<$user> if specified.  Honers DRYRUN flag.

=item B<run_backticked_command_always(cmd, [user])>

As B<run_backticked_command>, but does B<NOT> honor the DRYRUN flag.  Instead,
it ALWAYS runs.  Should only be used for commands which return information
without changing state of the system.

=back






