#Routines to do a "soft" quota check

package Glue::HPCC::Utils::SoftQuotaCheck;

use strict;
use warnings;

use Glue::HPCC::Cluster;
use Glue::Misc::Units qw(scale_size);
use Glue::HPCC::Utils::FileUtils qw(
		du_directory
		du_subdirs_of_dir_by_uid
		df_disk_containing_file
	);
use Glue::HPCC::Utils::MailUtils qw(send_template_email);
use Glue::HPCC::Project::Person;

use POSIX qw(floor strftime);
use Filesys::Df;
use Carp qw(croak carp);

#For debug mode
use Data::Dumper;

#----------------------------------------------------------
#		Globals
#----------------------------------------------------------


#One day in seconds
my $ONEDAY = 24*60*60;

#----------------------------------------------------------
#		Accessors
#----------------------------------------------------------

sub DEBUG($;$)
{	my $self = shift;
	my $new = shift;

	if ( defined $new )
	{	$new = lc $new;
		$new = 0 if $new eq 'no' || $new eq 'false' || $new eq 'off';
		$new = $new?1:0;
		$self->{DEBUG} = $new;
	}
	return $self->{DEBUG};
}

sub VERBOSE($;$)
{	my $self = shift;
	my $new = shift;

	if ( defined $new )
	{	$new = lc $new;
		$new = 0 if $new eq 'no' || $new eq 'false' || $new eq 'off';
		$new = $new?1:0;
		$self->{VERBOSE} = $new;
	}
	return $self->{VERBOSE};
}

sub hpc($;$)
{	my $self = shift;
	my $new = shift;
	my $me = __PACKAGE__ . '::hpc';

	if ( defined $new )
	{	croak "$me: Invalid value '$new', want Glue::HPCC::Cluster " .
			"instance, at " unless ref($new);
		$self->{hpc} = $new;
	}
	return $self->{hpc};
}
#----------------------------------------------------------
#		Constructors, etc
#----------------------------------------------------------

sub new($@)
{	my $class = shift;
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $self = shift;
	my %args = @_;

	my $me = __PACKAGE__ . '::new (parse_args)';
	my $tmp;

	if ( exists $args{hpc} )
	{	$tmp = delete $args{hpc};
		$self->hpc($tmp);
	}

	if ( exists $args{DEBUG} )
	{	$tmp = delete $args{DEBUG};
		$self->DEBUG($tmp);
	}

	if ( exists $args{VERBOSE} )
	{	$tmp = delete $args{VERBOSE};
		$self->VERBOSE($tmp);
	}

	if ( %args )
	{	my @tmp = keys %args;
		$tmp = join ", ", @tmp;
		croak "$me: Unrecognized arguments [ $tmp ] at ";
	}
}

sub _set_defaults($)
{	my $self = shift;
	my $me = __PACKAGE__ . '::new (set defaults)';

}

sub _init($)
{	my $self = shift;
	my $me = __PACKAGE__ . '::new (init)';

	my $tmp;

	$tmp = $self->hpc;
	croak "$me: Missing required parameter 'hpc' at " unless $tmp;
}

#----------------------------------------------------------
#		Debugging stuff
#----------------------------------------------------------

sub verbose($@)
#Print text, but only if in verbose mode
{	my $self = shift;
	my @args = @_;

	return unless $self->VERBOSE;

	my $tmp = join '', @args;
	chomp $tmp;
	print STDERR "[VERBOSE] $tmp\n";
}

sub debug($@)
#Print text, but only if in debug mode
{	my $self = shift;
	my @args = @_;

	return unless $self->DEBUG;

	my $tmp = join '', @args;
	chomp $tmp;
	print STDERR "[DEBUG] $tmp\n";
}

#----------------------------------------------------------
#		Time utilities
#----------------------------------------------------------

sub days_between($$$)
#Given two timestamps (seconds from 1970), compute number of days between
#them.  Rounds to nearest day.
{	my $class = shift;
	my $later = shift;
	my $earlier = shift;
	my $diff = $later - $earlier;
	$diff = 0 if $diff < 0;

	my $days = floor($diff/$ONEDAY + 0.5 );
	return $days;
}

my $date_format = "%a, %e %b %Y";
sub timestamp_to_date($$)
#Given a timestamp (seconds from 1970) convert to human readable format
{	my $class = shift;
	my $ts = shift;
	my $date = strftime($date_format, localtime($ts) );
	return $date;
}


#----------------------------------------------------------
#		Ordinal  utilities
#----------------------------------------------------------

my @ordinals =
(	[ '1st', 'first', ],
	[ '2nd', 'second', ],
	[ '3rd', 'third', ],
	[ '4th', 'fourth', ],
	[ '5th', 'fifth', ],
	[ '6th', 'sixth', ],
	[ '7th', 'seventh', ],
	[ '8th', 'eighth', ],
	[ '9th', 'ninth', ],
	[ '10th', 'tenth', ],
	[ '11th', 'eleventh', ],
	[ '12th', 'twelth', ],
);

sub to_ordinals($$)
#Given a positive integer, returns the abbreviated and written out ordinal
#format.  E.g., for 1 returns (1st, first), for 2 (2nd, second), etc.
{	my $class = shift;
	my $number = shift;
	return unless defined $number;
	$number = floor($number + 0.5); #Round number
	return unless $number > 0;

	my $rec = $ordinals[$number-1];
	if ( $rec && ref($rec) eq 'ARRAY' )
	{	return @$rec;
	}
	#If past the list given, we don't write out the second part.
	my $tmp;
	if ( $number % 10 == 1 )
	{	$tmp = $number . "st";
	} elsif ( $number % 10 == 2 )
	{	$tmp = $number . "nd";
	} elsif ( $number % 10 == 3 )
	{	$tmp = $number . "rd";
	} else
	{	$tmp = $number . "th";
	}
	return ( $tmp, $tmp );
}

#----------------------------------------------------------
#		Status/logfile utilities
#----------------------------------------------------------


sub read_status_from_logfile($$)
#Given name of status logfile, read it.
#Status log file is of format:
#key_object:sizeBytes:timestamp1stWarning:warnings:timestamp1stError:errors
#where
#key_object is either username (if -owner-combine) or full dir path
#sizeBytes is size in bytes
#timestamp1stWarning, timestamp1stError is time when 1st warning/error message
#	sent for this offense (seconds from 1970)
#warnings/errors is number of warning/error messages sent for this offense
#
#Returns hash keyed on key_object (user/directory name), value is hash ref
#with keys:
#	key_object, user, dir = key_object 
#	sizeBytes = size in bytes
#	firstWarning, firstError = timestamp of first Warning/Error sent
#	warnings, errors = count of warning/error messages sent
{	my $self = shift;
	my $statuslog = shift;
	my $me = __PACKAGE__ . '::read_status_from_logfile';

	my $status = {};
	return $status unless $statuslog && -r $statuslog;
	my $res = open(LOG,"<$statuslog");
	unless ( $res )
	{	carp "$me: Unable to open status log file $statuslog, " .
			"treating as empty at ";
		return $status;
	}

	my ($line, $key, $size, $ts_warned, $ts_error, $warnings, $errors);

	while ( $line = <LOG> )
	{	chomp $line;
		next unless $line;
		$line=~s/^\s*$//; $line=~s/\s*$//;
		next unless $line;
		next if $line=~/^#/;
		($key, $size, $ts_warned, $warnings, $ts_error, $errors) 
			 = split /\s*\|\s*/, $line;
		$status->{$key} = 
			{	#key_object=>$key,
				user => $key,
				#dir => $key,
				sizeBytes => $size,
				firstWarning => $ts_warned,
				firstError => $ts_error,
				warnings => $warnings,
				errors => $errors,
			};
	}
	return $status;
}
	
sub write_status_log_line($@)
#Writes line to status log file
#Takes key => value input pairs after implicit invocant
{	my $self = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::write_status_log_line';

	my $FH = delete $args{FH};
	my $user = delete $args{user};
	my $size= delete $args{size};
	my $firstWarned = delete $args{firstWarned};
	my $warnings = delete $args{warnings};
	my $firstError = delete $args{firstError};
	my $errors = delete $args{errors};

	if ( %args )
	{	my @tmp = keys %args;
		my $tmp = join ", ", @tmp;
		croak "$me: Unrecognized arguments [ $tmp ], aborting at ";
	}

	my $str = "$user|$size|$firstWarned|$warnings";
	$str .= "|$firstError|$errors" if $firstError;
	print $FH "$str\n";
}
	

sub write_status_to_logfile($@)
#Outputs the updated status log file
#Takes key => value pairs after the implicit invocant
{	my $self = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::write_status_to_logfile';

	my $status = delete $args{status};
	my $rootdir = delete $args{rootdir}|| 'unknown';
	my $quota = delete $args{quota} || 'unknown';
	my $logfile = delete $args{logfile};

	croak "$me: Missing required param 'logfile' at " unless $logfile;
	unless ( $status )
	{	carp "$me: No status given, treating as empty hash at ";
		$status = {};
	}

	if (%args)
	{	my @tmp = keys %args;
		my $tmp = join ", ", @tmp;
		croak "$me: Unrecognized arguments [$tmp], aborting at ";
	}

	open(LOG,">$logfile") or 
	    croak "$me: Unable to open status log $logfile for writing: $!";

	print LOG <<EOF;
#Status log for $0
#
#Checking directory(ies) $rootdir, quota/count = $quota
#
#user/dirname|size in bytes|timestamp of 1st warning|# warnings|timestamp of first Error|# errors
#
EOF
	my ($user, $oldrec, $size, $firstwarn, $firsterr, $warnings, $errors);
	my $now = time;
	foreach my $user (keys %$status)
	{	my $strec = $status->{$user};
		$size = $strec->{size};
		$firstwarn = $strec->{firstWarning} || $now;
		$firsterr = $strec->{firstError} || '';
		$warnings = $strec->{warnings} || 0;
		$errors = $strec->{errors} || 0;
		$self->write_status_log_line(
			FH => \*LOG, 
			user => $user, 
			size => $size, 
			firstWarned => $firstwarn,
			warnings => $warnings, 
			firstError => $firsterr, 
			errors => $errors,
		);
	}
	close(LOG) || die "Error closing status log after write: $!";
}

#----------------------------------------------------------
#		Disk usage utilities
#----------------------------------------------------------

sub fix_path($$)
#Given a raw path, return the "fixed" path
#This deals with things like allowing us to run on NFS fileserver w/out going
#through automounter, but reporting automounter paths
#
#This needs work, and can use being generalized for multiple clusters
{	my $self = shift;
	my $fpath = shift;

	$fpath=~s'^/export/data/'/data/dt-';
	return $fpath;
}


#----------------------------------------------------------
#		Email utilities
#----------------------------------------------------------

sub _send_warnviol_email($@)
#Send violation emails to users who have been over quota for more
#than the grace period
#Takes key => value arguments:
#	durec: The dubyuid record for this uid. REQUIRED.
#	urec: the  [ username, uid, email, fullname ] rec for this user. REQ'D
#		for users in violation.  REQUIRED
#	strec: status record for this user
#	template: Actual template, as array ref of strings.  REQUIRED
#	me: who to impersonate in error messages.
#	grace: number of days grace period, REQUIRED.
#	quota: human readable quota string, REQUIRED
{	my $self = shift;
	my %args = @_;
	my $realme = __PACKAGE__ . '::_send_warnviol_email';

	my $durec = delete $args{durec};
	my $urec = delete $args{urec};
	my $strec = delete $args{strec};
	my $template = delete $args{template};
	my $me = delete $args{me} || $realme;
	my $grace = delete $args{grace};
	my $quota = delete $args{quota};

	croak "$me: Missing required parameter 'durec' at " unless $durec;
	croak "$me: Missing required parameter 'urec' at " unless $urec;
	croak "$me: Missing required parameter 'strec' at " unless $strec;
	croak "$me: Missing required parameter 'template' at " unless $template;
	croak "$me: Missing required parameter 'quota' at " unless $quota;
	croak "$me: Missing required parameter 'grace' at " unless defined $grace;

	if ( %args )
	{	my @tmp = keys %args;
		my $tmp = join ", ", @tmp;
		croak "$me: Unrecognized arguments [ $tmp ], aborting.\n";
	}

	croak "$me: Invalid param durec '$durec', expecting hash ref at "
		unless ref($durec) eq 'HASH';
	croak "$me: Invalid param urec '$urec', expecting array ref at "
		unless ref($urec) eq 'ARRAY';
	croak "$me: Invalid param strec '$strec', expecting hash ref at "
		unless ref($strec) eq 'HASH';
	croak "$me: Invalid param template '$template', expecting array ref at "
		unless ref($template) eq 'ARRAY';
	
	my $hpc = $self->hpc;
	my $cluster = $hpc->clustername;

	my ($uname, $uid, $email, $fullname) = @$urec;
	my $sizeBytes = $durec->{sizeBytes};
	my $fullpaths = $durec->{fullname};
	my $dirnames  = $durec->{dirnames};
	my $warncount = $strec->{warnings} || 0;
	my $errcount  = $strec->{errors} || 0;
	my $now = time;
	my $firstWarnTS = $strec->{firstWarning} || $now;
	my $firstErrTS = $strec->{firstError} || $now;
	my $firstWarn = $self->timestamp_to_date($firstWarnTS);
	my $warndays = $self->days_between($now, $firstWarnTS);
	my $days2fix = $grace - $warndays;
	$days2fix = 1 if $days2fix < 1;

	my $size = scale_size($sizeBytes, '', 'auto', "%3.1f");
	my $sizebin = scale_size($sizeBytes, '', 'autoi', "%3.1f");

	my %template_vars =
	(	USER => $uname,
		FULLNAME => $fullname,
		EMAIL => $email,
                CLUSTER => $cluster,
                SIZE => $size,
                SIZE_BIN => $sizebin,
                QUOTA => $quota,
                WARNINGS => $warncount + $errcount,
                FIRSTWARNDATE => $firstWarn,
                GRACE_REMAINING => $days2fix,
	);

	send_template_email($template, \%template_vars);
}

sub send_violation_emails($@)
#Send violation emails to users who have been over quota for more
#than the grace period.
#
#Takes key => value arguments:
#	dubyuid: hash ref as returned by du_subdirs_of_dir_by_uid.  REQUIRED
#	users: list ref of tuplets [ username, uid, email, fullname ] 
#		for users in violation.  REQUIRED
#	status: hash ref representing status log. REQUIRED
#	template: path to template file to send. REQUIRED
#	grace: number of days grace period, REQUIRED.
#	quota: human readable quota string, REQUIRED
#
#For each user in users list, email will be sent and status updated.
{	my $self = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::send_violation_emails';

	my $dubyuid = delete $args{dubyuid};
	my $users = delete $args{users};
	my $status = delete $args{status};
	my $templatefile = delete $args{template};
	my $grace = delete $args{grace};
	my $quota = delete $args{quota};

	croak "$me: Missing required parameter 'dubyuid' at " unless $dubyuid;
	croak "$me: Missing required parameter 'users' at " unless $users;
	croak "$me: Missing required parameter 'status' at " unless $status;
	croak "$me: Missing required parameter 'template' at " 
		unless $templatefile;
	croak "$me: Missing required parameter 'grace' at " unless defined $grace;

	if ( %args )
	{	my @tmp = keys %args;
		my $tmp = join ", ", @tmp;
		croak "$me: Unrecognized arguments [ $tmp ], aborting.\n";
	}

	croak "$me: Invalid param dubyuid '$dubyuid', expecting hash ref at "
		unless ref($dubyuid) eq 'HASH';
	croak "$me: Invalid param users '$users', expecting array ref at "
		unless ref($users) eq 'ARRAY';
	croak "$me: Invalid param status '$status', expecting hash ref at "
		unless ref($status) eq 'HASH';
	
	my $hpc = $self->hpc;

	#Read the template
	my $TFILE;
	open($TFILE, "<$templatefile") 
		or croak "$me: Error opening template file $templatefile: " .
			"$!, at ";
	my $template = [ <$TFILE> ];
	close($TFILE);
	my $now = time;
	
	foreach my $urec (@$users)
	{	my ($uname, $uid, $email, $fullname) = @$urec;
		my $durec = $dubyuid->{$uid};
		my $strec = $status->{$uname} || {};
		
		$self->_send_warnviol_email(
			durec => $durec,
			urec => $urec,
			strec => $strec,
			template => $template,
			grace => $grace,
			quota => $quota,
			me => $me,
		);

		#Update status record for user
		$strec->{firstError} = $now unless defined $strec->{firstError};
		$strec->{errors} = 0 unless defined $strec->{errors};
		$strec->{errors}++;
		$strec->{size} =  $durec->{sizeBytes};

		$status->{$uname} = $strec;	
	}
}

sub send_warning_emails($@)
#Send warning emails to users who are over quota but within grace period
#
#Takes key => value arguments:
#	dubyuid: hash ref as returned by du_subdirs_of_dir_by_uid.  REQUIRED
#	users: list ref of tuplets [ username, uid, email, fullname ] 
#		for users in violation.  REQUIRED
#	status: hash ref representing status log. REQUIRED
#	template: path to template file to send. REQUIRED
#	grace: number of days grace period, REQUIRED.
#	quota: human readable quota string, REQUIRED
#
#For each user in users list, email will be sent and status updated.
{	my $self = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::send_warning_emails';

	my $dubyuid = delete $args{dubyuid};
	my $users = delete $args{users};
	my $status = delete $args{status};
	my $templatefile = delete $args{template};
	my $grace = delete $args{grace};
	my $quota = delete $args{quota};

	croak "$me: Missing required parameter 'dubyuid' at " unless $dubyuid;
	croak "$me: Missing required parameter 'users' at " unless $users;
	croak "$me: Missing required parameter 'status' at " unless $status;
	croak "$me: Missing required parameter 'template' at " 
		unless $templatefile;
	croak "$me: Missing required parameter 'grace' at " unless defined $grace;

	if ( %args )
	{	my @tmp = keys %args;
		my $tmp = join ", ", @tmp;
		croak "$me: Unrecognized arguments [ $tmp ], aborting.\n";
	}

	croak "$me: Invalid param dubyuid '$dubyuid', expecting hash ref at "
		unless ref($dubyuid) eq 'HASH';
	croak "$me: Invalid param users '$users', expecting array ref at "
		unless ref($users) eq 'ARRAY';
	croak "$me: Invalid param status '$status', expecting hash ref at "
		unless ref($status) eq 'HASH';
	
	my $hpc = $self->hpc;

	#Read the template
	my $TFILE;
	open($TFILE, "<$templatefile") 
		or croak "$me: Error opening template file $templatefile: " .
			"$!, at ";
	my $template = [ <$TFILE> ];
	close($TFILE);
	my $now = time;
	
	foreach my $urec (@$users)
	{	my ($uname, $uid, $email, $fullname) = @$urec;
		my $durec = $dubyuid->{$uid};
		my $strec = $status->{$uname} || {};
		
		$self->_send_warnviol_email(
			durec => $durec,
			urec => $urec,
			strec => $strec,
			template => $template,
			grace => $grace,
			quota => $quota,
			me => $me,
		);

		#Update status record for user
		$strec->{firstWarning} = $now 
			unless defined $strec->{firstWarning};
		$strec->{warnings} = 0 unless defined $strec->{warnings};
		$strec->{warnings}++;
		$strec->{size} =  $durec->{sizeBytes};

		$status->{$uname} = $strec;	
	}
}


#----------------------------------------------------------
#		User utilities
#----------------------------------------------------------

sub lookup_uid($$)
#Given user's uid, return (username, email,  fullname)
#Tries PDB first, then gecos/passwd file, then defaults
{	my $self = shift;
	my $uid = shift;
	my $me = __PACKAGE__ . '::lookup_uid';

	my $hpc = $self->hpc;
	my ($uname, $email, $fname);
	
	my $rec = Glue::HPCC::Project::Person->person_by_uid($hpc,$uid);
	if ( $rec )
	{	if ( ref($rec) )
		{	#Found them
			$uname = $rec->username;
			$email = $rec->email;
			$fname = $rec->fullname;
		} else
		{	carp "$me: Error looking up user for uid $uid: $rec\n" .
				"Trying to continue from passwd file, at ";
			undef $rec;
		}
	}

	unless ( $rec )
	{	my @pwent = getpwuid($uid);
		$uname = $pwent[0];
		$email = "${uname}\@umd.edu";

		unless ( $uname )
		{	$uname = "uid$uid";
			$email = 'payerle@umd.edu';
		}
		
		my $fname = $pwent[6] || '';
		$fname=~s/,.*$//;
		$fname=$uname unless $fname;
	}


	return ( $uname, $email, $fname);
}

#----------------------------------------------------------
#		Quota/top utilities
#----------------------------------------------------------

sub do_quota_check($$$)
#Takes either a dubyuser hash ref, and a quota, and returns
#a list ref of keys for overquota dirs/users.
#
#Quota should be in Bytes.
{	my $self = shift;
	my $hash = shift;
	my $quota = shift;
	my $me = __PACKAGE__ . '::do_quota_check';

	my @keys = keys %$hash;
	my @over = grep { $hash->{$_}->{sizeBytes} > $quota } @keys;
	return [ @over ];
}

sub check_past_grace($$$$)
#Takes a list of uids which are overquota, 
#status hash read in, and the grace period.
#
#For each uid in overquota list, get username, fullname, and email if
#possible, and puts in either a "warning" list ref or a "violation"
#list ref depending on whether they have been overquota for more than
#grace days or not.  These two list refs (or [ username, uid, email, full ]
#tuplets) are returned.
{	my $self = shift;
	my $over = shift || [];
	my $status = shift || {};
	my $grace = shift;
	my $me = __PACKAGE__ . '::check_past_grace';

	my ($user, $rec, $firstwarned, $diff);
	my $warns = [];
	my $errs = [];
	my $now = time;
	#Take 1/2 day less than stated grace time for rounding
	#We only run daily, so should not matter
	my $grace_secs = $grace * $ONEDAY - 0.5 * $ONEDAY;

	foreach my $uid ( @$over )
	{	#Expand the uid
		my ($uname, $email, $full) = $self->lookup_uid($uid);
		my $urec = [ $uname, $uid, $email, $full ];

		if ( exists $status->{$uname} )
		{	#Have record of them in status log
			$rec = $status->{$uname};
			$firstwarned = $rec->{firstWarning};
			unless ( $firstwarned )
			{	#Not warned yet even though in status log
				warn "$me: Key $uname in status log but no " . 
				     "firstWarned, treating as new offense.\n";
				push @$warns, $urec;
				next;
			}
			$diff = $now - $firstwarned;
			if ( $diff > $grace_secs )
			{	#Warned too long ago, escalate offense 
				push @$errs, $urec;
			} else
			{	push @$warns, $urec;
			}
		} else
		{	#No record, new offense
		push @$warns, $urec;
		}
	}
	return ( $warns, $errs);
}



#----------------------------------------------------------
#		High level soft quota check routine
#----------------------------------------------------------

sub soft_quota_check($@)
#Returns a list of users using more than the specified soft quota in
#the specified directories.
#
#Takes key => value pairs, recognizing:
#	hpc: Glue::HPCC::Cluster instance, REQUIRED
#	rootdirs: A list ref of root directories to check
#		REQUIRED.  If a string is given, is converted to a list ref
#		with the string as the only element.  If list ref of list refs,
#		inner list refs should be [ path, type ] where type is one
#		of 'nonlustre' (default), 'lustre-shared' (e.g. for DT where
#		home dir share lustre with data), 'lustre'.
#	quotaGB: Quota, in GB, to check against, REQUIRED
#	playGB: Defaults to 0.  Do not consider user overquota until exceed
#		quotaGB + playGB.  The purpose is to avoid sending email
#		to someone at 10.1 GB when quota is 10 GB; particularly since
#		such emails get confusing as usage would be < 10 GiB.
#	play%: 	Like playGB, but this is a percent of quotaGB.  If both
#		playGB and play% given, which ever is larger is used.
#	excludedirs: A list ref of immediate subdirs of the rootdirs to
#		exclude from the check.  Default is []
#	age: Only files older than age (in days) should be included in totals.
#		Default is all files considered.
#	statuslog: File containing status/history of recent past infractions
#		to determine whether get warning or violation.  REQUIRED
#	grace: Number of days for grace period.  Default is 0.  If user is
#		in the list less than grace days (from statuslog), it will
#		be a warning; if over is a violation.  
#	percent-used: If provided, the du, etc. (and quota/top calculations)
#		are only done if the filesystem is at least percent-used full.
#		Default is to always perform tasks.
#	suppress-email: Boolean, if true, no email is sent
#	suppress-update: Boolean, if true, status file will not be updated
#	warning-template: path to template for warning messages.  
#		REQUIRED unless suppress-email is true.
#	violation-template: path to template for violation messages.
#		REQUIRED unless suppress-email is true.
#	nonunique: When dealing with fs type lustre-shared, we can make 
#		certain optimizations if we assume that a given uid will have
#		at most 1 directory (e.g. as is the case for home directories)
#		These optimizations might not be valid otherwise; setting this
#		flag and will skip the optimizations and always return correct
#		values.  
#
{	my $self = shift;
	my %args = @_;
	my $me = __PACKAGE__ . '::soft_quota_check';

	my $rootdirs = delete $args{rootdirs};
	my $quotaGB = delete $args{quotaGB};
	my $playGB = delete $args{playGB} || 0;
	my $playpct = delete $args{'play%'} || 0;
	my $excludedirs = delete $args{excludedirs} || [];
	my $ownercombine = 1;
	my $age = delete $args{age} || 0;
	my $statuslog = delete $args{statuslog};
	my $grace = delete $args{grace} || 0;
	my $pctused = delete $args{'percent-used'} || 0;
	my $noemail = delete $args{'suppress-email'};
	my $noupdate = delete $args{'suppress-update'};
	my $warn_temp = delete $args{'warning-template'};
	my $viol_temp = delete $args{'violation-template'};
	my $nonunique = delete $args{nonunique};

	croak "$me: Missing required parm 'rootdirs' at " unless $rootdirs;
	croak "$me: Missing required parm 'quotaGB' at " unless $quotaGB;
	croak "$me: Missing required parm 'statuslog' at " unless $statuslog;
	croak "$me: Missing required parm 'warning-template' at " 
		unless $warn_temp || $noemail;
	croak "$me: Missing required parm 'violation-template' at " 
		unless $viol_temp || $noemail;

	if ( %args )
	{	my @tmp = keys %args;
		my $tmp = join ', ', @tmp;
		croak "$me: Unrecognized parameters [ $tmp ], aborting at ";
	}

	$rootdirs = [ $rootdirs ] unless ref($rootdirs) eq 'ARRAY';
	my $rdirstr;

	my $dubyuid = {}; 

	my $quotaB = $quotaGB * 1000 * 1000 * 1000;
	my $playB = $playGB * 1000 * 1000 * 1000;
	my $tmp = $quotaB * $playpct / 100;
	$playB = $tmp if $tmp > $playB;

	#Collect usage stats

	ROOTDIR: foreach my $rootdir (@$rootdirs)
	{	my ($rdir, $rdtype);
		if ( ref($rootdir) eq 'ARRAY' )
		{	($rdir, $rdtype) = @$rootdir;
		} else
		{	$rdir = $rootdir;
			$rdtype = 'nonlustre';
		}

		if ( $rdtype eq 'lustre' && $age )
		{	my $newtype = 'lustre-shared';
			$self->debug("Treating rdtype $rdtype as $newtype ",
				"because age is set");
			$rdtype = $newtype;
		}

		if ( $rdtype eq 'lustre-shared' && $nonunique )
		{	my $newtype = 'lustre-shared';
			$self->debug("Nonunique flag set, treating " .
				"$rdtype as $newtype at ");
			$rdtype = $newtype;
		}
		if ( $rdtype eq 'lustre' )
		{	carp "$me: $rdtype not yet supported, " .
				"treating as non-lustre at ";
		}

		#Skip root dir if filesystem is lightly used
		my ($totalsize, $freepct, $usedpct) = 
			df_disk_containing_file($rdir);
		next ROOT if $usedpct < $pctused;

		$rdirstr .= ', ' if $rdirstr;
		$rdirstr .= $rdir;

		#This can take a while, calculate tstamp for each rootdir
		my $tstamp;
		if ( $age )
		{	$tstamp = time;
			$tstamp -= $age * $ONEDAY;
		}
		
		if ( $rdtype eq 'nonlustre' )
		{	#Cant use fs quota, so just do a du

			$dubyuid = du_subdirs_of_dir_by_uid(
				rootdir => $rdir,
				exclude => $excludedirs,
				options => { last_modified_before => $tstamp },
				);
			next ROOTDIR;
		}

		if ( $rdtype eq 'lustre-shared' )
		{	#We should be checking lustre usage, and only if raw
			#usage exceeds the quota should we do compute_du_by_dir

			die "$me: rdtype $rdtype not supported yet";
		}

		if ( $rdtype eq 'lustre' )
		{	#We should be checking lustre usage, and only if raw
			#usage exceeds the quota should we do compute_du_by_dir

			die "$me: rdtype $rdtype not supported yet";
		}

		croak "$me: Urecognized rdtype $rdtype at ";
	}

	my $overquota = $self->do_quota_check($dubyuid, $quotaB + $playB);
	my $status = $self->read_status_from_logfile($statuslog);
	my ($warns, $viols) = $self->check_past_grace($overquota, 
		$status, $grace);

	#Remove from status log anyone who is not currently overquota
	my @over_unames = map { $_->[0] } ( @$warns, @$viols );
	my %over_unames = map { $_ => undef } @over_unames;
	my @fixed_unames = grep { ! exists $over_unames{$_} } (keys %$status);
	delete @$status{@fixed_unames};

	my ($urec, $uname, $strec, $uid, $durec);
			
	if ( $noemail )
	{	#Update firstWarning/firstError in status log as needed
		my $now = time;
		foreach $urec (@$viols)
		{	$uname = $urec->[0];
			$uid = $urec->[1];
			$strec = $status->{$uname};
			unless ( $strec )
			{	carp "$me: User $uname in viols list, but " .
					"no status record, at ";
				$strec = {};
			}
			$strec->{warnings}=0 unless defined $strec->{warnings};
			$strec->{errors}=0 unless defined $strec->{errors};
			$strec->{firstWarning}=$now 
				unless defined $strec->{firstWarning};
			$strec->{firstError}=$now 
				unless defined $strec->{firstError};
			$durec = $dubyuid->{$uid};
			$strec->{size} =  $durec->{sizeBytes};
			$status->{$uname} = $strec;
		}
		foreach $urec (@$warns)
		{	$uname = $urec->[0];
			$uid = $urec->[1];
			$strec = $status->{$uname} || {};
			$strec->{warnings}=0 unless defined $strec->{warnings};
			$strec->{firstWarning}=$now 
				unless defined $strec->{firstWarning};
			delete $strec->{errors};
			delete $strec->{firstError};
			$durec = $dubyuid->{$uid};
			$strec->{size} =  $durec->{sizeBytes};
			$status->{$uname} = $strec;
		}
	} else
	{	$self->send_violation_emails(
			dubyuid => $dubyuid,
			users => $viols,
			status => $status,
			template => $viol_temp,
			grace => $grace,
			quota => "$quotaGB GB",
		);
		$self->send_warning_emails(
			dubyuid => $dubyuid,
			users => $warns,
			status => $status,
			template => $warn_temp,
			grace => $grace,
			quota => "$quotaGB GB",
		);
	}	

	#Write the status log back
	$self->write_status_to_logfile(
		logfile => $statuslog, 
		status => $status,
		rootdir => $rdirstr,
		quota => "$quotaGB GB",
	) unless $noupdate;

	return ($status, $viols, $warns);
}


1;

