#!/usr/local/bin/perl
#
#Template toolkit utilities

package Glue::HPCC::Utils::TemplateUtils;
use strict;
use warnings;

use Carp qw(carp croak);
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw( 
	get_template_instance
	expand_template
);
use Template;


my $_template_instance;
my $TEMPLATE_ROOT = '/cell_root/software/HPCC/templates';
#warn "Using v1.1 template directory explicitly at ";
#$TEMPLATE_ROOT = '/cell_root/software/HPCC/1.1/templates';



sub get_template_instance(;$)
{	my $options = shift || {};
	my $force = delete $options->{force};

	return $_template_instance if $_template_instance && ! $force;
	my $me = __PACKAGE__ . '::get_template_instance';

	croak "$me: Invalid value '$options' for 'options'; expecting "
		. "a hash ref at " unless $options && ref($options) eq 'HASH';

	$_template_instance = Template->new(
		INCLUDE_PATH => $TEMPLATE_ROOT,
		EVAL_PERL => 0,
		ABSOLUTE => 1,
		RELATIVE => 1,
		%$options,
		);
	return $_template_instance if $_template_instance;

	croak "$me: Error creating instance of Template: " .
		Template->error . "\n";
}


sub expand_template($$)
#Expands a text template.
#Takes a Template Toolkit template and a macro hash.
#Macro hash is a hash ref of variable names and what they should be
#expanded to, which is passed to Template Toolkit
#Returns resultant text as a text string
#
#Template can be given as any of the following:
#	scalar: filename containing template, relative to TEMPLATE_ROOT
#	scalar ref: a text string containing the template
#	array ref: a list of text strings, one per line, forming the template
#	glob ref: assumed to be input filehandle from which template is to be read
{	my $template = shift;
	my $macros = shift || {};
	my $me = __PACKAGE__ . '::expand_template';

	croak "$me: Must supply a template" unless $template;

	if ( ref($template) eq 'ARRAY' )
	{	chomp @$template;
		my $str = join "\n", @$template;
		$template = \$str;
	}
	
	my $output;

	my $tt = get_template_instance;
	my $ok = $tt->process($template,$macros, \$output);
	unless ( $ok )
	{	croak "$me: Error processing template: " . 
			$tt->error . " at ";
	}
	return $output;
}



1;
__END__

=head1 NAME

Glue::HPCC::Utils::TemplateUtils

=head1 SYNOPSIS

  use Glue::HPCC::Utils::TemplateUtils qw( get_template_instance expand_template );

  my $tt = get_template_instance( { INCLUDE_PATH=> '.' } );
  my $macros = { ME => 'tom', YOU => 'someone else' };
  my $template = "my-template.tmplate";

  my $string = expand_template($template, $macros);

  ...


=head1 DESCRIPTION

Assorted utilities wrapping TemplateToolkit

Usually used indirectly through MailUtils::send_template_email.

=over 4

=item B<get_template_instance($options)>

Returns an instance of Template Toolkit, using a cached value if possible.
B<$options> is an optional hash of options to the template toolkit instance.
By default, the TT instance is created with options suitable for most HPCC
purposes.  NOTE: if a cached TT instance is found, this will be used regardless
of whether it was created with the specified options or not.  The special 
psuedo-option 'force' will cause any cached TT instance to be ignored and a 
new one (with specified options) to be created.  All other values in $options
hash ref should be options to the new method of Template module.  (Use this
method before calling any other methods using TT to change default TT settings).

The default options set the INCLUDE_PATH to the HPCC template root directory,
allow absolute and relative templates, and disables evaluation of perl in
the template.
		
=item B<expand_template($template, $macros)>

Expands the given template using the macros defined in %$macros.  $template
can take any form recognized by the TT process method, e.g. a scalar (representing
a filename containing the template), a scalar ref (representing a text string
containing the template), an array ref (a list of text strings, one per line,
representing the template), or a glob ref (assumed to be an input filehandle from
which the template is to be read).

$macros is a hash ref, the key being the name of the macro, and the value being
the value given to it.

=back






