#!/usr/local/bin/perl
#
#Utilities for dealing with time

package Glue::HPCC::Utils::TimeUtils;
use strict;
use warnings;

use Carp qw(carp croak);
use Time::Local qw(timelocal timelocal_nocheck);

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw( 
	tma_of_start_of_day
	tma_of_start_of_month
	tma_of_start_of_quarter
	tma_of_start_of_tomorrow
	tma_of_start_of_next_month
	tma_of_start_of_next_quarter
	normalize_tma

	tstamp_from_tma
	tstamp_start_of_day
	tstamp_start_of_month
	tstamp_start_of_quarter
	tstamp_start_of_tomorrow
	tstamp_start_of_next_month
	tstamp_start_of_next_quarter

	fraction_of_quarter
);

#----------------------------------------------------------------
#		Routines acting on struct tm array refs (tma's)
#These all act on struct tm array refs, e.g. refs to Perl arrays mimicing
#the C struct tm object.  I.e., a ref to what localtime returns in 
#array context.  They also return tma references (modified copies of input)
#
#normalize_tma is the only one which should work as expected with non-normalized
#tmas.  Eg. 2014-Jan-1 25:00:00 would return 2014-Jan-2 01:00:00
#
#The rest should all return normalized tma's if given a normalized tma as
#input.  And they will likely fail to work properly on at least some
#non-normalized tmas.
#
#E.g. tma_of_start_of_day(2014-Jan-1 25:00:00) = 2014-Jan-1 00:00:00
#	does NOT equal
#tma_of_start_of_day(tma_normalize(2014-Jan-1 25:00:00) = 2014-Jan-2 00:00:00
#
#
#----------------------------------------------------------------


sub normalize_tma($)
#Normalize a struct tm array ref.
#Convert back to secs from epoch, then back to struct tm
#returns array ref to normalized struct tm array
{	my $tma = shift;
	my $me = 'normalize_tma';

	my ($sec, $min, $hour, $mday, $mon, $year ) = @$tma;
	my $tstamp = timelocal_nocheck($sec, $min, 
		$hour, $mday, $mon, $year);
	my $normtma = [ localtime $tstamp ];
	return $normtma;
}

sub tma_of_start_of_day(;$)
#Takes a  ref to struct tm-like array.  If omitted, uses localtime
#Returns a ref to copy of the same but set to the start of that day.
#
#The returned array should be normalized if input one was.  Might get
#incorrect answer if array not normalized.  (E.g. 2014-Jan-1 30:00:00 will
#return 2014-Jan-1 00:00:00, not 2014-Jan-2 00:00:00)
{	my $tma1 = shift || [ localtime ];
	my $me = 'tma_of_start_of_day';

	croak "$me: Invalid input: expect 'struct tm' array reference, "
		. " but got '$tma1' at " unless $tma1 && ref($tma1) eq 'ARRAY';

	#Set to  midnight, in the AM of specified day
	my $tma = [ @$tma1 ];
	$tma->[0] = 0; #sec
	$tma->[1] = 0; #min
	$tma->[2] = 0; #hour

	return $tma;
}


sub tma_of_start_of_month(;$)
#Takes a  ref to struct tm-like array.  If omitted, uses localtime
#Returns a ref to copy of the same but set to the start of first day of
#that month.
#
#The returned array should be normalized if input one was.  Might get
#incorrect answer if array not normalized.  (E.g. 2014-Jan-32 will
#return 2014-Jan-1 00:00:00 not 2014-Feb-1 00:00:00)
{	my $tma1 = shift || [ localtime ];
	my $me = 'tma_of_start_of_month';

	croak "$me: Invalid input: expect 'struct tm' array reference, "
		. " but got '$tma1' at " unless $tma1 && ref($tma1) eq 'ARRAY';

	#Set to midnight, in the AM of specified day
	my $tma = tma_of_start_of_day($tma1);
	$tma->[3] = 1; #day of month

	return $tma;
}

sub tma_of_start_of_quarter(;$)
#Takes a  ref to struct tm-like array.  If omitted, uses localtime
#Returns a ref to copy of the same but set to the start of first day of
#that quarter
#
#The returned array should be normalized if input one was.  Might get
#incorrect answer if array not normalized.  (E.g. 2014-Mar-32 will
#return 2014-Jan-1 00:00:00 not 2014-Apr-1 00:00:00)
{	my $tma1 = shift || [ localtime ];
	my $me = 'tma_of_start_of_quarter';

	croak "$me: Invalid input: expect 'struct tm' array reference, "
		. " but got '$tma1' at " unless $tma1 && ref($tma1) eq 'ARRAY';

	#Set to 1 sec past midnight, in the AM of specified day
	my $tma = tma_of_start_of_month($tma1);
	my $mon = $tma->[4];

	if ( $mon >= 9 ) #Oct, Nov, Dec (remember the -1 in struct tm)
	{	$mon = 9;
	} elsif ( $mon >=6 ) #July, Aug, Sep
	{	$mon = 6;
	} elsif ( $mon >=3 ) #April, May, June
	{	$mon = 3;
	} else
	{	$mon = 0; #Jan, Feb, Mar
	}
	$tma->[4] = $mon;

	return $tma;
}

sub tma_of_start_of_tomorrow(;$)
#Takes a  ref to struct tm-like array.  If omitted, uses localtime
#Returns a ref to copy of the same but set to the start of tomorrow
#
#The returned array will be normalized if input was.  Might get
#incorrect answer if array not normalized.  (E.g. 2014-Jan-1 30:00:00 will
#return 2014-Jan-2 00:00:00, not 2014-Jan-3 00:00:00).  In some cases,
#we always normalize results.
{	my $tma1 = shift || [ localtime ];
	my $me = 'tma_of_start_of_tomorrow';

	croak "$me: Invalid input: expect 'struct tm' array reference, "
		. " but got '$tma1' at " unless $tma1 && ref($tma1) eq 'ARRAY';

	#Set to  midnight, in the AM of today
	my $tma = tma_of_start_of_day($tma1);
	#increment mday
	$tma->[3]++;
	if ( $tma->[3] > 28 )
	{	#We _might_ have crossed a month boundary, normalize
		$tma = normalize_tma($tma);
	}
	return $tma;
}

sub tma_of_start_of_next_month(;$)
#Takes a  ref to struct tm-like array.  If omitted, uses localtime
#Returns a ref to copy of the same but set to the start of first day of
#next month.
#
#The returned array will be normalized if input was.  Might get
#incorrect answer if array not normalized.  (E.g. 2014-Jan-35 will
#return 2014-Feb-1 00:00:00, not 2014-Mar-1 00:00:00). 
{	my $tma1 = shift || [ localtime ];
	my $me = 'tma_of_start_of_next_month';

	croak "$me: Invalid input: expect 'struct tm' array reference, "
		. " but got '$tma1' at " unless $tma1 && ref($tma1) eq 'ARRAY';

	#Set to  start of this month
	my $tma = tma_of_start_of_month($tma1);
	#increment month
	$tma->[4]++;
	if ( $tma->[4] > 11 )
	{	#We crossed a year boundary
		#Set month to Jan
		$tma->[4] = 0;
		#increment year
		$tma->[5]++;
	}
	return $tma;
}

sub tma_of_start_of_next_quarter(;$)
#Takes a  ref to struct tm-like array.  If omitted, uses localtime
#Returns a ref to copy of the same but set to the start of first day of
#next quarter
#
#The returned array will be normalized if input was.  Might get
#incorrect answer if array not normalized.  (E.g. 2014-Mar-35 will
#return 2014-Apr-1 00:00:00, not 2014-Jul-1 00:00:00). 
{	my $tma1 = shift || [ localtime ];
	my $me = 'tma_of_start_of_next_quarter';

	croak "$me: Invalid input: expect 'struct tm' array reference, "
		. " but got '$tma1' at " unless $tma1 && ref($tma1) eq 'ARRAY';

	#Set to  start of this quarter
	my $tma = tma_of_start_of_quarter($tma1);
	#Add 3 to month
	$tma->[4] += 3;
	if ( $tma->[4] > 11 )
	{	#We crossed a year boundary
		#Set month to Jan
		$tma->[4] = 0;
		#increment year
		$tma->[5]++;
	}
	return $tma;
}

#----------------------------------------------------------------
#		Routines acting on timestamps (tstamps, secs from epoch)
#
#These all deal with tstamps, e.g. seconds from 1970 (epoch)
#They mostly act on tstamps, and all return tstamps.
#The tstamp_start* will take a tstamp or a tma (tma's are refs)
#
#Basically mimic the corresponding tma routines, but returns the
#
#
#----------------------------------------------------------------

sub tstamp_from_tma($)
#Converts  a tma array ref to a tstamp.
#Does NOT assume is normalized
{	my $tma = shift;
	my $me = 'tstamp_from_tma';

	croak "$me: Invalid input: expect 'struct tm' array reference, "
		. " but got '$tma' at " unless $tma && ref($tma) eq 'ARRAY';


	my ($sec, $min, $hour, $mday, $mon, $year ) = @$tma;
	my $tstamp = timelocal_nocheck($sec, $min, 
		$hour, $mday, $mon, $year);
	return $tstamp;
}

sub tstamp_start_of_day(;$)
#Given a tstamp (defaults to time), returns the tstamp for
#the start of that day.
{	my $tstamp1 = shift || time;
	my $tma1 = $tstamp1;
	$tma1 = [ localtime($tstamp1) ] unless ref($tma1) eq 'ARRAY';
	my $tma = tma_of_start_of_day($tma1);
	return tstamp_from_tma($tma);
}

sub tstamp_start_of_month(;$)
#Given a tstamp (defaults to time), returns the tstamp for
#the start of that month
{	my $tstamp1 = shift || time;
	my $tma1 = $tstamp1;
	$tma1 = [ localtime($tstamp1) ] unless ref($tma1) eq 'ARRAY';
	my $tma = tma_of_start_of_month($tma1);
	return tstamp_from_tma($tma);
}

sub tstamp_start_of_quarter(;$)
#Given a tstamp (defaults to time), returns the tstamp for
#the start of that quarter.
{	my $tstamp1 = shift || time;
	my $tma1 = $tstamp1;
	$tma1 = [ localtime($tstamp1) ] unless ref($tma1) eq 'ARRAY';
	my $tma = tma_of_start_of_quarter($tma1);
	return tstamp_from_tma($tma);
}

sub tstamp_start_of_tomorrow(;$)
#Given a tstamp (defaults to time), returns the tstamp for
#the start of tomorrow
{	my $tstamp1 = shift || time;
	my $tma1 = $tstamp1;
	$tma1 = [ localtime($tstamp1) ] unless ref($tma1) eq 'ARRAY';
	my $tma = tma_of_start_of_tomorrow($tma1);
	return tstamp_from_tma($tma);
}

sub tstamp_start_of_next_month(;$)
#Given a tstamp (defaults to time), returns the tstamp for
#the start of the next month
{	my $tstamp1 = shift || time;
	my $tma1 = $tstamp1;
	$tma1 = [ localtime($tstamp1) ] unless ref($tma1) eq 'ARRAY';
	my $tma = tma_of_start_of_next_month($tma1);
	return tstamp_from_tma($tma);
}

sub tstamp_start_of_next_quarter(;$)
#Given a tstamp (defaults to time), returns the tstamp for
#the start of the next quarter
{	my $tstamp1 = shift || time;
	my $tma1 = $tstamp1;
	$tma1 = [ localtime($tstamp1) ] unless ref($tma1) eq 'ARRAY';
	my $tma = tma_of_start_of_next_quarter($tma1);
	return tstamp_from_tma($tma);
}

#----------------------------------------------------------------
#		Routines determining fraction of month/quarter/etc
#
#----------------------------------------------------------------

sub fraction_of_quarter(;$)
#Takes a tstamp (defaults to time) and returns a list
#(month_in_quarter, fraction_of_month, fraction_of_quarter)
#
#We determine which quarter the given tstamp is in, and then return
#month_in_quarter: 0,1, or 2 depending on whether is first, 2nd, or 3rd month
#	of the quarter
#fraction_of_month: between 0 and 1, the fraction of the current month that
#	has past
#fraction_of_quarter: between 0 and 1, the fraction of the current quarter
#	that has past
{	my $tstamp = shift || time;
	my $me = 'fraction_of_quarter';

	my $tma = $tstamp;
	$tma = [ localtime $tstamp ] unless ref($tma) eq 'ARRAY';
	if ( ref($tstamp) eq 'ARRAY' ) 
	{	#We were given a tma, convert to tstamp
		$tstamp = tstamp_from_tma($tma);
	}

	my $tma_start_quarter= tma_of_start_of_quarter($tma);
	my $ts_start_quarter = tstamp_from_tma($tma_start_quarter);
	my $ts_start_month =   tstamp_start_of_month($tma);
	my $ts_next_quarter =  tstamp_start_of_next_quarter($tma);
	my $ts_next_month =    tstamp_start_of_next_month($tma);

	my $mon_quarter_start = $tma_start_quarter->[4];
	my $mon_current       = $tma->[4];

	my $mon_of_quarter = $mon_current - $mon_quarter_start;
	die "$me: ASSERT ERROR: mon_of_quarter < 0 " if $mon_of_quarter < 0;
	die "$me: ASSERT ERROR: mon_of_quarter > 2 " if $mon_of_quarter > 2;

	my $secs_in_quarter = $ts_next_quarter - $ts_start_quarter;
	my $secs_in_month = $ts_next_month - $ts_start_month;
	die "$me: ASSERT ERROR: secs_in_quarter <= 0 " if $secs_in_quarter <= 0;
	die "$me: ASSERT ERROR: secs_in_month <= 0 " if $secs_in_month <= 0;

	my $secs_of_month = $tstamp - $ts_start_month;
	my $secs_of_quarter = $tstamp - $ts_start_quarter;
	die "$me: ASSERT ERROR: secs_of_quarter < 0 " if $secs_of_quarter < 0;
	die "$me: ASSERT ERROR: secs_of_month < 0 " if $secs_of_month < 0;

	my $quarter_frac = $secs_of_quarter/$secs_in_quarter;
	my $month_frac = $secs_of_month/$secs_in_month;
	die "$me: ASSERT ERROR: quarter_frac < 0 " if $quarter_frac < 0;
	die "$me: ASSERT ERROR: quarter_frac > 0 " if $quarter_frac > 1;
	die "$me: ASSERT ERROR: month_frac < 0 " if $month_frac < 0;
	die "$me: ASSERT ERROR: month_frac > 0 " if $month_frac > 1;
	
	return ($mon_of_quarter, $month_frac, $quarter_frac);
}

1;
__END__

=head1 NAME

Glue::HPCC::Utils::TimeUtils- Utils for dealing with time

=head1 SYNOPSIS

  use Glue::HPCC::Utils::TimeUtils qw( set_tma_to_start_of_day);

  my @lta = localtime; #current time
  my $daystart = tma_of_start_of_day(\@lta); #midnight this morning
  my $monstary = tma_of_start_of_month(\@lta); #1st day of this month
  ...


=head1 DESCRIPTION

Assorted utilities for dealing with time.

Routines with B<tma> in the name act on a reference to an array with
a B<struct tm> list, e.g. a reference to an array as returned by
B<localtime> function.



