#!/usr/local/bin/perl
#
#Unix group processing utilities

package Glue::HPCC::Utils::UnixGroup;
use strict;
use warnings;

use Carp qw(carp croak);
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw( 
	unix_group_exists 
	user_in_unix_group
	add_user_to_unix_group
	remove_user_from_unix_group
	set_users_primary_unix_group
	delete_unix_group
);

use Glue::HPCC::Utils::ShellUtils qw(run_system_command);
use Glue::HPCC::Utils::UserAccounts qw(user_exists);
use Net::Hesiod;

my $GRPCMD = "/usr/local/scripts/grp";
my $hesiod = new Net::Hesiod;


sub unix_group_exists($)
#Returns true (Hesiod group entry) if named group exists in Hesiod, 
#false otherwise.  
{	my $gname = shift;
	my $grpent = $hesiod->resolve($gname, "group");
	return $grpent;
}

sub user_in_unix_group($$)
#Returns true if specified user is in specified unix group
#Either as primary group or an extended group
#Returns false otherwise (including errors, although will warn)
{	my $user = shift;
	my $group = shift;
	my $me = 'user_in_unix_group';

	my $pwent = user_exists($user);
	unless ( $pwent )
	{	carp "$me: Invalid user '$user', not in hesiod at ";
		return; #continue
	}
	my @pwent = split /:/, $pwent;
	my $primary_gid = $pwent[3];

	my $grent = unix_group_exists($group);
	unless ( $grent )
	{	carp "$me: Invalid group '$group', not in hesiod at ";
		return; #continue
	}
	my @grent = split /:/, $grent;
	my $gid = $grent[2];
	return 1 if $primary_gid == $gid;

	my $membstr = $grent[3];
	my @members = ();
	@members = split /,/, $membstr if $membstr;
	my @temp = grep { $_ eq $user } @members;
	return 1 if scalar(@temp);
	return;
}
	
sub add_user_to_unix_group(@)
#Takes key=>value pairs for arguments.  Recognized arguments are:
#	user: User to add to $group.  REQUIRED.
#	group: Group to to add $user to.  REQUIRED.
#	runas: User to runas
#	force: Whether to use --force flag to grp
#
#Honors dryrun (via run_system_command)
{	my %args = @_;
	my $me = 'add_user_to_unix_group';

	my $user = delete $args{user};
	my $group = delete $args{group};
	my $runas = delete $args{runas};
	my $force = delete $args{force};

	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "$me: Unexpected args [ $tmp ], aborting at ";
	}

	unless ( $user )
	{	croak "$me: Missing required argument 'user' at ";
	}
	unless ( $group )
	{	croak "$me: Missing required argument 'group' at ";
	}


	my $force_flag = '';
	$force_flag = '--force' if $force;
	my $cmd = "$GRPCMD $force_flag add -user $user -group $group";
	run_system_command($cmd, $runas);
}

sub remove_user_from_unix_group(@)
#Takes key=>value pairs for arguments.  Recognized arguments are:
#	user: User to add to $group.  REQUIRED.
#	group: Group to to add $user to.  REQUIRED.
#	runas: User to runas
#	force: Whether to use --force flag to grp
#
#Honors dryrun (via run_system_command)
{	my %args = @_;
	my $me = 'remove_user_from_unix_group';

	my $user = delete $args{user};
	my $group = delete $args{group};
	my $runas = delete $args{runas};
	my $force = delete $args{force};

	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "$me: Unexpected args [ $tmp ], aborting at ";
	}

	unless ( $user )
	{	croak "$me: Missing required argument 'user' at ";
	}
	unless ( $group )
	{	croak "$me: Missing required argument 'group' at ";
	}


	my $force_flag = '';
	$force_flag = '--force' if $force;
	my $cmd = "$GRPCMD $force_flag remove -user $user -group $group";
	run_system_command($cmd, $runas);
}

sub set_users_primary_unix_group(@)
#Takes key=>value pairs for arguments.  Recognized arguments are:
#	user: User to add to $group.  REQUIRED.
#	group: Group to to add $user to.  REQUIRED.
#	runas: User to runas
#	force: Whether to use --force flag to grp
#
#Honors dryrun (via run_system_command)
{	my %args = @_;
	my $me = 'set_users_primary_unix_group';

	my $user = delete $args{user};
	my $group = delete $args{group};
	my $runas = delete $args{runas};
	my $force = delete $args{force};

	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "$me: Unexpected args [ $tmp ], aborting at ";
	}

	unless ( $user )
	{	croak "$me: Missing required argument 'user' at ";
	}
	unless ( $group )
	{	croak "$me: Missing required argument 'group' at ";
	}

	my $grent = unix_group_exists($group);
	unless ( $grent )
	{	croak "$me: Invalid group $group, not in hesiod, at ";
	}
	my @grent = split /:/, $grent;
	my $gid = $grent[2];

	my $pwent = user_exists($user);
	unless ( $pwent )
	{	croak "$me: Invalid user '$user', not in hesiod at ";
	}
	my @pwent = split /:/, $pwent;
	my $primary_gid = $pwent[3];

	if ( $primary_gid == $gid )
	{	carp "$me: User $user already has primary gid=$group at ";
		return;
	}

	my $force_flag = '';
	$force_flag = '--force' if $force;
	my $cmd = "$GRPCMD $force_flag primary $user $group";
	run_system_command($cmd, $runas);
}
	
sub delete_unix_group(@)
#Takes key=>value pairs for arguments.  Recognized arguments are:
#	group: Group to to delete.  REQUIRED.
#	runas: User to runas
#	force: Whether to use --force flag to grp
#
#Honors dryrun (via run_system_command)
{	my %args = @_;
	my $me = 'delete_unix_group';

	my $group = delete $args{group};
	my $runas = delete $args{runas};
	my $force = delete $args{force};

	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "$me: Unexpected args [ $tmp ], aborting at ";
	}

	unless ( $group )
	{	croak "$me: Missing required argument 'group' at ";
	}


	my $force_flag = '';
	$force_flag = '--force' if $force;
	my $cmd = "$GRPCMD $force_flag delete -group $group";
	run_system_command($cmd, $runas);
}

1;
__END__

=head1 NAME

Glue::HPCC::Utils::UnixGroup- Utils for Unix Groups

=head1 SYNOPSIS

  use Glue::HPCC::Utils::UnixGroup qw( group_exists );

  my $res = group_exists("dt-fooproject");
  ...


=head1 DESCRIPTION

Assorted utilities for dealing with Unix groups.
Most (all?) commands go through Glue::HPCC:Utils::ShellUtils and so honor
its DRYRUN setting.

=over 4

=item B<unix_group_exists($name)>

Returns true if group named B<$name> exists in Hesiod, false otherwise.

=item B<user_in_unix_group($user, $group)>

Returns true is user B<$user> is in Unix group B<$group>, false otherwise.

=item B<add_user_to_unix_group( user=>USER, group=>GROUP, [ runas=>RUNAS], [force=>1])>

Adds USER to unix group GROUP.  If force set, adds --force flag to the grp
command.  Runs grp command as RUNAS if provided.

=item B<set_users_primary_unix_group( user=>USER, group=>GROUP, [ runas=>RUNAS], [force=>1])>

Sets the primary unix group for  USER to  GROUP.  If force set, adds 
--force flag to the grp
command.  Runs grp command as RUNAS if provided.

=item B<remove_user_to_unix_group( user=>USER, group=>GROUP, [ runas=>RUNAS], [force=>1])>

Removes USER from unix group GROUP.  If force set, adds --force flag to the grp
command.  Runs grp command as RUNAS if provided.

=item B<delete_unix_group( group=>GROUP, [ runas=>RUNAS], [force=>1])>

Deletes the unix group GROUP.  If force set, adds --force flag to the grp command.
Runs grp command as RUNAS if provided.

=back






