#!/usr/local/bin/perl
#
#Unix user account processing utilities

package Glue::HPCC::Utils::UserAccounts;
use strict;
use warnings;

use Carp qw(carp croak);
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw( 
	user_exists 
	uid_from_username
	username_from_uid
	user_exists_in_AFS_shadow 
	make_user_home_directory
);

use Net::Hesiod;
use Glue::HPCC::Utils::AfsUtils qw(afs_sysname);
use Glue::HPCC::Utils::ShellUtils qw(run_system_command);

my $hesiod = new Net::Hesiod;
my $MK_NEWHOME = "/local/etc/mk_newhome";

sub user_exists($)
#Returns true (/etc/passwd entry) if named user exists, false otherwise
{	my $uname = shift;
	my $pwent = $hesiod->resolve($uname, "passwd");
	return $pwent;
}

sub user_exists_in_AFS_shadow($)
#Returns true if named user exists in shadow passwd file, false otherwise
{	my $uname = shift;

	my $sysname = afs_sysname;
	my $shadow_sys = $sysname;
	$shadow_sys =~ s/^.*_//;

	my $shadow_root = "/export/${shadow_sys}_shadow";
	my $shad_pwsrc = $shadow_root . '/glue.umd.edu/system/common/etc/pwdfile/@sys';
	unless ( -f $shad_pwsrc )
	{	die "Unable to find source for passwords in AFS shadow ($shad_pwsrc) at ";
	}

	my $FILE;
	open($FILE, "<$shad_pwsrc")  or die "Error opening PW stub in AFS shadow ($shad_pwsrc): $! at";

	my $line;
	while ( $line = <$FILE> )
	{	chomp $line;
		$line =~ s/:.*$//;
		return if ( $line eq $uname );
	}
	return;
}

sub uid_from_username($)
#Given username, returns uid.
{	my $username = shift;
	my $pwent = user_exists($username);
	return unless $pwent;
	my @tmp = split /:/, $pwent;
	return $tmp[2];
}

sub username_from_uid($)
#Given uid, returns username
{	my $uid = shift;
	my $name = getpwuid($uid);
	return unless $name;
	return $name;
}

sub make_user_home_directory(@)
#Takes key=>value pairs for arguments.  Recognized arguments are:
#	user: User to add to $group.  REQUIRED.
#	homeroot: Root for home directories.  REQUIRED.
#	uid: UID for user.  Defaults to that from Hesiod passwd entry.
#	gid: GID for user's primary group.  Defaults to that from Hesiod passwd entry.
#
#Honors dryrun (via run_system_command)
{	my %args = @_;

	my $user = delete $args{user};
	my $homeroot = delete $args{homeroot};
	my $uid = delete $args{uid};
	my $gid = delete $args{gid};
	my $runas = 'root';  #This MUST run as root

	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unexpected args [ $tmp ], aborting at ";
	}

	unless ( $user )
	{	croak "Missing required argument 'user' at ";
	}
	unless ( $homeroot )
	{	croak "Missing required argument 'hoemroot' at ";
	}

	
	my $pwent = user_exists($user);
	unless ( $pwent )
	{	croak "User $user not found in hesiod, aborting at ";
	}

	my @pwent = split /:/, $pwent;
	unless ( defined $uid )
	{	$uid = $pwent[2];
	}
	unless ( defined $gid )
	{	$gid = $pwent[3];
	}

	my $cmd = "$MK_NEWHOME $homeroot $user $uid $gid";
	run_system_command($cmd, $runas);
}


1;
__END__

=head1 NAME

Glue::HPCC::Utils::UnixGroup- Utils for Unix Groups

=head1 SYNOPSIS

  use Glue::HPCC::Utils::UnixGroup qw( group_exists );

  my $res = group_exists("dt-fooproject");
  ...


=head1 DESCRIPTION

Assorted utilities for dealing with Unix groups.
Most (all?) commands go through Glue::HPCC:Utils::ShellUtils and so honor
its DRYRUN setting.

=over 4

=item B<user_exists($name)>

Returns true if user named B<$name> exists in Hesiod, false otherwise.
Actually, returns their /etc/passwd entry from Hesiod.

=item B<user_exists_in_shadow($name)>

Returns true if user named B<$name> exists in passwd file in AFS shadow, false otherwise.

=item B<make_user_home_directory( user=>USER, [ uid=>UID ], [ gid=>GID ], homeroot=>ROOT, >

Runs B<mk_newhome> command, to create a home directory
for user named USER under homeroot.  If UID and/or GID not provided, will
be taken from B</etc/passwd>.  This will always run as root.  DRYRUN is honored
via B<run_system_command>.

=back






