#!/usr/local/bin/perl
#
#Allow code re-use for delusers, archive_*, etc
#scripts for stuff related to the "canned" reasons
#for revoking user access to project/cluster
#

package Glue::HPCC::Utils::UserRevocationReasons;

use strict;
use warnings;

use Carp qw(carp croak);

#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

#Recognized options in the option hash
my @known_options = qw( nologreason noconfirm reasonoptional usage_sub );
my %known_options = map { $_ => undef } @known_options;

my @known_arguments =
(	#Custom reasons

	{	name => 'reason',
		argstr => 'reason=s',
		#excludeOnOption
		allowMulti => 1,
		helpText => <<EOF,
--reason REASON: This will set the reason (as emailed to PoCs/etc)
	to REASON.  It can be repeated, in which case each occurrence
	will add a line. 
EOF
		onlyWithLogReasons => 1,
	},
		
	{	name => 'extra',
		argstr => 'extra=s',
		allowMulti => 1,
		helpText => <<EOF,
--extra REASON: This will append the specified custom reason REASON
	to the list of reasons (as emailed to PoCs/etc).  It can
	be repeated to add multiple lines of reasons.
EOF
	},

	{	name => 'logreason',
		argstr => 'logreason|log|log-reason|log_reason=s',
		allowMulti => 1,
		helpText => <<EOF,
--logreason REASON: This sets the reason (as logged in PDB logs) to
	REASON.  It can be repeated, in which case REASON is appended
	to any previous value with a semi-colon delimiter.  If omitted,
	the default is to use the reason (as emailed to PoCs, etc) with
	semi-colon delimiters between lines.  It cannot be used with
	the --logextra flag.
EOF
		excludeOnOption => 'nologreason',
	},

	{	name => 'logextra',
		argstr => 'logextra|log-extra|log_extra=s',
		allowMulti => 1,
		helpText => <<EOF,
--logextra REASON: If --logreason was not given, the reason for the logs
	is defaulted from the reason displayed to users.  In that case,
	one can give this flag to append extra REASON to the log reason
	default (with semi-colon delimitter).  It may be repeated to 
	append multiple lines.
EOF
		excludeOnOption => 'nologreason',
	},

	#Canned reasons

	{	name => 'userreq',
		argstr => 'userrequest|user-request|user_request:s',
		hasExtra => 1,
		helpText => <<EOF,
--userreq [EXTRA]: This generates a 'canned' reason for when the user
	requested their access be revoked.
EOF
		text => 'User requested removal from allocation.',
	},

	{	name => 'pocreq',
		argstr => 'pocrequest|poc-request|poc_request:s',
		hasExtra => 1,
		helpText => <<EOF,
--pocreq [EXTRA]: This generates a 'canned' reason for when a PoC requests
	to have an users access revoked.
EOF
		text => 'A Point-of-Contact for the allocation requested removal of user.',
	},

	{	name => 'nonumd',
		argstr => 'nonumd|noumd|non-umd|non_umd|no_umd|no-umd:s',
		hasExtra => 1,
		helpText => <<EOF,
--nonumd [EXTRA]: This generates a 'canned' reason for when an user is no
	longer associated with campus.
EOF
		text => 'User is no longer associated with the University.',
	},

	{	name => 'nonglue',
		argstr => 'nonglue|noglue|non-glue|non_glue|no_glue|no-glue:s',
		hasExtra => 1,
		helpText => <<EOF,
--nonglue [EXTRA]: This generates a 'canned' reason for when an user no
	longer has a valid Glue/TerpConnect account.
EOF
		text => 'User no longer has a valid TerpConnect/Glue account.',
	},

	{	name => 'allocexp',
		argstr => 'allocation-expired|allocation_expired|expired-allocation|expired_allocation|alloc-expired|alloc_expired|exp-allocation|exp_allocation|allocexp|expalloc:s',
		hasExtra => 1,
		helpText => <<EOF,
--alloc-expired [EXTRA]: This generates a 'canned' reason for when the 
	allocation account the user belonged to has expired/been disabled.
EOF
		text => 'Allocation has expired/been disabled.',
	},

);


#-------------------------------------------------------------------
#	Methods
#-------------------------------------------------------------------

sub help_text($;$)
#This returns a text string suitable for use as a help text for
#the reason-related arguments.  It accepts an optional options_hash
#to control behavior.
{	my $class = shift;
	my $ohash = shift || {};
	my $me = __PACKAGE__ . '::help_text';

	#Check for unknown options
	my @temp = keys %$ohash;
	@temp = grep { ! exists $known_options{$_} } @temp;
	if ( @temp )
	{	my $temp = join ", ", @temp;
		croak "$me: Unrecognized values [ $temp ]  in options hash at ";
	}

	my @lines = ();

	ARG: foreach my $argrec (@known_arguments)
	{	my $name = $argrec->{name};
		my $help = $argrec->{helpText};
		my $excl = $argrec->{excludeOnOption};
		next ARG if $excl && $ohash->{$excl};

		my $onlyLogs = $argrec->{onlyWithLogReasons};
		if ( $onlyLogs )
		{	#Add text to help saying CANNOT be used with
			#anything but --logreason/--logextra
			if ( $ohash->{nologreason} )
			{	#But we don't use --logreason
				$help .= <<EOF;
	This option cannot be used with any other reason 
	related options.
EOF
			} else
			{	$help .= <<EOF;
	This option cannot be used with any other reason 
	related options except --logreason or --logextra.
EOF
			}
		}

		my $hasExtra = $argrec->{hasExtra};
		if ( $hasExtra )
		{	$help .= <<EOF;
	If given, EXTRA will be appended (in parentheses) to the canned reason.
EOF
		}


		push @lines, $help;
	}

	my $help = join "\n", @lines;
	return $help;
}
	

sub make_arglist($$;$)
#This will generate an argument spec list for Getopt::Long::GetOptions
#for reason-related arguments (subject to options_hash).  The arglust
#will result in any values for command line arguments being stored
#in $urevoc_flags hash ref.
{	my $class = shift;
	my $urevoc_flags = shift;
	my $ohash = shift || {};
	my $me = __PACKAGE__ . '::make_arglist';

	croak "$me: Missing required option urevoc_flags at "
		unless $urevoc_flags;
	croak "$me: Invalid value '$urevoc_flags' for urevoc_flags; " .
		"expecting a hash ref, at "
		unless ref($urevoc_flags) eq 'HASH';

	#Check for unknown options
	my @temp = keys %$ohash;
	@temp = grep { ! exists $known_options{$_} } @temp;
	if ( @temp )
	{	my $temp = join ", ", @temp;
		croak "$me: Unrecognized values [ $temp ]  in options hash at ";
	}

	my $arglist = [];

	ARG: foreach my $argrec (@known_arguments)
	{	my $name = $argrec->{name};
		my $excl = $argrec->{excludeOnOption};
		next ARG if $excl && $ohash->{$excl};
		my $argstr =  $argrec->{argstr};
		my $multi =  $argrec->{allowMulti};

		my $sub;
		if ( $multi )
		{	$sub = sub 
			{	my $k = shift; my $v = shift;
				$urevoc_flags->{$name} = []
				   	unless $urevoc_flags->{$name} &&
					ref($urevoc_flags->{$name}) eq 'ARRAY';
				my $tmp = $urevoc_flags->{$name};
				push @$tmp, $v;
			}
		} else
		{	$sub = sub
			{	my $k = shift; my $v = shift;
				if ( exists $urevoc_flags->{$name} )
				{	die "Illegal attempt to repeat --$name flag";
				}
				$urevoc_flags->{$name} = $v;
			}
		}
		push @$arglist, $argstr => $sub;
	}
				
	return $arglist;
}

sub process_arghash($$;$)
#This takes the $urevoc_flags hash ref (used in make_arglist and populated by
#GetOptions) and processes it (according to options in options_hash) to
#produce and return ( $reasons, $logreason).  $reasons is array ref, $logreason
#a string.
{	my $class = shift;
	my $urevoc_flags = shift || {};
	my $ohash = shift || {};
	my $me = __PACKAGE__ . '::process_arghash';

	croak "$me: Missing required option urevoc_flags at "
		unless $urevoc_flags;
	croak "$me: Invalid value '$urevoc_flags' for urevoc_flags; " .
		"expecting a hash ref, at "
		unless ref($urevoc_flags) eq 'HASH';

	#Check for unknown options
	my @temp = keys %$ohash;
	@temp = grep { ! exists $known_options{$_} } @temp;
	if ( @temp )
	{	my $temp = join ", ", @temp;
		croak "$me: Unrecognized values [ $temp ]  in options hash at ";
	}


	my $reasons = [];
	my $logreason = [];
	my $extra = [];
	my $logextra = [];
	my $tmp;

	ARG: foreach my $argrec (@known_arguments)
	{	my $name = $argrec->{name};
		my $excl = $argrec->{excludeOnOption};
		next ARG if $excl && $ohash->{$excl};

		my $value = $urevoc_flags->{$name};
		next ARG unless defined $value;

		#Some stuff needs special handling
		if ( $name eq 'reason' )
		{	if ( ref($value) eq 'ARRAY' )
			{	push @$reasons, @$value;
			} else
			{	push @$reasons, $value;
			}
			next ARG;
		}

		if ( $name eq 'logreason' )
		{	if ( ref($value) eq 'ARRAY' )
			{	push @$logreason, @$value;
			} else
			{	push @$logreason, $value;
			}
			next ARG;
		}

		if ( $name eq 'extra' )
		{	$tmp = $urevoc_flags->{reason};
			die "Flags --reason and --extra are mutually exclusive.\n"
				if defined $tmp;
			if ( ref($value) eq 'ARRAY' )
			{	push @$extra, @$value;
			} else
			{	push @$extra, $value;
			}
			next ARG;
		}
			
		if ( $name eq 'logextra' )
		{	$tmp = $urevoc_flags->{logreason};
			die "Flags --logreason and --logextra are mutually exclusive.\n"
				if defined $tmp;
			if ( ref($value) eq 'ARRAY' )
			{	push @$logextra, @$value;
			} else
			{	push @$logextra, $value;
			}
			next ARG;
		}
			
		#'canned' messages
		my $text =  $argrec->{text};

		$tmp = $urevoc_flags->{reason};
		die "Flags --reason and --$name are mutually exclusive.\n"
			if defined $tmp;

		$text .= " ($value)" if $value;
		push @$reasons, $text;
	}	

	#Combine extra into reason
	if ( $extra && ref($extra) eq 'ARRAY' && scalar(@$extra) )
	{	push @$reasons, @$extra;
	}

	#Make sure that we have a reason after all that
	my $reas_opt = $ohash->{reasonoptional};
	my $usage = $ohash->{usage_sub};
	unless ( $reasons && ref($reasons) eq 'ARRAY' && scalar(@$reasons) )
	{	#No reason given 
		#Is this a fatal error
		unless ( $reas_opt )
		{	#Yes
			if ( $usage && ref($usage) eq 'CODE' )
			{	&$usage;
			} else
			{	print STDERR "Please use --help option for usage help.\n";
			}
			die "You must specify a reason (explicitly or use a 'canned' reason)\n";
		}
	}

	#Handle defaulting of logreason, etc
	$logreason = [ @$reasons ] unless ( scalar(@$logreason) );
	if ( scalar(@$logextra) )
	{	push @$logreason, @$logextra;
	}

	#Convert to string
	$logreason = join "; ", @$logreason;

	#Delete it if not doing logreason
	my $dont_logreason =  $ohash->{nologreason};
	$logreason = undef if $dont_logreason;

	unless ( $ohash->{noconfirm} )
	{	#Have user confirm the resulting reason
		print "The following reason will be used in email to PoCs/etc:\n";
		print "----------------------------------------------------\n";
		foreach (@$reasons) { print "\t$_\n"; }
		print "----------------------------------------------------\n";

		print "Is this reasonable to send? Typing anything but 'y' will abort. ";
		my $ans = <STDIN>;
		chomp $ans;
		unless ( $ans =~ /^\s*y(es?)?\s*$/i )
		{	#Did not type yes
			die "Aborting at user request.\n";
		}
	}

	if ( $dont_logreason )
	{	return $reasons;
	} else
	{	return ($reasons, $logreason);
	}
}

1;
__END__

=head1 NAME

Glue::HPCC::Utils::UserRevocationReasons 

=head1 SYNOPSIS

  use Glue::HPCC::Utils::UserRevocationReasons;
  use Getopt::Long;

  my $urevoc_options = { ... } ;

  sub usage()
  {	my $reasons_help = Glue::HPCC::Utils::UserRevocationReasons->make_help(
		$urevoc_options);
	print STDERR <<EOF;

	...
	$reasons_help
	...
  EOF
  }

  #Main
  my $help;
  my $urevoc_flags = {};
  my $arglist = Glue::HPCC::Utils::UserRevocationReasons->make_arglist(
	$urevoc_flags, $urevoc_options);
  my $res = GetOptions(
	'help!' => \$help,
	@$arglist,
	...
   );

   my ($reasons, $logreason) = 
	Glue::HPCC::Utils::UserRevocationReasons->process_arghash(
		$urevoc_flags, $urevoc_options);

   ...

=head1 DESCRIPTION

=head2 Overview

This is a collection of utility methods for use by command line scripts
like B<delusers> and B<archive_lustre_dirs> and other scripts related
to the revocation of user access to HPC projects and/or clusters.  

In particular, such scripts typically should be sending email to the PoCs
and/or users affected, and such messages will typically include reasons
for the action.  So, unlike most other HPC related scripts, the "reason"
argument is not hidden away in some log somewhere, but is actually exposed
to the users, and so should be a bit more formal in language.  We also
wish to support multi-line reasons in the emails, and would like to have
consistency in the most common reasons that are used.

We also want flexibility to customize the messages, or generate one-off
messages, etc.  And to enable differences between logged message and what
is sent to the users.  And it has to be easy to use.  And I don't want to
rewrite/cut-and-paste code between multiple command line scripts for this.

=head2 What end-user sees

For the end-user, e.g. someone using the B<delusers> command, they will
see that the B<delusers> command has options like

=over 4

=item B<--reason TEXT>

this sets the reason (as displayed to users) to the specified TEXT.  It
can be repeated, in which each occurance adds a new line of TEXT to the
reason.  It B<cannot> be used with any other "reason" arguments below
B<except> B<--logreason> or B<--logextra>.

=item B<--userreq [EXTRA]>

This generates a "canned" message indicating that the user requested to be
removed from the project/cluster.  If EXTRA is given, it will be appended
to the canned message.

=item B<--pocreq [EXTRA]>

This generates a "canned" message indicating that a PoC requested the user be
removed from the project/cluster.  If EXTRA is given, it will be appended
to the canned message.

=item B<--nonumd [EXTRA]>

This generates a "canned" message indicating the user had access revoked
because he was no longer associated with the University.  If EXTRA is given,
it will be appended to the canned message.

=item B<--noglue [EXTRA]>

This generates a "canned" message indicating the user had access revoked
because he no longer has a valid Glue account.  If EXTRA is given,
it will be appended to the canned message.

=item B<--extra EXTRA>

This will append EXTRA to the list of reasons.  It may be given multiple
times, with a new line appended each time.

=item B<--logreason TEXT>

This sets the reason (as included in PDB logs) to TEXT.  It can be given 
multiple times, in which case the various TEXTs are concatenated together
with a semi-colon ";" delimiter.  It B<cannot> be used with B<--logextra>.
If omitted, the reason text (as displayed to PoCs, etc.) will be used to
default the logreason by concantenating with a semi-colon delimiter.

=item B<--logextra EXTRA>

If B<--logreason> was not given, the reason as included in logs is defaulted
from the reason displayed to users.  In that case, one can give B<--logextra>
to append EXTRA to the log reason in addition to the default.  It can be
given multiple times to default multiple lines of EXTRA (with semi-colon delimiter).

=back

The arguments B<--reason>, B<--extra>, B<--logreason>, B<--logextra> may be
repeated.  The arguments B<--userreq>, B<--pocreq>, B<--nonumd>, and B<--noglue>
may B<not> be repeated, but any combination of them can be used, possibly in
addition to B<--extra> and either B<--logreason> or B<--logextra>.

=head2 The routines

There are three basic routines in this package.  They are all class methods
(mainly to reduce namespace pollution).  All routines accept an optional
B<options_hash> as discussed in the next section.  The B<options_hash> controls
the behavior of the routines.  Generally, all routines should be given the 
same options_hash.


First, B<help_text> will return a text string suitable for inclusion in
a commands usage subroutine.  It only takes the optional options_hash to control
which flags are displayed in the help text.

The method B<make_arglist> has a required hash reference argument, B<$urevoc_flags>,
and optional options_hash argument.  This will return a list ref which is suitable
for being part of the argument list passed to the B<GetOptions> method of B<Getopt::Long>.
This will define the various command parameters B<GetOptions> should recognize.  You
are free to add additional options before or afterwards (as long as no name conflicts).
The B<$urevoc_flags> hash ref will be where the various values for the command line
options are stored by B<GetOptions>.

Finally, B<process_arghash> takes both the B<$urevoc_flags> hash ref argument
(same as from B<make_arglist>) and an optional B<options_hash>.  It processes
the values in B<$urevoc_flags>,  which presumably have been populated with values
from B<GetOptions>, and returns a 2 element list (not list ref).  The first element
is the B<reason> generated, a list ref, one element per line, for display to PoCs/users/etc.,
and the second is a text string with the B<logreason> (i.e. the reason to be used for
PDB logs).  The B<logreason> value will be undef if the options_hash instructs
the code that log reasons are not to be generated.  

=head2 The options hash

This is an optional hash ref that can be provided to all of the three routines above
to control their behavior.  It is strongly advised to pass the same options hash to
all three routines.  Not all options are relevant to all routines, but they will 
silently ignore options not relevant to them.  

The following options are recognized:

=over 4

=item B<nologreason> (Boolean)

If set, the B<logreason> and B<logextra> options will be omitted in B<help_text>
and B<make_arglist>, and B<process_arghash> will return undef for logreason.
The default is to include those options and process them.

=item B<noconfirm> (Boolean)

If set, B<process_arghash> will not prompt the user to confirm that the 
proposed reason text looks OK.  The default is to prompt, because most
other HPC command line utilities use a B<--reason> flag to set a terse reason
for use in logs only, and this command wants something which will be shown 
to users.

=item B<reasonoptional> (Boolean)

If set, B<process_arghash> will not complain if no reason is set.  The
default is to require a reason (whether given explicitly with --reason,
or in some combination of 'canned' reasons and --extra), and to raise a
fatal exception otherwise.

=item B<usage_sub> (Coderef)

If given, this should be a code ref to an "usage" subroutine, which
B<process_arghash> might invoke in certain error situations.  If omitted,
a simple "Use --help option to get usage help" will be printed instead.

=back

   my $help;
   my $urevoc_flags = {};
   my $urevoc_arglist = Glue::HPCC::Utils::UserRevocationReasons->make_arglist(
		$urevoc_flags, $urevoc_options);
   use Getopt::Long;
   my $res = GetOptions(
	'help!' => \$help,
	@$urevoc_arglist,
	'verbose!' => \$verbose,
   );

   my ($reason, $logreason) = Glue::HPCC::Utils::UserRevocationReasons->process_arghash(
		$urevoc_flags, $urevoc_options);



