#!/usr/local/bin/perl
#
#Perl wrappers for Lustre lfs command

package Lustre::LFSWrapper;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);


#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $DEFAULT_LFS_CMD = '/usr/bin/lfs'; #Default

#-------------------------------------------------------------------
#	Class methods
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @rw_accessors = qw( 
	verbose 
	dryrun 
	lfs
	filesystem
);

__PACKAGE__->mk_accessors(@rw_accessors);


my @required_parms = qw(lfs
);

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	$class = ref($class) if ref($class);
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my ($arg, $meth, $val);
	RWARG: foreach $arg (@rw_accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	#Warn about unknown arguments
	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;

	unless ( $obj->lfs )
	{	$obj->lfs($DEFAULT_LFS_CMD);
	}
	return;
}

sub _init($)
{	my $obj = shift;

	my ($fld, $meth, $val);
	foreach $fld (@required_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld";
		}
	}

}

#-------------------------------------------------------------------
#	Class methods to Run basic lfs commands
#-------------------------------------------------------------------

sub _noshell_backticks($$@)
#Calls an external command using pipes and forks so no shell gets invoked
#Returns ($err, @out) where $err is the error is the exit status of
#the command, and @out is the list of output returned, line by line.
#If $mode is 0, only STDOUT is returned in @out, 
#If $mode is non-zero, STDERR is dupped onto STDOUT and also returned.
{	my $class = shift;
	my $mode = shift;
	my @cmd = @_;

	my ($err, @out, $PIPE, $res);
	my $chd_excode=254;

	if ( $res = open($PIPE, "-|" )  )
	{	#Parent
		if ( ! defined $res )
		{	my $tmp = join ' ', @cmd;
			die "Pipe to '$tmp' failed: $!";
		}
		@out = <$PIPE>;
		$res = close $PIPE;
		$err = $?;
		if ( $err && ( ($err >> 8) == $chd_excode ) )
		{	#We (probably?) got an exception running exec in
			#child process (e.g. taint issues).  Re-raise it
			my $exc = join '', @out;
			die $exc;
		}
		return ($err, @out);
	} else
	{	#Child
		#Duplicate stderr onto stdout if so requested
		if ( $mode )
		{	unless ( open(STDERR, '>&STDOUT') )
			{	print "Cannot dup stderr to stdout in child";
				die "Cannot dup stderr to stdout in child";
			}
		}
		#Wrap exec in an eval, and exit on error, not die.
		#Otherwise, if _noshell_backticks_ is put in an eval block,
		#an exception raised by exec (e.g. taint issues) will result in
		#BOTH child and parent continuing from the user's eval block
		#See e.g. http://www.perlmonks.org/?node=166538
		eval { exec { $cmd[0] } @cmd; #Not subject to shell escapes
		};
		exit $chd_excode;
	}
}

sub _run_generic_lfs_cmd_always($@)
#Run a generic lfs cmd, returning results
#Does NOT honors dryrun.
#This should only be called directly for commands which do NOT
#modify anything
{	my $self = shift;
	my @args = @_;

	my $cmd = $self->lfs;
	my $mode = 1;
	if ( $self->verbose )
	{	#Verbose mode:
		#Output command before executing it
		my @tmp = ($self->lfs, @args);
		my $tmpcmd = join ' ', @tmp;
		print STDERR "[VERBOSE] $tmpcmd\n";
	}
	my ($err, @out ) = $self->_noshell_backticks($mode, $cmd, @args);
	return ($err, @out);
}

sub run_generic_lfs_cmd($@)
#Run a generic lfs cmd, returning results
#Honors dryrun flag.
#
#Returns list ref of output, line by line on success.
#Returns non-ref error message on error.
{	my $self = shift;
	my @args = @_;

	if ( $self->dryrun )
	{	my @tmp = ( $self->lfs, @args);
		my $tmpcmd = join ' ', @tmp;
		print STDERR "[DRYRUN] $tmpcmd\n";
		return;
	}

	my ($err, @out) = $self->_run_generic_lfs_cmd_always(@args);
	if ( $err )
	{	my $errstr = "Exit code: $err";
		my $output = join "\n", @out;
		$errstr .= "\n$output" if $output;
		return $errstr;
	}
	#return $err if $err;
	return [ @out ];
	
}

sub run_generic_lfs_quota_query($@)
#Runs a lfs quota command to display disk usage/limits
#This does NOT handle lfs quota -t commands
#
#Arguments to this routine are passed straight to lfs
#
#On error, returns non-ref error message/error code
#On success, returns a list ref, one element per line of output.
#Elements are hash refs, with keys:
#	mode: user or group
#	name: user or group name
#	uid: user id (only for mode=user)
#	gid: group id (only for mode=group)
#	filesystem: name of filesystem/OBD_UUID
#	kbytes_used
#	quota_kbytes
#	limit_kbytes
#	grace_blocks
#	files_count
#	quota_files
#	limit_files
#	grace_files
#If called in array context, returns the above list ref and a second
#list ref with error messages from lfs quota commands.
#
#Does NOT honor dryrun
#Do NOT pass -q option, we use headers to parse info
{	my $self = shift;
	my @args = @_;
	my $me = 'run_generic_lfs_quota_query';

	my $lines = $self->run_generic_lfs_cmd(@args);

	#REturn if empty or error
	return $lines unless $lines && ref($lines) eq 'ARRAY';

	my ( $mode, $name, $uid, $gid, $fsystem);
	my ( $kbytes_used, $quota_kbytes, $limit_kbytes, $grace_blocks);
	my ( $files_count, $quota_files, $limit_files, $grace_files);

	my $expecting='diskquotas';
	my @tmp;
	my $tmp;
	my $errors = [];
	
	my $results = [];
	LINE: foreach my $line (@$lines)
	{	chomp $line;

		if ( $expecting eq 'filesystem' )
		{	if ( $line =~ /^Disk quotas for/ )
			{	$expecting = 'diskquotas';
			} else
			{	$line =~ s/\s*$//;
				$fsystem = $line;
				$expecting = 'quota';
				next LINE;
			}
		}

		if ( $expecting eq 'diskquotas' )
		{	croak "$me: Error parsing output; expecting " .
				"'Disk quotas for ...' but got '$line' at "
				unless $line =~ /^Disk quotas for/;
			$line =~ s/^Disk quotas for //;
			@tmp = split /\s+/, $line;
			$mode = $tmp[0];
			$name = $tmp[1];
			$tmp = $tmp[3];
			$tmp =~ s/\)://;
			if ( $mode eq 'user' )
			{	$uid = $tmp;
				$gid = undef;
			} elsif ( $mode eq 'group' )
			{	$gid = $tmp;
				$uid = undef;
			} else
			{	croak "$me: Error parsing output: invalid mode "
					. "expecting 'user' or 'group' but "
					. "got $mode at ";
			}
			$expecting = 'headerline';
			next LINE;
		}

		if ( $expecting eq 'headerline' )
		{	croak("$me: expecting headers, got $line at ")
				unless $line =~ /Filesystem\s+kbytes/;
			$expecting = 'quota';
			next LINE;
		}

		last LINE if ( $line =~ /^Total allocated inode/ );

		#Problem: if filesystem name too long, lfs splits record
		#across 2 lines, first with just filesystem name, 2nd
		#with the quota data.
		#expecting='quota' handles filesystem + quota, or just
		#	filesystem name
		#expecting='quota2' handles quota info (using existing $fsystem
		#	variable)
		#Problem2: is osts are disabled, etc. can get a 
		#	'quotacl ost0 failed.' message.
		if ( $expecting eq 'quota' )
		{	if ( $line =~ /^quotactl/ )
			{	#Error
				push @$errors, $line;
				next LINE;
			}
			$line =~ s/^\s*//;
			@tmp = split /\s+/, $line;
			$fsystem = shift @tmp;
			$expecting = 'quota2';
			$line = join ' ', @tmp; #In case have quota info
			next LINE unless @tmp;
		}

		if ( $expecting eq 'quota2' )
		{	#Note: 'quota' will fall through to here if it
			#included both fsname + quota, with $expecting
			#and $line hacked to be what we expect

			$line =~ s/^\s*//;
			@tmp = split /\s+/, $line;
			unshift @tmp, $fsystem;
			
			$kbytes_used  = $tmp[1];
			$quota_kbytes = $tmp[2];
			$limit_kbytes = $tmp[3];
			$grace_blocks = $tmp[4];
			$files_count  = $tmp[5];
			$quota_files  = $tmp[6];
			$limit_files  = $tmp[7];
			$grace_files  = $tmp[8];

			$kbytes_used = undef 
				if $kbytes_used && $kbytes_used eq '-';
			$quota_kbytes = undef 
				if $quota_kbytes && $quota_kbytes eq '-';
			$limit_kbytes = undef 
				if $limit_kbytes && $limit_kbytes eq '-';
			$grace_blocks = undef 
				if $grace_blocks && $grace_blocks eq '-';
			$files_count = undef 
				if $files_count && $files_count eq '-';
			$quota_files = undef 
				if $quota_files && $quota_files eq '-';
			$limit_files = undef 
				if $limit_files && $limit_files eq '-';
			$grace_files = undef 
				if $grace_files && $grace_files eq '-';

			push @$results,
				{	mode => $mode,
					name => $name,
					gid => $gid, uid => $uid,
					filesystem => $fsystem,
					kbytes_used  => $kbytes_used,
					quota_kbytes => $quota_kbytes,
					limit_kbytes => $limit_kbytes,
					grace_blocks => $grace_blocks,
					files_count  => $files_count,
					quota_files  => $quota_files,
					limit_files  => $limit_files,
					grace_files  => $grace_files,
				};
			$expecting = 'quota';
			next LINE;
		}

		die "Should not reach here ";
	}

	if ( wantarray )
	{	return ($results, $errors);
	} else
	{	return $results;
	}
}

#-------------------------------------------------------------------
#	High level commands
#-------------------------------------------------------------------

sub _group_quota_results_by_uuid($$)
#Takes a list ref as returned by 
# run_generic_lfs_quota_query and collects results into
#a list ref of hash refs with keys (name, space, files) where
#name has value of the name of the filesystem and/or UUID of the MDT/OST
#and space and files are hash refs with keys (used, quota, limit, grace)
#refering to the usage, quota, limit and grace for disk space (for the
#space key) or file counts (for the file key).  For space key, all but
#grace are in kB; for file key, file counts.
{	my $self = shift;
	my $rawlist = shift || [];

	my $newlist = [];

	foreach my $rec (@$rawlist)
	{	my $name = $rec->{filesystem};
		my $space= {
			used => $rec->{kbytes_used}, 
			quota => $rec->{quota_kbytes}, 
			limit => $rec->{limit_kbytes}, 
			grace => $rec->{grace_blocks},
		};
		my $files = { 
			used => $rec->{files_count}, 
			quota => $rec->{quota_files}, 
			limit => $rec->{limit_files}, 
			grace => $rec->{grace_files},
		};
		push @$newlist, 
			{ name=> $name, space=>$space, files => $files };
	}

	return $newlist;
}

sub get_usage_for_user($$;$$)
#Returns usage for specified user $user on specified filesystem/etc
#filesystem defaults to $self->filesystem
#
#User can be username or uid.
#	
#REturns a hash ref with keys (space, files, byost).
#Space and files keys have as values hash refs with keys (for space/files
#respectively)
#	used: the amount of space used (kb)/amount of files used (file count)
#	quota: the quota for space used (kb)/files (file count)
#	limit: the limit on the space used(kb)/files (file count)
#	grace: the grace on space used(blocks)/files (file count)
#
#byost is only present if $byost_flag was given, and is a hash ref keyed
#on the OST/MDT UUID name, with value being a hash ref with keys
#(name, space, files).  Name is the UUID of OST/MDT, and space/files are
#as above (but limited to the specific OST/MDT; usually only used will
#be defined).
#
#On error, either aborts or returns non-ref true value (text describing
#the error)
{	my $self = shift;
	my $user = shift;
	my $filesystem = shift || $self->filesystem;
	my $byost_flag = shift;
	my $me = __PACKAGE__ . '::get_usage_for_user';

	croak "$me: Missing filesystem parameter and no default in lfs wrapper at "
		unless $filesystem;
	croak "$me: Missing parameter user at " unless defined $user;

	return unless defined $user;
	my @lsf_flags = ( 'quota', '-u', $user );
	push @lsf_flags, '-v' if $byost_flag;
	push @lsf_flags, $filesystem;
	
	my $results = $self->run_generic_lfs_quota_query(@lsf_flags);
	croak "$me: Error running lfs quota for user $user: $results at "
		unless $results && ref($results) eq 'ARRAY';

	#Convert to $byost type results
	$results = $self->_group_quota_results_by_uuid($results);
	croak "$me: Error processing results for quota for user $user: $results at "
		unless $results && ref($results) eq 'ARRAY';

	my @temp = grep { $_->{name} eq $filesystem } @$results;
	unless ( @temp )
	{	warn "No data found for user $user and filesystem $filesystem at ";
		return undef;
	}
	if ( scalar(@temp) > 1 )
	{	warn "Multiple data found for user $user and filesystem $filesystem, only using first.";
	}
	my $res = $temp[0];
	my $space = $res->{space};
	my $files = $res->{files};

	@temp = grep { $_->{name} ne $filesystem } @$results;
	my $byost = { map { $_->{name} => $_ } @temp };

	$res = { space => $space, files => $files };
	$res->{byost} = $byost if $byost_flag;

	return $res;
}

sub get_usage_for_group($$;$$)
#Returns usage for specified group $group on specified filesystem/etc
#filesystem defaults to $self->filesystem
#
#User can be groupname or gid.
#	
#REturns a hash ref with keys (space, files, byost).
#Space and files keys have as values hash refs with keys (for space/files
#respectively)
#	used: the amount of space used (kb)/amount of files used (file count)
#	quota: the quota for space used (kb)/files (file count)
#	limit: the limit on the space used(kb)/files (file count)
#	grace: the grace on space used(blocks)/files (file count)
#
#byost is only present if $byost_flag was given, and is a hash ref keyed
#on the OST/MDT UUID name, with value being a hash ref with keys
#(name, space, files).  Name is the UUID of OST/MDT, and space/files are
#as above (but limited to the specific OST/MDT; usually only used will
#be defined).
#
#On error, either aborts or returns non-ref true value (text describing
#the error)
{	my $self = shift;
	my $group = shift;
	my $filesystem = shift || $self->filesystem;
	my $byost_flag = shift;
	my $me = __PACKAGE__ . '::get_usage_for_user';

	croak "$me: Missing filesystem parameter and no default in lfs wrapper at "
		unless $filesystem;
	croak "$me: Missing parameter group at " unless defined $group;

	return unless defined $group;
	my @lsf_flags = ( 'quota', '-g', $group );
	push @lsf_flags, '-v' if $byost_flag;
	push @lsf_flags, $filesystem;
	
	my $results = $self->run_generic_lfs_quota_query(@lsf_flags);
	croak "$me: Error running lfs quota for group $group: $results at "
		unless $results && ref($results) eq 'ARRAY';

	#Convert to $byost type results
	$results = $self->_group_quota_results_by_uuid($results);
	croak "$me: Error processing results for quota for group $group: $results at "
		unless $results && ref($results) eq 'ARRAY';

	my @temp = grep { $_->{name} eq $filesystem } @$results;
	unless ( @temp )
	{	warn "No data found for group $group and filesystem $filesystem at ";
		return undef;
	}
	if ( scalar(@temp) > 1 )
	{	warn "Multiple data found for group $group and filesystem $filesystem, only using first.";
	}
	my $res = $temp[0];
	my $space = $res->{space};
	my $files = $res->{files};

	@temp = grep { $_->{name} ne $filesystem } @$results;
	my $byost = { map { $_->{name} => $_ } @temp };

	$res = { space => $space, files => $files };
	$res->{byost} = $byost if $byost_flag;

	return $res;
}
#
1;
__END__

=head1 NAME

Lustre::LFSWrapper

=head1 SYNOPSIS

  use Lustre::LFSWrapper;

  my $lfs = Lustre::LFSWrapper->new( lfs => '/usr/bin/lfs' );

  my $lustre = "/lustre";
  my $user = 'joey';
  my $group = 'friends';

  my $jusage = $lfs->get_usage_for_user($user, $lustre);
  my $fusage = $lfs->get_usage_for_group($group, $lustre);

  my $jspace = $jusage->{space};
  print "$user is using ", $jspace->{used}, " kB\n";
  my $ffiles = $jfsage->{files};
  print "$group has ", $ffiles->{used}, " files/inodes used\n";

  ...


=head1 DESCRIPTION

This is a wrapper for the Lustre B<lfs> command.  One can set the path
to the B<lfs> command when instantiating the object or via accessors.
It is implemented as wrappers to the actual B<lfs> command, not API calls.
This means it does not require rebuilding with Lustre upgrades, although
might require updating code if the B<lfs> interface changes.


=head2 Constructor and accessors, etc.

The first step in using this class is to instantiate an instance of the
class with the constructor B<new>.  The constructor does not require any
explicit arguments, but optionally will accept any of the following:

=over 4

=item B<verbose>: Boolean, if set, operate in verbose mode. Default false.

=item B<dryrun>: Boolean, if set, operate in dry-run mode. Default false.

=item B<lfs>: The path to the lfs command. Default is '/usr/bin/lfs'.

=item B<filesystem>: The default path to the lustre filesystem.  No default.

=back

All of the above four options to new have corresponding accessors/mutators
of the same name which can be used to query/set their values after 
instantiation.

In B<dryrun> mode, no commands which will actually make changes to the
filesystem will be run, but instead the command which would have been run
will just be printed out to stderr, for testing/debugging.  Query-type commands
which do not make any actual changes _will_ run as normal.

In B<verbose> mode, the actual B<lfs> commands will be printed out before
being executed (actually, this has some intelligence and knows about the
B<dryrun> flag, so it won't print commands twice in B<dryrun> mode).

The B<lfs> mutator allows one to set the path to the actual lfs command
to be used.

The B<filesystem> data member is the default path to the lustre filesystem,
which will be used in higher level commands if no lustre filesystem is
explicity given.  There is no default for this data member, which means that
if it is not set, one must explicitly provide the path to the lustre filesystem
to higher level methods.


=head2 Higher level methods

The B<get_usage_for_user($user, $filesystem, $byost)> method will query 
the lustre usage for a specified user.  The B<$user> value can be either
the username or the uid, and is the only required explicit parameter.
The B<$filesystem> value is the path to the lustre filesystem.  If it omitted,
the B<filesystem> data member of the instance will be used if it was set.
It is required that either the B<$filesystem> parameter be set in the
function call, or that the B<filesystem> data member of the instance be set,
and the method will raise an exception otherwise.

The B<$byost> parameter is a boolean flag.  If set, the method will get and
return usage results by individual OST (and MDT) UUIDs.

On non-fatal errors, the method will return a non-reference error string.
On success, the method will return a hash reference with keys 'space',
'files', and (if the $byost flag was set) 'byost'.  The values corresponding
to these keys are again hash references, as described below:

The 'space' and 'files' hash references contain the usage statistics for this
user with respect to disk space (e.g. kB used) and inodes, respectively.
They will have keys:

=over 4

=item B<used>: the amount of space used (in kB) or the number of files/inodes used (file count)

=item B<quota>: The quota on the space used (in kB) or inodes (file count)

=item B<limit>: The limit on the space used (in Kb) or inodes (file count)

=item B<grace>: The grace setting on the space used (in Kb) or inodes (file count)

=back

If the B<$byost> flag was given and set, there will at the top level be a 
'byost' hash key in addition to the 'space' and 'files' hash keys.  This
will return another hash reference, keyed on the OST/MDT UUID name.  Each
of these is in turn another hash reference, which can have a 'space' and/or
'files' key (typically there will only be one of the two).  These 'space' and/or
'files' keys will have hash references like the top level 'space' and 'files'
hash keys, except that the values returned are only for that OST/MDT.

The B<get_usage_for_group($group, $filesystem, $byost)> method is similar to
B<get_usage_for_user>, except that it gets the usage by unix group.  The
B<$group> parameter can give either the group name or gid.  The B<$filesystem>
and B<$byost> parameters are the same as for the B<get_usage_for_user> case,
and the usage hash returned are also the same, except that the report usage
for the Unix group, not the user.

=head1 EXPORT

None.  OO interface only.

=head1 SEE ALSO


=head1 AUTHOR

Tom Payerle, payerle@umd.edu

=head1 COPYRIGHT AND LICENSE

Copyright 2014-2017, University of Maryland

This library is free software; you can redistribute it and/or modify it
under the same terms as Perl itself, either Perl version 5.8.4 or, at
your option, any later version of Perl 5 you may have available.
