#!/usr/local/bin/perl
#
#Perl wrappers for Slurm sacct cmd

package Slurm::Sacct;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);


#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

my $VERBOSE;
#my $SACCT_CMD = 'sacct'; #Default to sacct in PATH
my $SACCT_CMD = '/usr/local/bin/sacct'; #Default to STD GLUE sacct

#-------------------------------------------------------------------
#	Class methods
#-------------------------------------------------------------------

sub verbose($;$)
{	my $class = shift;
	my $new = shift;
	if ( defined $new )
	{	$VERBOSE = $new?1:0;
	}
	return $VERBOSE;
}

sub sacct($;$)
{	my $class = shift;
	my $new = shift;
	if ( $new )
	{	$SACCT_CMD = $new;
	}
	return $SACCT_CMD;
}

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @rw_accessors = qw(
);

my @sacct_fields_in_order = qw(
	Account
	AllocCPUS
	AssocID
	AveCPU
	AveCPUFreq
	AveDiskRead
	AveDiskWrite
	AvePages
	AveRSS
	AveVMSize
	BlockID
	CPUTime
	CPUTimeRAW
	Cluster
	Comment
	ConsumedEnergy
	DerivedExitCode
	Elapsed
	Eligible
	End
	ExitCode
	GID
	Group
	JobID
	JobIDRaw
	JobName
	Layout
	MaxDiskRead
	MaxDiskReadNode
	MaxDiskReadTask
	MaxDiskWrite
	MaxDiskWriteNode
	MaxDiskWriteTask
	MaxPages
	MaxPagesNode
	MaxPagesTask
	MaxRSS
	MaxRSSNode
	MaxRSSTask
	MaxVMSize
	MaxVMSizeNode
	MaxVMSizeTask
	MinCPU
	MinCPUNode
	MinCPUTask
	NCPUS
	NNodes
	NTasks
	NodeList
	Partition
	Priority
	QOSRAW
	ReqCPUFreq
	ReqCPUS
	ReqMem
	Reserved
	ResvCPU
	ResvCPURAW
	Start
	State
	Submit
	Suspended
	SystemCPU
	Timelimit
	TotalCPU
	UID
	User
	UserCPU
	WCKey
	WCKeyID
);

my @ro_accessors = (@sacct_fields_in_order);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);


my @required_parms = qw(
);

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	$class = ref($class) if ref($class);
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my ($arg, $meth, $val);
	RWARG: foreach $arg (@rw_accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	ROARG: foreach $arg (@ro_accessors)
	{	next ROARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next ROARG unless defined $val;
		$meth = $arg;
		$obj->set($meth,$val);
	}


	#Warn about unknown arguments
	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;
	return;
}

sub _init($)
{	my $obj = shift;

	my ($fld, $meth, $val);
	foreach $fld (@required_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld";
		}
	}

}

#-------------------------------------------------------------------
#	Class methods to Run basic sacct commands
#-------------------------------------------------------------------

sub _noshell_backticks($$@)
#Calls an external command using pipes and forks so no shell gets invoked
#Returns ($err, @out) where $err is the error is the exit status of
#the command, and @out is the list of output returned, line by line.
#If $mode is 0, only STDOUT is returned in @out, 
#If $mode is non-zero, STDERR is dupped onto STDOUT and also returned.
{	my $class = shift;
	my $mode = shift;
	my @cmd = @_;

	my ($err, @out, $PIPE, $res);

	if ( $res = open($PIPE, "-|" )  )
	{	#Parent
		if ( ! defined $res )
		{	my $tmp = join ' ', @cmd;
			die "Pipe to '$tmp' failed: $!";
		}
		@out = <$PIPE>;
		$res = close $PIPE;
		$err = $?;
		return ($err, @out);
	} else
	{	#Child
		#Duplicate stderr onto stdout if so requested
		if ( $mode )
		{	unless ( open(STDERR, '>&STDOUT') )
			{	print "Cannot dup stderr to stdout in child";
				die "Cannot dup stderr to stdout in child";
			}
		}

		exec { $cmd[0] } @cmd; #Not subject to shell escapes
		die "Failed to exec '" . (join ' ', @cmd) . "' in child.\n";
	}
}

sub _run_generic_sacct_cmd_always($@)
#Run a generic sacct cmd, returning results
#Does NOT honors dryrun.
#This should only be called directly for commands which do NOT
#modify the database
{	my $class = shift;
	my @args = @_;

	my $cmd = $class->sacct;
	my $mode = 1;
	if ( $class->verbose )
	{	#Verbose mode:
		#Output command before executing it
		my @tmp = ($class->sacct, @args);
		my $tmpcmd = join ' ', @tmp;
		print STDERR "[VERBOSE] $tmpcmd\n";
	}
	my ($err, @out ) = $class->_noshell_backticks($mode, $cmd, @args);
	return ($err, @out);
}

sub run_generic_sacct_cmd($@)
#Run a generic sacct cmd, returning results
#No dryrun mode, as sacct is view only anyway.  Also, no --readonly flag.
#
#Returns list ref of output, line by line on success.
#Returns non-ref error message on error.
{	my $class = shift;
	my @args = @_;

	my ($err, @out) = $class->_run_generic_sacct_cmd_always(@args);
	if ( $err )
	{	my $errstr = "Exit code: $err";
		my $output = join "\n", @out;
		$errstr .= "\n$output" if $output;
		return $errstr;
	}
	return $err if $err;
	return [ @out ];
	
}

sub run_generic_sacct_list_command($@)
#Runs a sacct list command, returns output as a list ref of list refs
# of Slurm::Sacct objects corresponding to items returned from the command
#
#On error, returns non-ref error message/error code
#
#Appends --parsable2 and --noheader to the command
#Does NOT honor dryrun (as --readonly)
{	my $class = shift;
	my @cmd = @_;

	my @fields = @sacct_fields_in_order;
	my $fields_str = join ",", @fields;
	push @cmd, '--parsable2', "--format=$fields_str";
	my $lines = $class->run_generic_sacct_cmd(@cmd);

	return $lines unless $lines && ref($lines) eq 'ARRAY';

	my $results = [];
	my $line = shift @$lines;
	chomp $line;
	my @hfields =  split /\|/, $line;
	LINE: foreach $line (@$lines)
	{	chomp $line;
		next LINE unless $line =~ /\|/;
		my @values = split /\|/, $line;
		my $rec = { map { $hfields[$_] => $values[$_] }
				(0 .. scalar(@hfields) -1 ) };
		my $obj = $class->new(%$rec);
		push @$results, $obj;
	}

	return $results;
}

#-------------------------------------------------------------------
#	High level commands
#-------------------------------------------------------------------

sub sacct_list($@)
#Gets sacct output.
#This is basically a wrapper to 
#run_generic_sacct_list_command($@)
#
#Input parameters are arguments to sacct command.  See man page
#They should be of the form
#-F value
#or
#--flag=value
#We do some massaging to make into that format, e.g.
#	flag, value => --flag=value
#	'F value' => -F, value
#
#We also prepend --allusers to the argument list.
#Only other valid parameters to sacct not in above format are:
#	--brief, --completion, --duplicates, --helpformat, --help, --long,
#	--allclusters, --noheader, --parsable, --parsable2, --truncate,
#	--usage, --verbose, --version, and --allocations.
#
#Of the above, only -X/--allocations seems of interest, so we handle 
#that as a special case.
#
#On success, returns a list ref of Slurm::Sacct objects corresponding
#to the output of the sacct command.  Possibly an empty list.
#
#On error, either aborts or returns non-ref true value (text describing
#the error)
{	my $class = shift;
	my @where = @_;
	my $me = 'sacct_list';

	my @newargs = ( '--allusers', );
	my ($arg, $val, @tmp);
	ARG: while ( $arg = shift @where )
	{	#Strip leading -'s, we will add back proper number later
		$arg =~ s/^\s*\-*\s*//;

		#Handle this as a special case
		if ( $arg eq 'X' || $arg eq 'allocations' )
		{	$arg = '--allocations';
			push @newargs, $arg;
			next ARG;
		}

		#Do we have a proper 1 char arg
		if ( length($arg) eq 1 )
		{	$arg = "-$arg";
			$val = shift @where;
			push @newargs, $arg, $val;
			next ARG;
		}

		#do we have f=val
		if ( $arg =~ /^.\s*=\s*/ )
		{	@tmp = split /\s*=\s*/, $arg;
			$arg = $tmp[0]; $arg = "-$arg";
			$val = $tmp[1];
			push @newargs, $arg, $val;
			next ARG;
		}

		#do we have "f val"
		if ( $arg =~ /^.\s+/ )
		{	@tmp = split /\s+/, $arg;
			$arg = $tmp[0]; $arg = "-$arg";
			$val = $tmp[1];
			push @newargs, $arg, $val;
			next ARG;
		}

		#Do we have a proper flag=value
		if ( $arg =~ /^\w\w+\s*=\s*/ )
		{	$arg = "--$arg";
			$arg =~ s/\s*=\s*/=/;
			push @newargs, $arg;
			next ARG;
		}

		#Do we have a  "flag value"
		if ( $arg =~ /^\w\w+\s+/ )
		{	@tmp = split /\s+/, $arg;
			$arg = $tmp[0]; $arg = "--$arg";
			$val = $tmp[1];
			push @newargs, "$arg=$val";
			next ARG;
		}

		#Do we have a  "flag",  "value"
		if ( $arg =~ /^\w\w+$/ )
		{	$arg = "--$arg";
			$val = shift @where;
			push @newargs, "$arg=$val";
			next ARG;
		}


		die "Improper looking argument '$arg' at ";
	}
	my $results = $class->run_generic_sacct_list_command(@newargs);
	return $results;
}
	
1;
__END__

=head1 NAME

Slurm::Sacct

=head1 SYNOPSIS

  use Slurm::Sacct;

  my $Sacct = 'Slurm::Sshare';
  $Sacct->verbose(1); #Print sacct commands as running them

  my $jobs = $Sacct->sacct_list(-S => '2014-01-01', -E => '2014-01-10' );

  foreach $job (@$jobs)
  {	print $job->JobID, $job->User, $job->Partition,
		$job->NCPUS, $job->Elapsed, "\n";
  }

  ...


=head1 DESCRIPTION

This is a wrapper for the Slurm B<sacct> command.  Some class methods allow
one to set the path to the B<sacct> command, the verbosity, and to run
B<sacct> commands.

The primary method is the class method B<sacct_list> which runs the
B<sacct> command with appropriate arguments, and returns a list ref
of the rows returned by the B<sacct> command as instances of this class.
The B<sacct_list> method takes key => value pairs as arguments, and
currently recognized the following keys:

=over 4

=item B<clusters>: A list of clusters to run the command on.  At least one cluster is REQUIRED.

=item B<accounts>: A list of accounts to report on.  The B<--accounts> argument to the B<sacct> command.

=item B<users>: A list of users to report on.  The special value B<ALL>, if given as the only user, will result in the B<-all> flag passed to B<sacct>, otherwise this becomes the B<--users> argument to B<sacct>.

=back

For all the lists above, you can give either a list ref or a scalar CSV string.
At least one cluster must be given.  

The instances have the following read-only accessors:

=over 4

=item B<cluster>: 

The cluster the result is from.  This is NOT part of the standard B<sacct>
output, but we run the B<sacct> command once for each cluster specified (and
we require at least one cluster to be specified), and this reflects the 
cluster the record was from.

=item B<account>: The account.

=item B<user>: The user.

=item B<raw_shares>: The raw shares assigned to the user/account.

=item B<normalized_shares>: 

The shares assigned to the user/account normalized to the total number of
shares.

=item B<raw_usage>: 

The number of cpu-seconds for all jubs charged against the account by the
user (or all users of the account if users field is empty).   This will 
decay over time if B<PriorityDecayHalfLife> defined in slurm configuration.

=item B<normalized_usage>

The usage normalized to the total number of cpu-seconds of all jobs run on
the cluster, subject to the B<PriorityDecayHalfLife> when defined.

=item B<effective_usage>

The effective usage augments the normalized usage to account for usage from
sibling accounts.

=item B<fairshare>: The fair share factor

=item B<grpcpumins>: The CPU-minutes limit on the account.

=item B<cpurunmins>: 

The number of CPU minutes accumulated by jobs currently running against the
account.

=back

=head2 Miscellaneous class methods

=over 4

=item B<sacct>:

The path to the Slurm B<sacct> command.  Defaults to just "sacct", i.e. 
will look for it in your path.

=item B<verbose>:

If this is set, the module works in B<verbose> mode.  Basically, the various
B<sacct> commands will be printed out just before they are executed.  


=item B<$sacct->run_generic_sacct_cmd(@sacct_args)>

This runs B<sacct> with the arguments given in B<@sacct_args>,

=item B<$sacct->run_generic_sacct_list_command(@sacct_args)>

This behaves much like B<run_generic_safe_sacct_cmd>, except that it also
adds the flags B<-parseable2> and B<-noheader> to the argument list.  This
is intended to wrap commands to collect information using B<sacct> and
present to your script in a machine-usable fashion. 

=back

The B<run*> methods above are the methods to actually run B<sacct> commands.  
These use forking and pipes to avoid invoking an Unix shell for additional 
safety (i.e., shell expansions will not work).  The intent is to be able to 
programatically issue B<sacct> commands and get the results.  The functions
take the B<sacct> command, with each word as a separate argument to
the function.  The actual B<sacct> command should NOT be given, as we will
use the one set with the B<sacct> class method.

The return value for all these commands is either:

=over 4

=item A list reference, containing the output of the command with each element being one line of output, if the command was successful

=item A non-reference value containing an error message if the command failed.

=back

B<NOTE:> Since the B<sacct> command only views information, there is 
NO B<dryrun> option.  There is a B<verbose> class method, and if that class
data member is set, the actual B<sacct> commands will be printed before
being executed.

=head1 EXPORT

None.  OO interface only.

=head1 SEE ALSO

Slurmdb

Slurm::Sacctmgr

=head1 AUTHOR

Tom Payerle, payerle@umd.edu

=head1 COPYRIGHT AND LICENSE

Copyright 2014, University of Maryland

This library is free software; you can redistribute it and/or modify it
under the same terms as Perl itself, either Perl version 5.8.4 or, at
your option, any later version of Perl 5 you may have available.
