#!/usr/local/bin/perl
#
#Perl wrappers for Slurm scontrol command

package Slurm::Scontrol;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);


#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @rw_accessors = qw(
	dryrun
	verbose
);

my @ro_accessors = qw(
	scontrol
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);


my @required_parms = qw(
	scontrol
);

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my ($arg, $meth, $val);
	RWARG: foreach $arg (@rw_accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	ROARG: foreach $arg (@ro_accessors)
	{	next ROARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next ROARG unless defined $val;
		$meth = $arg;
		$obj->set($meth,$val);
	}


	#Warn about unknown arguments
	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;

	my ($tmp);
	$tmp = $obj->scontrol;
	#$obj->set('scontrol', 'scontrol') unless $tmp; #try to find in path

	#Standard Glue path
	$obj->set('scontrol', '/usr/local/slurm/bin/scontrol') unless $tmp; 

	return;
}

sub _init($)
{	my $obj = shift;

	my ($fld, $meth, $val);
	foreach $fld (@required_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld";
		}
	}

}

#-------------------------------------------------------------------
#	Basic scontrol commands
#-------------------------------------------------------------------

sub _noshell_backticks($$@)
#Calls an external command using pipes and forks so no shell gets invoked
#Returns ($err, @out) where $err is the error is the exit status of
#the command, and @out is the list of output returned, line by line.
#If $mode is 0, only STDOUT is returned in @out, 
#If $mode is non-zero, STDERR is dupped onto STDOUT and also returned.
{	my $obj = shift;
	my $mode = shift;
	my @cmd = @_;

	my ($err, @out, $PIPE, $res);

	if ( $res = open($PIPE, "-|" )  )
	{	#Parent
		if ( ! defined $res )
		{	my $tmp = join ' ', @cmd;
			die "Pipe to '$tmp' failed: $!";
		}
		@out = <$PIPE>;
		$res = close $PIPE;
		$err = $?;
		return ($err, @out);
	} else
	{	#Child
		#Duplicate stderr onto stdout if so requested
		if ( $mode )
		{	unless ( open(STDERR, '>&STDOUT') )
			{	print "Cannot dup stderr to stdout in child";
				die "Cannot dup stderr to stdout in child";
			}
		}

		exec { $cmd[0] } @cmd; #Not subject to shell escapes
		die "Failed to exec '" . (join ' ', @cmd) . "' in child.\n";
	}
}

sub _run_generic_scontrol_cmd_always($@)
#Run a generic scontrol cmd, returning results
#Does NOT honors dryrun.
#This should only be called directly for commands which do NOT
#modify the database
{	my $obj = shift;
	my @args = @_;

	my $cmd = $obj->scontrol;
	my $mode = 1;
	if ( $obj->verbose )
	{	#Verbose mode:
		#Output command before executing it
		my @tmp = ($obj->scontrol, @args);
		my $tmpcmd = join ' ', @tmp;
		print STDERR "[VERBOSE] $tmpcmd\n";
	}
	my ($err, @out ) = $obj->_noshell_backticks($mode, $cmd, @args);
	return ($err, @out);
}

sub run_generic_scontrol_cmd($@)
#Run a generic scontrol cmd, returning results
#Honors dryrun
#Returns list ref of output, line by line on success.
#Returns non-ref error message on error.
{	my $obj = shift;
	my @args = @_;

	if ( $obj->dryrun )
	{	unshift @args, $obj->scontrol;
		my $cmd = join ' ', @args;
		print STDERR "[DRYRUN] $cmd\n";
		return [];
	}
	my ($err, @out) = $obj->_run_generic_scontrol_cmd_always(@args);
	if ( $err )
	{	my $errstr = "Exit code: $err";
		my $output = join "\n", @out;
		$errstr .= "\n$output" if $output;
		return $errstr;
	}
	#return $err if $err;
	return [ @out ];
	
}

sub run_generic_safe_scontrol_cmd($@)
#Run a "safe" generic scontrol cmd, returning results
#which should not modify accounting info.
#Does NOT honor dryrun, as command is "safe"
#
#NOTE: the safeness of the command is NOT verified.
#If you call this method, you are promising that the command supplied is safe.
#
#Returns list ref of output, line by line on success.
#Returns non-ref error message on error.
{	my $obj = shift;
	my @args = @_;

	#push @args, '--readonly';
	my ($err, @out) = $obj->_run_generic_scontrol_cmd_always(@args);
	if ( $err )
	{	my $errstr = "Exit code: $err";
		my $output = join "\n", @out;
		$errstr .= "\n$output" if $output;
		return $errstr;
	}
	#return $err if $err;
	return [ @out ];
}

sub _process_scontrol_show_record($$)
#Processes the concatenated record from scontrol show command
{	my $obj = shift;
	my $line = shift || '';

	my @fldvals = split /\s+/, $line;

	my $rec = {};
	my $lastfld;
	FLDVAL: foreach my $fldval (@fldvals)
	{	unless ( $fldval =~ /=/ )
		{	#We got something w/out an = sign; tack it on
			#to the previous fld
			if ( $lastfld )
			{	$rec->{$lastfld} .= " $fldval";
			} else
			{	die "Bad output from scontrol show '$line' at ";
			}
			next FLDVAL;
		}
		my ( $fld, $val) = split /=/, $fldval, 2;
		warn "Missing value for $fld at " unless defined $val;
		warn "Duplicate field $fld at " if exists $rec->{$fld};
		$rec->{$fld} = $val;
		$lastfld = $fld;
	}

	return $rec;
}

sub run_generic_scontrol_show_command($@)
#Runs a scontrol show command, returns output as a list ref of hash refs
#of the fields (key=>values) in order. 
#Standard scontrol list format is assumed:
#	blank lines ignored
#	lines starting with white space are considered a continuation of
#		the previous line
#	output is KEY=VALUE pairs on a (possibly continued) line
#
#On error, returns non-ref error message/error code
#
#Does NOT honor dryrun (as is readonly)
{	my $obj = shift;
	my @cmd = @_;

	my $lines = $obj->run_generic_safe_scontrol_cmd(@cmd);

	return $lines unless $lines && ref($lines) eq 'ARRAY';

	my $results = [];
	my $curline='';
	LINE: foreach my $line (@$lines)
	{	chomp $line;
		next LINE unless $line;

		if ( $line =~ /^\s+/ )
		{	#Got a continuation line
			$line =~ s/^\s*//;
			$line =~ s/\s*$//;
			$curline .= " $line" if $line;
			next LINE;
		}

		#Not a continuation line

		#Process old line if any first
		if ( $curline )
		{	my $rec = $obj->_process_scontrol_show_record($curline);
			push @$results, $rec;
		}

		#Start work on new line; just store as might be continued
		$line =~ s/\s*$//;
		$curline = $line;
		next LINE;
	}
	#Process last line
	if ( $curline )
	{	my $rec2 = $obj->_process_scontrol_show_record($curline);
		push @$results, $rec2;
	}

	return $results;
}

#Higher level functions

sub expand_hostnames($$)
#Takes a slurm nodename description (e.g. compute-f10-[1-5],compute-f09-7'
#and returns the expanded list, e.g. compute-f10-1, compute-f10-2, compute-f10-3, ...
#
#Description can be a simple string, or a list ref of strings.  In latter case,
#each string is expanded separately, and the total is returned.
{	my $self = shift;
	my $nodedesc = shift;

	return unless $nodedesc;
	$nodedesc = [ $nodedesc ] unless ref($nodedesc) eq 'ARRAY';

	my $nodes = [];
	foreach my $ndesc (@$nodedesc)
	{	my $results = $self->run_generic_safe_scontrol_cmd('show',
			'hostnames', $ndesc);
		unless ( $results && ref($results) eq 'ARRAY' )
		{	die "Error expanding $ndesc: $results at ";
		}
		chomp @$results;
		push @$nodes, @$results;
	}
	return $nodes;
}

1;
__END__

=head1 NAME

Slurm::Scontrol - Perl wrapper to scontrol command

=head1 SYNOPSIS

  use Slurm::Scontrol;

  my $scontrol = Slurm::Scontrol->new( scontrol=>'path to scontrol command');

  my $scontrol->dryrun(1);
  my $scontrol->verbose(1);

  my $results;

  #This one won't run in dryrun mode
  $results = $scontrol->run_generic_scontrol_cmd('create', 
		'Reservation=my-reservation', 'Account=my-account',
		'Nodes=my-node', 'StartTime=2028-12-31', 
		'Duration=1');
  if ( $results && ref($results) eq 'ARRAY' )
  {	print "Results of create reservation are: ",
		(join "\n", @$results), "\n\n";
  } else
  {	print STDERR "Error running create cluster: $results\n";
  }

  #This one will run in dryrun mode. 
  $scontrol->run_generic_safe_scontrol_cmd('show', 'nodes');
  if ( $results && ref($results) eq 'ARRAY' )
  {	print "Results of show nodes (human readable) are: ",
		(join "\n", @$results), "\n\n";
  } else
  {	print STDERR "Error running list cluster: $results\n";
  }

  #This also runs in dryrun mode, but does some parsing of the results
  #into a list ref of hash refs
  #are added to the command silently
  $scontrol->run_generic_scontrol_list_command('show','nodes');
  if ( $results && ref($results) eq 'ARRAY' )
  {	print "CPUs on nodes are:";
	foreach my $rec (@$results)
	{	$node = $rec->{NodeName}; 
		$cpus = $rec->{CPUTot};
		print "$node : $cpus\n";
	}
  } else
  {	print STDERR "Error running show node: $results\n";
  }

  #Or the same as above, in more OO fashion
  use Slurm::Scontrol::Node;
  $results = Slurm::Scontrol::Node->scontrol_show_node($scontrol);
  if ( $results && ref($results) eq 'ARRAY' )
  {	print "CPUs on nodes are:";
	foreach my $rec (@$results)
	{	$node = $rec->NodeName; 
		$cpus = $rec->CPUTot;
		print "$node : $cpus\n";
	}
  } else
  {	print STDERR "Error running show node: $results\n";
  }

  ...


=head1 DESCRIPTION

B<Slurm::Scontrol> and subclasses provide wrappers to the Slurm B<scontrol> 
command.  

=head2 CONSTRUCTOR and DATA members

Interface is object orientated, mainly to reduce namespace pollution.
The constructor B<new> takes key => value pairs to set the data members, 
which are:

=over 4

=item B<scontrol>:

The path to the Slurm B<scontrol> command.  Can only be set at time of
construction.  Defaults to just "scontrol", i.e. will look for it in your
path.

=item B<dryrun>:

If this is set, the module works in B<dryrun> mode.  See below for details.
Default is false.

=item B<verbose>:

If this is set, the module works in B<verbose> mode.  Basically, the various
B<scontrol> commands will be printed out just before they are executed.  
Commands which do not actually run due to B<dryrun> mode will not be printed
out due to B<verbose> mode, but will get printed out due to being B<dryrun>
mode (i.e. they will get printed once, and only once).
Default is false.

=back

Accessors exist for all of the above.
The B<scontrol> field is read-only, and can only be set at time of
construction.  B<dryrun> and B<verbose> are read-write, and can be set
at any time by providing a defined boolean value as an explicit argument
to the accessor.  Usual perl boolean semantics apply, except that you
cannot provide an undef value to the accessor to unset the fields (as that
would be interpretted as a read accessor, not a mutator).

The main methods are the methods to actually run B<scontrol> commands.  These
use forking and pipes to avoid invoking an Unix shell for additional safety
(i.e., shell expansions will not work).  The intent is to be able to 
programatically issue B<scontrol> commands and get the results.  The functions
take the B<scontrol> command, with each word as a separate argument to
the function.  The actual B<scontrol> command should NOT be given, as we will
use the one provided to the constructor.

The return value for all these commands, with the exception of
B<run_generic_scontrol_show_command>, is either:

=over 4

=item A list reference, containing the output of the command with each element being one line of output, if the command was successful

=item A non-reference value containing an error message if the command failed.

=back

For B<run_generic_scontrol_show_command>, the output on errors is the
same as above.  On success, it parses the output of the show command and
returns a list ref of hash refs with the key=> value values returned by
the show command.

=head2 RUN-type commands

The "run" commands are:

=over 4

=item B<$scontrol->run_generic_scontrol_cmd(@scontrol_args)>

This runs B<scontrol> with the arguments given in B<@scontrol_args>,
UNLESS we are in dryrun mode.  In dryrun mode, the command is only printed
out (with a B<[DRYRUN]> prefix).

=item B<$scontrol->run_generic_safe_scontrol_cmd(@scontrol_args)>

This behaves much like B<run_generic_scontrol_cmd>, except that it will
always run, even in dryrun mode.  
It is intended for commands which do B<NOT> change anything in Slurm.
Because the command does not change anything, it can safely run in B<dryrun>
mode.  This is useful for B<scontrol> commands that your script runs to
gather information.  B<NOTE>: The code does not verify that the command does
not update anything in Slurm; it blindly trusts the caller to have done so.

=item B<$scontrol->run_generic_scontrol_show_command(@scontrol_args)>

This behaves much like B<run_generic_safe_scontrol_cmd>, except that it tries
to parse the output.  Note that the 'list' part of the command is NOT assumed
and still needs to be provided.  It returns a list ref of hash refs of the
key=value pairs returned by the scontrol command.

=back

In addition, the method B<expand_hostnames> will take a node description
(e.g. 'compute-f10-[1-2],compute-f09-7' and return a list ref with the
individual node names (e.g. [ 'compute-f10-1', 'compute-f10-2', 'compute-f09-7' ]).
The node description can be a string or a list ref of strings (in the 
latter case, each element get expanded via C<scontrol show hostname>
individually, and results appended to the list returned).


=head2 DRYRUN MODE

For testing of scripts, it is nice to have a B<dryrun> mode in which the
user can see what commands would be executed, but without the commands actually
executing.  To do this, have a B<dryrun> flag in your code, and set
the B<dryrun> data method of your B<Slurm::Scontrol> instance.  This can
be done either at time of construction or afterwards with the 
B<dryrun> mutator.  B<NOTE:> Although standard Perl boolean semantics are
honored, you cannot use the B<dryrun> mutator to unset the data method; as

  $scontrol->dryrun(undef);

will be considered a request to get the value of B<dryrun>, not to set it
to undef (false).  Use instead something like:

  $scontrol->dryrun(0);

Actually, only commands which modify Slurm should not run; running
show commands and the like are  harmless and you normally want those to
run (otherwise your code won't get far enough to reach the parts that really
need debugging).  So we provide a B<run_generic_scontrol_cmd> method which
honors the B<dryrun> setting, and will only actually execute the command
when B<dryrun> is false, as well as B<run_generic_safe_scontrol_cmd>, which
does B<NOT> honor the B<dryrun> setting, and always runs.

The B<run_generic_scontrol_list_command> is a variant of the 
B<run_generic_safe_scontrol_cmd>, intended for the somewhat common operation
of requesting information from B<scontrol>.  It tries to parse the output
as well.

=head2 VERBOSE MODE

If the B<verbose> data method is set, either at construction time or via
the B<verbose> mutator afterwards, the B<scontrol> commands will be printed
out to STDERR before being executed.  This is again useful for debugging.

B<Note:> If both B<verbose> and B<dryrun> modes are set, only the "safe"
commands are printed out due to the instance being in B<verbose> mode.
The "unsafe" commands get intercepted by B<dryrun> mode first, before 
B<verbose> mode can print them out.  But the B<dryrun> mode prints them out
anyway, so the result is the command gets printed out exactly once, which is
what you probably want anyway.

=head1 EXPORT

None.  OO interface only.

=head1 SEE ALSO

Slurmdb

Slurm::Scontrol::Node

=head1 AUTHOR

Tom Payerle, payerle@umd.edu

=head1 COPYRIGHT AND LICENSE

Copyright 2014, University of Maryland

This library is free software; you can redistribute it and/or modify it
under the same terms as Perl itself, either Perl version 5.8.4 or, at
your option, any later version of Perl 5 you may have available.
