#!/usr/local/bin/perl
#
#Base class for scontrol entities

package Slurm::Scontrol::EntityBase;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);


use Data::Dumper;
#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

sub _ro_fields($)
{	return [];
}

sub _rw_fields($)
{	return [];
}

sub _required_fields($)
{	return [];
}

sub _special_fields($)
#This should be overloaded to return any fields with special data types
#for this class of entity
{	my $class = shift;
	return {};
}

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	my @args = @_;
	$class = ref($class) if ref($class);

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my $accessors = $obj->_rw_fields;
	my ($arg, $meth, $val);
	RWARG: foreach $arg (@$accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	$accessors = $obj->_ro_fields;
	ROARG: foreach $arg (@$accessors)
	{	next ROARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next ROARG unless defined $val;
		$meth = $arg;
		$obj->set($meth,$val);
	}


	#Warn about unknown arguments
	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;

	return;
}

sub _init($)
{	my $obj = shift;

	my ($fld, $meth, $val);
	my $req_parms = $obj->_required_fields;
	foreach $fld (@$req_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld";
		}
	}

}

sub new_from_scontrol_show_record($$)
{	my $class = shift;
	my $hash = shift;

	return $class->new(%$hash);
}
	
	
#-------------------------------------------------------------------
#	Lookup entity with scontrol
#-------------------------------------------------------------------

sub _scontrol_entity_name($)
{	my $class = shift;
	$class = ref($class) if ref($class);

	my $base = $class;
	$base =~ s/^.*://;
	$base = lc $base;
	return $base;
}

sub _scontrol_show_cmd($)
{	my $class = shift;
	$class = ref($class) if ref($class);

	my $base = $class->_scontrol_entity_name;
	return [ 'show', $base, ];
}

sub scontrol_show($$;$)
#Does scontrol show to get a list all of the entities of this type
#matching given ID (or all if ID is missing)
{	my $class = shift;
	my $scontrol = shift;
	my $id = shift;

	my $me = 'scontrol_show';
	croak "No/invalid Slurm::Scontrol object passed to $me at "
		unless $scontrol && ref($scontrol);

	my $cmd = $class->_scontrol_show_cmd;
	my @cmd = @$cmd;
	push @cmd, $id if defined $id;

	my $list = $scontrol->run_generic_scontrol_show_command(@cmd);
	unless ( $list && ref($list) )
	{	croak "Error running show cmd for $class: $list at ";
	}

	my @objects = ();
	foreach my $rec (@$list)
	{	my $obj = $class->new_from_scontrol_show_record($rec);
		push @objects, $obj;
	}

	return [@objects];
}

sub _scontrol_name_field($)
{	my $class = shift;
	$class = ref($class) if ref($class);
	die "Class $class did not overload _scontrol_name_field ";
}

sub scontrol_show_me($$)
#Takes an instance of an entity, and does an scontrol list to find the
#current value in SlurmDB for this entity instance
#Returns undef if no matches (I don't exist)
#Returns a new instance with that info
#On error, returns a non-ref error string
{	my $obj = shift;
	my $scontrol = shift;

	my $me = 'scontrol_show_me';
	croak "No/invalid Slurm::Scontrol object passed to $me at "
		unless $scontrol && ref($scontrol);

	my $namefld = $obj->_scontrol_name_field;
	my $name = $obj->$namefld;
	my $list = $obj->scontrol_show($scontrol, $name);
	return $list unless $list && ref($list) eq 'ARRAY';
	return unless scalar(@$list); #No matches, I don't exist

	if ( scalar(@$list) > 1 )
	{	my $class = ref($obj);
		return "Multiple  objects of type $class found with name $name";
	}

	my $obj2 = $list->[0];
	return $obj2;
}

1;
__END__

=head1 NAME

Slurm::Sacctmgr::EntityBase

=head1 SYNOPSIS

  package Slurm::Sacctmgr::Account;
  use base qw(Slurm::Sacctmgr::EntityBase);

  sub _ro_fields($)
  {	my $class = shift;
	return qw( account description organization coordinators);
  }

  sub _special_fields($)
  {	my $class = shift;
	return { 'coordinators' => 'csv' };
  }

  ...


=head1 DESCRIPTION

This is the base class for entities managed by sacctmgr.
It provides common constructors, etc.

Child classes should overload the following methods to customize
behavior:

=over 4

=item B<_ro_fields> 

Should return a list ref of the names of the
read-only fields for this entity.  Default is empty.

=item B<_rw_fields>

Should return a list ref of the names of the
read-write fields for this entity.  Default is empty.

=item B<_special_fields>

Should return a hash ref, keyed on field name, with value 
a string indicating the type of the field, for any and all
fields of this entity class which are non-trivial.  See
below for list of known data types.  Default is empty.

=item B<_required_fields>

A list of required fields for the constructor to make the object.
Default is empty.

=item B<_sacctmgr_fields_in_order>

A list ref of field names in order sacctmgr will return them
Must be overridden.

=item B<_sacctmgr_entity_name>

The name of sacctmgr entity associated with this Perl class.
Default is just derived from the class name.

=item B<_sacctmgr_list_cmd>

The sacctmgr list command to list entities of this type.
Default is "list $entity format=$fmtstr" where 
$entity is B<_sacctmgr_entity_name> and $fmtstr is the comma separated
concatenation of the field names in B<_sacctmgr_fields_in_order>

=item B<_sacctmgr_add_cmd>

The sacctmgr command to add entities of this type.
Default is "-i add $entity" where $entity is B<_sacctmgr_entity_name>.

=item B<_sacctmgr_modify_cmd>

The sacctmgr command to modify entities of this type.
Default is "-i modify $entity" where $entity is B<_sacctmgr_entity_name>.

=item B<_sacctmgr_delete_cmd>

The sacctmgr command to delete entities of this type.
Default is "-i delete $entity" where $entity is B<_sacctmgr_entity_name>.

=item B<_my_sacctmgr_where_clause>

The "where" clause to give to the sacctmgr list command to list
just the record corresponding to this instance in the database.
The default is "name=>$NameValue" where NameValue is the value of
the field named B<_sacctmgr_name_field>.  Should be a hash ref
of key => value pairs.

=item B<_sacctmgr_name_field>

Used to default B<_my_sacctmgr_where_clause> above.

=item B<_my_sacctmgr_add_clause>

The clause to give to the sacctmgr list command to add this
entity to the database with sacctmgr.  Should be a hash ref of
key => value pairs.  The default is just the list of parameter names
(from B<_sacctmgr_fields_in_order>) and there values, with the name
of B<_sacctmgr_name_field> replaced by 'name', and the key => value
pairs for undef values omitted.

=back

The data types for the B<_special_fields> hash are:

=over 4

=item B<csv>: 

The should be converted to a list ref from a comma separated string

=back 4

Most useful methods are:

=over 4

=item B<new_from_sacctmgr_list_record($rec)>

This is an alternative constructor, which  
takes a list ref e.g. from one of the list returned by a 
Slurm::Sacctmgr::run_generic_sacctmgr_list_command and returns
a new instance of specified type.  Elements of list are required to
be in the order given by B<_sacctmgr_fields_in_order>.

=item B<sacctmgr_list>($sacctmgr, [ where=>val, [where2=>val2 ... ])

This calls Slurm::Sacctmgr::run_generic_list_command to list the
entities of our type.  Optional where clauses given as key => value pairs
can be given, otherwise it should list all entities.  The returned list
is converted to a list of Perl objects of the appropriate type.
On error, a non-ref error string is returned.

=item B<sacctmgr_list_me($sacctmgr)>

This is an instance method, and calls B<sacctmgr_list> with the 
appropriate where clause to return only the record representing the
current instance.  This is returned as a new instance.  If no matching
record found, returns undef. On error, 
a non-ref error string is returned.

=item B<new_from_sacctmgr_by_name($sacctmgr, $name)>

Returns a new instance of the Perl object for given entity type, obtained
from looking up in sacctmgr for the record with the given name.  If no
such record found, returns undef.  Returns non-ref true value (error message)
on error.  Otherwise, returns the instance.

=item B<sacctmgr_add>($sacctmgr, [ fld1=>val, [fld2=>val2 ... ])

This calls Slurm::Sacctmgr::run_generic_command to add an entity
of the desired type with the specified fields.  In addition to the
standard field definitions, if the pseudo-field '--ok-if-previously-exists'
is passed and true, the method will return "successfully" even if B<sacctmgr>
errors because the entity already exists.

=item B<sacctmgr_add_me>($sacctmgr, [ extra1 => val, [ extra2 => val2 ...]])

This is an instance method, and calls B<sacctmgr_add> with the 
appropriate parameters to add the entity represented by the Perl object.
If any "extra" arguments are provided, this will also be passed to B<sacctmgr>.
B<NOTE:> The extra arguments will override data members if there is a conflict.

=item B<sacctmgr_modify($sacctmgr, $where, $update)>

This calls Slurm::Sacctmgr::run_generic_command to modify one or
more entities of given type.  The entities to update are selected
by the key=>value pairs in the $where hash ref; if omitted will act on
ALL entities.  The fields to update are given by key=>value pairs in
the $update hash ref.

=item B<sacctmgr_modify_me($sacctmgr, [ fld1 => val1 [, fld2=>val2 ... ])>

This is an instance method, and calls B<sacctmgr_modify> 
to modify the entity represented by the Perl object as indicated.

=item B<sacctmgr_save_me>($sacctmgr, [ extra1 => val1, [ extra2=>val2 ... ]])

This is an instance method, and calls B<sacctmgr_modify> to update the
entity with the same name to the values of the Perl object.  Obviously
cannot be used to change the name of an object.  If no entity exists in
sacctmgr db, does same as B<sacctmgr_add_me>.
B<NOTE:> The extra arguments will override data members if there is a conflict.

=item B<sacctmgr_delete>($sacctmgr, [ fld1=>val, [fld2=>val2 ... ])

This calls Slurm::Sacctmgr::run_generic_command to delete one or more
entities matching the specified fields.  It requires some where clause
or will abort.

=item B<sacctmgr_delete_me($sacctmgr)>

This is an instance method, and calls B<sacctmgr_modify> 
to delete the entity record in the DB corresponding to the invocant
Perl object.

=back

