#!/usr/local/bin/perl
#
#Perl wrappers for Slurm sinfo command

package Slurm::Sinfo;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);
use Data::Dumper;
use Slurm::Sinfo::Node;


#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @rw_accessors = qw(
	verbose
);

my @ro_accessors = qw(
	sinfo
);

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_ro_accessors(@ro_accessors);


my @required_parms = qw(
	sinfo
);


#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my ($arg, $meth, $val);
	RWARG: foreach $arg (@rw_accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	ROARG: foreach $arg (@ro_accessors)
	{	next ROARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next ROARG unless defined $val;
		$meth = $arg;
		$obj->set($meth,$val);
	}


	#Warn about unknown arguments
	if ( scalar(keys %args) )
	{	my $tmp = join ", ", (keys %args);
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;

	my ($tmp);
	$tmp = $obj->sinfo;
	#$obj->set('sinfo', 'sinfo') unless $tmp; #try to find in path

	#Standard Glue path
	$obj->set('sinfo', '/usr/local/slurm/bin/sinfo') unless $tmp; 

	return;
}

sub _init($)
{	my $obj = shift;

	my ($fld, $meth, $val);
	foreach $fld (@required_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld";
		}
	}

}

#-------------------------------------------------------------------
#	Basic sinfo commands
#-------------------------------------------------------------------

sub _noshell_backticks($$@)
#Calls an external command using pipes and forks so no shell gets invoked
#Returns ($err, @out) where $err is the error is the exit status of
#the command, and @out is the list of output returned, line by line.
#If $mode is 0, only STDOUT is returned in @out, 
#If $mode is non-zero, STDERR is dupped onto STDOUT and also returned.
{	my $obj = shift;
	my $mode = shift;
	my @cmd = @_;

	my ($err, @out, $PIPE, $res);

	if ( $res = open($PIPE, "-|" )  )
	{	#Parent
		if ( ! defined $res )
		{	my $tmp = join ' ', @cmd;
			die "Pipe to '$tmp' failed: $!";
		}
		@out = <$PIPE>;
		$res = close $PIPE;
		$err = $?;
		return ($err, @out);
	} else
	{	#Child
		#Duplicate stderr onto stdout if so requested
		if ( $mode )
		{	unless ( open(STDERR, '>&STDOUT') )
			{	print "Cannot dup stderr to stdout in child";
				die "Cannot dup stderr to stdout in child";
			}
		}
		exec { $cmd[0] } @cmd; #Not subject to shell escapes
		die "Failed to exec '" . (join ' ', @cmd) . "' in child.\n";
	}
}

sub _run_sinfo_cmd($@)
#Run a sinfo cmd, returning results as array of lines
{	my $obj = shift;
	my @args = @_;

	my $cmd = $obj->sinfo;
	my $mode = 0;
	if ( $obj->verbose )
	{	#Verbose mode:
		#Output command before executing it
		my @tmp = ($cmd, @args);
		my $tmpcmd = join ' ', @tmp;
		print STDERR "[VERBOSE] $tmpcmd\n";
	}
	my ($err, @out ) = $obj->_noshell_backticks($mode, $cmd, @args);
	if ( $err )
	{	my $errstr = "Exit code: $err";
		my $output = join "\n", @out;
		$errstr .= "\n$output" if $output;
		return $errstr;
	}
	#return $err if $err;
	return [ @out ];
	
}

sub run_sinfo_cmd($@)
#Runs specified squeue command
#We add the -o option with list of known sinfo fields
# to end of any user specified options,
#and parse the results in a list ref of Slurm::Sinfo::Node objects
{	my $obj = shift;
	my @args = @_;

	my $me = 'run_sinfo_cmd';
	my $fields = Slurm::Sinfo::Node->formats_in_sinfo_order;
	my $fldstr = join '|%', @$fields;
	$fldstr = '%' . $fldstr;
	#push @args, '-o', "%all", '--noheader';
	#push @args, '-O', "$fldstr", '--noheader';
	push @args, '-o', "$fldstr", '--noheader';

	my $output = $obj->_run_sinfo_cmd(@args);
	return $output unless $output && ref($output) eq 'ARRAY';

	$output = Slurm::Sinfo::Node->parse_sinfo_output($output);

	return $output;
}
	

1;
__END__

=head1 NAME

Slurm::Sinfo - Perl wrapper to sinfo command

=head1 SYNOPSIS

  use Slurm::Sinfo;

  my $sinfo = Slurm::Sinfo->new( sinfo=>'path to sinfo command');

  my $sinfo->verbose(1);

  #Returns a list ref of Slurm::Sinfo::Node objects for j
  #the 'standard' partition
  my $results = $sinfo->run_squeue_cmd('-p', 'standard');
  unless ( $results && ref($results) eq 'ARRAY' )
  {	print STDERR "Error running squeue command: $results\n";
  }

  foreach my $rec (@$results)
  {	my $avail = $rec->availability;
	my $cpus = $rec->cpus;
	my $nlist = $rec->nodelist;
	print "$avail $cpus $nodelist\n";
  }

  ...


=head1 DESCRIPTION

B<Slurm::Sinfo> and subclasses provide wrappers to the Slurm B<sinfo> 
command.  

B<NOTE:> because the sinfo command is informational only, there is
NO dryrun option.

=head2 CONSTRUCTOR and DATA members

Interface is object orientated, mainly to reduce namespace pollution.
The constructor B<new> takes key => value pairs to set the data members, 
which are:

=over 4

=item B<sinfo>:

The path to the Slurm B<sinfo> command.  Can only be set at time of
construction.  Defaults to just "sinfo", i.e. will look for it in your
path.

=item B<verbose>:

If this is set, the module works in B<verbose> mode.  Basically, the various
B<sinfo> commands will be printed out just before they are executed.  
Default is false.

=back

Accessors exist for all of the above.

The B<sinfo> field is read-only, and can only be set at time of
construction.  B<verbose> is read-write, and can be set
at any time by providing a defined boolean value as an explicit argument
to the accessor.  Usual perl boolean semantics apply, except that you
cannot provide an undef value to the accessor to unset the fields (as that
would be interpretted as a read accessor, not a mutator).

The main method is B<run_sinfo_command>, which actually runs the
B<sinfo> command with arguments given.  These
use forking and pipes to avoid invoking an Unix shell for additional safety
(i.e., shell expansions will not work).  The intent is to be able to 
programatically issue B<sinfo> commands and get the results.  The functions
take the B<sinfo> command, with each word as a separate argument to
the function.  The actual B<sinfo> command should NOT be given, as we will
use the one provided to the constructor.

On success, the return value of the B<run_sinfo_command> is a list ref of
B<Slurm::Sinfo::Node> instances describing the data returned 
by the squeue command.  On non-fatal errors, a non-ref error string is returned.

=head2 VERBOSE MODE

If the B<verbose> data method is set, either at construction time or via
the B<verbose> mutator afterwards, the B<squeue> commands will be printed
out to STDERR before being executed.  This is again useful for debugging.

=head1 EXPORT

None.  OO interface only.

=head1 SEE ALSO

Slurm::Squeue::Job

=head1 AUTHOR

Tom Payerle, payerle@umd.edu

=head1 COPYRIGHT AND LICENSE

Copyright 2014-2015, University of Maryland

This library is free software; you can redistribute it and/or modify it
under the same terms as Perl itself, either Perl version 5.8.4 or, at
your option, any later version of Perl 5 you may have available.
