#!/usr/local/bin/perl
#
#Perl wrappers for Slurm sinfo command
#Represents a node, combined over all partitions

package Slurm::Sinfo::CombinedNode;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);
use Data::Dumper;


#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @required_parms = qw(
	nodename
);


#These are in order returned by %all format of sinfo command
my @rw_accessors = qw(
	nodename

	availability
	state
	reason
	reason_timestamp
	user_who_set_reason
	uid_who_set_reason

	node_allocated
	node_idle
	node_other
	node_mixed

	slurmd_version

	cpus
	cpus_allocated
	cpus_idle
	cpus_other
	sockets
	cores
	threads

	memory
	tmp_disk
	features
	gres
	cpu_load

	partitions
);
	
my @ro_accessors = qw();

__PACKAGE__->mk_accessors(@rw_accessors);

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my ($arg, $meth, $val);
	RWARG: foreach $arg (@rw_accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	ROARG: foreach $arg (@ro_accessors)
	{	next ROARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next ROARG unless defined $val;
		$meth = $arg;
		$obj->set($meth,$val);
	}


	#Warn about unknown arguments
	if ( scalar(keys %args) )
	{	my $tmp = join "', '", (keys %args);
		$tmp = "'$tmp'";
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;
	return;
}

sub _init($)
{	my $obj = shift;

	my ($fld, $meth, $val);
	foreach $fld (@required_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld";
		}
	}

}

sub new_from_sinfo_node_obj($$)
{	my $class = shift;
	my $node = shift;
	my $me = __PACKAGE__ . '::new_from_sinfo_node_obj';
	
	unless ( $node && ref($node) && $node->isa('Slurm::Sinfo::Node') )
	{	die "$me: Invalid parameter node, expecting Slurm::Sinfo::Node reference at";
	}

	my $hostnames = $node->hostnames;
	my @tmp = split /\s*,\s*/, $hostnames;
	die "$me: No hostname found at " unless @tmp;
	die "$me: Multiple hostnames ($hostnames) found at " if scalar(@tmp) > 1;

	my $nodename = $hostnames;

	my %hash = (
		nodename => $nodename,
		availability => $node->availability,
		state => $node->state,
		reason => $node->reason,
		reason_timestamp => $node->reason_timestamp,
		user_who_set_reason => $node->user_who_set_reason,
		uid_who_set_reason => $node->uid_who_set_reason,
		
		node_allocated => ($node->nodes_allocated)?1:0,
		node_idle => ($node->nodes_idle)?1:0,
		node_other => ($node->nodes_other)?1:0,
		node_mixed => ($node->state eq 'mixed')?1:0,

		slurmd_version => $node->slurmd_version,

		cpus => $node->cpus_total,
		cpus_allocated => $node->cpus_allocated,
		cpus_idle => $node->cpus_idle,
		cpus_other => $node->cpus_other,

		sockets => $node->sockets,
		cores => $node->cores,
		threads => $node->threads,

		memory => $node->memory,
		tmp_disk => $node->tmp_disk,

		cpu_load => $node->cpu_load,
	);

	my $tmp = $node->features;
	if ( $tmp )
	{	$hash{features} = [ split /\s*;\s*/, $tmp ];
	} else
	{	$hash{features} = [];
	}

	$tmp = $node->gres;
	if ( $tmp )
	{	@tmp = split /\s*,\s*/, $tmp;
		$tmp = {};
		foreach my $gresrec (@tmp)
		{	my ($gres, $val) = split /:/, $gresrec, 2;
			$tmp->{$gres} = $val;
		}
		$hash{gres} = $tmp;
	} else
	{	$hash{gres} = {};
	}

	$tmp = $node->partition;
	$hash{partitions} = [ $tmp ];
	
	return $class->new(%hash);
}
		
#These are the fields that must match when merging sinfo node objects
#Either a scalar (field has same name in CombinedNode and Node) or
#array ref (CombinedNode name, Node field name )
my @_require_matching_fields = 
(	[ 'nodename', 'hostnames' ],
	[ 'cpus', 'cpus_total' ],
	'sockets',
	'cores',
	'threads',
	'memory',
	'tmp_disk',
);


sub merge_with_sinfo_node_obj($$)
{	my $self = shift;
	my $node = shift;
	my $me = __PACKAGE__ . '::merge_with_sinfo_node_obj';
	
	my $nodename = $self->nodename;

	unless ( $node && ref($node) && $node->isa('Slurm::Sinfo::Node') )
	{	die "$me: Invalid parameter node, expecting Slurm::Sinfo::Node reference at";
	}



	#Make sure fields which must match actually do match
	foreach my $fieldrec (@_require_matching_fields)
	{	my ( $nfld, $cnfld);
		next unless $fieldrec;
		if ( ref($fieldrec) eq 'ARRAY' )
		{	($cnfld, $nfld) = @$fieldrec;
		} else
		{	$cnfld = $nfld = $fieldrec;
		}

		my $nval = $node->$nfld;
		my $cnval = $self->$cnfld;
		unless ( $cnval eq $nval )
		{	die "$me: Attempt to merge to $nodename but field mismatch: cnfld=$cnfld, cnval=$cnval, nval=$nval at ";
		}
	}

	#Fields to merge
	my $partition = $node->partition;
	my $parts = $self->partitions || [];
	push @$parts, $partition;
	$self->partitions($parts);
	
	$parts = $self->partitions;

}

sub make_combined_node_hash($$)
#This takes a list ref of Slurm::Sinfo::Node instances, and creates
#a CombinedNode instance for each distinct node represented.  If multiple
#records exist for a Node (as will be the case if can be accessed on multiple
#partititions), these will be consolidated (using merge_with_sinfo_node_obj).
#
#The results are then used to form a hash ref based on nodename, which is
#returned.
#
#On non-fatal errors, returns a non-ref error string.
{	my $class = shift;
	my $nodelist = shift;
	my $me = __PACKAGE__ . '::make_combined_node_hash';

	return "$me: Null node list at " unless $nodelist;
	return "$me: Invalid nodelist '$nodelist', expecting list ref" 
		unless ref($nodelist) eq 'ARRAY';
	return [] unless scalar(@$nodelist);

	my $cnhash = {};

	my $i=0;
	foreach my $node (@$nodelist)
	{	$i++;
		unless ( $node && ref($node) && $node->isa('Slurm::Sinfo::Node') )
		{	die "$me: Invalid element (# $i) in nodelist: '$node', " .
				"expecting a Slurm::Sinfo::Node instance, at ";
		}
		my $nodename = $node->hostnames;
		my $cnode = $cnhash->{$nodename};
		if ( $cnode )
		{	#print STDERR "Merging $nodename with existing...\n";
			$cnode->merge_with_sinfo_node_obj($node);
		} else
		{	#print STDERR "Making new instance for $nodename...\n";
			$cnode = Slurm::Sinfo::CombinedNode->new_from_sinfo_node_obj(
				$node);
		}
		$nodename = $cnode->nodename;
		$cnhash->{$nodename} = $cnode;
	}
	return $cnhash;
}


1;
__END__

=head1 NAME

Slurm::Sinfo::CombinedNode - represents sinfo data about a node (for all partitions)

=head1 SYNOPSIS

  use Slurm::Sinfo;
  use Slurm::Sinfo::Node;
  use Slurm::Sinfo::CombinedNode;

  my $sinfo = Slurm::Sinfo->new( sinfo=>'path to sinfo command');

  my $nodelist = $sinfo->run_sinfo_cmd;
  my $nodehash = Slurm::Sinfo::CombinedNode->make_combined_node_hash($nodelist);

  foreach $node (keys %$nodehash)
  {	print $node->nodename, $node->cpus_total, "\n";

=head1 DESCRIPTION

B<Slurm::Sinfo> and subclasses provide wrappers to the Slurm B<sinfo> 
command.  This <CombinedNode> class is designed for consolidating information
from the B<Slurm::Sinfo::Node> class (which is partition-node specific).


=head1 EXPORT

None.  OO interface only.

=head1 SEE ALSO

Slurm::Squeue::Job

=head1 AUTHOR

Tom Payerle, payerle@umd.edu

=head1 COPYRIGHT AND LICENSE

Copyright 2014, University of Maryland

This library is free software; you can redistribute it and/or modify it
under the same terms as Perl itself, either Perl version 5.8.4 or, at
your option, any later version of Perl 5 you may have available.
