#!/usr/local/bin/perl
#
#Perl wrappers for Slurm sinfo command
#Represents a node/partition information

package Slurm::Sinfo::Node;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);
use Data::Dumper;


#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @required_parms = qw(
);


#These are in order returned by %all format of sinfo command
#Probably needs updating/modifying based on Slurm version.
#This is what -o %all gave for slurm v14 ???
my @fields_in_sinfo_order = qw(
	availability
	cpus
	tmp_disk
	features
	groups
	share
	timelimit
	memory
	hostnames
	node_addresses
	priority
	root_only
	job_size_nodes
	state
	user_who_set_reason
	slurmd_version
	scheduling_weight
	sockets_cores_threads
	allocated_idle_nodes
	max_cpus_per_node
	cpus_allocated_idle_other_total
	node_count
	reason
	nodes_allocated_idle_other_total
	gres
	reason_timestamp
	default_walltime
	preemption_mode
	nodelist
	cpu_load
	partition1
	partition
	allowed_allocating_nodes
	state
	user_with_uid
	sockets
	cores
	threads
);

#This converts field names from e.g. @fields_in_sinfo_order to specifiers
#for the sinfo -O / --Format option. 
my %Format_specifiers_by_field =
(	
	allocmem =>  'allocmem',
	allowed_allocating_nodes => 'allocnodes',
	availability => 'available',
	cpus => 'cpus',
	cpu_load => 'cpusload',
	freemem => 'freemem',
	cpus_allocated_idle_other_total => 'cpusstate',
	cores => 'cores',
	default_walltime => 'defaulttime',
	tmp_disk => 'disk',
	features => 'features',
	features_act => 'features_act',
	groups => 'groups',
	gres => 'gres',
	max_cpus_per_node => 'maxcpuspernode',
	memory => 'memory',
	node_count => 'nodes',
	node_addresses => 'nodeaddr',
	allocated_idle_nodes => 'nodeai',
	nodes_allocated_idle_other_total => 'nodeaiot',
	hostnames => 'nodehost',
	nodelist => 'nodelist',
	share => 'oversubscribe', #Should rename this?
	partition => 'partition',
	partition1 => 'partitionname',
	preemption_mode => 'preemptmode',
	priorityjobfactor => 'priorityjobfactor',
	priority => 'prioritytier',
	reason => 'reason',
	root_only => 'root',
	job_size_nodes => 'size',
	state => 'statecompact',
	statelong => 'statelong',
	sockets => 'sockets',
	sockets_cores_threads => 'socketcorethread',
	timelimit => 'time',
	reason_timestamp => 'timestamp',
	threads => 'threads',
	user_who_set_reason => 'user',
	user_with_uid => 'userlong',
	slurmd_version => 'version',
	scheduling_weight => 'weight',
);

#This converts field names from e.g. @fields_in_sinfo_order to specifiers
#for the sinfo -o / --format option. 
my %format_specifiers_by_field =
(	
	availability => 'a',
	allocated_idle_nodes => 'A',
	features_act => 'b',
	max_cpus_per_node => 'B',
	cpus => 'c',
	cpus_allocated_idle_other_total => 'C',
	tmp_disk => 'd',
	node_count => 'D',
	freemem => 'e',
	reason => 'E',
	features => 'f',
	nodes_allocated_idle_other_total => 'F',
	groups => 'g',
	gres => 'G',
	share => 'h', 
	reason_timestamp => 'H',
	priorityjobfactor => 'I',
	timelimit => 'l',
	default_walltime => 'L',
	memory => 'm',
	preemption_mode => 'M',
	hostnames => 'n',
	nodelist => 'N',
	node_addresses => 'o',
	cpu_load => 'O',
	priority => 'p',
	partition => 'P',
	root_only => 'r',
	partition1 => 'R',
	job_size_nodes => 's',
	allowed_allocating_nodes => 'S',
	state => 't',
	statelong => 'T',
	user_who_set_reason => 'u',
	user_with_uid => 'U',
	slurmd_version => 'v',
	scheduling_weight => 'w',
	sockets => 'X',
	cores => 'Y',
	threads => 'Z',
	sockets_cores_threads => 'z',
);

my @Formats_in_sinfo_order = 
	map {  $Format_specifiers_by_field{$_} || $_ } 
		@fields_in_sinfo_order;

my @formats_in_sinfo_order = 
	map {  $format_specifiers_by_field{$_} || $_ } 
		@fields_in_sinfo_order;


my @rw_accessors = (
	@fields_in_sinfo_order,
	'uid_who_set_reason',
	'nodes_allocated',
	'nodes_idle',
	'nodes_other',
	'nodes_total',
	'cpus_allocated',
	'cpus_idle',
	'cpus_other',
	'cpus_total',
);

my @ro_accessors = qw();

__PACKAGE__->mk_accessors(@rw_accessors);


#-------------------------------------------------------------------
#	Share list of fields with Slurm::Sinfo
#-------------------------------------------------------------------

sub known_sinfo_cmd_fields($)
{	my $class =shift;
	return [ @fields_in_sinfo_order ];
}

sub Formats_in_sinfo_order($)
{	my $class =shift;
	return [ @Formats_in_sinfo_order ];
}

sub formats_in_sinfo_order($)
{	my $class =shift;
	return [ @formats_in_sinfo_order ];
}

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my ($arg, $meth, $val);
	RWARG: foreach $arg (@rw_accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	ROARG: foreach $arg (@ro_accessors)
	{	next ROARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next ROARG unless defined $val;
		$meth = $arg;
		$obj->set($meth,$val);
	}


	#Warn about unknown arguments
	if ( scalar(keys %args) )
	{	my $tmp = join "', '", (keys %args);
		$tmp = "'$tmp'";
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;
	return;
}

sub _init($)
{	my $obj = shift;

	my ($fld, $meth, $val);
	foreach $fld (@required_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld";
		}
	}

}

sub new_from_sinfo_output($$)
{	my $class = shift;
	my $line = shift;

	chomp $line;
	#Trim leading/trailing whitespace
	$line   =~ s/^\s*//; $line =~ s/\s*$//;

	my @values = split /\s*\|\s*/, $line;
	my @fields = @fields_in_sinfo_order;
	my %hash = ();

	FIELD: foreach my $field (@fields)
	{	my $val = shift @values;
		$field = lc $field;
		next FIELD unless $field;
		next FIELD unless defined $val;

		#Special cases
		next FIELD if ( $field eq 'user_who_set_reason' 
			&& $val eq 'Unknown' );
		next FIELD if ( $field eq 'user_with_uid' 
			&& $val eq 'Unknown' );
		$val = $class->_canonicalize_boolean($val)
			if ( $field eq 'root_only' 
			);

		$hash{$field} = $val;
	}

	#Special, derived fields
	my $tmp;

	$tmp = $hash{user_with_uid};
	if ( $tmp && $tmp =~ /\((\d+)\)/ )
	{	$hash{uid_who_set_reason} = $1;
	}
	
	$tmp = $hash{nodes_allocated_idle_other_total};
	if ( $tmp && $tmp =~ '^(\d+)/(\d+)/(\d+)/(\d+)$' )
	{	$hash{nodes_allocated} = $1;
		$hash{nodes_idle} = $2;
		$hash{nodes_other} = $3;
		$hash{nodes_total} = $4;
	}

	$tmp = $hash{cpus_allocated_idle_other_total};
	if ( $tmp && $tmp =~ '^(\d+)/(\d+)/(\d+)/(\d+)$' )
	{	$hash{cpus_allocated} = $1;
		$hash{cpus_idle} = $2;
		$hash{cpus_other} = $3;
		$hash{cpus_total} = $4;
	}

	return $class->new(%hash);
}
		

#-------------------------------------------------------------------
#	Class methods
#-------------------------------------------------------------------

sub _canonicalize_boolean($$)
{	my $class = shift;
	my $val = shift;
	return unless defined $val; #undef => undef

	return 0 unless $val; #0 or empty string is false
	$val = lc $val;
	return 0 if $val eq 'no';

	return 1;
}

sub parse_sinfo_output($$)
{	my $class = shift;
	my $output_list = shift;
	my $me = 'parse_sinfo_output';

	return "$me: Null output list" unless $output_list;
	return "$me: Invalid output list '$output_list', expecting list ref" 
		unless ref($output_list) eq 'ARRAY';
	return [] unless scalar(@$output_list);

	my $nodes = [];
	foreach my $line (@$output_list)
	{	my $node = $class->new_from_sinfo_output($line);
		push @$nodes, $node;
	}
	return $nodes;
}


1;
__END__

=head1 NAME

Slurm::Sinfo::Node - Represents a node/partition as returned from sinfo

=head1 SYNOPSIS

  use Slurm::Sinfo;

  my $sinfo = Slurm::Sinfo->new( sinfo=>'path to sinfo command');

  my $nodes = $sinfo->run_sinfo_cmd();

=head1 DESCRIPTION

B<Slurm::Sinfo> and subclasses provide wrappers to the Slurm B<sinfo> 
command.  This represents the node/partition data returned by
sinfo command.

Objects are normally created with the B<new_from_sinfo_output>, which
takes the output of sinfo command with "-o '%all'" argument, or as is
the more common case, from B<parse_sinfo_output>.  This latter takes
the full output of sinfo command, one line per array element, and
returns the result of running B<new_from_sinfo_output> on each line.

The resultant B<Node> objects have the following fields available:

=over 4

=item B<availability>: 

The availability of the node, e.g. 'up', etc.

=item B<hostnames>: 

The hostnames of the nodes represented.  Typically, (i.e. when
using B<Slurm::Sinfo::run_sinfo_cmd>, which uses a '-o %all' option
to sinfo), there should be a separate record for each host, so this
should be a single name.  

=item B<node_addresses>: 

Like B<hostnames>, the addresses for the node.  Typically only one.

=item B<nodelist>: 

Like B<hostnames>, typcially only one.

=item B<state>: 

The state of the node.

=item B<reason>: 

This contains the text reason the node is in an unavailable state.

=item B<reason_timestamp>: 

This contains the timestamp for reason, if any.

=item B<user_who_set_reason>: 

This contains the name of the user who set the reason.

=item B<uid_who_set_reason>: 

This contains the uid of the user who set the reason.

=item B<slurmd_version>:

The version of slurmd running on node(s).

=item B<node_count>: 

The count of nodes this record represents.  Typically, just 1.

=item B<nodes_allocated>:

The number of nodes in allocated state.  Typically 0 or 1.

=item B<nodes_idle>:

The number of nodes in idle state.  Typically 0 or 1.

=item B<nodes_other>:

The number of nodes in other (neither allocated nor idle) state.  Typically 0 or 1.

=item B<nodes_total>:

The total number of nodes.  Typically 1.

=item B<allocated_idle_nodes>: 

A string giving number of "allocated/idle" nodes.

=item B<nodes_allocated_idle_other_total>: 

A string giving number of "allocated/idle/other/total" nodes.  This
is the raw sinfo output, used to get B<nodes_allocated>, B<nodes_idle>,
B<nodes_other>, B<nodes_total>, above.

=item B<cpus>: 

The number of cpus represeted by the record.

=item B<cpus_allocated>:

The number of cpus in the allocated state.

=item B<cpus_idle>:

The number of cpus in the idle state.

=item B<cpus_other>:

The number of cpus in "other" (neither allocated nor idle) state.

=item B<cpus_total>:

The total number of cpus.

=item B<max_cpus_per_node>: 

The maximum number of cpus per node.  Typically same as cpus.

=item B<cpus_allocated_idle_other_total>: 

A string with CPU counts for "allocated/idle/other/total".  This
is the raw sinfo output used to generate B<cpus_allocated>, etc. above.

=item B<sockets>: 

The number of sockets per node.

=item B<cores>: 

The number of cores per socket.

=item B<threads>: 

The number of threads per core.

=item B<sockets_cores_threads>: 

A string "S:C:T" with the above sockets:core:threads info.

=item B<memory>: 

Size of real memory in MB.

=item B<tmp_disk>: 

The amount of temporary disk space (per node) in MB.

=item B<features>: 

A comma delimited list of features available on the node.

=item B<gres>: 

Generic Consumable Resources available on this node, comma
delimited, with colons indicating quantities.

=item B<groups>: 

A comma delimited list of groups who can use these resources for
this partitition.

=item B<share>: 

Whether jobs can share these resources.  'no' means they
are never shared.  'exclusive' means dedicated to jobs, 'force'
means always can be shared, and 'yes' means shared based on 
job resource allocation.

=item B<timelimit>: 

Maximum time limit for any user job.

=item B<default_walltime>: 

Default walltime limit for jobs.

=item B<priority>: 

Priority.

=item B<root_only>: 

Boolean.  If true, only root can allocate resources in this partition.

=item B<job_size_nodes>: 

Minimum and maximum node count that can be allocated to any user job.

=item B<scheduling_weight>: 

Scheduling weight.

=item B<preemption_mode>: 

Preemption mode

=item B<cpu_load>: 

CPU load.

=item B<partition>: 

Partition name.

=item B<allowed_allocating_nodes>: 

The nodes which can allocate this resource.

=back

=head1 EXPORT

None.  OO interface only.

=head1 SEE ALSO

Slurm::Squeue::Job

=head1 AUTHOR

Tom Payerle, payerle@umd.edu

=head1 COPYRIGHT AND LICENSE

Copyright 2014, University of Maryland

This library is free software; you can redistribute it and/or modify it
under the same terms as Perl itself, either Perl version 5.8.4 or, at
your option, any later version of Perl 5 you may have available.
