#!/usr/local/bin/perl
#
#Perl wrappers for Slurm squeue command
#Represents a single job

package Slurm::Squeue::Job;
use strict;
use warnings;
use base qw(Class::Accessor);
use Carp qw(carp croak);
use Data::Dumper;
use POSIX qw(floor); #Fordelta_seconds_to_elapsed
use Time::ParseDate;


#-------------------------------------------------------------------
#	Globals
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#	Accessors
#-------------------------------------------------------------------

my @required_parms = qw(
	jobid
);


my @rw_accessors = qw(
	account
	array_job_id
	array_task_id
	command
	comment
	contiguous
	core_spec
	cores
	cores_per_socket
	cpus
	dependency
	end_time
	exc_nodes
	exec_host
	features
	gres
	group
	jobid
	licenses
	min_cpus
	min_memory
	min_tmp_disk
	name
	nice
	nodelist
	nodes
	over_subscribe
	partition
	priority
	qos
	reason
	req_nodes
	reservation
	schednodes
	shared
	sockets
	sockets_per_node
	st
	start_time
	state
	submit_time
	threads
	threads_per_core
	time
	time_left
	timelimit
	time_limit
	user
	username
	wckey
	work_dir
);

#Calculated fields
my @calc_rw_accessors = qw(
	end_tstamp
	start_tstamp
	submit_tstamp
);
	
my @ro_accessors = qw();

__PACKAGE__->mk_accessors(@rw_accessors);
__PACKAGE__->mk_accessors(@calc_rw_accessors);

#-------------------------------------------------------------------
#	Constructors, etc
#-------------------------------------------------------------------

sub new($;@)
{	my $class = shift;
	my @args = @_;

	my $obj = {};
	bless $obj, $class;

	$obj->_parse_args(@args);
	$obj->_set_defaults;
	$obj->_init;

	return $obj;
}

sub _parse_args($@)
{	my $obj = shift;
	my %args = @_;

	my ($arg, $meth, $val);
	RWARG: foreach $arg (@rw_accessors, @calc_rw_accessors)
	{	next RWARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next RWARG unless defined $val;
		$meth = $arg;
		$obj->$meth($val);
	}

	ROARG: foreach $arg (@ro_accessors)
	{	next ROARG unless exists $args{$arg};
		$val = delete $args{$arg};
		next ROARG unless defined $val;
		$meth = $arg;
		$obj->set($meth,$val);
	}


	#Warn about unknown arguments
	if ( scalar(keys %args) )
	{	my $tmp = join "', '", (keys %args);
		$tmp = "'$tmp'";
		croak "Unrecognized arguments [ $tmp ] to constructor at ";
	};
}

sub _set_defaults($)
{	my $obj = shift;
	return;
}

sub _init($)
{	my $obj = shift;

	my ($fld, $meth, $val);
	foreach $fld (@required_parms)
	{	$meth = $fld;
		$val = $obj->$meth;
		unless ( defined $val )
		{	croak "Missing required argument $fld";
		}
	}

}

sub new_from_squeue_output($$$)
{	my $class = shift;
	my $header = shift;
	my $line = shift;

	chomp $header;
	chomp $line;
	$header =~ s/^\s*//; $header =~ s/\s*$//;
	$line   =~ s/^\s*//; $line =~ s/\s*$//;

	#print STDERR "header='$header'\n";
	#print STDERR "line='$line'\n";

	my @fields = split /\s*\|\s*/, $header;

	#FEATURES can possibly be of form [ a|b|c ]
	# Change to [a;b;c]
	my @tmp = grep { $fields[$_] eq 'FEATURES' } ( 0 .. scalar(@fields)-1);
	my $feature_idx = $tmp[0];
	
	my @values = split /\s*\|\s*/, $line;
	if ( defined $feature_idx )
	{	@values = split /\s*\|\s*/, $line, $feature_idx+1;
		my $rest = pop @values;
		if ( $rest =~ /^\[/ )
		{	#Need to fix
			my $tmp = $rest;
			$tmp =~ s/\]\|.*$/]/;
			$tmp =~ s/\|/;/g;
			push @values, $tmp;
			$rest =~ s/^\[[^\]]*\]\|//;
			@tmp = split /\s*\|\s*/, $rest;
			push @values, @tmp;
		} else
		{	@tmp = split /\s*\|\s*/, $rest;
			push @values, @tmp;
		}
	}
	#foreach my $i (0 .. scalar(@fields) -1 )
	#{	print STDERR "$i: $fields[$i] => $values[$i]\n";
	#}

	my %hash = ();
	my $nodes_reason_string;

	FIELD: foreach my $field (@fields)
	{	my $val = shift @values;
		$field = lc $field;
		next FIELD unless $field;
		next FIELD unless defined $val;

		#Handle special cases

		if ( $field eq 'user' )
		{	#Same header for user name and user id
			if ( $val =~ /^\d+$/ )
			{	#value looks like id
				$hash{user} = $val;
				$hash{username} = $val unless $hash{username};
			} else
			{	$hash{username} = $val;
			}
			next FIELD;
		}

		if ( $field eq 'nodelist(reason)' )
		{	#Do this last
			$nodes_reason_string = $val;
			next FIELD;
		}

		if ( $field eq 's:c:t' )
		{	my ($sockets, $cores, $threads) = split /:/, $val;
			$hash{sockets} = $sockets;
			$hash{cores} = $cores;
			$hash{threads} = $threads;
			next FIELD;
		}

		if ( $field eq 'jobid' )
		{	#%all produces two fields labelled 'jobid'
			#First is the true job id ?
			#Second seems to be dup of array_job_id ???
			#Just skip second for now
			next FIELD if exists $hash{jobid};
		}

		$hash{$field} = $val;
	}

	#And some calculated types
	$hash{end_tstamp} = $class->tstamp_from_time_field($hash{end_time});
	$hash{start_tstamp} = $class->tstamp_from_time_field($hash{start_time});
	$hash{submit_tstamp} = 
		$class->tstamp_from_time_field($hash{submit_time});

	return $class->new(%hash);
}
		

#-------------------------------------------------------------------
#	Class methods
#-------------------------------------------------------------------

sub parse_squeue_output_into_jobs($$)
{	my $class = shift;
	my $output_list = shift;
	my $me = 'parse_squeue_output_into_jobs';

	return "$me: Null output list" unless $output_list;
	return "$me: Invalid output list '$output_list', expecting nonempty list ref" 
		unless ref($output_list) eq 'ARRAY' && scalar(@$output_list);

	my $header = shift @$output_list;
	my $jobs = [];
	foreach my $line (@$output_list)
	{	my $job = $class->new_from_squeue_output($header, $line);
		push @$jobs, $job;
	}
	return $jobs;
}

#-------------------------------------------------------------------
#	Calculated values
#-------------------------------------------------------------------

sub tstamp_from_time_field($$)
#Converts a timefield (e.g. '2017-03-21T05:36:06') to second from epoch
#Returns undef if not a valid timefield
{	my $class = shift;
	my $timestr = shift;

	return unless $timestr;

	my $tstamp = parsedate($timestr,
		WHOLE=>1,
		TIME_REQUIRED=>1,
		DATE_REQUIRED=>1,
		NO_RELATIVE=>1,
		VALIDATE=>1,
	);

	return $tstamp;
}

sub elapsed_to_delta_seconds($$)
#Converts an elapsed time structure (days-hours:mins:secs) to seconds
{	my $class = shift;
	my $elapsed = shift;
	return unless $elapsed;

	my $delta = 0;
	my $onemin = 60;
	my $onehr = 60*$onemin;
	my $oneday = 24*$onehr;

	my $tmpelapsed = $elapsed;
	if ( $tmpelapsed =~ /^(\d+)-/ )
	{	my $days = $1;
		$tmpelapsed =~ s/^(\d+)-//;
		$delta += $days * $oneday;
	}

	my @tmp = split /:/, $tmpelapsed;
	my $hrs = $tmp[0];
	my $mins = $tmp[1];
	my $secs = $tmp[2];
	$delta += $onehr * $hrs;
	$delta += $onemin * $mins;
	$delta += $secs;

	return $delta;
}

sub delta_seconds_to_elapsed($$;$)
#Takes a delta in seconds and returns elapsed time as
# (days-)hours:mins:seconds
#Flags determines some options: Recognized are:
#	force_days: always includes days-, even if 0 days
{	my $class = shift;
	my $delta = shift;
	my $flags = shift || {};
	my $me = __PACKAGE__ . '::delta_seconds_to_elapsed';

	my $secs = $delta % 60;
	$delta = floor($delta/60);
	my $mins = $delta % 60;
	$delta = floor($delta/60);
	my $hours = $delta %24;
	$delta = floor($delta/24);
	my $days = $delta;

	my $hrminsec_fmt = "%02d:%02d:%02d";
	my $fmt;
	my @args;
	if ( $days || $flags->{force_days} )
	{	$fmt = "%02d-$hrminsec_fmt";
		@args = ($days, $hours, $mins, $secs);
	} else
	{	$fmt = $hrminsec_fmt;
		@args = ($hours, $mins, $secs);
	}

	my $etime = sprintf $fmt, @args;
	return $etime;
}


sub time_left_seconds($)
#Converts time_left (as days=hours:mins:secs) to seconds
{	my $self = shift;
	return unless defined $self && ref($self);
	my $time_left = $self->time_left;
	return unless defined $time_left;

	my $dtime = $self->elapsed_to_delta_seconds($time_left);
	return $dtime;
}

sub seconds_waited($)
#How many seconds did job spend waiting in the queue.
#If has started, start_time - submit_time
#If it has not started yet, now - submit_time
{	my $self = shift;

	my $state = $self->st;
	my $startts = time;
	my $submitts = $self->submit_tstamp;
	#Should always have a submit time, right?
	return unless defined $submitts;

	if ( $self->start_tstamp && $state ne 'PD' )
	{	#If has a start time, use it unless job is pending
		$startts = $self->start_tstamp;
	}
	
	my $delta_secs = $startts - $submitts;
	$delta_secs = 0 if $delta_secs < 0;
	return $delta_secs;
}

sub wait_seconds($)
#How many seconds did job expected to wait in the queue.
#Same as seconds_waited if job has started (start_time - submit_time)
#If job is pending but has a start_time, return start_time - submit_time
#If job is pending but has no start_time, return now - start_time
{	my $self = shift;

	my $state = $self->state;
	my $startts = time;
	my $submitts = $self->submit_tstamp;
	#Should always have a submit time, right?
	return unless defined $submitts;

	if ( $self->start_tstamp )
	{	#If has a start time, use it 
		$startts = $self->start_tstamp;
	}
	
	my $delta_secs = $startts - $submitts;
	$delta_secs = 0 if $delta_secs < 0;
	return $delta_secs;
}



1;
__END__

=head1 NAME

Slurm::Squeue - Perl wrapper to squeue command

=head1 SYNOPSIS

  use Slurm::Squeue;

  my $squeue = Slurm::Squeue->new( squeue=>'path to squeue command');

  my $squeue->verbose(1);

=head1 DESCRIPTION

B<Slurm::Squeue> and subclasses provide wrappers to the Slurm B<squeue> 
command.  


=head1 EXPORT

None.  OO interface only.

=head1 SEE ALSO

Slurm::Squeue::Job

=head1 AUTHOR

Tom Payerle, payerle@umd.edu

=head1 COPYRIGHT AND LICENSE

Copyright 2014, University of Maryland

This library is free software; you can redistribute it and/or modify it
under the same terms as Perl itself, either Perl version 5.8.4 or, at
your option, any later version of Perl 5 you may have available.
